Package: visPedigree
Type: Package
Title: Tidying, Analysis, and Fast Visualization of Animal and Plant
        Pedigrees
Version: 1.9.0
Authors@R: person("Sheng","Luan", email="luansheng@gmail.com", role = c("aut","cre"))
Description: Provides tools for the analysis and visualization of animal and
    plant pedigrees. Analytical methods include equivalent complete generations,
    generation intervals, effective population size (via inbreeding, coancestry,
    and demographic approaches), founder and ancestor contributions, partial
    inbreeding, genetic diversity indices, and additive (A), dominance (D), and
    epistatic (AA) relationship matrices. Core algorithms — ancestry tracing,
    topological sorting, inbreeding coefficients, and matrix construction — are
    implemented in C++ ('Rcpp', 'RcppArmadillo') and 'data.table', scaling to
    pedigrees with over one million individuals. Matrix-free pedigree
    traversals apply additive relationship matrices or their inverses to
    vectors and matrices without materializing dense square matrices. Pedigree
    graphs are rendered via 'igraph' with support for compact full-sib family
    display; relationship matrices can be visualized as heatmaps. Supports
    complex mating systems, including selfing and pedigrees in which the same
    individual can appear as both sire and dam.
License: GPL-3
Encoding: UTF-8
LazyData: true
Depends: R (>= 4.1.0)
Imports: data.table (>= 1.14.0), igraph (>= 1.3.0), Matrix, methods,
        Rcpp, lattice
LinkingTo: Rcpp, RcppArmadillo
Suggests: nadiv (>= 2.18.0), testthat (>= 3.0.0), knitr, rmarkdown,
        devtools, roxygen2, pkgdown
URL: https://github.com/luansheng/visPedigree,
        https://luansheng.github.io/visPedigree/
BugReports: https://github.com/luansheng/visPedigree/issues
VignetteBuilder: knitr
Config/testthat/edition: 3
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-07-15 00:06:43 UTC; luansheng
Author: Sheng Luan [aut, cre]
Maintainer: Sheng Luan <luansheng@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-15 05:30:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-07-15 07:22:44 UTC; unix
Archs: visPedigree.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("[",pedmat)
S3method("[",tidyped)
S3method(as.matrix,pedmat)
S3method(as.vector,pedmat)
S3method(dim,pedmat)
S3method(length,pedmat)
S3method(plot,pedhalflife)
S3method(plot,pedstats)
S3method(plot,tidyped)
S3method(print,pedcontrib)
S3method(print,pedhalflife)
S3method(print,pediv)
S3method(print,pedmat)
S3method(print,pedstats)
S3method(print,splitped)
S3method(print,summary.pedmat)
S3method(print,summary.tidyped)
S3method(print,tidyped)
S3method(summary,pedmat)
S3method(summary,splitped)
S3method(summary,tidyped)
S3method(t,pedmat)
export(as_tidyped)
export(expand_pedmat)
export(has_candidates)
export(has_inbreeding)
export(inbreed)
export(is_tidyped)
export(pedancestry)
export(pedcontrib)
export(pedecg)
export(pedfclass)
export(pedgenint)
export(pedhalflife)
export(pediv)
export(pedmat)
export(pedmeta)
export(pedne)
export(pedpartial)
export(pedprod)
export(pedrel)
export(pedstats)
export(pedsubpop)
export(query_relationship)
export(splitped)
export(summary_pedmat)
export(tidyped)
export(vismat)
export(visped)
import(data.table)
import(igraph)
importFrom(Rcpp,sourceCpp)
importFrom(data.table,as.data.table)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,dev.cur)
importFrom(grDevices,dev.off)
importFrom(grDevices,pdf)
importFrom(grDevices,svg)
importFrom(graphics,strwidth)
importFrom(igraph,E)
importFrom(igraph,V)
importFrom(igraph,add_vertices)
importFrom(igraph,as_adj_list)
importFrom(igraph,components)
importFrom(igraph,degree)
importFrom(igraph,distances)
importFrom(igraph,edge_attr)
importFrom(igraph,ego)
importFrom(igraph,ends)
importFrom(igraph,graph_from_data_frame)
importFrom(igraph,graph_from_edgelist)
importFrom(igraph,is_dag)
importFrom(igraph,layout_with_sugiyama)
importFrom(igraph,neighborhood)
importFrom(igraph,neighbors)
importFrom(igraph,norm_coords)
importFrom(igraph,plot.igraph)
importFrom(igraph,shortest_paths)
importFrom(igraph,subcomponent)
importFrom(igraph,topo_sort)
importFrom(igraph,vcount)
importFrom(igraph,vertex_attr)
importFrom(igraph,which_loop)
importFrom(lattice,histogram)
importFrom(lattice,levelplot)
importFrom(lattice,panel.abline)
importFrom(lattice,panel.levelplot)
importFrom(stats,as.dist)
importFrom(stats,as.formula)
importFrom(stats,hclust)
importFrom(stats,median)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(utils,head)
importFrom(utils,modifyList)
useDynLib(visPedigree, .registration = TRUE)

# visPedigree: Tidying, Analysis, and Fast Visualization of Animal and Plant Pedigrees <img src="https://raw.githubusercontent.com/luansheng/visPedigree/main/man/figures/logo.png" align="right" height="138" alt="" />

`visPedigree` provides tools for the analysis and visualization of animal and plant pedigrees. Analytical methods cover equivalent complete generations, generation intervals, effective population size, founder and ancestor contributions, partial inbreeding, genetic diversity indices, and relationship matrices (A, D, AA). Core algorithms — ancestry tracing, topological sorting, inbreeding coefficients, and matrix construction — are implemented in C++ (Rcpp/RcppArmadillo) and `data.table`, scaling to pedigrees with over one million individuals. Pedigree graphs are rendered via `igraph`; relationship matrices can be visualized as heatmaps. Complex mating systems are supported, including selfing and pedigrees in which the same individual can appear as both sire and dam.

<p align="center">
  <img src="https://raw.githubusercontent.com/luansheng/visPedigree/main/man/figures/ped_example.png" width="800" alt="Example pedigree visualization produced by visPedigree">
</p>

## Key Features

- **Pedigree Standardization**: Standardizes pedigree records, handles selfing and pedigrees in which the same individual can appear as both sire and dam, detects pedigree loops, prepares pedigrees for downstream analysis, and efficiently splits disconnected sub-populations.
- **Comprehensive Pedigree Analysis**: Computes pedigree summaries, equivalent complete generations, generation intervals, effective population size, founder and ancestor contributions, partial inbreeding, relationship matrices, and inbreeding coefficients.
- **High-Throughput Matrix Calculation**: Calculates Additive (A), Dominance (D), and Additive-by-Additive (AA) relationship matrices and their inverses.
- **Matrix-Free Relationship Products**: Computes $Ax$, $AX$, $A^{-1}x$, and $A^{-1}X$ directly from an ordered pedigree with `pedprod()`, avoiding construction of the dense additive relationship matrix.
- **Advanced Visualization**: Renders scalable pedigree graphs via `igraph` with compact representations for large full-sib families, plus heatmap displays for relationship matrices.
- **High Performance**: Core algorithms — ancestry tracing, topological sorting, inbreeding calculation, and matrix construction — are implemented in C++ (Rcpp/RcppArmadillo) with `data.table` for tabular operations, scaling to pedigrees with over one million individuals.

## Installation

### Stable version from CRAN

```R
install.packages("visPedigree")
```

### Development version from GitHub

```R
# install.packages("devtools")
devtools::install_github("luansheng/visPedigree", build_vignettes = TRUE)
```

## Documentation

### Standard workflows

Recommended reading order:

1. [How to tidy a pedigree](https://luansheng.github.io/visPedigree/articles/tidy-pedigree.html)
2. [Efficient visPedigree Workflows](https://luansheng.github.io/visPedigree/articles/efficient-visPedigree-workflows.html)
3. [How to draw a pedigree](https://luansheng.github.io/visPedigree/articles/draw-pedigree.html)
4. [Pedigree Analysis and Population Genetics](https://luansheng.github.io/visPedigree/articles/pedigree-analysis.html)
5. [Relationship Matrix](https://luansheng.github.io/visPedigree/articles/relationship-matrix.html)

### Developer documentation

6. [tidyped Class Structure and Extension Notes](https://luansheng.github.io/visPedigree/articles/tidyped-structure.html)

## Quick Start

```R
library(visPedigree)

# Example 1: Tidy and visualize a small pedigree
# Use compact = TRUE to condense eligible full-sib groups into a green-grey
# family-summary rectangle labelled FS×N.
cands <- c("Y", "Z1", "Z2")
small_ped |>
  tidyped(cand = cands) |>
  visped(compact = TRUE)

# Example 2: Relationship Matrices (v1.0.0+)
# Compute the additive relationship matrix (A) using high-performance C++ algorithms.
# Use compact = TRUE to accelerate calculation for pedigrees with large full-sib families.
mat_a <- simple_ped |> tidyped() |> pedmat(method = "A", compact = TRUE)
# Visualize the relationship matrix as a heatmap with histograms
vismat(mat_a)

# Apply A to candidate contribution weights without constructing A
tp_simple <- tidyped(simple_ped)
candidate_weights <- c(J5X804 = 0.5, J3Y620 = 0.5)
weighted_relationship <- pedprod(tp_simple, candidate_weights)

# Example 3: Inbreeding & Highlighting
# Calculate inbreeding coefficients (f) and display them on the graph (e.g., "ID (0.05)").
# Specific individuals can be highlighted to track their inheritance paths.
simple_ped |>
  tidyped(inbreed = TRUE) |>
  visped(highlight = "J5X804", trace = "up", showf = TRUE, compact = TRUE)

# Example 4: Pedigree Analysis (v1.4.0+)
# Summarize pedigree structure
tp <- tidyped(big_family_size_ped)
tp |>
  pedstats(timevar="Year")

# Summarize diversity-related indicators
ref_ind <- tp[Gen == max(Gen), Ind]
tp |>
  pediv(reference = ref_ind)

# Example 5: Pedigree Splitting
# Automatically split the pedigree into independent sub-populations (connected groups).
# Each group is returned as a standalone tidyped object for separate analysis.
split_list <- simple_ped |> tidyped() |> splitped()
summary(split_list[[1]])
```

## Citation

Luan Sheng (2026). visPedigree: Tidying, Analysis, and Fast Visualization of Animal and Plant Pedigrees. R package version 1.9.0, https://github.com/luansheng/visPedigree.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(visPedigree)
library(data.table)

data(deep_ped, package = "visPedigree")
data(big_family_size_ped, package = "visPedigree")
data(small_ped, package = "visPedigree")
data(inbred_ped, package = "visPedigree")

tp_deep <- tidyped(deep_ped)
tp_small <- tidyped(small_ped)
tp_inbred <- tidyped(inbred_ped)

## ----pedstats-overview--------------------------------------------------------
stats_deep <- pedstats(tp_deep)

stats_deep$summary
tail(stats_deep$ecg)
stats_deep$gen_intervals

# Visualize ancestral depth (Equivalent Complete Generations)
plot(stats_deep, type = "ecg", metric = "ECG")

## ----pedecg-example-----------------------------------------------------------
ecg_deep <- pedecg(tp_deep)

ecg_deep[order(-ECG)][1:10]

## ----pedgenint-basic----------------------------------------------------------
tp_time <- tidyped(big_family_size_ped)

genint_year <- pedgenint(tp_time, timevar = "Year", unit = "year")
genint_year

# Visualize generation intervals
# Note: we can also use stats <- pedstats(tp_time, timevar = "Year") followed by plot(stats)
plot(genint_year)

## ----pedgenint-cycle----------------------------------------------------------
genint_cycle <- pedgenint(tp_time, timevar = "Year", unit = "year", cycle = 1.2)

genint_cycle[Pathway %in% c("SS", "SD", "DS", "DD", "Average")]

## ----pedsubpop-example--------------------------------------------------------
ped_demo <- data.table(
  Ind = c("A", "B", "C", "D", "E", "F", "G", "H"),
  Sire = c(NA, NA, "A", NA, NA, "E", NA, NA),
  Dam  = c(NA, NA, "B", NA, NA, NA, NA, NA),
  Sex = c("male", "female", "male", "female", "male", "female", "male", "female"),
  Batch = c("L1", "L1", "L1", "L1", "L2", "L2", "L3", "L3")
)

tp_demo <- tidyped(ped_demo)

pedsubpop(tp_demo)
pedsubpop(tp_demo, by = "Batch")

## ----pediv-setup--------------------------------------------------------------
ref_pop <- tp_deep[Gen >= max(Gen) - 1, Ind]
length(ref_pop)

## ----pediv-analysis-----------------------------------------------------------
div_res <- pediv(tp_deep, reference = ref_pop, top = 10, seed = 42L)

div_res$summary
div_res$ancestors

## ----pediv-genediv------------------------------------------------------------
# GeneDiv is in the summary alongside MeanCoan
div_res$summary[, .(fg, MeanCoan, GeneDiv)]

## ----pediv-shannon------------------------------------------------------------
# Shannon metrics are included in pediv() output
div_res$summary[, .(NFounder, feH, fe, NAncestor, faH, fa)]

# The ratio feH/fe reveals long-tail founder diversity
div_res$summary[, .(rho_founder = feH / fe, rho_ancestor = faH / fa)]

## ----pedcontrib-shannon-------------------------------------------------------
# pedcontrib() provides the same metrics via its summary
contrib_res <- pedcontrib(tp_deep, reference = ref_pop, mode = "both")
contrib_res$summary[c("n_founder", "f_e_H", "f_e", "n_ancestor", "f_a_H", "f_a")]

## ----pedhalflife-analysis-----------------------------------------------------
hl <- pedhalflife(tp_deep, timevar = "Gen")
print(hl)

## ----pedhalflife-timeseries---------------------------------------------------
hl$timeseries

## ----pedhalflife-subset-------------------------------------------------------
hl_recent <- pedhalflife(tp_deep, timevar = "Gen",
                         at = tail(sort(unique(tp_deep$Gen)), 4))
print(hl_recent)

## ----pedhalflife-plot, fig.width=6, fig.height=4------------------------------
plot(hl, type = "log")

## ----pedhalflife-plot-raw, fig.width=6, fig.height=4--------------------------
plot(hl, type = "raw")

## ----pedrel-example-----------------------------------------------------------
tp_small$BirthYear <- 2010 + tp_small$Gen

rel_by_gen <- pedrel(tp_small, by = "Gen")
rel_by_gen

## ----pedrel-reference---------------------------------------------------------
ref_ids <- c("Z1", "Z2", "X", "Y")

pedrel(tp_small, by = "Gen", reference = ref_ids)

## ----pedrel-coancestry--------------------------------------------------------
coan_by_gen <- pedrel(tp_small, by = "Gen", scale = "coancestry")
coan_by_gen

## ----inbreed-trend------------------------------------------------------------
tp_inbred_f <- inbreed(tp_inbred)

f_trend <- tp_inbred_f[, .(MeanF = mean(f, na.rm = TRUE)), by = Gen]
f_trend

## ----pedfclass-example--------------------------------------------------------
pedfclass(tp_inbred)

## ----pedfclass-breaks---------------------------------------------------------
pedfclass(tp_inbred, breaks = c(0.03125, 0.0625, 0.125, 0.25))

## ----pedancestry-example------------------------------------------------------
ped_line <- data.table(
  Ind = c("A", "B", "C", "D", "E", "F", "G"),
  Sire = c(NA, NA, NA, NA, "A", "C", "E"),
  Dam  = c(NA, NA, NA, NA, "B", "D", "F"),
  Sex = c("male", "female", "male", "female", "male", "female", "male"),
  Line = c("Line1", "Line1", "Line2", "Line2", NA, NA, NA)
)

tp_line <- tidyped(ped_line)

anc <- pedancestry(tp_line, foundervar = "Line")
anc

## ----pedpartial-example-------------------------------------------------------
partial_f <- pedpartial(tp_inbred, ancestors = c("A", "B"))
partial_f

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6.5,
  fig.height = 6.5,
  dpi = 300,
  out.width = "100%"
)
library(visPedigree)

## ----smallped, fig.show="hold"------------------------------------------------
tidy_small_ped <-
  tidyped(ped = small_ped,
          cand = c("Y", "Z1", "Z2"))
visped(tidy_small_ped, 
        compact = TRUE, 
        cex=0.5, 
        symbolsize=10, 
        file = tempfile(fileext = ".pdf"))

## ----symbol-schemes, fig.show="hold"------------------------------------------
visped(tidy_small_ped, 
      compact = TRUE,
      cex=0.5, 
      symbolsize=10)
visped(tidy_small_ped, 
        compact = TRUE, 
        shapeby = "role",
        cex=0.5, 
        symbolsize=10)

## ----vissimpleped, fig.show="hold"--------------------------------------------
tidy_simple_ped <- tidyped(simple_ped)
visped(tidy_simple_ped, cex=0.3, symbolsize=10)

## ----vissimpleped_genlab, fig.show="hold"-------------------------------------
visped(tidy_simple_ped, cex = 0.3, symbolsize = 10, genlab = TRUE)

## ----vissimpleped_custom_genlab, fig.show="hold"------------------------------
generation_labels <- paste0("Gen_", sort(unique(tidy_simple_ped$Gen)))
visped(tidy_simple_ped, cex = 0.3, symbolsize = 10, genlab = generation_labels, genlabcex=0.6)

## ----vissimpleped_genlabcex, fig.show="hold"----------------------------------
# cex controls individual label size; genlabcex controls generation label size
visped(tidy_simple_ped, cex = 0.3, symbolsize = 15, genlab = TRUE, genlabcex = 0.8)

## -----------------------------------------------------------------------------
suppressMessages(visped(tidy_simple_ped, showgraph = FALSE, file = tempfile(fileext = ".pdf")))

## -----------------------------------------------------------------------------
suppressMessages(visped(tidy_simple_ped, showgraph = FALSE, file = tempfile(fileext = ".svg")))

## ----custom-labels, fig.show="hold"-------------------------------------------
tidy_simple_ped[, ShortLabel := paste0("N", seq_len(.N))]
visped(tidy_simple_ped, labelvar = "ShortLabel", cex = 0.3, symbolsize = 10)

# A row-aligned vector is also accepted
visped(tidy_simple_ped, labelvar = rep("1x", nrow(tidy_simple_ped)),
       cex = 0.3, symbolsize = 10)

## ----highlight1---------------------------------------------------------------
visped(tidyped(small_ped), highlight = c("Y", "Z1"))

## ----highlight_trace----------------------------------------------------------
# Highlight individual "Y" and all its ancestors and descendants
visped(tidyped(small_ped), highlight = "Y", trace = TRUE)

## ----highlight2---------------------------------------------------------------
visped(tidyped(small_ped), 
       highlight = list(ids = c("Y", "Z1"), 
                        frame.color = "#4caf50", 
                        color = "#81c784"))

## ----showinbreed--------------------------------------------------------------
library(data.table)
test_ped <- data.table(
  Ind = c("A", "B", "C", "D", "E"),
  Sire = c(NA, NA, "A", "C", "C"),
  Dam = c(NA, NA, "B", "B", "D"),
  Sex = c("male", "female", "male", "female", "male")
)
tidy_test_ped_inbreed <- tidyped(test_ped, inbreed = TRUE)
visped(tidy_test_ped_inbreed, showf = TRUE)

## ----plant-selfing-draw-------------------------------------------------------
library(data.table)

# P1 × P2 → F1;  F1 selfs → F2_1, F2_2;  F2_1 selfs → F3
plant_ped <- data.table(
  Ind  = c("P1", "P2", "F1", "F2_1", "F2_2", "F3"),
  Sire = c(NA,   NA,   "P1", "F1",   "F1",   "F2_1"),
  Dam  = c(NA,   NA,   "P2", "F1",   "F1",   "F2_1")
)

tp_plant <- tidyped(plant_ped, selfing = TRUE, inbreed = TRUE)

visped(tp_plant, compact = TRUE, showf = TRUE, cex=0.5, symbolsize=10)

## ----deepped, eval=FALSE------------------------------------------------------
# cand_J11_labels <- deep_ped[(substr(Ind, 1, 3) == "K11"), Ind]
# visped(
#   tidyped(deep_ped,
#     cand = cand_J11_labels,
#     tracegen = 3),
#   cex=0.08,
#   symbolsize=5.5
#   )

## ----reduceped1---------------------------------------------------------------
cand_J11_labels <- deep_ped[(substr(Ind,1,3) == "K11"),Ind]
visped(
  tidyped(
    deep_ped,
    cand = cand_J11_labels,
    trace = "up",
    tracegen = 3
  ),
  cex=0.08, 
  symbolsize=5.5,
  compact = TRUE,
  showgraph = TRUE,
  file = tempfile(fileext = ".pdf")
)

## ----reduceped2---------------------------------------------------------------
visped(
  tidyped(
    deep_ped,
    cand = cand_J11_labels,
    trace = "up",
    tracegen = 3
  ),
  compact = TRUE,
  cex = 0.83,
  showgraph = FALSE,
  file = tempfile(fileext = ".pdf")
)

## ----reduceped3---------------------------------------------------------------
suppressMessages(visped(
  tidyped(
    deep_ped,
    cand = cand_J11_labels, 
    tracegen = 3),
  compact = TRUE,
  outline = TRUE,
  showgraph = TRUE,
  file = tempfile(fileext = ".pdf")
))

## ----pedofoneind--------------------------------------------------------------
suppressWarnings(
  K110550H_ped <- tidyped(deep_ped, cand = "K110550H")
)
if(interactive()) {
  visped(K110550H_ped, compact = TRUE)
}

## ----optiMate-----------------------------------------------------------------
cand_2007_G8_labels <-
  big_family_size_ped[(Year == 2007) & (substr(Ind, 1, 2) == "G8"), Ind]
suppressWarnings(
  cand_2007_G8_tidy_ped_ancestor_2 <-
    tidyped(
      big_family_size_ped,
      cand = cand_2007_G8_labels,
      trace = "up",
      tracegen = 2
    )
)
sire_label <-
  unique(cand_2007_G8_tidy_ped_ancestor_2[Ind %in% cand_2007_G8_labels,
                                          Sire])
dam_label <-
  unique(cand_2007_G8_tidy_ped_ancestor_2[Ind %in% cand_2007_G8_labels,
                                          Dam])
sire_dam_label <- unique(c(sire_label, dam_label))
sire_dam_label <- sire_dam_label[!is.na(sire_dam_label)]
sire_dam_ped <-
  cand_2007_G8_tidy_ped_ancestor_2[Ind %in% sire_dam_label]
sire_dam_ped <-
  sire_dam_ped[, FamilyID := paste(Sire, Dam, sep = "")]
family_size <- sire_dam_ped[, .N, by = c("FamilyID")]
fullsib_family_label <- unique(sire_dam_ped$FamilyID)
suppressMessages(
  visped(
    cand_2007_G8_tidy_ped_ancestor_2,
    compact = TRUE,
    outline = TRUE,
    showgraph = TRUE
  )
)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6.5,
  fig.height = 6,
  dpi = 96,
  out.width = "100%"
)

## ----setup--------------------------------------------------------------------
library(visPedigree)
library(data.table)

data(simple_ped, package = "visPedigree")

## ----tidy-once----------------------------------------------------------------
tp_master <- tidyped(simple_ped)

class(tp_master)
is_tidyped(tp_master)
pedmeta(tp_master)

## ----fast-trace---------------------------------------------------------------
tp_up <- tidyped(tp_master, cand = "J5X804", trace = "up", tracegen = 2)
tp_down <- tidyped(tp_master, cand = "J0Z990", trace = "down")

has_candidates(tp_up)
tp_up[, .(Ind, Sire, Dam, Cand)]

## ----fast-trace-pattern, eval = FALSE-----------------------------------------
# # expensive once
# # tp_master <- tidyped(raw_ped)
# 
# # cheap many times
# # tp_a <- tidyped(tp_master, cand = ids_a, trace = "up")
# # tp_b <- tidyped(tp_master, cand = ids_b, trace = "all", tracegen = 3)
# # tp_c <- tidyped(tp_master, cand = ids_c, trace = "down")

## ----dt-modify----------------------------------------------------------------
tp_work <- copy(tp_master)
tp_work[, phenotype := seq_len(.N)]

class(tp_work)
head(tp_work[, .(Ind, phenotype)])

## ----incomplete-subset--------------------------------------------------------
ped_year <- data.table(
  Ind = c("A", "B", "C", "D"),
  Sire = c(NA, NA, "A", "C"),
  Dam = c(NA, NA, "B", "B"),
  Year = c(2000, 2000, 2005, 2006)
)

tp_year <- tidyped(ped_year)
sub_dt <- tp_year[Year > 2005]

class(sub_dt)
sub_dt

## ----incomplete-subset-error, error = TRUE------------------------------------
try({
inbreed(sub_dt)
})

## ----explicit-tracing---------------------------------------------------------
valid_sub_tp <- tidyped(tp_year, cand = "D", trace = "up")

class(valid_sub_tp)
valid_sub_tp[, .(Ind, Sire, Dam, Cand)]

## ----explicit-analysis--------------------------------------------------------
inbreed(valid_sub_tp)[Ind == "D", .(Ind, f)]

## ----split-vs-summary---------------------------------------------------------
sub_tps <- splitped(tp_master)
length(sub_tps)
class(sub_tps[[1]])

pedsubpop(tp_master)

## ----accessors----------------------------------------------------------------
tp_f <- inbreed(tp_master)

is_tidyped(tp_f)
has_inbreeding(tp_f)
has_candidates(tp_f)
pedmeta(tp_f)

## ----best-practice, eval = FALSE----------------------------------------------
# # 1. build one validated master object
# # tp_master <- tidyped(raw_ped)
# 
# # 2. add analysis-specific columns in place
# # tp_master[, phenotype := pheno_vector]
# # tp_master[, cohort := year_vector]
# 
# # 3. extract valid candidate sub-pedigrees explicitly
# # tp_sel <- tidyped(tp_master, cand = selected_ids, trace = "up", tracegen = 3)
# 
# # 4. run downstream analysis on either the full master or traced sub-pedigree
# # pedstats(tp_master)
# # pedmat(tp_sel)
# # inbreed(tp_sel)
# # visped(tp_sel)
# 
# # 5. split only when disconnected components really matter
# # comps <- splitped(tp_master)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6.5,
  fig.height = 6.5,
  dpi = 300,
  out.width = "100%"
)
library(visPedigree)
library(Matrix)

## ----basic_calc---------------------------------------------------------------
# Load example pedigree and tidy it
data(small_ped)
tped <- tidyped(small_ped)

# Calculate Additive Relationship Matrix (A)
mat_A <- pedmat(tped, method = "A")

# Calculate Dominance Relationship Matrix (D)
mat_D <- pedmat(tped, method = "D")

# Calculate inbreeding coefficients (f)
vec_f <- pedmat(tped, method = "f")

## ----sparse_check-------------------------------------------------------------
class(mat_A)

## ----pedprod_Ax---------------------------------------------------------------
# Load example pedigree and tidy it
data(small_ped)
tped <- tidyped(small_ped)

# Equal contributions from two candidates; other individuals have zero weight
weights <- c(Z1 = 0.5, Z2 = 0.5)

# Relationship of every pedigree individual to the weighted candidate group
Ax <- pedprod(tped, weights)
head(Ax)

# Average additive relationship c' A c and average coancestry
mean_relationship <- sum(weights * Ax[names(weights)])
mean_coancestry <- mean_relationship / 2
c(mean_relationship = mean_relationship, mean_coancestry = mean_coancestry)

## ----pedprod_named_unnamed----------------------------------------------------
# Named: only listed individuals get non-zero values
named_x <- c(Z1 = 0.3, Z2 = 0.4, A = 0.3)
pedprod(tped, named_x)[1:6]  # B, C, etc. are automatically zero

# Unnamed: must match the pedigree size
unnamed_x <- rep(1, nrow(tped))
length(pedprod(tped, unnamed_x))

## ----pedprod_Ainv-------------------------------------------------------------
# Ainv * vector
x <- rnorm(nrow(tped))
Ainv_x <- pedprod(tped, x, method = "Ainv")
head(Ainv_x)

# Verify against explicit computation (small pedigree only)
A <- pedmat(tped, method = "A", sparse = FALSE)
Ainv <- pedmat(tped, method = "Ainv", sparse = FALSE)
all.equal(
  unname(Ainv_x),
  unname(drop(Ainv %*% x)),
  tolerance = 1e-12
)

## ----pedprod_Ainv_identity----------------------------------------------------
# Ainv * (A * x) should recover x (to machine precision)
A_x <- pedprod(tped, x, method = "A")
Ainv_Ax <- pedprod(tped, A_x, method = "Ainv")
all.equal(unname(Ainv_Ax), unname(x), tolerance = 1e-12)

## ----pedprod_AX---------------------------------------------------------------
# Three ways to weight the SAME two candidates (Z1 and Z2 are full sibs)
schemes <- cbind(
  Equal      = c(Z1 = 0.5, Z2 = 0.5),
  Z1_only    = c(Z1 = 1.0, Z2 = 0.0),
  Weighted   = c(Z1 = 0.7, Z2 = 0.3)
)
AX <- pedprod(tped, schemes)

# The schemes diverge at the candidates themselves (and their descendants);
# a shared ancestor is equally related to any weighting of two full sibs.
AX[c("Z1", "Z2"), ]

# Expected average relationship within each contributing group, 0.5 * c'A c.
# Under random mating this equals the mean inbreeding of the resulting progeny.
group_coancestry <- 0.5 * colSums(schemes * AX[rownames(schemes), ])
round(group_coancestry, 5)

## ----pedprod_ocs--------------------------------------------------------------
# Weighted candidate contributions
candidates <- setNames(c(0.25, 0.25, 0.25, 0.25), c("Z1", "Z2", "A", "B"))
Ac <- pedprod(tped, candidates)

# Average coancestry of the selected group: 0.5 * c' A c
c_accepted <- sum(candidates * Ac[names(candidates)]) / 2
c_accepted

## ----pedprod_blup-------------------------------------------------------------
# Simulated breeding values as a matrix right-hand side
set.seed(20260704)
Z_design <- cbind(
  trait1 = rnorm(nrow(tped)),
  trait2 = rnorm(nrow(tped))
)
rownames(Z_design) <- tped$Ind

# Ainv * Z in one traversal — no Ainv ever stored
Ainv_Z <- pedprod(tped, Z_design, method = "Ainv")
dim(Ainv_Z)
Ainv_Z[1:5, ]

## ----pedprod_geneflow---------------------------------------------------------
# One unit contribution vector per founder: column f isolates founder f
founder_ids <- tped[Gen == 1, Ind]
founder_design <- diag(length(founder_ids))
dimnames(founder_design) <- list(founder_ids, founder_ids)

# A %*% e_f is founder f's expected genetic contribution to every individual
footprint <- pedprod(tped, founder_design, method = "A")

# Founder composition of one candidate: contributions sum to one
round(footprint["Z1", ], 4)

# Average founder contribution to the youngest generation, ranked
young <- tped[Gen == max(Gen), Ind]
founder_share <- sort(colMeans(footprint[young, , drop = FALSE]), decreasing = TRUE)
round(founder_share, 4)

## ----pedprod_large_proof, eval=FALSE------------------------------------------
# # Pedigrees beyond the dense-A guard still work with pedprod()
# n <- 50000L
# ids <- paste0("I", seq_len(n))
# raw <- data.frame(
#   Ind  = ids,
#   Sire = c(NA_character_, ids[-n]),
#   Dam  = NA_character_,
#   stringsAsFactors = FALSE
# )
# tped_large <- tidyped(raw)
# 
# # pedprod works; pedmat would error
# result <- pedprod(tped_large, setNames(1, tail(ids, 1)))
# length(result)  # 50000

## ----matrix_summary-----------------------------------------------------------
tail(summary(mat_A),10)

## ----query--------------------------------------------------------------------
# Query relationship between Z1 and Z2
query_relationship(mat_A, "Z1", "Z2")

# Query multiple pairs
query_relationship(mat_A, c("Z1", "A"), c("Z2", "B"))

## ----compact_calc-------------------------------------------------------------
# Calculate compacted A matrix
mat_compact <- pedmat(tped, method = "A", compact = TRUE)

# The result is a 'pedmat' object containing the compacted matrix
print(mat_compact[11:20,11:20])

## ----expand-------------------------------------------------------------------
# Expand to full 28x28 matrix
mat_full <- expand_pedmat(mat_compact)
dim(mat_full)

# Query still works the same way
query_relationship(mat_compact, "Z1", "Z2")

## ----heatmap, fig.width=6, fig.height=6---------------------------------------
# Heatmap of the A matrix (with default clustering reorder)
vismat(mat_A, labelcex = 0.5)

## ----heatmap_compact, fig.width=6, fig.height=6-------------------------------
# Compact matrix: expanded automatically (message printed)
vismat(mat_compact,labelcex=0.5)

## ----heatmap_no_reorder, fig.width=6, fig.height=6----------------------------
vismat(mat_A, reorder = FALSE, labelcex = 0.5)

## ----heatmap_ids, fig.width=5, fig.height=5-----------------------------------
target_ids <- rownames(as.matrix(mat_A))[1:8]
vismat(mat_A, ids = target_ids,
       main = "Relationship Heatmap — First 8 Individuals")

## ----heatmap_group, fig.width=6, fig.height=6---------------------------------
# Mean relationship between generations
vismat(mat_A, ped = tped, by = "Gen",
       main = "Mean Relationship Between Generations")

## ----heatmap_family, fig.width=6, fig.height=6--------------------------------
# Mean relationship between full-sib families
# (founders without a family assignment are excluded automatically)
vismat(mat_A, ped = tped, by = "Family",
       main = "Mean Relationship Between Full-Sib Families")

## ----histogram, fig.width=6, fig.height=4-------------------------------------
# Distribution of relationship coefficients
vismat(mat_A, type = "histogram")

## ----large_ped_tip, fig.width=9, fig.height=8---------------------------------
data(big_family_size_ped)

tp_big <- tidyped(big_family_size_ped)
last_gen <- max(tp_big$Gen, na.rm = TRUE)

# Compute the compact A matrix for the entire pedigree
mat_big_compact <- pedmat(tp_big, method = "A", compact = TRUE)

# Focus on all individuals in the last generation that belong to a family
ids_last_gen <- tp_big[Gen == last_gen & !is.na(Family), Ind]

# vismat() aggregates directly from the compact matrix — no expansion needed
vismat(
       mat_big_compact,
       ped = tp_big,
       ids = ids_last_gen,
       by = "Family",
       labelcex = 0.3,
       main = paste("Mean Relationship Between All Families in Generation", last_gen)
)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6.5,
  fig.height = 6.5,
  dpi = 96,
  out.width = "100%"
)
library(visPedigree)

## ----gettingdataset,eval=FALSE------------------------------------------------
# data(package="visPedigree")

## ----simpleped----------------------------------------------------------------
head(simple_ped)
tail(simple_ped)
# The number of individuals in the pedigree dataset
nrow(simple_ped)
# Individual records with missing parents
simple_ped[Sire %in% c("0", "*", "NA", NA) |
             Dam %in% c("0", "*", "NA", NA)]

## ----error=TRUE---------------------------------------------------------------
try({
x <- data.table::copy(simple_ped)
x[ID == "J2F588", Sire := "J0Z167"]
y <- tidyped(x)
})

## ----tidyped------------------------------------------------------------------
tidy_simple_ped <- tidyped(simple_ped)
head(tidy_simple_ped)
tail(tidy_simple_ped)
nrow(tidy_simple_ped)

## -----------------------------------------------------------------------------
tidy_simple_ped_no_gen_num <-
  tidyped(simple_ped, addgen = FALSE, addnum = FALSE)
    head(tidy_simple_ped_no_gen_num)

## ----loop_detection, error=TRUE-----------------------------------------------
try({
# loop_ped contains cycles (e.g., V -> T -> R -> P -> M -> V)
# Attempting to tidy it will result in an error
try(tidyped(loop_ped))
})

## ----writeped,eval=FALSE------------------------------------------------------
# saved_ped <- data.table::copy(tidy_simple_ped)
# saved_ped[is.na(Sire), Sire := "0"]
# saved_ped[is.na(Dam), Dam := "0"]
# data.table::fwrite(
#   x = saved_ped,
#   file = tempfile(fileext = ".csv"),
#   sep = ",",
#   quote = FALSE
# )

## -----------------------------------------------------------------------------
tidy_simple_ped_J5X804_ancestors <-
  tidyped(ped = tidy_simple_ped_no_gen_num, cand = "J5X804")
  tail(tidy_simple_ped_J5X804_ancestors)

## -----------------------------------------------------------------------------
tidy_simple_ped_J5X804_ancestors_2 <-
  tidyped(ped = tidy_simple_ped_no_gen_num,
  cand = "J5X804",
  tracegen = 2)
  print(tidy_simple_ped_J5X804_ancestors_2)

## -----------------------------------------------------------------------------
tidy_simple_ped_J0Z990_offspring <-
  tidyped(ped = tidy_simple_ped_no_gen_num, cand = "J0Z990", trace = "down")
  print(tidy_simple_ped_J0Z990_offspring)

## ----intped-------------------------------------------------------------------
tidy_simple_ped_with_int <-
  tidyped(ped = tidy_simple_ped_no_gen_num, addnum = TRUE)
head(tidy_simple_ped_with_int)

## ----inbreed------------------------------------------------------------------
# Create a simple inbred pedigree
library(data.table)
test_ped <- data.table(
  Ind = c("A", "B", "C", "D", "E"),
  Sire = c(NA, NA, "A", "C", "C"),
  Dam = c(NA, NA, "B", "B", "D"),
  Sex = c("male", "female", "male", "female", "male")
)
# Option 1: Calculate during tidying
tidy_test <- tidyped(test_ped, inbreed = TRUE)
head(tidy_test)

# Option 2: Calculate after tidying
tidy_test <- inbreed(tidyped(test_ped))

## ----genmethod----------------------------------------------------------------
# Default behavior (Top-Down): J2Y434 is at Gen 3
tidy_top <- tidyped(simple_ped, genmethod = "top")
tidy_top[Ind == "J2Y434"]

# Bottom-Up behavior: J2Y434 is at Gen 6
tidy_bottom <- tidyped(simple_ped, genmethod = "bottom")
tidy_bottom[Ind == "J2Y434"]

## ----summary------------------------------------------------------------------
# Summarize the tidied pedigree
summary(tidy_simple_ped)

## ----splitped-----------------------------------------------------------------
# Split the pedigree into components
sub_pedigrees <- splitped(tidy_simple_ped)

# View summary of the split result
summary(sub_pedigrees)

# Access a specific sub-pedigree
# first_sub <- sub_pedigrees[[1]]

## ----plant-selfing------------------------------------------------------------
library(data.table)

# A plant pedigree with self-fertilization:
# P1 (male) × P2 (female) → F1
# F1 selfs → F2_1, F2_2
# F2_1 selfs → F3
plant_ped <- data.table(
  Ind  = c("P1", "P2", "F1", "F2_1", "F2_2", "F3"),
  Sire = c(NA,   NA,   "P1", "F1",   "F1",   "F2_1"),
  Dam  = c(NA,   NA,   "P2", "F1",   "F1",   "F2_1")
)

# Without selfing = TRUE: error — F1 and F2_1 appear as both Sire and Dam
try(tidyped(plant_ped))

## ----plant-selfing-tidy-------------------------------------------------------
tp_plant <- tidyped(plant_ped, selfing = TRUE, inbreed = TRUE)
tp_plant[]

# Check metadata
attr(tp_plant, "ped_meta")

## ----plant-selfing-summary----------------------------------------------------
summary(tp_plant)

## ----plant-selfing-inbreed----------------------------------------------------
# Inbreeding coefficients
tp_plant[, .(Ind, Sex, Gen, f)]

#' Matrix-Free Pedigree Relationship Products
#'
#' Computes products of the form \eqn{A x}, \eqn{A X}, \eqn{A^{-1} x}, or
#' \eqn{A^{-1} X} directly from an ordered pedigree, without constructing the
#' full additive relationship matrix \eqn{A}. This is useful for large
#' pedigrees where explicitly storing the dense \eqn{n \times n} matrix is
#' infeasible.
#'
#' @param ped A complete pedigree accepted by \code{\link{tidyped}}, preferably
#'   an existing \code{tidyped} object.
#' @param x A numeric or logical vector, or a numeric or logical matrix.
#'   An unnamed vector must have one value per pedigree individual in
#'   \code{IndNum} order. A named vector may contain a subset of individuals;
#'   omitted individuals receive value zero. For a matrix, the same rules apply
#'   to its rows and \code{rownames}.
#' @param method Character scalar. \code{"A"} (default) computes products with
#'   the additive relationship matrix. \code{"Ainv"} computes products with its
#'   inverse.
#'
#' @details
#' For an ordered pedigree, the additive relationship matrix can be written as
#' \deqn{A = T D T^\prime,}
#' where \eqn{T} is the transmission matrix and \eqn{D} contains Mendelian
#' sampling variances. \code{pedprod()} applies these factors through backward
#' and forward pedigree traversals. For an \eqn{n \times p} right-hand side,
#' computation requires \eqn{O(np)} time and working memory rather than
#' materializing the \eqn{O(n^2)} relationship matrix.
#'
#' Named inputs are aligned by individual ID. Every supplied name must occur in
#' \code{ped$Ind}, and duplicate, missing, or empty names are rejected. Missing
#' values and non-finite input values are also rejected to avoid propagating an
#' invalid value through all descendants or relatives.
#'
#' @return For vector input, a named numeric vector in pedigree order. For
#'   matrix input, a numeric matrix whose rows are in pedigree order; input
#'   column names are preserved.
#'
#' @references
#' Colleau, J. J. (2002). An indirect approach to the extensive calculation of
#' relationship coefficients. \emph{Genetics Selection Evolution}, 34,
#' 409--421. \doi{10.1051/gse:2002015}
#'
#' Colleau, J. J., Palhiere, I., Rodriguez-Ramilo, S. T., & Legarra, A.
#' (2017). A fast indirect method to compute functions of genomic relationships
#' concerning genotyped and ungenotyped individuals, for diversity management.
#' \emph{Genetics Selection Evolution}, 49, 87.
#' \doi{10.1186/s12711-017-0363-9}
#'
#' @seealso \code{\link{pedmat}} for explicitly constructing relationship
#'   matrices and \code{\link{pedrel}} for group relationship summaries.
#'
#' @examples
#' tped <- tidyped(small_ped)
#'
#' # Equal contributions from two candidates; unspecified individuals are zero.
#' contributions <- c(Z1 = 0.5, Z2 = 0.5)
#' relationship_to_group <- pedprod(tped, contributions)
#'
#' # Average additive relationship among the weighted candidates: c' A c.
#' sum(contributions * relationship_to_group[names(contributions)])
#'
#' # Multiple contribution schemes can be evaluated in one call.
#' schemes <- cbind(
#'   Equal = c(A = 0.5, B = 0.5),
#'   SireA = c(A = 1.0, B = 0.0)
#' )
#' pedprod(tped, schemes)
#'
#' @export
pedprod <- function(ped, x, method = c("A", "Ainv")) {
  method <- match.arg(method)

  if (inherits(ped, "splitped")) {
    stop(
      "pedprod() does not support 'splitped' objects directly. ",
      "Apply pedprod() to one complete pedigree group at a time.",
      call. = FALSE
    )
  }

  ped <- data.table::copy(ped)
  if (is_tidyped(ped) ||
      all(c("Ind", "Sire", "Dam", "Sex", "Gen",
            "IndNum", "SireNum", "DamNum") %in% names(ped))) {
    ped <- ensure_complete_tidyped(ped, "pedprod()")
  } else {
    ped <- tidyped(ped, addnum = TRUE)
  }

  if (!all(c("IndNum", "SireNum", "DamNum") %in% names(ped))) {
    ped[, IndNum := .I]
    ped[, SireNum := match(Sire, Ind, nomatch = 0L)]
    ped[, DamNum := match(Dam, Ind, nomatch = 0L)]
  }

  data.table::setorder(ped, IndNum)
  rhs <- prepare_pedprod_rhs(x, ped$Ind)
  f_res <- cpp_calculate_inbreeding(ped$SireNum, ped$DamNum)

  result <- switch(
    method,
    A = cpp_multiply_A(ped$SireNum, ped$DamNum, f_res$dii, rhs$values),
    Ainv = cpp_multiply_Ainv(ped$SireNum, ped$DamNum, f_res$dii, rhs$values)
  )

  rownames(result) <- ped$Ind
  colnames(result) <- rhs$colnames

  if (rhs$vector) {
    result <- as.numeric(result[, 1L])
    names(result) <- ped$Ind
  }

  result
}

# Prepare a vector or matrix right-hand side in pedigree row order.
prepare_pedprod_rhs <- function(x, ids) {
  is_vector <- is.atomic(x) && is.null(dim(x))
  is_matrix <- is.matrix(x)

  if (!is_vector && !is_matrix) {
    stop("'x' must be a numeric or logical vector or matrix.", call. = FALSE)
  }
  if (!is.numeric(x) && !is.logical(x)) {
    stop("'x' must be numeric or logical.", call. = FALSE)
  }
  if (length(x) == 0L || (is_matrix && ncol(x) == 0L)) {
    stop("'x' must contain at least one value.", call. = FALSE)
  }
  if (anyNA(x) || any(!is.finite(x))) {
    stop("'x' must contain only finite, non-missing values.", call. = FALSE)
  }

  n <- length(ids)

  if (is_vector) {
    x_names <- names(x)
    if (is.null(x_names)) {
      if (length(x) != n) {
        stop(
          sprintf("Unnamed 'x' must have length %d; got %d.", n, length(x)),
          call. = FALSE
        )
      }
      values <- matrix(as.numeric(x), ncol = 1L)
    } else {
      validate_pedprod_ids(x_names, ids, "names(x)")
      values <- matrix(0, nrow = n, ncol = 1L)
      values[match(x_names, ids), 1L] <- as.numeric(x)
    }

    return(list(values = values, vector = TRUE, colnames = NULL))
  }

  x_rows <- rownames(x)
  if (is.null(x_rows)) {
    if (nrow(x) != n) {
      stop(
        sprintf("Unlabelled matrix 'x' must have %d rows; got %d.", n, nrow(x)),
        call. = FALSE
      )
    }
    values <- matrix(
      as.numeric(x),
      nrow = nrow(x),
      ncol = ncol(x),
      dimnames = list(NULL, colnames(x))
    )
  } else {
    validate_pedprod_ids(x_rows, ids, "rownames(x)")
    values <- matrix(
      0,
      nrow = n,
      ncol = ncol(x),
      dimnames = list(NULL, colnames(x))
    )
    values[match(x_rows, ids), ] <- x
  }

  list(values = values, vector = FALSE, colnames = colnames(x))
}

# Validate user-supplied individual identifiers before zero-filling/alignment.
validate_pedprod_ids <- function(input_ids, pedigree_ids, label) {
  if (anyNA(input_ids) || any(input_ids == "")) {
    stop(sprintf("%s must not contain missing or empty IDs.", label), call. = FALSE)
  }
  if (anyDuplicated(input_ids)) {
    stop(sprintf("%s must contain unique individual IDs.", label), call. = FALSE)
  }

  unknown <- setdiff(input_ids, pedigree_ids)
  if (length(unknown) > 0L) {
    stop(
      sprintf(
        "%s contains unknown individual IDs: %s.",
        label,
        paste(utils::head(unknown, 5L), collapse = ", ")
      ),
      call. = FALSE
    )
  }

  invisible(TRUE)
}

test_that("pedprod matches explicit A and Ainv products", {
  tped <- tidyped(small_ped)
  A <- pedmat(tped, method = "A", sparse = FALSE)
  Ainv <- pedmat(tped, method = "Ainv", sparse = FALSE)

  set.seed(20260704)
  x <- stats::rnorm(nrow(tped))
  X <- cbind(First = x, Second = rev(x), Indicator = seq_len(nrow(tped)) %% 2)

  expect_equal(
    unname(pedprod(tped, x, method = "A")),
    unname(drop(A %*% x)),
    tolerance = 1e-12
  )
  expect_equal(
    unname(pedprod(tped, X, method = "A")),
    unname(A %*% X),
    tolerance = 1e-12
  )
  expect_equal(
    unname(pedprod(tped, x, method = "Ainv")),
    unname(drop(Ainv %*% x)),
    tolerance = 1e-12
  )
  expect_equal(
    unname(pedprod(tped, X, method = "Ainv")),
    unname(Ainv %*% X),
    tolerance = 1e-12
  )
})

test_that("pedprod aligns named vectors and matrices and zero-fills omissions", {
  tped <- tidyped(small_ped)
  A <- pedmat(tped, method = "A", sparse = FALSE)

  weights <- c(Z2 = 0.7, A = 0.3)
  full_weights <- setNames(numeric(nrow(tped)), tped$Ind)
  full_weights[names(weights)] <- weights

  result <- pedprod(tped, weights)
  expect_named(result, tped$Ind)
  expect_equal(
    unname(result),
    unname(drop(A %*% full_weights)),
    tolerance = 1e-12
  )

  schemes <- rbind(
    Z2 = c(One = 0.7, Two = 0.2),
    A = c(One = 0.3, Two = 0.8)
  )
  full_schemes <- matrix(
    0,
    nrow = nrow(tped),
    ncol = ncol(schemes),
    dimnames = list(tped$Ind, colnames(schemes))
  )
  full_schemes[rownames(schemes), ] <- schemes

  matrix_result <- pedprod(tped, schemes)
  expect_equal(rownames(matrix_result), tped$Ind)
  expect_equal(colnames(matrix_result), colnames(schemes))
  expect_equal(
    unname(matrix_result),
    unname(A %*% full_schemes),
    tolerance = 1e-12
  )
})

test_that("pedprod handles founders, one known parent, selfing, and disconnected groups", {
  raw <- data.frame(
    Ind = c("F1", "F2", "H1", "S1", "G1", "G2", "H2"),
    Sire = c(NA, NA, "F1", "H1", "S1", NA, "G1"),
    Dam = c(NA, NA, NA, "H1", "F2", NA, "G2"),
    stringsAsFactors = FALSE
  )
  tped <- tidyped(raw, selfing = TRUE)
  A <- pedmat(tped, method = "A", sparse = FALSE)
  Ainv <- pedmat(tped, method = "Ainv", sparse = FALSE)
  x <- seq_len(nrow(tped)) / nrow(tped)

  expect_equal(
    unname(A %*% Ainv),
    diag(nrow(tped)),
    tolerance = 1e-12
  )
  expect_equal(
    unname(pedprod(tped, x)),
    unname(drop(A %*% x)),
    tolerance = 1e-12
  )
  expect_equal(
    unname(pedprod(tped, x, method = "Ainv")),
    unname(drop(Ainv %*% x)),
    tolerance = 1e-12
  )
})

test_that("pedprod accepts raw pedigrees and tidyped objects without numeric columns", {
  raw <- as.data.frame(small_ped)
  tped_without_num <- tidyped(raw, addnum = FALSE)
  x <- rep(1, nrow(tped_without_num))

  raw_result <- pedprod(raw, x)
  tidyped_result <- pedprod(tped_without_num, x)

  expect_equal(raw_result, tidyped_result)
  expect_named(tidyped_result, tidyped(raw)$Ind)
})

test_that("pedprod validates right-hand sides and pedigree structure", {
  tped <- tidyped(small_ped)
  n <- nrow(tped)

  expect_error(pedprod(tped, numeric(n - 1L)), "must have length")
  expect_error(pedprod(tped, matrix(1, nrow = n - 1L)), "must have.*rows")
  expect_error(pedprod(tped, letters[seq_len(n)]), "numeric or logical")
  expect_error(pedprod(tped, replace(numeric(n), 1L, NA_real_)), "finite")
  expect_error(pedprod(tped, replace(numeric(n), 1L, Inf)), "finite")
  expect_error(pedprod(tped, numeric()), "at least one")
  expect_error(pedprod(tped, matrix(numeric(), nrow = n, ncol = 0L)), "at least one")

  unknown <- c(A = 0.5, UNKNOWN = 0.5)
  expect_error(pedprod(tped, unknown), "unknown individual IDs")

  duplicated <- c(A = 0.5, A = 0.5)
  expect_error(pedprod(tped, duplicated), "unique individual IDs")

  missing_name <- c(0.5, 0.5)
  names(missing_name) <- c("A", NA_character_)
  expect_error(pedprod(tped, missing_name), "missing or empty IDs")

  incomplete <- suppressWarnings(tped[-1L])
  expect_error(
    pedprod(incomplete, rep(1, nrow(incomplete))),
    "structurally complete pedigree"
  )

  expect_error(pedprod(splitped(tped), rep(1, n)), "does not support 'splitped'")
  expect_error(pedprod(tped, rep(1, n), method = "D"))
})

test_that("pedprod remains available beyond the dense A size guard", {
  skip_on_cran()

  n <- 25001L
  ids <- paste0("I", seq_len(n))
  raw <- data.frame(
    Ind = ids,
    Sire = c(NA_character_, ids[-n]),
    Dam = NA_character_,
    stringsAsFactors = FALSE
  )
  tped <- suppressMessages(tidyped(raw))

  result <- pedprod(tped, setNames(1, ids[n]))

  expect_length(result, n)
  expect_true(all(is.finite(result)))
  expect_error(pedmat(tped, method = "A"), "too large")
})

#' Calculate inbreeding coefficients
#'
#' \code{inbreed} function calculates the inbreeding coefficients for all individuals in a tidied pedigree.
#'
#' This function takes a pedigree tidied by the \code{\link{tidyped}} function and
#' calculates the inbreeding coefficients using an optimized C++ implementation of
#' the Sargolzaei & Iwaisaki (2005) LAP (Longest Ancestral Path) bucket algorithm.
#' This method is the fastest known direct algorithm for computing all inbreeding
#' coefficients: it replaces the O(\eqn{N^2}) linear scan of Meuwissen & Luo (1992)
#' with O(1) bucket pops and selective ancestor clearing, giving
#' \eqn{O(\sum m_i)} total work where \eqn{m_i} is the number of distinct
#' ancestors of individual \eqn{i}. At \eqn{N = 1{,}000{,}000}, the kernel
#' completes in approximately 0.12 s — over 10\eqn{\times} faster than the previous
#' Meuwissen & Luo (1992) implementation and on par with the \pkg{pedigreemm}
#' reference C implementation of the same algorithm.
#' It is the core engine used by both \code{tidyped(..., inbreed = TRUE)} and
#' \code{pedmat(..., method = "f")}, ensuring consistent results across the package.
#'
#' @param ped A \code{tidyped} object.
#' @param ... Additional arguments (currently ignored).
#' @return A \code{tidyped} object with an additional column \strong{f}.
#' @examples
#' library(visPedigree)
#' data(simple_ped)
#' ped <- tidyped(simple_ped)
#' ped_f <- inbreed(ped)
#' ped_f[f > 0, .(Ind, Sire, Dam, f)]
#' @export
#' @import data.table
inbreed <- function(ped, ...) {
  ped <- ensure_complete_tidyped(ped, "inbreed()")

  meta <- attr(ped, "ped_meta")

  # tidyped guarantees IndNum == seq_len(N) and rows are sorted by IndNum,
  # so no re-sorting is needed before calling the C++ engine.
  # We still copy to avoid mutating the caller's object in place.
  ped_work <- copy(ped)

  res_f <- cpp_calculate_inbreeding(ped_work$SireNum, ped_work$DamNum)

  ped_work[, f := res_f$f]

  # Propagate ped_meta from original object
  result <- new_tidyped(ped_work)
  if (!is.null(meta)) data.table::setattr(result, "ped_meta", meta)

  return(result)
}

test_that("inbreed works correctly", {
  ped <- data.frame(
    Ind = c("A", "B", "C", "D", "E", "F"),
    Sire = c(NA, NA, "A", "A", "C", "C"),
    Dam = c(NA, NA, "B", "B", "D", "D")
  )
  tped <- tidyped(ped)
  
  res <- inbreed(tped)
  
  expect_s3_class(res, "tidyped")
  expect_true("f" %in% names(res))
  
  # A and B are founders, f = 0
  expect_equal(res$f[res$Ind == "A"], 0)
  expect_equal(res$f[res$Ind == "B"], 0)
  
  # C and D are offspring of A and B, f = 0
  expect_equal(res$f[res$Ind == "C"], 0)
  expect_equal(res$f[res$Ind == "D"], 0)
  
  # E and F are offspring of full siblings C and D, f = 0.25
  expect_equal(res$f[res$Ind == "E"], 0.25)
  expect_equal(res$f[res$Ind == "F"], 0.25)
})

test_that("inbreed handles missing parents", {
  ped <- data.frame(
    Ind = c("A", "B", "C", "D"),
    Sire = c(NA, NA, "A", "C"),
    Dam = c(NA, NA, NA, "B")
  )
  tped <- tidyped(ped)
  
  res <- inbreed(tped)
  
  expect_s3_class(res, "tidyped")
  expect_true("f" %in% names(res))
  expect_equal(res$f, c(0, 0, 0, 0))
})

test_that("inbreed validates input", {
  # ensure_tidyped errors on data.frame missing core columns
  expect_error(inbreed(data.frame(Ind="A", Sire=NA, Dam=NA)), "tidyped object")
})

#' Fade colors by appending a reduced alpha value
#'
#' Converts any R color specification to `#RRGGBB4D` form.
#' Handles hex colors (`#RRGGBB`, `#RRGGBBAA`) and named colors (e.g. `"red"`).
#'
#' @param x Character vector of colors.
#' @return Character vector of faded hex colors.
#' @keywords internal
fade_cols <- function(x) {
  vapply(x, function(col) {
    nc <- nchar(col)
    if (nc == 7 && substring(col, 1, 1) == "#") {
      # #RRGGBB -> append 4D
      paste0(col, "4D")
    } else if (nc == 9 && substring(col, 1, 1) == "#") {
      # #RRGGBBAA -> replace AA with 4D
      paste0(substring(col, 1, 7), "4D")
    } else {
      # Named color or other format -> convert via col2rgb
      rgb_vals <- tryCatch(
        grDevices::col2rgb(col),
        error = function(e) return(col)
      )
      if (is.character(rgb_vals)) return(rgb_vals)
      sprintf("#%02X%02X%02X4D", rgb_vals[1], rgb_vals[2], rgb_vals[3])
    }
  }, character(1), USE.NAMES = FALSE)
}

VISPED_COLORS <- c(
  male = "#119ecc",
  female = "#f4b131",
  monoecious = "#26a69a",
  unknown = "#d9d9d9",
  unknown_frame = "#777777",
  family_fill = "#9cb383",
  family_frame = "#5f7650",
  focal_frame = "#9c27b0",
  relative_frame = "#ab47bc"
)

#' Create a vectorized regular-polygon igraph shape
#' @param sides Number of polygon sides.
#' @return An igraph shape plotting function.
#' @keywords internal
#' @noRd
make_visped_polygon <- function(sides) {
  force(sides)

  function(coords, v = NULL, params) {
    get_param <- function(name) {
      value <- params("vertex", name)
      if (length(value) != 1L && !is.null(v)) {
        value <- value[v]
      }
      rep(value, length.out = nrow(coords))
    }

    if (nrow(coords) == 0L) {
      return(invisible(NULL))
    }

    fill <- get_param("color")
    frame <- get_param("frame.color")
    frame_width <- get_param("frame.width")
    size <- get_param("size")
    drawable <- is.finite(size) & size > 0
    if (!any(drawable)) {
      return(invisible(NULL))
    }

    coords <- coords[drawable, , drop = FALSE]
    fill <- fill[drawable]
    frame <- frame[drawable]
    frame_width <- frame_width[drawable]
    size <- size[drawable]
    frame[frame_width <= 0] <- NA_character_
    frame_width[frame_width <= 0] <- 1
    stars <- matrix(rep(size, each = sides), nrow = nrow(coords), byrow = TRUE)

    # A small number of frame widths is typical; batching keeps polygon drawing vectorized.
    width_groups <- split(seq_len(nrow(coords)), frame_width)
    for (idx in width_groups) {
      graphics::symbols(
        x = coords[idx, 1],
        y = coords[idx, 2],
        stars = stars[idx, , drop = FALSE],
        inches = FALSE,
        add = TRUE,
        bg = fill[idx],
        fg = frame[idx],
        lwd = frame_width[idx[1L]]
      )
    }

    invisible(NULL)
  }
}

#' Register visPedigree shapes required by sex-based symbol coding
#' @param needed Character vector containing "diamond" and/or "hexagon".
#' @return Invisibly returns NULL.
#' @keywords internal
#' @noRd
register_visped_shapes <- function(needed) {
  shape_names <- c(
    diamond = "visped_diamond",
    hexagon = "visped_hexagon"
  )
  needed <- intersect(needed, names(shape_names))
  available <- igraph::shapes()

  for (shape in needed) {
    name <- shape_names[[shape]]
    if (name %in% available) {
      next
    }

    sides <- if (shape == "diamond") 4L else 6L
    igraph::add_shape(
      name,
      clip = igraph::shapes("circle")$clip,
      plot = make_visped_polygon(sides)
    )
    available <- c(available, name)
  }

  invisible(NULL)
}

#' Styling and finalizing pedigree graph
#' @import data.table
#' @keywords internal
get_highlight_ids <- function(ped, highlight, trace) {
  focal_ids <- NULL
  relative_ids <- NULL
  trace_edges <- data.table(parent = character(0), child = character(0))
  
  # Default colors
  colors <- list(
    focal_frame = VISPED_COLORS[["focal_frame"]],
    focal_fill = NULL,
    rel_frame = VISPED_COLORS[["relative_frame"]],
    rel_fill = NULL
  )
  
  if (!is.null(highlight)) {
    if (is.list(highlight)) {
      focal_ids <- highlight$ids
      if (!is.null(highlight$frame.color)) colors$focal_frame <- highlight$frame.color
      if (!is.null(highlight$color)) colors$focal_fill <- highlight$color
      if (!is.null(highlight$rel.frame.color)) colors$rel_frame <- highlight$rel.frame.color
      if (!is.null(highlight$rel.color)) colors$rel_fill <- highlight$rel.color
    } else {
      focal_ids <- highlight
    }
    
    all_inds <- ped$Ind
    focal_ids <- intersect(focal_ids, all_inds)
    
    if (!isFALSE(trace) && length(focal_ids) > 0) {
      trace_dir <- if (isTRUE(trace)) "all" else trace
      meta <- attr(ped, "ped_meta")
      selfing_val <- if (!is.null(meta)) isTRUE(meta$selfing) else FALSE
      relatives_ped <- suppressWarnings(tidyped(ped, cand = focal_ids, trace = trace_dir, selfing = selfing_val))
      relative_ids <- setdiff(unique(relatives_ped$Ind), focal_ids)
      
      # Build trace_edges: parent-child pairs that are ON the trace path.
      # For trace="all", we must call tidyped separately for "up" and "down"
      # to avoid cross-path edges (e.g. N is X's ancestor via Q->T->U,
      # but N->Z1/Z2 is NOT on X's trace path even though both N and Z1/Z2
      # are highlighted).
      if (trace_dir == "all") {
        ped_up <- suppressWarnings(tidyped(ped, cand = focal_ids, trace = "up"))
        ped_down <- suppressWarnings(tidyped(ped, cand = focal_ids, trace = "down"))
        trace_edges <- rbind(
          ped_up[!is.na(Sire), .(parent = Sire, child = Ind)],
          ped_up[!is.na(Dam), .(parent = Dam, child = Ind)],
          ped_down[!is.na(Sire), .(parent = Sire, child = Ind)],
          ped_down[!is.na(Dam), .(parent = Dam, child = Ind)]
        )
      } else {
        trace_edges <- rbind(
          relatives_ped[!is.na(Sire), .(parent = Sire, child = Ind)],
          relatives_ped[!is.na(Dam), .(parent = Dam, child = Ind)]
        )
      }
      trace_edges <- unique(trace_edges)
    }
  }
  
  list(focal = focal_ids, relatives = relative_ids, all_ids = c(focal_ids, relative_ids),
       colors = colors, trace_edges = trace_edges)
}

#' Apply node styles (color, shape, highlighting)
#' @import data.table
#' @keywords internal
apply_node_styles <- function(ped_node, highlight_info, shapeby = "sex") {
  shape = frame.color = color = size = label.color = frame.width = label.font = highlighted = NULL
  
  # Default styles (set all at once)
  ped_node[, `:=`(
    shape = "circle", 
    frame.color = VISPED_COLORS[["unknown_frame"]],
    color = VISPED_COLORS[["unknown"]],
    size = 15, 
    label.color = "#0d0312", 
    frame.width = 0.2, 
    label.font = 1, 
    highlighted = FALSE
  )]
  
  ped_node[sex == "male", `:=`(
    frame.color = VISPED_COLORS[["male"]],
    color = VISPED_COLORS[["male"]]
  )]
  ped_node[sex == "female", `:=`(
    frame.color = VISPED_COLORS[["female"]],
    color = VISPED_COLORS[["female"]]
  )]
  ped_node[sex == "monoecious", `:=`(
    frame.color = VISPED_COLORS[["monoecious"]],
    color = VISPED_COLORS[["monoecious"]]
  )]
  ped_node[is.na(sex) | sex == "unknown", `:=`(
    frame.color = VISPED_COLORS[["unknown_frame"]],
    color = VISPED_COLORS[["unknown"]]
  )]

  if (shapeby == "sex") {
    real_nodes <- ped_node$nodetype == "real"
    needed <- character()
    if (any(real_nodes & (is.na(ped_node$sex) | ped_node$sex == "unknown"))) {
      needed <- c(needed, "diamond")
    }
    if (any(real_nodes & ped_node$sex == "monoecious", na.rm = TRUE)) {
      needed <- c(needed, "hexagon")
    }
    register_visped_shapes(needed)

    ped_node[nodetype == "real" & sex == "male", shape := "square"]
    ped_node[nodetype == "real" & (is.na(sex) | sex == "unknown"),
             shape := "visped_diamond"]
    ped_node[nodetype == "real" & sex == "monoecious",
             shape := "visped_hexagon"]
  }

  # Family summaries are not individuals and therefore have no sex.
  ped_node[nodetype == "compact", `:=`(
    shape = "rectangle",
    frame.color = VISPED_COLORS[["family_frame"]],
    color = VISPED_COLORS[["family_fill"]]
  )]
  
  # Virtual nodes: circle with tiny size and transparency to fix edge gaps
  ped_node[nodetype == "virtual", `:=`(
    shape = "circle", 
    label = "", 
    size = 0.001, 
    frame.width = 0, 
    color = "#FFFFFF00", 
    frame.color = "#FFFFFF00"
  )]
  
  # Apply highlighting
  h_ids <- highlight_info$all_ids
  if (length(h_ids) > 0) {
    focal <- highlight_info$focal
    relatives <- highlight_info$relatives
    colors <- highlight_info$colors
    
    # Use binary search / keys for matching if many ids
    if (length(relatives) > 0) {
      ped_node[label %in% relatives & nodetype == "real", highlighted := TRUE]
    }
    if (length(focal) > 0) {
      ped_node[label %in% focal & nodetype == "real", highlighted := TRUE]
    }
    
    h_familynums <- ped_node[highlighted == TRUE, unique(familynum)]
    ped_node[id %in% h_familynums & nodetype == "virtual", highlighted := TRUE]
    
    # Batch update non-highlighted
    ped_node[highlighted == FALSE & nodetype %in% c("real", "compact"), `:=`(
      color = fade_cols(color), 
      frame.color = fade_cols(frame.color), 
      label.color = fade_cols(label.color)
    )]
    
    # Highlighted styles
    if (length(relatives) > 0) {
      ped_node[label %in% relatives & nodetype == "real", `:=`(
        frame.color = colors$rel_frame, 
        frame.width = 0.5,
        color = if (!is.null(colors$rel_fill)) colors$rel_fill else color
      )]
    }
    
    if (length(focal) > 0) {
      ped_node[label %in% focal & nodetype == "real", `:=`(
        frame.color = colors$focal_frame, 
        frame.width = 1, 
        label.font = 2,
        color = if (!is.null(colors$focal_fill)) colors$focal_fill else color
      )]
    }
  }
  return(ped_node[])
}

#' Finalize graph and reindex IDs
#' @import data.table
#' @keywords internal
finalize_graph <- function(ped_node, ped_edge, highlight_info, trace, showf) {
  old_ids <- ped_node$id
  ped_node[, id := seq_len(.N)]
  ped_edge[, from := ped_node$id[match(from, old_ids)]]
  ped_edge[, to := ped_node$id[match(to, old_ids)]]
  
  real_max <- max(ped_node[nodetype %in% c("real", "compact")]$id, na.rm = TRUE)
  
  tonodecolor = i.color = i.highlighted = from_highlighted = role = NULL
  ped_edge[ped_node, ":="(tonodecolor = i.color, to_highlighted = i.highlighted), on = .(to = id)]
  ped_edge[ped_node, from_highlighted := i.highlighted, on = .(from = id)]
  
  h_ids <- highlight_info$all_ids
  has_trace <- !isFALSE(trace) && length(highlight_info$relatives) > 0
  trace_edges <- highlight_info$trace_edges
  
  # Role-based edge coloring for family→parent edges
  ped_edge[role == "sire", color := VISPED_COLORS[["male"]]]
  ped_edge[role == "dam", color := VISPED_COLORS[["female"]]]
  ped_edge[role == "selfing", color := VISPED_COLORS[["monoecious"]]]
  
  # If highlighting is active, fade edges based on endpoint highlight status
  if (length(h_ids) > 0) {
    # Fade if target parent is not highlighted
    ped_edge[role %in% c("sire", "dam", "selfing") & to_highlighted == FALSE,
             color := fade_cols(color)]
    # Further fade if source family node is not highlighted
    ped_edge[role %in% c("sire", "dam", "selfing") & from_highlighted == FALSE,
             color := fade_cols(color)]
  }
  
  if (length(h_ids) > 0 && has_trace) {
    # Use trace_edges to precisely control which family->parent edges are highlighted.
    # A family->parent edge should only be highlighted if there exists at least one
    # child of that family whose parent-child relationship is on the trace path.
    if (!is.null(trace_edges) && nrow(trace_edges) > 0) {
      # Get the label of the 'to' node (the parent)
      ped_edge[ped_node, to_label := i.label, on = .(to = id)]
      # Get the label of the 'from' node (the family node's label = familylabel)
      ped_edge[ped_node, from_label := i.label, on = .(from = id)]
      
      # For each individual->family edge (from <= real_max), record child-family mapping
      child_to_family <- ped_edge[from <= real_max, .(child_id = from, family_id = to)]
      child_to_family[ped_node, child_label := i.label, on = .(child_id = id)]
      
      # For family->parent edges that are currently highlighted (from_highlighted == TRUE),
      # check if the parent-child relationship is actually on the trace path
      fam_parent_idx <- which(ped_edge$from > real_max & ped_edge$from_highlighted == TRUE)
      
      for (i in fam_parent_idx) {
        fam_id <- ped_edge$from[i]
        parent_label <- ped_edge$to_label[i]
        
        # Find children connected to this family node
        children_labels <- child_to_family[family_id == fam_id, child_label]
        
        # Check if any child-parent pair is in trace_edges
        on_path <- any(trace_edges[parent == parent_label & child %in% children_labels, .N] > 0)
        
        if (!on_path) {
          # This edge is NOT on the trace path - fade it
          ped_edge[i, color := fade_cols(ped_edge$tonodecolor[i])]
        }
      }
      
      # Clean up temporary columns
      ped_edge[, c("to_label", "from_label") := NULL]
    }
    
    # For individual->family edges: highlight only if this individual
    # appears as a child in trace_edges (i.e., the upward connection
    # to its parents is on the trace path).
    ped_edge[ped_node, from_label := i.label, on = .(from = id)]
    ped_edge[
      from <= real_max & from_highlighted == TRUE,
      color := fifelse(from_label %in% trace_edges$child, "#333333", "#3333334D")
    ]
    ped_edge[, from_label := NULL]
  } else if (length(h_ids) == 0) {
    # No highlighting: edges already follow parent node color
  }
  # When highlighting without trace (has_trace == FALSE), keep individual->family edges faded
  
  ped_edge[from <= real_max, ":="(curved = 0, arrow.size = 0, arrow.width = 0, arrow.mode = 0)]
  
  new_names_edge <- c(c("from", "to"), setdiff(colnames(ped_edge), c("from", "to", "tonodecolor", "role")))
  ped_edge <- ped_edge[, ..new_names_edge][order(from, to)]
  
  new_names_node <- c("id", setdiff(colnames(ped_node), "id"))
  ped_node <- ped_node[, ..new_names_node][order(layer, id)]
  
  # Ensure Ind column exists for layout matching (clean label before modification)
  ped_node[, Ind := label]

  if ("DisplayLabel" %in% colnames(ped_node)) {
    display_label <- ped_node[["DisplayLabel"]]
    display_idx <- ped_node[["nodetype"]] == "real" & !is.na(display_label) & display_label != ""
    set(ped_node, i = which(display_idx), j = "label", value = display_label[display_idx])
  }
  
  if (showf && "f" %in% colnames(ped_node)) {
    ped_node[nodetype %in% c("real", "compact") & !is.na(f) & f > 0, 
             label := paste0(label, "\n[", round(f, 4), "]")]
  }
  list(node = ped_node, edge = ped_edge)
}

library(testthat)
library(data.table)
library(visPedigree)

make_sex_shape_ped <- function() {
  tidyped(
    data.table(
      Ind = c("M", "F", "H", "U", "HC"),
      Sire = c(NA, NA, NA, "M", "H"),
      Dam = c(NA, NA, NA, "F", "H")
    ),
    selfing = TRUE
  )
}

test_that("visped works with basic tidy input", {
  tidy_ped <- tidyped(simple_ped, addgen = TRUE, addnum = TRUE)
  # Check return structure
  res <- visped(tidy_ped, showgraph = FALSE, file = tempfile())
  expect_type(res, "list")
  expect_named(res, c("g", "layout", "canvas_width", "canvas_height", "node_size", "best_cex"))
  expect_s3_class(res$g, "igraph")
})

test_that("visped handles raw input by auto-tidying", {
  # Raw data frame
  res <- visped(simple_ped, showgraph = FALSE, file = tempfile())
  expect_s3_class(res$g, "igraph")
  
  # Check dimensions - simple_ped has 52 individuals in example
  expect_gt(igraph::vcount(res$g), 0)
})

test_that("visped writes SVG output when requested", {
  tidy_ped <- tidyped(simple_ped, addgen = TRUE, addnum = TRUE)
  svg_file <- tempfile(fileext = ".svg")

  res <- visped(tidy_ped, showgraph = FALSE, file = svg_file)

  expect_s3_class(res$g, "igraph")
  expect_true(file.exists(svg_file))
  expect_gt(file.info(svg_file)$size, 0)
})

test_that("visped supports custom generation labels in displayed-layer order", {
  tidy_ped <- tidyped(simple_ped, addgen = TRUE, addnum = TRUE)
  default_info <- visPedigree:::prepare_visped_geninfo(
    data.table(gen = c(1L, 3L, 5L), y = c(0.1, 0.5, 0.9))
  )
  custom_info <- visPedigree:::prepare_visped_geninfo(
    data.table(gen = c(1L, 3L, 5L), y = c(0.1, 0.5, 0.9)),
    c("Founders", "Cycle 1", "Current")
  )

  expect_equal(default_info$label, c("G1", "G3", "G5"))
  expect_equal(custom_info$label, c("Founders", "Cycle 1", "Current"))

  custom_labels <- paste0("Layer ", seq_len(length(unique(tidy_ped$Gen))))
  expect_silent(
    suppressMessages(visped(
      tidy_ped,
      genlab = custom_labels,
      showgraph = FALSE,
      file = tempfile(fileext = ".pdf")
    ))
  )
})

test_that("visped validates custom generation labels", {
  tidy_ped <- tidyped(simple_ped, addgen = TRUE, addnum = TRUE)

  expect_error(
    visped(tidy_ped, genlab = c("one", "two"), showgraph = FALSE, file = tempfile()),
    "one label for each displayed generation"
  )
  expect_error(
    visped(tidy_ped, genlab = c("1" = "Founders"), showgraph = FALSE, file = tempfile()),
    "unnamed character vector"
  )
  expect_error(
    visped(tidy_ped, genlab = character(), showgraph = FALSE, file = tempfile()),
    "TRUE, FALSE, or a non-empty unnamed character vector"
  )
  expect_error(
    visped(tidy_ped, genlab = 1, showgraph = FALSE, file = tempfile()),
    "TRUE, FALSE, or a non-empty unnamed character vector"
  )
})

test_that("visped reports the effective generation-label cex", {
  tidy_ped <- tidyped(simple_ped, addgen = TRUE, addnum = TRUE)

  expect_message(
    visped(
      tidy_ped,
      genlab = TRUE,
      genlabcex = 1.2,
      showgraph = FALSE,
      file = tempfile(fileext = ".pdf")
    ),
    "Generation label cex: 1.2"
  )
})

test_that("visped supports custom individual labels", {
  tidy_ped <- tidyped(simple_ped, addgen = TRUE, addnum = TRUE)
  tidy_ped[, DisplayID := paste0("node_", seq_len(.N))]

  res <- visped(tidy_ped, labelvar = "DisplayID", showgraph = FALSE, file = tempfile())
  vertices <- as.data.table(igraph::as_data_frame(res$g, what = "vertices"))
  real_nodes <- vertices[nodetype == "real"]

  expected <- tidy_ped$DisplayID[match(real_nodes$Ind, tidy_ped$Ind)]
  expect_equal(real_nodes$label, expected)
})

test_that("visped accepts row-aligned labels without losing ID alignment", {
  raw_ped <- copy(as.data.table(simple_ped))
  raw_ped <- raw_ped[rev(seq_len(nrow(raw_ped)))]
  original_names <- names(raw_ped)
  raw_ids <- as.character(raw_ped[[1L]])
  display_labels <- paste0("node_", raw_ids)
  display_labels[1:2] <- c(NA_character_, "")
  focal <- "J5X804"

  res <- visped(
    raw_ped,
    labelvar = display_labels,
    highlight = focal,
    trace = "all",
    showgraph = FALSE,
    file = tempfile()
  )
  vertices <- as.data.table(igraph::as_data_frame(res$g, what = "vertices"))
  real_nodes <- vertices[nodetype == "real"]
  expected <- display_labels[match(real_nodes$Ind, raw_ids)]
  expected[is.na(expected) | expected == ""] <-
    real_nodes$Ind[is.na(expected) | expected == ""]

  expect_equal(real_nodes$label, expected)
  expect_true(real_nodes[Ind == focal, highlighted])
  expect_identical(names(raw_ped), original_names)
})

test_that("visped rejects missing custom label columns", {
  tidy_ped <- tidyped(simple_ped, addgen = TRUE, addnum = TRUE)

  expect_error(
    visped(tidy_ped, labelvar = "MissingLabel", showgraph = FALSE, file = tempfile()),
    "specified by 'labelvar' was not found"
  )
  expect_error(
    visped(tidy_ped, labelvar = c("one", "two")),
    "one label per pedigree row"
  )
  expect_error(
    visped(tidy_ped, labelvar = seq_len(nrow(tidy_ped))),
    "must be NULL"
  )
})

test_that("visped parameter 'compact' works", {
  # Use a specific family known to have > 2 siblings
  # Sire="ZZ5", Dam="Z86" from big_family_size_ped
  target_sire <- "ZZ5"
  target_dam <- "Z86"
  
  # Select the family + parents
  raw_ped <- as.data.table(big_family_size_ped)
  family_inds <- raw_ped[Sire == target_sire & Dam == target_dam, Ind]
  
  # Prepare subset
  subset_ped <- raw_ped[Ind %in% c(target_sire, target_dam, family_inds)]
  
  tidy_fam <- tidyped(subset_ped, addgen = TRUE, addnum = TRUE)
  
  # Without compact
  res_full <- visped(tidy_fam, compact = FALSE, showgraph = FALSE, file = tempfile())
  
  # With compact
  res_compact <- visped(
    tidy_fam,
    compact = TRUE,
    labelvar = paste0("node_", tidy_fam$Ind),
    showgraph = FALSE,
    file = tempfile()
  )
  
  # Check if "compact" nodes exist
  node_types <- igraph::V(res_compact$g)$nodetype
  expect_true("compact" %in% node_types)
  compact_labels <- igraph::V(res_compact$g)[nodetype == "compact"]$label
  real_labels <- igraph::V(res_compact$g)[nodetype == "real"]$label
  expect_true(all(grepl("^FS\u00d7[0-9]+$", compact_labels)))
  expect_true(all(grepl("^node_", real_labels)))
  expect_true(all(igraph::V(res_compact$g)[nodetype == "compact"]$shape == "rectangle"))
  expect_equal(
    igraph::V(res_compact$g)[nodetype == "compact"]$size2,
    igraph::V(res_compact$g)[nodetype == "compact"]$size
  )
  expect_true(all(
    igraph::V(res_compact$g)[nodetype == "compact"]$color == "#9cb383"
  ))
  expect_true(all(
    igraph::V(res_compact$g)[nodetype == "compact"]$frame.color == "#5f7650"
  ))
  
  # Compact graph should have fewer nodes (real id nodes replaced by one compact node)
  expect_lt(igraph::vcount(res_compact$g), igraph::vcount(res_full$g))
})

test_that("visped shapeby sex is the default and preserves sex colors", {
  tidy_ped <- make_sex_shape_ped()
  res_default <- visped(tidy_ped, showgraph = FALSE, file = tempfile())
  res_sex <- visped(
    tidy_ped,
    shapeby = "sex",
    showgraph = FALSE,
    file = tempfile()
  )

  default_nodes <- as.data.table(igraph::as_data_frame(res_default$g, what = "vertices"))
  sex_nodes <- as.data.table(igraph::as_data_frame(res_sex$g, what = "vertices"))
  expect_equal(default_nodes$shape, sex_nodes$shape)
  expect_equal(sex_nodes[Ind == "F", shape], "circle")
  expect_equal(sex_nodes[Ind == "M", shape], "square")
  expect_equal(sex_nodes[Ind == "U", shape], "visped_diamond")
  expect_equal(sex_nodes[Ind == "H", shape], "visped_hexagon")
  expect_equal(sex_nodes[Ind == "M", color], "#119ecc")
  expect_equal(sex_nodes[Ind == "F", color], "#f4b131")
  expect_equal(sex_nodes[Ind == "H", color], "#26a69a")
  expect_equal(sex_nodes[Ind == "U", color], "#d9d9d9")
  expect_equal(sex_nodes[Ind == "U", frame.color], "#777777")
})

test_that("visped shapeby role keeps individual records as circles", {
  tidy_ped <- make_sex_shape_ped()
  res <- visped(
    tidy_ped,
    shapeby = "role",
    showgraph = FALSE,
    file = tempfile()
  )
  nodes <- as.data.table(igraph::as_data_frame(res$g, what = "vertices"))

  expect_true(all(nodes[nodetype == "real", shape] == "circle"))
  expect_equal(nodes[Ind == "M", color], "#119ecc")
  expect_equal(nodes[Ind == "F", color], "#f4b131")
  expect_equal(nodes[Ind == "H", color], "#26a69a")
  expect_equal(nodes[Ind == "U", color], "#d9d9d9")
  expect_equal(nodes[Ind == "U", frame.color], "#777777")
})

test_that("visped shapeby sex maps all supported sex classes", {
  tidy_ped <- make_sex_shape_ped()
  res <- visped(
    tidy_ped,
    shapeby = "sex",
    showgraph = FALSE,
    file = tempfile()
  )
  nodes <- as.data.table(igraph::as_data_frame(res$g, what = "vertices"))

  expect_equal(nodes[Ind == "F", shape], "circle")
  expect_equal(nodes[Ind == "M", shape], "square")
  expect_equal(nodes[Ind == "U", shape], "visped_diamond")
  expect_equal(nodes[Ind == "H", shape], "visped_hexagon")
  expect_true(all(c("visped_diamond", "visped_hexagon") %in% igraph::shapes()))
  expect_silent(visPedigree:::register_visped_shapes(c("diamond", "hexagon")))
  expect_error(
    visped(tidy_ped, shapeby = "unsupported"),
    "'arg' should be one of"
  )
})

test_that("custom shapes skip zero-size edge-only rendering pass", {
  zero_params <- function(type, name) {
    switch(
      name,
      color = "#000000",
      frame.color = "#000000",
      frame.width = 0,
      size = 0
    )
  }
  coords <- matrix(c(0.5, 0.5), nrow = 1L)

  expect_silent(visPedigree:::make_visped_polygon(4L)(coords, params = zero_params))
  expect_silent(visPedigree:::make_visped_polygon(6L)(coords, params = zero_params))
})

test_that("compact labels cannot be confused with numeric individual IDs", {
  sibling_ids <- paste0("S", seq_len(10L))
  raw_ped <- data.table(
    Ind = c("10", "F", "P1", "P2", sibling_ids, "C"),
    Sire = c(NA, NA, NA, NA, rep("P1", 10L), "10"),
    Dam = c(NA, NA, NA, NA, rep("P2", 10L), "F")
  )
  res <- visped(
    tidyped(raw_ped),
    compact = TRUE,
    showgraph = FALSE,
    file = tempfile()
  )
  nodes <- as.data.table(igraph::as_data_frame(res$g, what = "vertices"))

  expect_true("10" %in% nodes[nodetype == "real", label])
  expect_true("FS\u00d710" %in% nodes[nodetype == "compact", label])
  expect_equal(nodes[label == "FS\u00d710", shape], "rectangle")
})

test_that("highlighted and parent individuals are not compacted", {
  raw_ped <- data.table(
    Ind = c("P1", "P2", "A", "B", "C", "D"),
    Sire = c(NA, NA, "P1", "P1", "P1", "A"),
    Dam = c(NA, NA, "P2", "P2", "P2", "B")
  )
  tidy_ped <- tidyped(raw_ped)
  res <- visped(
    tidy_ped,
    compact = TRUE,
    highlight = "C",
    showgraph = FALSE,
    file = tempfile()
  )
  nodes <- as.data.table(igraph::as_data_frame(res$g, what = "vertices"))

  expect_true(all(c("A", "B", "C") %in% nodes[nodetype == "real", Ind]))
  expect_false(any(nodes[nodetype == "compact", label] == "FS\u00d73"))
})

test_that("sex-based custom shapes render to PDF and SVG", {
  tidy_ped <- make_sex_shape_ped()
  for (extension in c(".pdf", ".svg")) {
    output <- tempfile(fileext = extension)
    expect_silent(
      suppressMessages(visped(
        tidy_ped,
        shapeby = "sex",
        showgraph = FALSE,
        file = output
      ))
    )
    expect_true(file.exists(output))
    expect_gt(file.info(output)$size, 0)
  }
})

test_that("visped parameter 'outline' works", {
  tidy_ped <- tidyped(simple_ped)
  res_outline <- visped(tidy_ped, outline = TRUE, showgraph = FALSE, file = tempfile())
  
  # In outline mode, node sizes are very small
  sizes <- igraph::V(res_outline$g)$size
  # Non-highlighted nodes get size 0.0001
  expect_true(any(sizes <= 0.001))
})

test_that("visped parameter 'cex' works", {
  tidy_ped <- tidyped(simple_ped)
  
  # User specified cex
  my_cex <- 0.75
  res <- visped(tidy_ped, cex = my_cex, showgraph = FALSE, file = tempfile())
  
  # Check label.cex attribute
  real_nodes <- igraph::V(res$g)[nodetype == "real"]
  expect_true(all(real_nodes$label.cex == my_cex))
})

test_that("visped parameter 'showf' displays inbreeding coefficients", {
  # Force add f column if not present or calc it
  tidy_f <- tidyped(simple_ped, inbreed = TRUE)
  
  # Check functionality without conditional dependency if tidyped handles it
  # If tidyped uses nadiv internally, we rely on it being available
  if ("f" %in% colnames(tidy_f)) {
      res <- visped(tidy_f, showf = TRUE, showgraph = FALSE, file = tempfile())
      labels <- igraph::V(res$g)$label
      # Should contain brackets like "[0.003]"
      expect_true(any(grepl("\\[", labels)))
  }
  
  # Test automatic calculation when f is missing
  tidy_no_f <- copy(tidy_f)
  tidy_no_f[, f := NULL]

  expect_message(
    res_auto <- visped(
      tidy_no_f,
      showf = TRUE,
      showgraph = FALSE,
      file = tempfile()
    ),
    "Calculated inbreeding coefficients automatically"
  )

  labels_auto <- igraph::V(res_auto$g)$label
  expect_true(any(grepl("\\[", labels_auto)))
})

test_that("visped parameter 'showf' warns on incomplete pedigrees", {
  tp_full <- tidyped(simple_ped)
  tp_bad <- suppressWarnings(tp_full[Gen > 2])

  expect_false(is_tidyped(tp_bad))
  expect_false("f" %in% names(tp_bad))

  warn_msgs <- character(0)
  res_bad <- withCallingHandlers(
    visped(
      tp_bad,
      showf = TRUE,
      showgraph = FALSE,
      file = tempfile()
    ),
    warning = function(w) {
      warn_msgs <<- c(warn_msgs, conditionMessage(w))
      invokeRestart("muffleWarning")
    }
  )

  expect_true(any(grepl(
    "cannot be computed automatically because the pedigree is structurally incomplete",
    warn_msgs,
    fixed = TRUE
  )))

  labels_bad <- igraph::V(res_bad$g)$label
  expect_false(any(grepl("\\[", labels_bad)))
})

test_that("visped parameter 'highlight' works with vector", {
  tidy_ped <- tidyped(simple_ped)
  target <- "J5X804"
  
  res <- visped(tidy_ped, highlight = target, showgraph = FALSE, file = tempfile())
  
  hl_nodes <- igraph::V(res$g)[highlighted]
  expect_gt(length(hl_nodes), 0)
  expect_true(target %in% hl_nodes$label)
})

test_that("visped parameter 'highlight' works with list", {
  tidy_ped <- tidyped(simple_ped)
  hl_list <- list(ids = c("J5X804", "J3Y620"), frame.color = "red", color = "yellow")
  
  res <- visped(tidy_ped, highlight = hl_list, showgraph = FALSE, file = tempfile())
  
  hl_nodes <- igraph::V(res$g)[highlighted]
  expect_gt(length(hl_nodes), 0)
})


test_that("visped parameter 'trace' works", {
  tidy_ped <- tidyped(simple_ped)
  target <- "J5X804"
  
  # Trace up
  res_up <- visped(tidy_ped, highlight = target, trace = "up", showgraph = FALSE, file = tempfile())
  
  # Trace down
  res_down <- visped(tidy_ped, highlight = target, trace = "down", showgraph = FALSE, file = tempfile())
  
  # The set of highlighted nodes should differ for this individual (who has parents and offspring)
  h_up <- igraph::V(res_up$g)[highlighted]$label
  h_down <- igraph::V(res_down$g)[highlighted]$label
  
  expect_false(setequal(h_up, h_down))
})

test_that("visped parameter 'file' generates PDF", {
  tidy_ped <- tidyped(simple_ped)
  tmp <- tempfile(fileext = ".pdf")
  
  expect_message(visped(tidy_ped, file = tmp), "Pedigree saved to")
  expect_true(file.exists(tmp))
  unlink(tmp)
})

test_that("visped filters isolated individuals", {
  # Manually construct isolated individual cleanly
  ped_iso <- data.table(
    Ind = c("A", "B", "C", "ISO1"),
    Sire = c(NA_character_, "A", "A", NA_character_),
    Dam = c(NA_character_, NA_character_, NA_character_, NA_character_),
    Sex = c("male", "male", "female", "male")
  )
  
  expect_message(res <- visped(ped_iso, showgraph = FALSE, file = tempfile()), "Removed 1 isolated individuals")
  
  # Check graph does not contain ISO1
  expect_false("ISO1" %in% igraph::V(res$g)$label)
})

test_that("visped parameter 'showgraph' controls plotting", {
    # It's hard to test side effects (plotting) but we can ensure showgraph=FALSE returns silently
    tidy_ped <- tidyped(simple_ped)
    expect_error(visped(tidy_ped, showgraph = FALSE), "Both 'showgraph' and 'file' are disabled")
})

test_that("visped parameter 'pagewidth' works", {
  tidy_ped <- tidyped(simple_ped)
  
  # Default case (should handle standard width)
  res_default <- visped(tidy_ped, showgraph = FALSE, file = tempfile())
  
  # Custom width limit
  # Force a very small width limit to ensure it constrains the output
  small_limit <- 15
  res_small <- visped(tidy_ped, pagewidth = small_limit, showgraph = FALSE, file = tempfile())
  
  expect_lte(res_small$canvas_width, small_limit)
  expect_lte(res_small$canvas_height, small_limit)
  
  # Ensure it doesn't crash on large limits
  res_large <- visped(tidy_ped, pagewidth = 1000, showgraph = FALSE, file = tempfile())
  expect_true(res_large$canvas_width <= 1000)
})

test_that("visped validates new parameters", {
  tidy_ped <- tidyped(simple_ped)
  
  # Validate pagewidth
  expect_error(visped(tidy_ped, pagewidth = -10), "'pagewidth' must be a single positive number")
  expect_error(visped(tidy_ped, pagewidth = "invalid"), "'pagewidth' must be a single positive number")
  
  # Validate symbolsize
  expect_error(visped(tidy_ped, symbolsize = 0), "'symbolsize' must be a single positive number")
  expect_error(visped(tidy_ped, symbolsize = "big"), "'symbolsize' must be a single positive number")
  
  # Validate maxiter
  expect_error(visped(tidy_ped, maxiter = -100), "'maxiter' must be a single positive integer")
  expect_error(visped(tidy_ped, maxiter = "many"), "'maxiter' must be a single positive integer")
  
  # Validate compact
  expect_error(visped(tidy_ped, compact = "yes"), "'compact' must be TRUE or FALSE")
  expect_error(visped(tidy_ped, compact = NA), "'compact' must be TRUE or FALSE")
  expect_error(visped(tidy_ped, compact = NULL), "'compact' must be TRUE or FALSE")
  
  # Validate outline
  expect_error(visped(tidy_ped, outline = 1), "'outline' must be TRUE or FALSE")
  
  # Validate showgraph
  expect_error(visped(tidy_ped, showgraph = "no"), "'showgraph' must be TRUE or FALSE")
  
  # Validate showf
  expect_error(visped(tidy_ped, showf = NA), "'showf' must be TRUE or FALSE")
  
  # Validate cex
  expect_error(visped(tidy_ped, cex = -1), "'cex' must be NULL or a single positive number")
  expect_error(visped(tidy_ped, cex = "small"), "'cex' must be NULL or a single positive number")
  
  # Validate file
  expect_error(visped(tidy_ped, file = TRUE), "'file' must be NULL or a single character string")
  expect_error(visped(tidy_ped, file = NA_character_), "'file' must be NULL or a single character string")

  # Validate highlight
  expect_error(visped(tidy_ped, highlight = 123), "'highlight' must be NULL, a character vector, or a list")

  # Sanitization test
  # Should not crash with NA or empty string
  expect_no_error(suppressMessages(visped(tidy_ped, highlight = c("J5X804", NA, ""), showgraph = FALSE, file = tempfile())))
  expect_no_error(suppressMessages(visped(tidy_ped, highlight = list(ids = c("J5X804", NA, "")), showgraph = FALSE, file = tempfile())))
  
  # Validate trace
  expect_error(visped(tidy_ped, trace = "left"), "'trace' must be TRUE, FALSE, 'up', 'down', or 'all'")
  expect_error(visped(tidy_ped, trace = 1), "'trace' must be TRUE, FALSE, 'up', 'down', or 'all'")

  # Validate genlabcex
  expect_error(visped(tidy_ped, genlabcex = 0), "'genlabcex' must be NULL or a single positive number")
  expect_error(visped(tidy_ped, genlabcex = "large"), "'genlabcex' must be NULL or a single positive number")
})

test_that("visped parameter 'genlabcex' works independently of label cex", {
  tidy_ped <- tidyped(simple_ped)
  tmp <- tempfile(fileext = ".pdf")

  expect_no_error(
    visped(
      tidy_ped,
      showgraph = FALSE,
      file = tmp,
      genlab = TRUE,
      genlabcex = 1.4,
      cex = 0.4
    )
  )

  expect_true(file.exists(tmp))
  unlink(tmp)
})

# --- Internal: dgeMatrix fast-coercion cache ---
# Wrap a base matrix as dgeMatrix without any N² symmetry/sparsity scan.
# Used by the sparse=TRUE path of pedmat() when dense storage is appropriate.
# inherits(result, "Matrix") == TRUE; as.matrix() round-trips correctly.
# Class definition is cached on first call to avoid repeated namespace lookups.
.dgeMatrix_cache <- new.env(parent = emptyenv())
.to_dgeMatrix <- function(mat) {
  if (is.null(.dgeMatrix_cache$cls))
    .dgeMatrix_cache$cls <- methods::getClass("dgeMatrix", where = asNamespace("Matrix"))
  dn <- dimnames(mat)
  if (is.null(dn)) dn <- list(NULL, NULL)
  methods::new(.dgeMatrix_cache$cls, x = as.double(mat),
               Dim = as.integer(dim(mat)), Dimnames = dn)
}

#' Genetic Relationship Matrices and Inbreeding Coefficients
#'
#' @description
#' Optimized calculation of additive (A), dominance (D), epistatic (AA) 
#' relationship matrices, their inverses, and inbreeding coefficients (f).
#' Uses Rcpp with Meuwissen & Luo (1992) algorithm for efficient computation.
#'
#' @param ped A tidied pedigree from \code{\link{tidyped}}. Must be a single
#'   pedigree, not a splitped object. For splitped results, use 
#'   \code{pedmat(ped_split$GP1, ...)} to process individual groups.
#' @param method Character, one of:
#' \itemize{
#'   \item \code{"A"}: Additive (numerator) relationship matrix (default)
#'   \item \code{"f"}: Inbreeding coefficients (returns named vector)
#'   \item \code{"Ainv"}: Inverse of A using Henderson's rules (O(n) complexity)
#'   \item \code{"D"}: Dominance relationship matrix
#'   \item \code{"Dinv"}: Inverse of D (requires matrix inversion)
#'   \item \code{"AA"}: Additive-by-additive epistatic matrix (A # A)
#'   \item \code{"AAinv"}: Inverse of AA
#' }
#' @param sparse Logical, if \code{TRUE} returns sparse Matrix (recommended for
#'   large pedigrees). Default is \code{TRUE}.
#' @param invert_method Character, method for matrix inversion (Dinv/AAinv only):
#' \itemize{
#'   \item \code{"auto"}: Auto-detect and use optimal method (default)
#'   \item \code{"sympd"}: Force Cholesky decomposition (faster for SPD matrices)
#'   \item \code{"general"}: Force general LU decomposition
#' }
#' @param threads Integer. Number of OpenMP threads to use. Use 0 to keep
#'   the system/default setting. Currently, multi-threading is explicitly implemented for:
#' \itemize{
#'   \item \code{"D"}: Dominance relationship matrix (significant speedup).
#'   \item \code{"Ainv"}: Inverse of A (only for large pedigrees, n >= 5000).
#' }
#' For \code{"Dinv"}, \code{"AA"}, and \code{"AAinv"}, parallelism depends on the 
#' linked BLAS/LAPACK library (e.g., OpenBLAS, MKL, Accelerate) and is not 
#' controlled by this parameter. Methods \code{"A"} and \code{"f"} are single-threaded.
#' @param compact Logical, if \code{TRUE} compacts full-sibling families by
#'   selecting one representative per family. This dramatically reduces matrix
#'   dimensions for pedigrees with large full-sib groups. See Details.
#' 
#' @details 
#' \strong{API Design:}
#' 
#' Only a single method may be requested per call. This design prevents
#' accidental heavy computations. If multiple matrices are needed, call
#' \code{pedmat()} separately for each method.
#'
#' \strong{Compact Mode (\code{compact = TRUE}):}
#' 
#' Full-siblings share identical relationships with all other individuals.
#' Compact mode exploits this by selecting one representative per full-sib
#' family, dramatically reducing matrix size. For example, a pedigree with
#' 170,000 individuals might compact to 1,800 unique relationship patterns.
#' 
#' Key features:
#' \itemize{
#'   \item \code{\link{query_relationship}(x, id1, id2)}: Query any individual
#'         pair, including merged siblings (automatic lookup)
#'   \item \code{\link{expand_pedmat}(x)}: Restore full matrix dimensions
#'   \item \code{\link{vismat}(x)}: Visualize directly (auto-expands compact)
#' }
#' 
#' \strong{Performance Notes:}
#' \itemize{
#'   \item \strong{Ainv}: O(n) complexity using Henderson's rules. Fast for any size.
#'   \item \strong{Dinv/AAinv}: O(n³) matrix inversion. Practical limits:
#'     \itemize{
#'       \item n < 500: ~10-20 ms
#'       \item n = 1,000: ~40-60 ms  
#'       \item n = 2,000: ~130-150 ms
#'       \item n > 2,000: Consider using \code{compact = TRUE}
#'     }
#'   \item \strong{Memory}: Sparse matrices use ~O(nnz) memory; dense use O(n²)
#' }
#'
#' @return Returns a matrix or vector with S3 class \code{"pedmat"}.
#' 
#' \strong{Object type by method:}
#' \itemize{
#'   \item \code{method="f"}: Named numeric vector of inbreeding coefficients
#'   \item All other methods: Sparse or dense matrix (depending on \code{sparse})
#' }
#' 
#' \strong{S3 Methods:}
#' \itemize{
#'   \item \code{print(x)}: Display matrix with metadata header
#'   \item \code{\link{summary_pedmat}(x)}: Detailed statistics (size, compression, mean, density)
#'   \item \code{dim(x)}, \code{length(x)}, \code{Matrix::diag(x)}, \code{t(x)}: Standard operations
#'   \item \code{x[i, j]}: Subsetting (behaves like underlying matrix)
#'   \item \code{as.matrix(x)}: Convert to base matrix
#' }
#' 
#' \strong{Accessing Metadata (use \code{attr()}, not \code{$}):}
#' \itemize{
#'   \item \code{attr(x, "ped")}: The pedigree used (or compact pedigree if \code{compact=TRUE})
#'   \item \code{attr(x, "ped_compact")}: Compact pedigree (when \code{compact=TRUE})
#'   \item \code{attr(x, "method")}: Calculation method used
#'   \item \code{attr(x, "call_info")}: Full calculation metadata (timing, sizes, etc.)
#'   \item \code{names(attributes(x))}: List all available attributes
#' }
#' 
#' \strong{Additional attributes when \code{compact = TRUE}:}
#' \itemize{
#'   \item \code{attr(x, "compact_map")}: data.table mapping individuals to representatives
#'   \item \code{attr(x, "family_summary")}: data.table summarizing merged families
#'   \item \code{attr(x, "compact_stats")}: Compression statistics (ratio, n_families, etc.)
#' }
#' 
#' @seealso 
#' \code{\link{tidyped}} for preparing pedigree data,
#' \code{\link{pedprod}} for matrix-free products with A or Ainv,
#' \code{\link{query_relationship}} for querying individual pairs,
#' \code{\link{expand_pedmat}} for restoring full dimensions,
#' \code{\link{vismat}} for visualization,
#' \code{\link{inbreed}} for simple inbreeding calculation
#' 
#' @examples
#' # Basic usage with small pedigree
#' library(visPedigree)
#' tped <- tidyped(small_ped)
#' 
#' # --- Additive Relationship Matrix (default) ---
#' A <- pedmat(tped)
#' A["A", "B"]      # Relationship between A and B
#' Matrix::diag(A)  # Diagonal = 1 + F (inbreeding)
#' 
#' # --- Inbreeding Coefficients ---
#' f <- pedmat(tped, method = "f")
#' f["Z1"]  # Inbreeding of individual Z1
#' 
#' # --- Using summary_pedmat() ---
#' summary_pedmat(A)   # Detailed matrix statistics
#' 
#' # --- Accessing Metadata ---
#' attr(A, "ped")              # Original pedigree
#' attr(A, "method")           # "A"
#' names(attributes(A))        # All available attributes
#' 
#' # --- Compact Mode (for large full-sib families) ---
#' A_compact <- pedmat(tped, method = "A", compact = TRUE)
#' 
#' # Query relationships (works for any individual, including merged sibs)
#' query_relationship(A_compact, "Z1", "Z2")  # Full-sibs Z1 and Z2
#' 
#' # View compression statistics
#' attr(A_compact, "compact_stats")
#' attr(A_compact, "family_summary")
#' 
#' # Expand back to full size
#' A_full <- expand_pedmat(A_compact)
#' dim(A_full)  # Original dimensions restored
#' 
#' # --- Inverse Matrices ---
#' Ainv <- pedmat(tped, method = "Ainv")  # Henderson's rules (fast)
#' 
#' # --- Dominance and Epistatic ---
#' D <- pedmat(tped, method = "D")
#' AA <- pedmat(tped, method = "AA")
#' 
#' # --- Visualization (requires display device) ---
#' \donttest{
#' vismat(A)                       # Heatmap of relationship matrix
#' vismat(A_compact)               # Works with compact matrices
#' vismat(A, by = "Gen")     # Group by generation
#' }
#' 
#' @references
#' Meuwissen, T. H. E., & Luo, Z. (1992). Computing inbreeding coefficients 
#' in large populations. Genetics Selection Evolution, 24(4), 305-313.
#' 
#' Henderson, C. R. (1976). A simple method for computing the inverse of a
#' numerator relationship matrix used in prediction of breeding values.
#' Biometrics, 32(1), 69-83.
#' 
#' @export
pedmat <- function(ped, method = "A", sparse = TRUE, invert_method = "auto", 
                     threads = 0, compact = FALSE) {
  # Check for splitped input - not supported

  if (inherits(ped, "splitped")) {
    stop(paste0(
      "pedmat() does not support 'splitped' objects directly.\n",
      "Use lapply() to process each group separately:\n",
      "  lapply(ped_split[grep('^GP', names(ped_split))], pedmat, method = '", 
      method, "')\n",
      "Or process a single group: pedmat(ped_split$GP1, method = '", method, "')"
    ), call. = FALSE)
  }
  
  # Coerce method to character
  method <- as.character(method)
  
  # Only single method allowed
  if (length(method) != 1) {
    stop(sprintf(
      paste0("Only a single method may be requested per call. ",
             "You requested %d methods: %s. ",
             "Call pedmat() separately for each method."),
      length(method), paste(method, collapse = ", ")
    ), call. = FALSE)
  }
  
  # Validate method
  all_methods <- c("A", "f", "Ainv", "D", "Dinv", "AA", "AAinv")
  if (!method %in% all_methods) {
    stop(sprintf("Invalid method: '%s'. Valid methods: %s", 
                 method, paste(all_methods, collapse = ", ")), call. = FALSE)
  }
  
  # Validate invert_method
  invert_method <- match.arg(invert_method, c("auto", "sympd", "general"))

  # Validate threads
  if (!is.numeric(threads) || length(threads) != 1 || is.na(threads) || threads < 0) {
    stop("'threads' must be a single non-negative integer.")
  }
  threads <- as.integer(threads)
  if (threads > 0) {
    if (!cpp_openmp_available()) {
      warning("OpenMP is not available; 'threads' will be ignored.", call. = FALSE)
    }
    prev_threads <- cpp_set_num_threads(threads)
    on.exit(cpp_set_num_threads(prev_threads), add = TRUE)
  }
  
  # Store original pedigree N for call_info
  n_original <- if (is_tidyped(ped)) nrow(ped) else {
    # Check if it's a data.frame/table before calling nrow
    if (is.data.frame(ped)) nrow(ped) else NA_integer_
  }

  # Handle compact mode
  if (compact) {
    # Store original pedigree for metadata
    ped_original <- if (is_tidyped(ped)) ped else tidyped(ped, addnum = TRUE)
    n_original <- nrow(ped_original)

    # Compact the pedigree
    compact_result <- compact_ped_for_matrix(ped_original)
    
    # For D/AA queries in compact mode, we also need A matrix for full-sibling calculations
    needs_A <- method %in% c("D", "AA")
    
    # Calculate the requested method
    result <- pedmat(
      ped = compact_result$ped_compact,
      method = method,
      sparse = sparse,
      invert_method = invert_method,
      threads = threads,
      compact = FALSE
    )
    
    # If D or AA, extract A matrix from the calculation process (if cached)
    # Note: D and AA calculations internally compute A first, so we can reuse it
    A_matrix <- NULL
    if (needs_A) {
      # Try to extract A from the calculation process (if cached)
      # Otherwise calculate it separately (fallback for safety)
      A_matrix <- tryCatch({
        attr(result, "A_intermediate")
      }, error = function(e) NULL)
      
      if (is.null(A_matrix)) {
        A_matrix <- pedmat(
          ped = compact_result$ped_compact,
          method = "A",
          sparse = sparse,
          compact = FALSE
        )
      }
      # Strip pedmat class for storage
      if (inherits(A_matrix, "pedmat")) {
        class(A_matrix) <- setdiff(class(A_matrix), "pedmat")
      }
    }
    
    # Store A matrix if calculated
    if (!is.null(A_matrix)) {
      attr(result, "A_matrix") <- A_matrix
    }
    
    # Update attributes for compaction
    ci <- attr(result, "call_info")
    ci$compact <- TRUE
    ci$n_original <- n_original
    ci$n_compact <- nrow(compact_result$ped_compact)
    ci$ped_original <- ped_original  # Save for expand_pedmat
    attr(result, "call_info") <- ci
    
    # Add compaction-specific attributes
    attr(result, "compact_map") <- compact_result$compact_map
    attr(result, "family_summary") <- compact_result$family_summary
    attr(result, "compact_stats") <- compact_result$compact_stats
    
    # Mark result as pedmat for compact mode
  if (isS4(result)) {
    attr(result, "pedmat_S4") <- TRUE
  } else {
    class(result) <- c("pedmat", class(result))
  }
  
  return(result)
  }

  # 1. Standardize pedigree
  if (is_tidyped(ped) || all(c("Ind", "Sire", "Dam", "Sex", "Gen", "IndNum", "SireNum", "DamNum") %in% names(ped))) {
    ped <- ensure_complete_tidyped(ped, sprintf("pedmat(method = \"%s\")", method))
  } else if (!is_tidyped(ped)) {
    ped <- tidyped(ped, addnum = TRUE)
  }
  
  # Ensure sorted by IndNum to match matrix indexing
  ped <- data.table::copy(ped)
  data.table::setorder(ped, IndNum)

  sire <- ped$SireNum
  dam <- ped$DamNum
  n <- nrow(ped)
  n_original <- n
  
  # Use a named list to store results
  output <- list()
  
  # Lazy-loading objects to avoid redundant calculations
  # Note: A_dense, D_dense, etc. will fail for 1M individuals. 
  # We only build them if specifically requested in 'method'.
  cache <- list(f_res = NULL, A_dense = NULL, D_dense = NULL, AA_dense = NULL)
  
  get_f_res <- function() {
    if (is.null(cache$f_res)) {
      # Reverting to highly optimized sequential version for f
      # It is very fast and has no parallel overhead or sorting overhead
      cache$f_res <<- cpp_calculate_inbreeding(sire, dam)
    }
    cache$f_res
  }
  
  get_A_dense <- function() {
    if (n > 25000) stop("Pedigree too large (>25k) for dense A matrix. Use 'Ainv' or sparse methods.")
    if (is.null(cache$A_dense)) cache$A_dense <<- cpp_calculate_A(sire, dam)
    cache$A_dense
  }

  get_D_dense <- function() {
    if (n > 25000) stop("Pedigree too large (>25k) for dense D matrix.")
    if (is.null(cache$D_dense)) {
      A_mat <- get_A_dense()
      cache$D_dense <<- cpp_calculate_D(sire, dam, A_mat)
    }
    cache$D_dense
  }

  get_AA_dense <- function() {
    if (n > 25000) stop("Pedigree too large (>25k) for dense AA matrix.")
    if (is.null(cache$AA_dense)) {
      A_mat <- get_A_dense()
      cache$AA_dense <<- cpp_calculate_AA(A_mat)
    }
    cache$AA_dense
  }

  # Process methods (ordered by dependency)
  all_methods <- c("A", "f", "Ainv", "D", "Dinv", "AA", "AAinv")
  active_methods <- intersect(all_methods, method)

  # Check for invalid methods
  invalid <- setdiff(method, all_methods)
  if (length(invalid) > 0) {
    stop(sprintf("Invalid method(s): %s", paste(invalid, collapse = ", ")))
  }

  # Process the single requested method (previously supported multiple)
  for (m in active_methods) {
    if (m == "f") {
      # Prefer getting f from existing A diag, else calculated
      if (!is.null(cache$A_dense)) {
        output$f <- diag(cache$A_dense) - 1.0
      } else {
        output$f <- get_f_res()$f
      }
      
    } else if (m == "A") {
      A_mat <- get_A_dense()
      # A matrices are nearly fully non-zero; converting to dgCMatrix via
      # Matrix::Matrix(dense, sparse=TRUE) scans all N² elements to detect
      # zeros — pure overhead with no storage benefit.  Return a dgeMatrix
      # (dense symmetric) which satisfies inherits(., "Matrix") and coerces
      # to a base matrix via as.matrix() without the N² scan.
      output$A <- if (sparse) .to_dgeMatrix(A_mat) else A_mat
      
    } else if (m == "D") {
      D_mat <- get_D_dense()
      output$D <- if (sparse) .to_dgeMatrix(D_mat) else D_mat
      # Store A matrix for compact mode reuse (on output list, not S4 matrix)
      if (!is.null(cache$A_dense)) {
        output$A_intermediate_D <- cache$A_dense
      }
      
    } else if (m == "Dinv") {
      D_mat <- get_D_dense()
      Dinv <- switch(invert_method,
        "auto" = cpp_invert_auto(D_mat),
        "sympd" = cpp_invert_sympd(D_mat),
        "general" = cpp_invert_dense(D_mat)
      )
      output$Dinv <- if (sparse) .to_dgeMatrix(Dinv) else Dinv

    } else if (m == "AA") {
      AA_mat <- get_AA_dense()
      output$AA <- if (sparse) .to_dgeMatrix(AA_mat) else AA_mat
      # Store A matrix for compact mode reuse (on output list, not S4 matrix)
      if (!is.null(cache$A_dense)) {
        output$A_intermediate_AA <- cache$A_dense
      }

    } else if (m == "AAinv") {
      AA_mat <- get_AA_dense()
      AAinv <- switch(invert_method,
        "auto" = cpp_invert_auto(AA_mat),
        "sympd" = cpp_invert_sympd(AA_mat),
        "general" = cpp_invert_dense(AA_mat)
      )
      output$AAinv <- if (sparse) .to_dgeMatrix(AAinv) else AAinv
      
    } else if (m == "Ainv") {
      f_res <- get_f_res()
      # Pass threads if we want, but currently C++ uses max
      triplets <- cpp_build_ainv_triplets(sire, dam, f_res$dii)
      Ainv <- Matrix::sparseMatrix(
        i = triplets$i, j = triplets$j, x = triplets$v,
        dims = c(n, n), symmetric = TRUE
      )
      output$Ainv <- if (sparse) Ainv else as.matrix(Ainv)
    }
  }

  # 3. Add dimension names to matrices
  ind_names <- ped$Ind
  for (m in names(output)) {
    obj <- output[[m]]
    if (is.matrix(obj) || inherits(obj, "Matrix")) {
      # Square matrices: set both dimensions
      rownames(obj) <- ind_names
      colnames(obj) <- ind_names
      output[[m]] <- obj
    } else if (is.numeric(obj) && length(obj) == n) {
      # Vectors (f)
      names(obj) <- ind_names
      output[[m]] <- obj
    }
  }

  # Construct return object: ALWAYS return pure matrix/vector
  # All metadata stored in attributes
  result <- if (length(active_methods) == 1) {
    output[[active_methods]]
  } else {
    # Multiple methods: return named list
    output
  }
  
  # Transfer A_intermediate from output list to result attributes (if exists)
  # This is for compact mode reuse
  if (length(active_methods) == 1) {
    if (active_methods == "D" && !is.null(output$A_intermediate_D)) {
      attr(result, "A_intermediate") <- output$A_intermediate_D
    } else if (active_methods == "AA" && !is.null(output$A_intermediate_AA)) {
      attr(result, "A_intermediate") <- output$A_intermediate_AA
    }
  }
  
  # Attach call information
  attr(result, "call_info") <- list(
    method = if(length(active_methods) == 1) active_methods else names(output),
    sparse = sparse,
    compact = FALSE,  # This branch is never compact
    n_original = n_original,
    n_compact = n_original,
    timestamp = Sys.time()
  )
  
  # Attach method name for easy access
  attr(result, "method") <- if(length(active_methods) == 1) active_methods else names(output)

  # Attach the pedigree used for calculation
  attr(result, "ped") <- ped
  
  # Store all methods if multiple were calculated
  if (length(active_methods) > 1) {
    attr(result, "mat_all") <- output
  }
  
  # Set S3 class - but only for non-S4 objects
  # For S4 objects (like Matrix), we keep them as-is and add pedmat_S4 marker
  if (isS4(result)) {
    attr(result, "pedmat_S4") <- TRUE
  } else {
    class(result) <- c("pedmat", class(result))
  }
  
  return(result)
}


#' Compact pedigree by merging full siblings for matrix calculation
#' 
#' @description
#' This internal function identifies full siblings (individuals sharing the same
#' sire and dam) and selects one representative per family. This can dramatically
#' reduce memory requirements when calculating relationship matrices for pedigrees
#' with large full-sibling families.
#' 
#' @param ped A tidyped object or pedigree data.
#' 
#' @return A list containing:
#' \itemize{
#'   \item \code{ped_compact}: The compacted pedigree (tidyped object)
#'   \item \code{compact_map}: Individual-level mapping table
#'   \item \code{family_summary}: Family-level summary table
#'   \item \code{compact_stats}: Compression statistics
#' }
#' 
#' @keywords internal
compact_ped_for_matrix <- function(ped) {
  # Ensure tidyped format
  if (!is_tidyped(ped)) {
    ped <- tidyped(ped, addnum = TRUE)
  }
  
  ped_dt <- data.table::as.data.table(data.table::copy(ped))
  n_original <- nrow(ped_dt)
  
  # Step 1: Create family identifiers
  # Using numeric keys for uniqueness and efficiency
  ped_dt[, family_key := paste(SireNum, DamNum, sep = "_")]
  ped_dt[SireNum == 0 | DamNum == 0, family_key := NA_character_]
  
  # Step 2: Create human-readable family labels
  ped_dt[!is.na(family_key), family_label := paste0(Sire, "x", Dam)]
  
  # Step 3: Calculate family sizes
  ped_dt[!is.na(family_key), family_size := .N, by = family_key]
  ped_dt[is.na(family_key), family_size := 1L]
  
  # Step 4: Identify individuals who are parents
  # Parents must be kept because their offspring reference them
  all_sires <- unique(ped_dt$Sire[!is.na(ped_dt$Sire) & ped_dt$Sire != ""])
  all_dams <- unique(ped_dt$Dam[!is.na(ped_dt$Dam) & ped_dt$Dam != ""])
  all_parents <- unique(c(all_sires, all_dams))
  
  # Mark who is a parent
  ped_dt[, IsParent := Ind %in% all_parents]
  
  # Step 5: Identify full-sibling families (family_size >= 2)
  fullsib_families <- ped_dt[!is.na(family_key) & family_size >= 2]
  
  if (nrow(fullsib_families) == 0) {
    # No full-sibling families to compact
    ped_compact <- data.table::copy(ped)
    return(list(
      ped_compact = ped_compact,
      compact_map = data.table(
        Ind = ped_dt$Ind,
        IndNum = ped_dt$IndNum,
        RepInd = ped_dt$Ind,
        RepIndNum = ped_dt$IndNum,
        FamilyID = NA_character_,
        SireNum = ped_dt$SireNum,
        DamNum = ped_dt$DamNum,
        Sire = ped_dt$Sire,
        Dam = ped_dt$Dam,
        FamilyLabel = NA_character_,
        FamilySize = 1L,
        IsRepresentative = TRUE,
        IsCompacted = FALSE
      ),
      family_summary = data.table(),
      compact_stats = list(
        n_original = n_original,
        n_compact = n_original,
        n_removed = 0L,
        n_families_compacted = 0L,
        compression_ratio = 1.0,
        memory_saved_pct = 0.0
      )
    ))
  }
  
  # Step 6: For each family, select representative from NON-PARENT members only
  # Parents are NEVER removed and NEVER represent others to avoid identity loss.
  
  compactable_members <- fullsib_families[IsParent == FALSE]
  
  if (nrow(compactable_members) == 0) {
    # No non-parent full-siblings to compact
    ped_compact <- data.table::copy(ped)
    return(list(
      ped_compact = ped_compact,
      compact_map = data.table(
        Ind = ped_dt$Ind,
        IndNum = ped_dt$IndNum,
        RepInd = ped_dt$Ind,
        RepIndNum = ped_dt$IndNum,
        FamilyID = NA_character_,
        SireNum = ped_dt$SireNum,
        DamNum = ped_dt$DamNum,
        Sire = ped_dt$Sire,
        Dam = ped_dt$Dam,
        FamilyLabel = NA_character_,
        FamilySize = 1L,
        IsRepresentative = TRUE,
        IsCompacted = FALSE
      ),
      family_summary = data.table(),
      compact_stats = list(
        n_original = n_original,
        n_compact = n_original,
        n_removed = 0L,
        n_families_compacted = 0L,
        compression_ratio = 1.0,
        memory_saved_pct = 0.0
      )
    ))
  }
  
  representatives <- compactable_members[, {
    list(RepInd = Ind[1],
         RepIndNum = IndNum[1])
  }, by = family_key]
  
  # Step 7: Create family lookup table with FamilyID
  family_lookup <- compactable_members[, .(
    SireNum = first(SireNum),
    DamNum = first(DamNum),
    Sire = first(Sire),
    Dam = first(Dam),
    FamilyLabel = first(family_label),
    FamilySize = .N,
    Gen = first(Gen)
  ), by = family_key]
  
  family_lookup[, FamilyID := sprintf("F%04d", .I)]
  
  # Add representative information
  family_lookup <- representatives[family_lookup, on = "family_key"]
  
  # Step 8: Build compact_map for ALL individuals
  # Map compactable members to their representative, others to themselves
  
  compact_map_fullsibs <- family_lookup[compactable_members, on = "family_key",
    .(Ind = i.Ind,
      OldIndNum = i.IndNum,
      RepInd = RepInd,
      OldRepIndNum = RepIndNum,
      FamilyID = FamilyID,
      SireNum = i.SireNum,
      DamNum = i.DamNum,
      Sire = i.Sire,
      Dam = i.Dam,
      FamilyLabel = FamilyLabel,
      FamilySize = FamilySize,
      IsParent = FALSE,
      IsRepresentative = (i.Ind == RepInd),
      IsCompacted = TRUE)
  ]
  
  # Non-compacted individuals (all parents + unique non-parents)
  non_compacted <- ped_dt[!(Ind %in% compactable_members$Ind)]
  compact_map_others <- data.table(
    Ind = non_compacted$Ind,
    OldIndNum = non_compacted$IndNum,
    RepInd = non_compacted$Ind,
    OldRepIndNum = non_compacted$IndNum,
    FamilyID = NA_character_,
    SireNum = non_compacted$SireNum,
    DamNum = non_compacted$DamNum,
    Sire = non_compacted$Sire,
    Dam = non_compacted$Dam,
    FamilyLabel = NA_character_,
    FamilySize = 1L,
    IsParent = non_compacted$IsParent,
    IsRepresentative = TRUE,
    IsCompacted = FALSE
  )
  
  compact_map <- rbind(compact_map_fullsibs, compact_map_others)
  data.table::setorder(compact_map, OldIndNum)
  
  # Step 9: Create compacted pedigree
  # Remove rule: Non-parent AND non-representative individuals
  # Keep rule: All parents + representatives
  remove_ids <- compact_map[IsParent == FALSE & IsRepresentative == FALSE, OldIndNum]
  ped_compact <- ped_dt[!(IndNum %in% remove_ids)]
  data.table::setorder(ped_compact, IndNum)
  
  # CRITICAL: Renumber IndNum to be consecutive (1, 2, 3, ...)
  # C++ code expects IndNum to be consecutive indices starting from 1
  old_to_new_map <- data.table(
    OldNum = ped_compact$IndNum,
    NewNum = seq_len(nrow(ped_compact))
  )
  
  # Update compact_map: map OldRepIndNum to NewRepIndNum
  compact_map <- old_to_new_map[compact_map, on = c(OldNum = "OldRepIndNum")]
  setnames(compact_map, "NewNum", "RepIndNum")
  compact_map[, OldNum := NULL]
  
  # Update SireNum and DamNum in compact_map to match compact matrix indices
  # First SireNum
  compact_map <- old_to_new_map[compact_map, on = c(OldNum = "SireNum")]
  setnames(compact_map, "NewNum", "SireNum")
  compact_map[is.na(SireNum), SireNum := 0L] # Handle founders / missing from map
  compact_map[, OldNum := NULL]
  
  # Then DamNum
  compact_map <- old_to_new_map[compact_map, on = c(OldNum = "DamNum")]
  setnames(compact_map, "NewNum", "DamNum")
  compact_map[is.na(DamNum), DamNum := 0L]
  compact_map[, OldNum := NULL]
  
  # Also store old IndNum for reference
  setnames(compact_map, "OldIndNum", "IndNum")
  
  # Update ped_compact: renumber IndNum, SireNum, DamNum
  ped_compact[, OldIndNum := IndNum]
  ped_compact <- old_to_new_map[ped_compact, on = c(OldNum = "OldIndNum")]
  ped_compact[, IndNum := NewNum]
  ped_compact[, c("NewNum", "OldNum") := NULL]
  
  # Update SireNum
  old_to_new_sire <- copy(old_to_new_map)
  setnames(old_to_new_sire, c("OldSireNum", "NewSireNum"))
  ped_compact <- old_to_new_sire[ped_compact, on = c(OldSireNum = "SireNum")]
  ped_compact[!is.na(NewSireNum), SireNum := NewSireNum]
  ped_compact[is.na(NewSireNum), SireNum := 0L]
  ped_compact[, c("OldSireNum", "NewSireNum") := NULL]
  
  # Update DamNum
  old_to_new_dam <- copy(old_to_new_map)
  setnames(old_to_new_dam, c("OldDamNum", "NewDamNum"))
  ped_compact <- old_to_new_dam[ped_compact, on = c(OldDamNum = "DamNum")]
  ped_compact[!is.na(NewDamNum), DamNum := NewDamNum]
  ped_compact[is.na(NewDamNum), DamNum := 0L]
  ped_compact[, c("OldDamNum", "NewDamNum") := NULL]
  
  # Remove OldIndNum if it still exists
  if ("OldIndNum" %in% names(ped_compact)) {
    ped_compact[, OldIndNum := NULL]
  }
  
  # Clean up extra columns added during compaction
  extra_cols <- c("family_key", "family_label", "family_size", "IsParent")
  for (col in extra_cols) {
    if (col %in% names(ped_compact)) {
      ped_compact[, (col) := NULL]
    }
  }
  
  # Count before tidyped (for accurate statistics)
  n_before_tidyped <- nrow(ped_compact)
  
  # Calculate family compression counts BEFORE tidyped (while we still have compact_map with removed individuals)
  family_compressed_counts <- compact_map[IsCompacted == TRUE, .(
    NCompressed = sum(IsParent == FALSE & IsRepresentative == FALSE)
  ), by = FamilyID]
  
  # CRITICAL: Use tidyped to ensure ped_compact is complete and properly numbered
  # This will add missing parents, renumber everything correctly, and compute Family field
  ped_compact <- tidyped(ped_compact, addnum = TRUE)
  
  # Verify Family and FamilySize columns exist in ped_compact
  if (!("Family" %in% names(ped_compact))) {
    warning("Family field missing in compact pedigree. This may cause issues with vismat(by = 'Family').")
  }
  
  # After tidyped, rebuild compact_map to include ALL individuals (kept + removed)
  # Save original compact_map information before rebuilding
  original_compact_map <- copy(compact_map)
  
  # Build compact_map for individuals in ped_compact (kept individuals)
  kept_compact_map <- ped_compact[, .(
    Ind,
    IndNum,
    SireNum,
    DamNum,
    Sire,
    Dam
  )]
  
  # Merge with original mapping information
  kept_compact_map <- original_compact_map[kept_compact_map, on = "Ind",
    .(Ind = i.Ind,
      IndNum = i.IndNum,
      SireNum = i.SireNum,
      DamNum = i.DamNum,
      Sire = i.Sire,
      Dam = i.Dam,
      RepInd = RepInd,
      FamilyID = FamilyID,
      FamilyLabel = FamilyLabel,
      FamilySize = FamilySize,
      IsCompacted = IsCompacted,
      IsParent = IsParent,
      IsRepresentative = IsRepresentative)
  ]
  
  # For individuals added by tidyped (not in original), set defaults
  kept_compact_map[is.na(RepInd), `:=`(
    RepInd = Ind,
    FamilyID = NA_character_,
    FamilyLabel = NA_character_,
    FamilySize = 1L,
    IsCompacted = FALSE,
    IsParent = TRUE,  # Added by tidyped means they're parents
    IsRepresentative = TRUE
  )]
  
  # Update RepIndNum based on new numbering in ped_compact
  rep_num_mapping <- ped_compact[, .(RepInd = Ind, RepIndNum = IndNum)]
  kept_compact_map <- rep_num_mapping[kept_compact_map, on = "RepInd"]
  
  # Build compact_map for removed individuals (not in ped_compact)
  removed_inds <- original_compact_map[!(Ind %in% ped_compact$Ind)]
  if (nrow(removed_inds) > 0) {
    # For removed individuals, keep original info but update RepIndNum
    removed_compact_map <- removed_inds[, .(
      Ind,
      IndNum = NA_integer_,  # No IndNum in ped_compact
      RepInd,
      SireNum,
      DamNum,
      Sire,
      Dam,
      FamilyID,
      FamilyLabel,
      FamilySize,
      IsCompacted,
      IsParent,
      IsRepresentative = FALSE  # Removed individuals are never representatives
    )]
    
    # Update RepIndNum for removed individuals
    removed_compact_map <- rep_num_mapping[removed_compact_map, on = "RepInd"]
    
    # Combine kept and removed
    new_compact_map <- rbind(kept_compact_map, removed_compact_map, fill = TRUE)
  } else {
    new_compact_map <- kept_compact_map
  }
  
  # CRITICAL: Re-map SireNum and DamNum to match the FINAL ped_compact numbering
  # because tidyped() might have reordered/added individuals.
  final_num_mapping <- ped_compact[, .(ParentID = Ind, NewParentNum = IndNum)]
  
  # Update SireNum
  new_compact_map[, SireNum := final_num_mapping$NewParentNum[match(Sire, final_num_mapping$ParentID)]]
  new_compact_map[is.na(SireNum), SireNum := 0L]
  
  # Update DamNum
  new_compact_map[, DamNum := final_num_mapping$NewParentNum[match(Dam, final_num_mapping$ParentID)]]
  new_compact_map[is.na(DamNum), DamNum := 0L]
  
  # Sort: kept individuals first (by IndNum), then removed individuals (by Ind)
  new_compact_map[!is.na(IndNum), sort_key := IndNum]
  new_compact_map[is.na(IndNum), sort_key := 1000000 + seq_len(.N)]
  data.table::setorder(new_compact_map, sort_key)
  new_compact_map[, sort_key := NULL]
  
  compact_map <- new_compact_map
  data.table::setorder(compact_map, IndNum)
  
  n_compact <- nrow(ped_compact)
  n_removed <- n_original - n_before_tidyped  # Use pre-tidyped count
  n_families <- nrow(family_lookup)
  
  # Step 9: Calculate statistics
  compression_ratio <- n_compact / n_original
  # Memory saved for dense matrix: (n_original^2 - n_compact^2) / n_original^2
  memory_saved_pct <- (1 - (n_compact / n_original)^2) * 100
  
  # By sex statistics
  sex_stats <- compact_map[, .(
    n_original = .N,
    n_compact = sum(IsRepresentative),
    n_removed = sum(!IsRepresentative)
  ), by = .(Sex = ifelse(is.na(Sire) & is.na(Dam), "founder", 
                         ifelse(Sire != "", "with_parents", "unknown")))]
  
  # Family size distribution
  family_size_breaks <- c(0, 2, 10, 50, 100, Inf)
  family_size_labels <- c("1", "2-10", "11-50", "51-100", "100+")
  family_lookup[, SizeCategory := cut(FamilySize, 
                                        breaks = family_size_breaks,
                                        labels = family_size_labels,
                                        right = TRUE)]
  
  family_size_dist <- family_lookup[, .(
    n_families = .N,
    n_individuals_total = sum(FamilySize),
    n_individuals_removed = sum(FamilySize - 1)
  ), by = SizeCategory]
  
  # Ensure result is not a tidyped object (just a data.table)
  class(family_size_dist) <- c("data.table", "data.frame")
  
  compact_stats <- list(
    n_original = n_original,
    n_compact = n_compact,
    n_removed = n_removed,
    n_families_compacted = n_families,
    compression_ratio = compression_ratio,
    memory_saved_pct = memory_saved_pct,
    by_sex = sex_stats,
    family_size_dist = family_size_dist
  )
  
  # Step 10: Prepare final family_summary (update RepIndNum with new numbering)
  # Match RepInd with the new IndNum from ped_compact
  rep_num_mapping <- ped_compact[, .(RepInd = Ind, NewRepIndNum = IndNum)]
  family_summary <- family_lookup[, .(FamilyID, SireNum, DamNum, Sire, Dam, 
                                      FamilyLabel, FamilySize, RepInd, Gen)]
  family_summary <- rep_num_mapping[family_summary, on = "RepInd"]
  setnames(family_summary, "NewRepIndNum", "RepIndNum")
  

  # Add NCompressed column using pre-calculated counts (before tidyped)
  family_summary <- family_compressed_counts[family_summary, on = "FamilyID"]
  family_summary[is.na(NCompressed), NCompressed := 0L]
  
  # Reorder columns for clarity
  data.table::setcolorder(family_summary, c("FamilyID", "FamilyLabel", "Sire", "Dam",
                                             "SireNum", "DamNum", "FamilySize", 
                                             "NCompressed", "RepInd", "RepIndNum", "Gen"))
  data.table::setorder(family_summary, FamilyID)
  
  # Ensure result is not a tidyped object
  class(family_summary) <- c("data.table", "data.frame")
  class(compact_map) <- c("data.table", "data.frame")
  
  return(list(
    ped_compact = ped_compact,
    compact_map = compact_map,
    family_summary = family_summary,
    compact_stats = compact_stats
  ))
}


#' Query Relationship Coefficients from a Pedigree Matrix
#' 
#' @description
#' Retrieves relationship coefficients between individuals from a pedmat object.
#' For compact matrices, automatically handles lookup of merged full-siblings.
#' 
#' @param x A pedmat object created by \code{\link{pedmat}}.
#' @param id1 Character, first individual ID.
#' @param id2 Character, second individual ID. If \code{NULL}, returns 
#'   the entire row of relationships for \code{id1}.
#' 
#' @return 
#' \itemize{
#'   \item If \code{id2} is provided: numeric value (relationship coefficient)
#'   \item If \code{id2} is \code{NULL}: named numeric vector (id1's row)
#'   \item Returns \code{NA} if individual not found
#' }
#' 
#' @details
#' For compact matrices (\code{compact = TRUE}), this function automatically

#' maps individuals to their family representatives. For methods A, D, and AA,
#' it can compute the correct relationship even between merged full-siblings
#' using the formula:
#' \itemize{
#'   \item \strong{A}: \eqn{a_{ij} = 0.5 * (a_{is} + a_{id})} where s, d are parents
#'   \item \strong{D}: \eqn{d_{ij} = a_{ij}^2} (for full-sibs in same family)
#'   \item \strong{AA}: \eqn{aa_{ij} = a_{ij}^2}
#' }
#' 
#' @note
#' Inverse matrices (Ainv, Dinv, AAinv) are \strong{not supported} because
#' inverse matrix elements do not represent meaningful relationship coefficients.
#' 
#' @seealso \code{\link{pedmat}}, \code{\link{expand_pedmat}}
#' 
#' @examples
#' tped <- tidyped(small_ped)
#' A <- pedmat(tped, method = "A", compact = TRUE)
#' 
#' # Query specific pair
#' query_relationship(A, "A", "B")
#' 
#' # Query merged full-siblings (works with compact)
#' query_relationship(A, "Z1", "Z2")
#' 
#' # Get all relationships for one individual
#' query_relationship(A, "A")
#' 
#' @export
query_relationship <- function(x, id1, id2 = NULL) {
  # Check if x is a pedmat (either S3 or S4 with marker)
  is_pedmat <- inherits(x, "pedmat") || !is.null(attr(x, "pedmat_S4"))
  if (!is_pedmat) {
    stop("x must be a pedmat object")
  }
  
  ci <- attr(x, "call_info")
  primary_method <- ci$method[1]
  
  # Inverse matrices are not supported - their elements don't represent
  # meaningful relationship coefficients
  if (primary_method %in% c("Ainv", "Dinv", "AAinv")) {
    stop(sprintf(
      paste0("query_relationship() does not support inverse matrices (%s). ",
             "Inverse matrix elements do not represent meaningful relationship ",
             "coefficients between individuals."),
      primary_method
    ), call. = FALSE)
  }
  
  # Handle vector case (e.g., inbreeding coefficients "f")
  # For vectors, id2 is ignored - we just look up the value for id1
  if (primary_method == "f" || (is.numeric(x) && is.null(dim(x)))) {
    # Strip class for indexing
    vec <- x
    if (inherits(vec, "pedmat")) {
      class(vec) <- setdiff(class(vec), "pedmat")
    }
    
    if (isTRUE(ci$compact)) {
      compact_map <- attr(x, "compact_map")
      map1 <- compact_map[Ind == id1]
      if (nrow(map1) == 0) return(NA_real_)
      rep1_idx <- map1$RepIndNum[1]
      return(vec[rep1_idx])
    } else {
      # Non-compact: use names directly
      if (id1 %in% names(vec)) {
        return(vec[id1])
      } else {
        return(NA_real_)
      }
    }
  }
  
  # Get raw matrix (already is the raw matrix)
  mat <- x
  if (inherits(mat, "pedmat")) {
    class(mat) <- setdiff(class(mat), "pedmat")
  }
  
  # Check if it's compact
  if (isTRUE(ci$compact)) {
    compact_map <- attr(x, "compact_map")
    
    # Use data.table for fast lookup
    map1 <- compact_map[Ind == id1]
    if (nrow(map1) == 0) return(NA_real_)
    rep1_idx <- map1$RepIndNum[1]
    rep1_ind <- map1$RepInd[1]
    
    if (is.null(id2)) {
      # Return row for representative
      # The returned vector keeps the matrix column names (compact pedigree individuals)
      res <- mat[rep1_idx, ]
      # Names are already set from matrix colnames, no need to modify
      return(res)
    }
    
    map2 <- compact_map[Ind == id2]
    if (nrow(map2) == 0) return(NA_real_)
    rep2_idx <- map2$RepIndNum[1]
    rep2_ind <- map2$RepInd[1]
    
    # The actual individuals are different but they share the same representative
    # (Full-sibling case)
    if (id1 != id2 && rep1_idx == rep2_idx) {
      repS_idx <- map1$SireNum[1]
      repD_idx <- map1$DamNum[1]
      
      # Helper to get relationship (handling missing parents as 0)
      # We need A values even if primary_method is D or AA
      get_A_val <- function(r1, r2) {
        if (r1 <= 0 || r2 <= 0) return(0.0)
        
        # Look for A matrix
        A_mat <- NULL
        if (primary_method == "A") {
            A_mat <- mat
        } else {
            # Check for stored A matrix in attributes
            # Try both "A_matrix" and "A_intermediate" for backward/internal compatibility
            A_mat <- attr(x, "A_intermediate")
            if (is.null(A_mat)) {
              A_mat <- attr(x, "A_matrix")
            }
            
            if (is.null(A_mat)) {
              mat_all <- attr(x, "mat_all")
              if (!is.null(mat_all) && !is.null(mat_all$A)) {
                A_mat <- mat_all$A
              }
            }
        }
        
        if (is.null(A_mat)) return(NA_real_)
        return(A_mat[r1, r2])
      }
      
      A_ss <- get_A_val(repS_idx, repS_idx)
      if (repS_idx <= 0) A_ss <- 1.0 
      A_dd <- get_A_val(repD_idx, repD_idx)
      if (repD_idx <= 0) A_dd <- 1.0
      A_sd <- get_A_val(repS_idx, repD_idx)
      
      
      # Calculate A_ij
      if (repS_idx <= 0 && repD_idx <= 0) {
        res_A <- 0.0
      } else {
        res_A <- 0.25 * (A_ss + A_dd + 2 * A_sd)
      }
      
      if (primary_method == "A") return(res_A)
      
      # For D: 0.25 * (A_SS * A_DD + A_SD^2)
      if (primary_method == "D") return(0.25 * (A_ss * A_dd + A_sd^2))
      
      # For AA: A_ij^2
      if (primary_method == "AA") return(res_A^2)
      
      return(res_A) 
    }
    
    return(mat[rep1_idx, rep2_idx])
  } else {
    # Non-compact: direct indexing
    if (is.null(id2)) return(mat[id1, ])
    return(mat[id1, id2])
  }
}

# --------------------------------------------------------------------------
# Internal helper: sibling off-diagonal value from parent relationships
# --------------------------------------------------------------------------
# Computes the relationship value between full-siblings who share the same
# compact representative. Used by expand_pedmat() and aggregate_compact_by_group().
#
# @param A_mat The compact A (additive) matrix.
# @param sire_idx Integer, row index of the sire in A_mat (0 = unknown).
# @param dam_idx  Integer, row index of the dam in A_mat (0 = unknown).
# @param method   Character, one of "A", "D", "AA".
# @return Numeric scalar: the sibling off-diagonal value.
calc_sib_offdiag <- function(A_mat, sire_idx, dam_idx, method = "A") {
  A_ss <- if (sire_idx > 0) A_mat[sire_idx, sire_idx] else 1.0
  A_dd <- if (dam_idx  > 0) A_mat[dam_idx,  dam_idx]  else 1.0
  A_sd <- if (sire_idx > 0 && dam_idx > 0) A_mat[sire_idx, dam_idx] else 0.0

  res_A_sib <- 0.25 * (A_ss + A_dd + 2 * A_sd)

  switch(method,
    "A"  = res_A_sib,
    "D"  = 0.25 * (A_ss * A_dd + A_sd^2),
    "AA" = res_A_sib^2,
    stop(sprintf("Unsupported method '%s' for sibling off-diagonal.", method))
  )
}

#' Expand a Compact Pedigree Matrix to Full Dimensions
#' 
#' @description
#' Restores a compact pedmat to its original dimensions by mapping each
#' individual to their family representative's values. For non-compact matrices,
#' returns the matrix unchanged.
#' 
#' @param x A pedmat object from \code{\link{pedmat}}.
#' 
#' @return 
#' Matrix or vector with original pedigree dimensions:
#' \itemize{
#'   \item Matrices: Row and column names set to all individual IDs
#'   \item Vectors (e.g., method="f"): Names set to all individual IDs
#' }
#' The result is \strong{not} a pedmat object (S3 class stripped).
#' 
#' @details
#' For compact matrices, full-siblings within the same family will have
#' identical relationship values in the expanded matrix because they shared
#' the same representative during calculation.
#' 
#' @seealso \code{\link{pedmat}}, \code{\link{query_relationship}}
#' 
#' @examples
#' tped <- tidyped(small_ped)
#' 
#' # Compact matrix
#' A_compact <- pedmat(tped, method = "A", compact = TRUE)
#' dim(A_compact)  # Reduced dimensions
#' 
#' # Expand to full size
#' A_full <- expand_pedmat(A_compact)
#' dim(A_full)  # Original dimensions restored
#' 
#' # Non-compact matrices are returned unchanged
#' A <- pedmat(tped, method = "A", compact = FALSE)
#' A2 <- expand_pedmat(A)
#' identical(dim(A), dim(A2))  # TRUE
#' 
#' @export
expand_pedmat <- function(x) {
  # NOTE: sibling off-diagonal correction uses calc_sib_offdiag() below.
  # Check if x is a pedmat (either S3 or S4 with marker)
  is_pedmat <- inherits(x, "pedmat") || !is.null(attr(x, "pedmat_S4"))
  if (!is_pedmat) {
    stop("x must be a pedmat object")
  }
  
  ci <- attr(x, "call_info")
  if (!isTRUE(ci$compact)) {
    # Already full size, just return the matrix
    result <- x
    if (inherits(result, "pedmat")) {
      class(result) <- setdiff(class(result), "pedmat")
    }
    return(result)
  }
  
  map <- attr(x, "compact_map")
  mat <- x
  if (inherits(mat, "pedmat")) {
    class(mat) <- setdiff(class(mat), "pedmat")
  }
  
  # Get original order based on Ind column
  # compact_map is sorted by OldIndNum (original pedigree order)
  # We need to get the original pedigree Ind order
  ped_full <- attr(x, "call_info")$ped_original
  if (is.null(ped_full)) {
    # Fallback: use map$Ind order as-is
    orig_ind_order <- map$Ind
  } else {
    orig_ind_order <- ped_full$Ind
  }
  
  # Filter map to match original order
  map_ordered <- map[match(orig_ind_order, map$Ind)]
  
  # Matrix-like or vector-like?
  if (!is.null(dim(mat))) {
    # Matrix expansion
    result_full <- mat[map_ordered$RepIndNum, map_ordered$RepIndNum]
    rownames(result_full) <- map_ordered$Ind
    colnames(result_full) <- map_ordered$Ind
    
    # Fix sibling off-diagonals for A, D, AA
    # In compact mode, full-siblings share a representative. Their relationship to
    # others is identical, but their relationship to each other (off-diagonal) 
    # is different from their relationship to self (diagonal).
    primary_method <- ci$method[1]
    if (primary_method %in% c("A", "D", "AA")) {
      counts <- table(map_ordered$RepIndNum)
      reps_with_sibs <- as.integer(names(counts[counts > 1]))
      
      if (length(reps_with_sibs) > 0) {
        # Get A matrix (required for sibling calculation)
        A_mat <- if (primary_method == "A") mat else attr(x, "A_intermediate")
        
        if (!is.null(A_mat)) {
          for (rep_idx in reps_with_sibs) {
            first_idx <- which(map_ordered$RepIndNum == rep_idx)[1]
            s_idx <- map_ordered$SireNum[first_idx]
            d_idx <- map_ordered$DamNum[first_idx]
            
            sib_val <- calc_sib_offdiag(A_mat, s_idx, d_idx, primary_method)
            
            member_indices <- which(map_ordered$RepIndNum == rep_idx)
            if (length(member_indices) > 1) {
              diag_val <- mat[rep_idx, rep_idx]
              result_full[member_indices, member_indices] <- sib_val
              diag_indices <- cbind(member_indices, member_indices)
              result_full[diag_indices] <- diag_val
            }
          }
        }
      }
    }
    return(result_full)
  } else {
    # Vector expansion
    result_full <- mat[map_ordered$RepIndNum]
    names(result_full) <- map_ordered$Ind
    return(result_full)
  }
}

# --- S3 Methods for pedmat class ---

#' @export
print.pedmat <- function(x, ...) {
  ci <- attr(x, "call_info")
  cat("Pedigree Matrix (", paste(ci$method, collapse=", "), ")\n", sep="")
  if (isTRUE(ci$compact)) {
    cat("Compacted: ", ci$n_original, " -> ", ci$n_compact, " individuals\n", sep="")
  } else {
    cat("Size: ", ci$n_original, " individuals\n", sep="")
  }
  
  cat("\nUse summary() for details, attr(x, 'ped') for pedigree data.\n\n")
  
  # Strip pedmat class and print using underlying method
  x_show <- x
  if (inherits(x_show, "pedmat")) {
    class(x_show) <- setdiff(class(x_show), "pedmat")
  }
  print(x_show, ...)
}

#' Summary Statistics for Pedigree Matrices
#' 
#' @description
#' Computes and displays summary statistics for a pedmat object.
#' 
#' @param x A pedmat object from \code{\link{pedmat}}.
#' 
#' @return An object of class \code{"summary.pedmat"} with statistics including
#'   method, dimensions, compression ratio (if compact), mean relationship, 
#'   and matrix density.
#' 
#' @details
#' Since pedmat objects are often S4 sparse matrices with custom attributes,
#' use this function instead of the generic \code{summary()} to ensure proper
#' display of pedigree matrix statistics.
#' 
#' @examples
#' tped <- tidyped(small_ped)
#' A <- pedmat(tped, method = "A")
#' summary_pedmat(A)
#' 
#' @seealso \code{\link{pedmat}}
#' @export
summary_pedmat <- function(x) {
  # Check if it's a pedmat
  is_pedmat <- inherits(x, "pedmat") || !is.null(attr(x, "pedmat_S4"))
  if (!is_pedmat) {
    stop("x must be a pedmat object")
  }
  
  ci <- attr(x, "call_info")
  
  stats <- list(
    method = ci$method,
    compact = isTRUE(ci$compact),
    n_original = ci$n_original,
    n_calculated = ci$n_compact
  )
  
  # Calculate matrix-level stats
  obj_clean <- x
  class(obj_clean) <- setdiff(class(obj_clean), "pedmat")
  
  if (!is.list(obj_clean)) {
    # For Matrix objects, use Matrix::mean or extract values
    stats$mean_val <- tryCatch({
      n <- as.numeric(nrow(obj_clean))
      denom <- n * n - n
      if (denom <= 0) return(NA_real_)
      if (inherits(obj_clean, "Matrix")) {
        total_sum <- sum(obj_clean)
        diag_sum <- sum(Matrix::diag(obj_clean))
        (total_sum - diag_sum) / denom
      } else {
        total_sum <- sum(obj_clean)
        diag_sum <- sum(diag(obj_clean))
        (total_sum - diag_sum) / denom
      }
    }, error = function(e) NA_real_)
    
    stats$sparsity <- if (inherits(obj_clean, "Matrix")) {
      # Works for both sparseMatrix (e.g. dgCMatrix) and dense Matrix (e.g. dgeMatrix).
      # speed branch returns dgeMatrix for A/D/AA to avoid the N² zero-scan
      # overhead of Matrix::Matrix(..., sparse=TRUE); nnzero() handles both cases.
      Matrix::nnzero(obj_clean) / (as.numeric(nrow(obj_clean)) * ncol(obj_clean))
    } else if (is.matrix(obj_clean)) {
      sum(obj_clean != 0L) / length(obj_clean)
    } else { NA_real_ }
  }
  
  if (isTRUE(ci$compact)) {
    stats$compact_stats <- attr(x, "compact_stats")
    fs <- attr(x, "family_summary")
    if (!is.null(fs)) {
      stats$top_families <- head(fs[order(-fs$FamilySize), ], 5)
    }
  }
  
  class(stats) <- "summary.pedmat"
  return(stats)
}

#' @export
summary.pedmat <- function(object, ...) {
  # Delegate to summary_pedmat
  summary_pedmat(object)
}

#' @export
print.summary.pedmat <- function(x, ...) {
  cat("Summary of Pedigree Matrix (", paste(x$method, collapse=", "), ")\n", sep="")
  cat("========================================\n")
  cat("Input Size:     ", x$n_original, " individuals\n")
  cat("Calculated Size:", x$n_calculated, " individuals\n")
  
  if (x$compact) {
    cat("\nCompaction Results:\n")
    if (!is.null(x$compact_stats)) {
      ratio <- x$compact_stats$compression_ratio
      cat("- Compression: ", round(ratio * 100, 1), "%\n", sep="")
      cat("- Families:    ", x$compact_stats$n_families_compacted, " families merged\n")
    }
    if (!is.null(x$top_families)) {
      cat("\nTop 5 families by size:\n")
      print(x$top_families[, c("FamilyLabel", "FamilySize")])
    }
  }
  
  if (!is.null(x$mean_val)) {
      cat("\nMatrix Properties:\n")
      cat("- Mean off-diagonal relationship: ", if(is.na(x$mean_val)) "NA" else round(x$mean_val, 6), "\n")
      cat("- Density (non-zero): ", if(is.na(x$sparsity)) "NA" else round(x$sparsity * 100, 2), "%\n", sep="")
  }
  cat("========================================\n")
}

# --- Generic functions to enable S3 dispatch ---



#' @export
#' @method [ pedmat
`[.pedmat` <- function(x, ...) {
  # Strip pedmat class and delegate to underlying method
  class(x) <- setdiff(class(x), "pedmat")
  x[...]
}

#' @export
#' @method dim pedmat
dim.pedmat <- function(x) {
  class(x) <- setdiff(class(x), "pedmat")
  dim(x)
}

#' @export
#' @method length pedmat
length.pedmat <- function(x) {
  class(x) <- setdiff(class(x), "pedmat")
  length(x)
}

#' @export
#' @method t pedmat
t.pedmat <- function(x) {
  class(x) <- setdiff(class(x), "pedmat")
  t(x)
}

#' @export
as.matrix.pedmat <- function(x, ...) {
  class(x) <- setdiff(class(x), "pedmat")
  if (inherits(x, "Matrix")) return(as.matrix(x))
  return(x)
}

#' @export
as.vector.pedmat <- function(x, mode = "any") {
  class(x) <- setdiff(class(x), "pedmat")
  as.vector(x, mode = mode)
}

library(testthat)
library(visPedigree)
library(data.table)

make_pedrel_test_pedigree <- function() {
  tidyped(data.table(
    Ind = c("A", "B", "C", "D", "E", "F", "G", "H"),
    Sire = c(NA, NA, "A", "A", "C", "C", "E", "E"),
    Dam = c(NA, NA, "B", "B", "D", "D", "F", "F"),
    Cohort = rep(1:4, each = 2)
  ))
}

explicit_group_summary <- function(group, ped, scale, reference = NULL) {
  ids <- ped[Cohort == group, Ind]
  if (!is.null(reference)) ids <- ids[ids %in% reference]
  A <- pedmat(ped, method = "A", sparse = FALSE)
  idx <- match(ids, rownames(A))
  A_group <- A[idx, idx, drop = FALSE]

  if (scale == "relationship") {
    (sum(A_group) - sum(diag(A_group))) /
      (length(idx) * (length(idx) - 1))
  } else {
    sum(A_group) / (2 * length(idx)^2)
  }
}

test_that("pedrel matrix-free summaries match an explicit A matrix", {
  ped <- make_pedrel_test_pedigree()

  rel <- pedrel(ped, by = "Cohort", scale = "relationship")
  coan <- pedrel(ped, by = "Cohort", scale = "coancestry")

  expected_rel <- vapply(
    rel$Cohort, explicit_group_summary, numeric(1),
    ped = ped, scale = "relationship"
  )
  expected_coan <- vapply(
    coan$Cohort, explicit_group_summary, numeric(1),
    ped = ped, scale = "coancestry"
  )

  expect_equal(rel$MeanRel, expected_rel, tolerance = 1e-12)
  expect_equal(coan$MeanCoan, expected_coan, tolerance = 1e-12)
  expect_true(all(rel$Status == "ok"))
  expect_true(all(coan$Status == "ok"))
})

test_that("pedrel reference summaries match the selected explicit submatrix", {
  ped <- make_pedrel_test_pedigree()
  reference <- c("C", "D", "G", "H")

  rel <- suppressWarnings(
    pedrel(ped, by = "Cohort", reference = reference)
  )
  coan <- suppressWarnings(
    pedrel(ped, by = "Cohort", reference = reference, scale = "coancestry")
  )

  valid_rel <- rel[Status == "ok"]
  valid_coan <- coan[Status == "ok"]
  expected_rel <- vapply(
    valid_rel$Cohort, explicit_group_summary, numeric(1),
    ped = ped, scale = "relationship", reference = reference
  )
  expected_coan <- vapply(
    valid_coan$Cohort, explicit_group_summary, numeric(1),
    ped = ped, scale = "coancestry", reference = reference
  )

  expect_equal(valid_rel$MeanRel, expected_rel, tolerance = 1e-12)
  expect_equal(valid_coan$MeanCoan, expected_coan, tolerance = 1e-12)
})

test_that("pedrel handles a reference consisting only of founders", {
  ped <- make_pedrel_test_pedigree()
  reference <- c("A", "B")

  rel <- suppressWarnings(
    pedrel(ped, by = "Cohort", reference = reference)
  )
  coan <- suppressWarnings(
    pedrel(ped, by = "Cohort", reference = reference, scale = "coancestry")
  )

  expect_equal(rel[Cohort == 1, MeanRel], 0)
  expect_equal(coan[Cohort == 1, MeanCoan], 0.25)
  expect_equal(rel[Cohort == 1, Status], "ok")
  expect_equal(coan[Cohort == 1, Status], "ok")
})

test_that("pedrel matches explicit A for one-parent and selfing pedigrees", {
  one_parent <- tidyped(data.table(
    Ind = c("A", "B", "C"),
    Sire = c(NA, "A", "A"),
    Dam = c(NA, NA, NA),
    Cohort = c(NA, 1L, 1L)
  ))
  selfed <- suppressMessages(tidyped(
    data.table(
      Ind = c("A", "B", "C", "D", "E"),
      Sire = c(NA, NA, "A", "C", "C"),
      Dam = c(NA, NA, "B", "C", "C"),
      Cohort = c(NA, NA, NA, 1L, 1L)
    ),
    selfing = TRUE
  ))

  for (ped in list(one_parent, selfed)) {
    rel <- pedrel(ped, by = "Cohort")
    coan <- pedrel(ped, by = "Cohort", scale = "coancestry")

    expect_equal(
      rel$MeanRel,
      explicit_group_summary(1L, ped, "relationship"),
      tolerance = 1e-12
    )
    expect_equal(
      coan$MeanCoan,
      explicit_group_summary(1L, ped, "coancestry"),
      tolerance = 1e-12
    )
  }
})

test_that("pedrel no longer constructs or scans a relationship matrix", {
  ped <- make_pedrel_test_pedigree()
  local_mocked_bindings(
    pedmat = function(...) stop("pedmat() must not be called"),
    .package = "visPedigree"
  )

  rel <- pedrel(ped, by = "Cohort")
  coan <- pedrel(ped, by = "Cohort", scale = "coancestry", compact = TRUE)

  expect_true(all(rel$Status == "ok"))
  expect_true(all(coan$Status == "ok"))
})

test_that("pedrel processes more groups than one product batch", {
  n_groups <- 40L
  raw <- data.table(
    Ind = c("Sire", "Dam", sprintf("F%03d", seq_len(2L * n_groups))),
    Sire = c(NA, NA, rep("Sire", 2L * n_groups)),
    Dam = c(NA, NA, rep("Dam", 2L * n_groups)),
    Cohort = c(NA_integer_, NA_integer_, rep(seq_len(n_groups), each = 2L))
  )
  ped <- tidyped(raw)

  rel <- pedrel(ped, by = "Cohort")
  coan <- pedrel(ped, by = "Cohort", scale = "coancestry")

  expect_equal(rel$MeanRel, rep(0.5, n_groups))
  expect_equal(coan$MeanCoan, rep(0.375, n_groups))
  expect_true(all(rel$Status == "ok"))
  expect_true(all(coan$Status == "ok"))
})

test_that("pedrel validates compact", {
  ped <- make_pedrel_test_pedigree()

  expect_error(pedrel(ped, by = "Cohort", compact = NA), "compact")
})

# test-summary-pedmat-vismat-vispstat.R
# Basic tests for summary_pedmat, vismat, and vispstat

# --- summary_pedmat ---
test_that("summary_pedmat works on A matrix", {
  tp <- tidyped(simple_ped)
  A <- pedmat(tp, method = "A")
  s <- summary_pedmat(A)

  expect_s3_class(s, "summary.pedmat")
  expect_equal(s$method, "A")
  expect_false(s$compact)
  expect_equal(s$n_original, nrow(tp))
})

test_that("summary.pedmat S3 dispatch works", {
  tp <- tidyped(simple_ped)
  A <- pedmat(tp, method = "A")
  # Use summary_pedmat directly; generic summary() may dispatch to
  # summary.matrix before summary.pedmat depending on class order.
  s <- summary_pedmat(A)

  expect_s3_class(s, "summary.pedmat")
})

test_that("summary_pedmat errors on non-pedmat input", {
  expect_error(summary_pedmat(matrix(1:4, 2, 2)), "must be a pedmat")
})

# --- vismat ---
test_that("vismat produces a heatmap from pedmat object", {
  tp <- tidyped(simple_ped)
  A <- pedmat(tp, method = "A")

  # vismat returns a lattice/trellis object
  p <- vismat(A, type = "heatmap")
  expect_true(inherits(p, "trellis"))
})

test_that("vismat produces a histogram from pedmat object", {
  tp <- tidyped(simple_ped)
  A <- pedmat(tp, method = "A")

  p <- vismat(A, type = "histogram")
  expect_true(inherits(p, "trellis"))
})

test_that("vismat works with tidyped input", {
  tp <- tidyped(simple_ped)

  p <- vismat(tp, type = "heatmap")
  expect_true(inherits(p, "trellis"))
})

test_that("vismat reorder=FALSE preserves original order", {
  tp <- tidyped(small_ped)
  A <- pedmat(tp, method = "A")
  p <- vismat(A, reorder = FALSE)
  expect_true(inherits(p, "trellis"))
})

test_that("vismat ids subsets the matrix", {
  tp <- tidyped(small_ped)
  A <- pedmat(tp, method = "A")
  target <- rownames(as.matrix(A))[1:5]
  p <- vismat(A, ids = target)
  expect_true(inherits(p, "trellis"))
})

test_that("vismat ids errors on non-existent IDs", {
  tp <- tidyped(small_ped)
  A <- pedmat(tp, method = "A")
  expect_error(vismat(A, ids = c("NONEXIST_1", "NONEXIST_2")),
               "None of the specified")
})

test_that("vismat by groups by generation", {
  tp <- tidyped(small_ped)
  A <- pedmat(tp, method = "A")
  expect_message(
    p <- vismat(A, ped = tp, by = "Gen"),
    "Aggregating"
  )
  expect_true(inherits(p, "trellis"))
})

test_that("matrix-free pedigree grouping matches explicit A aggregation", {
  tp <- tidyped(small_ped)
  A <- as.matrix(pedmat(tp, method = "A", sparse = FALSE))
  groups <- as.character(tp$Gen)
  levels <- sort(unique(groups))
  sizes <- tabulate(factor(groups, levels = levels), nbins = length(levels))

  row_agg <- rowsum(A, groups, reorder = TRUE)
  expected_sum <- t(rowsum(t(row_agg), groups, reorder = TRUE))
  expected <- expected_sum / outer(sizes, sizes)

  actual <- aggregate_pedigree_by_group(tp, tp$Ind, "Gen")
  expect_equal(actual, expected, tolerance = 1e-12)
})

test_that("vismat grouped tidyped path does not construct A", {
  tp <- tidyped(small_ped)
  testthat::local_mocked_bindings(
    pedmat = function(...) stop("pedmat() must not be called"),
    .package = "visPedigree"
  )

  expect_s3_class(
    suppressMessages(vismat(tp, by = "Gen", reorder = FALSE)),
    "trellis"
  )
})

test_that("vismat by errors without ped", {
  tp <- tidyped(small_ped)
  A <- as.matrix(pedmat(tp, method = "A"))
  # Plain matrix has no ped attribute
  expect_error(vismat(A, by = "Gen"), "'ped' must be provided")
})

test_that("vismat grouping deprecated parameter works", {
  tp <- tidyped(small_ped)
  A <- pedmat(tp, method = "A")
  expect_warning(
    p <- vismat(A, ped = tp, grouping = "Gen"),
    "deprecated"
  )
  expect_true(inherits(p, "trellis"))
})

test_that("vismat works with plain matrix input", {
  tp <- tidyped(small_ped)
  A <- pedmat(tp, method = "A")
  A_dense <- as.matrix(A)
  class(A_dense) <- "matrix"  # ensure no pedmat class
  p <- vismat(A_dense)
  expect_true(inherits(p, "trellis"))
})

test_that("vismat labelcex controls label font size", {
  tp <- tidyped(small_ped)
  A <- pedmat(tp, method = "A")
  p <- vismat(A, labelcex = 0.3)
  expect_true(inherits(p, "trellis"))
})

test_that("vismat auto-expands compact pedmat", {
  tp <- tidyped(small_ped)
  A_compact <- pedmat(tp, method = "A", compact = TRUE)
  ci <- attr(A_compact, "call_info")

  # Should produce a message about expanding
  expect_message(
    p <- vismat(A_compact),
    "Expanding compact matrix"
  )
  expect_true(inherits(p, "trellis"))
})

test_that("vismat errors on unsupported type", {
  tp <- tidyped(small_ped)
  A <- pedmat(tp, method = "A")
  expect_error(vismat(A, type = "scatter"), "not supported")
})

# --- vispstat ---
test_that("vispstat produces ECG histogram", {
  tp <- tidyped(simple_ped)
  stats <- pedstats(tp)

  p <- vispstat(stats, type = "ecg")
  expect_true(inherits(p, "trellis"))
})

test_that("vispstat ECG supports FullGen metric", {
  tp <- tidyped(simple_ped)
  stats <- pedstats(tp)

  p <- vispstat(stats, type = "ecg", metric = "FullGen")
  expect_true(inherits(p, "trellis"))
})

test_that("vispstat errors on non-pedstats input", {
  expect_error(vispstat(list()), "must be a pedstats object")
})

test_that("plot.pedstats dispatches to vispstat for ecg", {
  tp <- tidyped(simple_ped)
  stats <- pedstats(tp)

  p <- plot(stats, type = "ecg")
  expect_true(inherits(p, "trellis"))
})

test_that("vispstat genint produces bar chart when gen_intervals exist", {
  tp <- tidyped(simple_ped)
  # Assign synthetic birth years so pedgenint can compute intervals
  tp$Year <- tp$Gen * 3 + 2000
  stats <- pedstats(tp, timevar = "Year")
  skip_if(is.null(stats$gen_intervals), "No generation intervals computed")

  p <- vispstat(stats, type = "genint")
  expect_true(inherits(p, "trellis"))
})

test_that("vispstat genint errors when gen_intervals is NULL", {
  tp <- tidyped(simple_ped)
  stats <- pedstats(tp, genint = FALSE)

  expect_error(vispstat(stats, type = "genint"),
               "No generation interval data found")
})

test_that("plot.pedstats dispatches genint correctly", {
  tp <- tidyped(simple_ped)
  tp$Year <- tp$Gen * 3 + 2000
  stats <- pedstats(tp, timevar = "Year")
  skip_if(is.null(stats$gen_intervals), "No generation intervals computed")

  p <- plot(stats, type = "genint")
  expect_true(inherits(p, "trellis"))
})

# --- vismat compact + by (fast aggregation path) ---
test_that("vismat compact+by=Gen matches expand-then-aggregate", {
  tp <- tidyped(small_ped)
  A_compact <- pedmat(tp, method = "A", compact = TRUE)
  A_full    <- pedmat(tp, method = "A", compact = FALSE)

  # Fast path: aggregate directly from compact
  agg_compact <- visPedigree:::aggregate_compact_by_group(
    A_compact, tp$Ind, "Gen", tp
  )

  # Reference: expand, then aggregate manually
  mat_exp <- as.matrix(expand_pedmat(A_compact))
  grp <- tp$Gen[match(rownames(mat_exp), tp$Ind)]
  grps <- sort(unique(grp))
  agg_ref <- matrix(0, length(grps), length(grps),
                    dimnames = list(grps, grps))
  for (i in seq_along(grps)) {
    for (j in i:length(grps)) {
      idx_i <- which(grp == grps[i])
      idx_j <- which(grp == grps[j])
      agg_ref[i, j] <- mean(mat_exp[idx_i, idx_j])
      agg_ref[j, i] <- agg_ref[i, j]
    }
  }

  expect_equal(agg_compact, agg_ref, tolerance = 1e-12)
})

test_that("vismat compact+by=Family matches expand-then-aggregate", {
  tp <- tidyped(small_ped)
  A_compact <- pedmat(tp, method = "A", compact = TRUE)

  # Focus on gen 3 (has 3 families)
  ids_g3 <- tp[Gen == 3, Ind]
  agg_compact <- visPedigree:::aggregate_compact_by_group(
    A_compact, ids_g3, "Family", tp
  )

  # Reference
  mat_exp <- as.matrix(expand_pedmat(A_compact))
  sub_ids <- ids_g3
  sub_mat <- mat_exp[sub_ids, sub_ids, drop = FALSE]
  grp <- tp$Family[match(sub_ids, tp$Ind)]
  grps <- sort(unique(grp[!is.na(grp)]))
  keep <- !is.na(grp)
  sub_mat <- sub_mat[keep, keep, drop = FALSE]
  grp <- grp[keep]
  agg_ref <- matrix(0, length(grps), length(grps),
                    dimnames = list(grps, grps))
  for (i in seq_along(grps)) {
    for (j in i:length(grps)) {
      idx_i <- which(grp == grps[i])
      idx_j <- which(grp == grps[j])
      agg_ref[i, j] <- mean(sub_mat[idx_i, idx_j])
      agg_ref[j, i] <- agg_ref[i, j]
    }
  }

  expect_equal(agg_compact, agg_ref, tolerance = 1e-12)
})

test_that("vismat compact+by produces valid trellis plot", {
  tp <- tidyped(small_ped)
  A_compact <- pedmat(tp, method = "A", compact = TRUE)

  expect_message(
    p <- vismat(A_compact, ped = tp, by = "Gen"),
    "Aggregating"
  )
  expect_true(inherits(p, "trellis"))
})

# Test edge cases and boundary conditions

test_that("pedmat rejects multiple methods", {
  skip_if_not_installed("visPedigree")
  tped <- tidyped(small_ped)
  
  expect_error(
    pedmat(tped, method = c("A", "D")),
    "Only a single method may be requested"
  )
})

test_that("pedmat rejects invalid methods", {
  skip_if_not_installed("visPedigree")
  tped <- tidyped(small_ped)
  
  expect_error(
    pedmat(tped, method = "invalid"),
    "Invalid method"
  )
})

test_that("query_relationship rejects inverse matrices", {
  skip_if_not_installed("visPedigree")
  tped <- tidyped(small_ped)
  Ainv <- pedmat(tped, method = "Ainv")
  
  expect_error(
    query_relationship(Ainv, "A", "B"),
    "does not support inverse matrices"
  )
})

test_that("vismat rejects inverse matrices", {
  skip_if_not_installed("visPedigree")
  tped <- tidyped(small_ped)
  Ainv <- pedmat(tped, method = "Ainv")
  
  expect_error(
    vismat(Ainv),
    "does not support inverse matrices"
  )
})

test_that("compact mode expand_pedmat works for vectors", {
  skip_if_not_installed("visPedigree")
  tped <- tidyped(small_ped)
  
  # Get inbreeding coefficients in compact mode
  f_compact <- pedmat(tped, method = "f", compact = TRUE)
  f_full <- pedmat(tped, method = "f", compact = FALSE)
  
  # Expand compact version
  f_expanded <- expand_pedmat(f_compact)
  
  # Should have same length as original pedigree
  expect_equal(length(f_expanded), nrow(tped))
  
  # Names should match original individuals
  expect_equal(names(f_expanded), tped$Ind)
})

test_that("compact mode query_relationship handles full-siblings", {
  skip_if_not_installed("visPedigree")
  tped <- tidyped(small_ped)
  
  # C, D, E are full-siblings (AxB)
  A_compact <- pedmat(tped, method = "A", compact = TRUE)
  A_full <- pedmat(tped, method = "A", compact = FALSE)
  
  # Full-sibling relationship should be 0.5
  expect_equal(query_relationship(A_compact, "C", "D"), 0.5)
  expect_equal(query_relationship(A_compact, "C", "E"), 0.5)
  expect_equal(query_relationship(A_compact, "D", "E"), 0.5)
  
  # Should match full matrix
  expect_equal(
    query_relationship(A_compact, "C", "D"),
    query_relationship(A_full, "C", "D")
  )
})

test_that("D matrix compact mode full-sibling query works", {
  skip_if_not_installed("visPedigree")
  tped <- tidyped(small_ped)
  
  # D matrix in compact mode
  D_compact <- pedmat(tped, method = "D", compact = TRUE)
  D_full <- pedmat(tped, method = "D", compact = FALSE)
  
  # Query full-siblings
  d_cd_compact <- query_relationship(D_compact, "C", "D")
  d_cd_full <- query_relationship(D_full, "C", "D")
  
  # Should match (or be close due to formula)
  expect_equal(d_cd_compact, d_cd_full, tolerance = 1e-6)
})

test_that("sex conflict detection works", {
  skip_if_not_installed("visPedigree")
  
  # Create pedigree with sex conflict
  bad_ped <- data.frame(
    Ind = c("A", "B", "C"),
    Sire = c(NA, NA, "A"),
    Dam = c(NA, NA, "A"),  # A appears as both Sire and Dam!
    stringsAsFactors = FALSE
  )
  
  expect_error(
    tidyped(bad_ped),
    "Sex conflict detected"
  )
  
  # Same pedigree should work with selfing = TRUE
  res <- tidyped(bad_ped, selfing = TRUE)
  expect_s3_class(res, "tidyped")
  expect_equal(res[Ind == "A", Sex], "monoecious")
  expect_true(isTRUE(attr(res, "ped_meta")$selfing))
})

test_that("sex annotation conflict detection works", {
  skip_if_not_installed("visPedigree")
  
  # Create pedigree with explicit sex conflict
  bad_ped <- data.frame(
    Ind = c("A", "B", "C"),
    Sire = c(NA, NA, "A"),
    Dam = c(NA, NA, "B"),
    Sex = c("female", "male", NA),  # A is female but used as Sire!
    stringsAsFactors = FALSE
  )
  
  expect_error(
    tidyped(bad_ped),
    "Sex annotation conflicts"
  )
})

test_that("query_relationship id2=NULL returns correctly named vector", {
  skip_if_not_installed("visPedigree")
  tped <- tidyped(small_ped)
  A_compact <- pedmat(tped, method = "A", compact = TRUE)
  
  # Query row for individual A
  row_a <- query_relationship(A_compact, "A")
  
  # Should have names
  expect_true(!is.null(names(row_a)))
  
  # Length should match matrix dimension
  mat <- A_compact
  class(mat) <- setdiff(class(mat), "pedmat")
  expect_equal(length(row_a), ncol(mat))
  
  # Names should match matrix column names
  expect_equal(names(row_a), colnames(mat))
})

test_that("pedmat threads argument works correctly", {
  # Load small data
  tped <- tidyped(small_ped)
  
  # 1. Invalid input handling
  expect_error(pedmat(tped, threads = "invalid"), "'threads' must be a single non-negative integer")
  expect_error(pedmat(tped, threads = -1), "'threads' must be a single non-negative integer")
  
  # 2. OpenMP availability check
  # We can't easily force OpenMP on/off, but we can check if it warns when threads > 0 
  # and OpenMP is reported as unavailable
  
  # Check if OpenMP is available (internal helper)
  has_openmp <- visPedigree:::cpp_openmp_available()
  
  if (!has_openmp) {
    # Expect warning if threads requested but not available
    expect_warning(
      pedmat(tped, method = "A", threads = 2),
      "OpenMP is not available"
    )
    
    # Expect NO warning if threads=0 or 1 (default/serial)
    # Note: threads=1 might still trigger code paths but usually doesn't warn if implementation is smart,
    # but the current implementation warns if threads > 0 and no OpenMP.
    # Actually code says: if (threads > 0) { if (!available) warning... }
    # So threads=1 SHOULD warn if OpenMP is missing according to current logic? 
    # Let's verify code:
    # if (threads > 0) {
    #   if (!cpp_openmp_available()) {
    #     warning(...)
    #   }
    # }
    expect_warning(
        pedmat(tped, method="A", threads = 1),
        "OpenMP is not available"
    )
    
    expect_silent(pedmat(tped, method="A", threads = 0))
    
  } else {
    # If OpenMP IS available
    expect_silent(pedmat(tped, method = "A", threads = 2))
    expect_silent(pedmat(tped, method = "A", threads = 0))
    
    # We can't easily test speedup on small_ped, but we can verify it runs
    res <- pedmat(tped, method = "A", threads = 2)
    # Check if it is a valid pedmat object (S3 class or S4 with marker)
    is_pedmat <- inherits(res, "pedmat") || !is.null(attr(res, "pedmat_S4"))
    expect_true(is_pedmat)
  }
})

# Tests for compact pedigree matrix functionality

# Helper to create a pedigree with large full-sibling families
create_fullsib_ped <- function(n_sires = 3, n_dams = 4, offspring_per_family = 10) {
  ped_list <- list()
  ind_counter <- 1

  sires <- paste0("S", sprintf("%02d", seq_len(n_sires)))
  ped_list[[length(ped_list) + 1]] <- data.table::data.table(
    Ind = sires, Sire = NA, Dam = NA, Sex = "male"
  )

  dams <- paste0("D", sprintf("%02d", seq_len(n_dams)))
  ped_list[[length(ped_list) + 1]] <- data.table::data.table(
    Ind = dams, Sire = NA, Dam = NA, Sex = "female"
  )

  for (s in seq_len(n_sires)) {
    for (d in seq_len(n_dams)) {
      ids <- paste0("O", sprintf("%04d", ind_counter:(ind_counter + offspring_per_family - 1)))
      ind_counter <- ind_counter + offspring_per_family
      sexes <- rep(c("male", "female"), length.out = offspring_per_family)
      ped_list[[length(ped_list) + 1]] <- data.table::data.table(
        Ind = ids, Sire = sires[s], Dam = dams[d], Sex = sexes
      )
    }
  }
  data.table::rbindlist(ped_list)
}

test_that("compact pedigree reduces dimensions", {
  ped <- create_fullsib_ped(n_sires = 3, n_dams = 4, offspring_per_family = 20)
  tped <- tidyped(ped, addnum = TRUE)
  n_orig <- nrow(tped)

  f_compact <- pedmat(tped, method = "f", compact = TRUE)

  ci <- attr(f_compact, "call_info")
  expect_true(ci$compact)
  expect_equal(ci$n_original, n_orig)
  expect_true(ci$n_compact < n_orig)
})

test_that("compact f matches non-compact f for all individuals", {
  ped <- create_fullsib_ped(n_sires = 2, n_dams = 3, offspring_per_family = 5)
  tped <- tidyped(ped, addnum = TRUE)

  f_full <- pedmat(tped, method = "f", compact = FALSE)
  f_compact <- pedmat(tped, method = "f", compact = TRUE)

  # Expand compact and compare
  f_expanded <- expand_pedmat(f_compact)
  expect_equal(length(f_expanded), length(f_full))
  expect_equal(as.numeric(f_expanded), as.numeric(f_full), tolerance = 1e-10)
})

test_that("compact A matrix expand matches non-compact A", {
  ped <- create_fullsib_ped(n_sires = 2, n_dams = 2, offspring_per_family = 5)
  tped <- tidyped(ped, addnum = TRUE)

  A_full <- pedmat(tped, method = "A", compact = FALSE, sparse = FALSE)
  A_compact <- pedmat(tped, method = "A", compact = TRUE, sparse = FALSE)

  A_expanded <- expand_pedmat(A_compact)

  expect_equal(dim(A_expanded), dim(A_full))
  expect_equal(as.matrix(A_expanded), as.matrix(A_full), tolerance = 1e-10, ignore_attr = TRUE)
})

test_that("compact_map and family_summary are well-formed", {
  ped <- create_fullsib_ped(n_sires = 2, n_dams = 3, offspring_per_family = 8)
  tped <- tidyped(ped, addnum = TRUE)

  f_compact <- pedmat(tped, method = "f", compact = TRUE)

  compact_map <- attr(f_compact, "compact_map")
  family_summary <- attr(f_compact, "family_summary")
  compact_stats <- attr(f_compact, "compact_stats")

  # compact_map should cover all original individuals
  expect_true(nrow(compact_map) >= nrow(tped))

  # family_summary should have one row per full-sib family
  expect_true(nrow(family_summary) > 0)
  expect_true("FamilyID" %in% names(family_summary))
  expect_true("RepInd" %in% names(family_summary))

  # compact_stats should report compression
  expect_equal(compact_stats$n_original, nrow(tped))
  expect_true(compact_stats$compression_ratio < 1)
  expect_true(compact_stats$memory_saved_pct > 0)
})

test_that("query_relationship works for full-siblings in compact mode", {
  tped <- tidyped(small_ped)

  A_compact <- pedmat(tped, method = "A", compact = TRUE)
  A_full <- pedmat(tped, method = "A", compact = FALSE)

  # C, D, E are full-sibs (AxB)
  val_compact <- query_relationship(A_compact, "C", "D")
  val_full <- query_relationship(A_full, "C", "D")
  expect_equal(val_compact, val_full, tolerance = 1e-10)
})

test_that("query_relationship returns NA for unknown individuals", {
  tped <- tidyped(small_ped)
  A <- pedmat(tped, method = "A", compact = TRUE)
  expect_true(is.na(query_relationship(A, "NONEXISTENT", "A")))
})

test_that("expand_pedmat returns unchanged for non-compact", {
  tped <- tidyped(small_ped)
  A <- pedmat(tped, method = "A", compact = FALSE)
  A2 <- expand_pedmat(A)
  expect_equal(dim(A), dim(A2))
})

test_that("compact mode works with D matrix", {
  ped <- create_fullsib_ped(n_sires = 2, n_dams = 2, offspring_per_family = 5)
  tped <- tidyped(ped, addnum = TRUE)

  D_compact <- pedmat(tped, method = "D", compact = TRUE)
  D_full <- pedmat(tped, method = "D", compact = FALSE)

  # Query a specific pair
  val_compact <- query_relationship(D_compact, "S01", "D01")
  val_full <- query_relationship(D_full, "S01", "D01")
  expect_equal(val_compact, val_full, tolerance = 1e-6)
})

test_that("summary_pedmat works for compact matrices", {
  ped <- create_fullsib_ped(n_sires = 2, n_dams = 3, offspring_per_family = 10)
  tped <- tidyped(ped, addnum = TRUE)

  A_compact <- pedmat(tped, method = "A", compact = TRUE)
  s <- summary_pedmat(A_compact)

  expect_s3_class(s, "summary.pedmat")
  expect_true(s$compact)
  expect_true(!is.null(s$compact_stats))
  expect_output(print(s), "Compaction Results")
})

test_that("pedigree with no full-sibs returns uncompacted", {
  ped <- data.frame(
    Ind = c("A", "B", "C", "D"),
    Sire = c(NA, NA, "A", "A"),
    Dam = c(NA, NA, "B", NA),
    Sex = c("male", "female", "male", "female"),
    stringsAsFactors = FALSE
  )
  tped <- tidyped(ped, addnum = TRUE)

  f_compact <- pedmat(tped, method = "f", compact = TRUE)
  ci <- attr(f_compact, "call_info")

  # No full-sibs to compact, so n_compact should equal n_original
  expect_true(ci$n_compact <= ci$n_original)
})

# Tests for pedmat() core functionality

test_that("pedmat computes A matrix correctly for small pedigree", {
  tped <- tidyped(small_ped)
  A <- pedmat(tped, method = "A")

  # Should be a pedmat object (S4 sparse with marker)
  is_pedmat <- inherits(A, "pedmat") || !is.null(attr(A, "pedmat_S4"))
  expect_true(is_pedmat)

  # Diagonal should be 1 + F for all individuals
  f <- pedmat(tped, method = "f")
  expect_equal(as.numeric(Matrix::diag(A)), 1 + as.numeric(f), tolerance = 1e-10)

  # Symmetry
  A_mat <- as.matrix(A)
  expect_equal(A_mat, t(A_mat), tolerance = 1e-10)

  # Founders: self-relationship = 1 (no inbreeding)
  expect_equal(A_mat["A", "A"], 1.0)
  expect_equal(A_mat["B", "B"], 1.0)

  # Parent-offspring relationship = 0.5
  expect_equal(A_mat["A", "C"], 0.5)
  expect_equal(A_mat["B", "C"], 0.5)
})

test_that("pedmat computes f (inbreeding) correctly", {
  tped <- tidyped(small_ped)
  f <- pedmat(tped, method = "f")

  # Returns named vector
  expect_true(is.numeric(f))
  expect_true(!is.null(names(f)))
  expect_equal(length(f), nrow(tped))

  # Founders have f = 0
  founders <- tped$Ind[is.na(tped$Sire) & is.na(tped$Dam)]
  expect_true(all(f[founders] == 0))

  # All f values >= 0

  expect_true(all(f >= 0))
})

test_that("pedmat computes Ainv correctly", {
  tped <- tidyped(small_ped)
  A <- pedmat(tped, method = "A", sparse = FALSE)
  Ainv <- pedmat(tped, method = "Ainv", sparse = FALSE)

  # A %*% Ainv should approximate identity
  product <- as.matrix(A %*% Ainv)
  n <- nrow(product)
  I <- diag(n)
  expect_equal(unname(product), I, tolerance = 1e-6)
})

test_that("pedmat computes D matrix correctly", {
  tped <- tidyped(small_ped)
  D <- pedmat(tped, method = "D")

  is_pedmat <- inherits(D, "pedmat") || !is.null(attr(D, "pedmat_S4"))
  expect_true(is_pedmat)

  # D matrix should be symmetric
  D_mat <- as.matrix(D)
  expect_equal(D_mat, t(D_mat), tolerance = 1e-10)

  # Diagonal of D should be 1 for non-inbred individuals
  # D_ii = 1 - F_i^2 for most formulations, or 1 for founders
  expect_true(all(diag(D_mat) > 0))
})

test_that("pedmat computes Dinv correctly", {
  tped <- tidyped(small_ped)
  D <- pedmat(tped, method = "D", sparse = FALSE)
  Dinv <- pedmat(tped, method = "Dinv", sparse = FALSE)

  product <- as.matrix(D %*% Dinv)
  n <- nrow(product)
  I <- diag(n)
  expect_equal(unname(product), I, tolerance = 1e-4)
})

test_that("pedmat computes AA matrix correctly", {
  tped <- tidyped(small_ped)
  A <- pedmat(tped, method = "A", sparse = FALSE)
  AA <- pedmat(tped, method = "AA", sparse = FALSE)

  # AA = A # A (Hadamard product)
  A_mat <- as.matrix(A)
  AA_mat <- as.matrix(AA)
  expect_equal(AA_mat, A_mat * A_mat, tolerance = 1e-10, ignore_attr = TRUE)
})

test_that("pedmat computes AAinv correctly", {
  tped <- tidyped(small_ped)
  AA <- pedmat(tped, method = "AA", sparse = FALSE)
  AAinv <- pedmat(tped, method = "AAinv", sparse = FALSE)

  product <- as.matrix(AA %*% AAinv)
  n <- nrow(product)
  I <- diag(n)
  expect_equal(unname(product), I, tolerance = 1e-4)
})

test_that("pedmat sparse vs dense produce equivalent results", {
  tped <- tidyped(small_ped)

  A_sparse <- pedmat(tped, method = "A", sparse = TRUE)
  A_dense <- pedmat(tped, method = "A", sparse = FALSE)

  expect_true(inherits(A_sparse, "Matrix") || !is.null(attr(A_sparse, "pedmat_S4")))
  expect_true(is.matrix(as.matrix(A_dense)))

  expect_equal(as.matrix(A_sparse), as.matrix(A_dense), tolerance = 1e-10, ignore_attr = TRUE)
})

test_that("pedmat attaches correct metadata", {
  tped <- tidyped(small_ped)
  A <- pedmat(tped, method = "A")

  ci <- attr(A, "call_info")
  expect_equal(ci$method, "A")
  expect_false(ci$compact)
  expect_equal(ci$n_original, nrow(tped))
  expect_true(!is.null(ci$timestamp))

  expect_equal(attr(A, "method"), "A")
  expect_true(!is.null(attr(A, "ped")))
})

test_that("pedmat rejects splitped objects", {
  tped <- tidyped(small_ped)
  sp <- splitped(tped)

  expect_error(pedmat(sp, method = "A"), "does not support 'splitped'")
})

test_that("pedmat S3 methods work", {
  tped <- tidyped(small_ped)
  A <- pedmat(tped, method = "A")
  n <- nrow(tped)

  # dim
  expect_equal(dim(A), c(n, n))

  # subsetting
  sub <- A[1:3, 1:3]
  expect_equal(nrow(sub), 3)

  # summary_pedmat
  s <- summary_pedmat(A)
  expect_s3_class(s, "summary.pedmat")
  expect_equal(s$method, "A")

  # S4 sparse matrices use pedmat_S4 marker; S3 vectors use pedmat class
  f <- pedmat(tped, method = "f")
  expect_output(print(f), "Pedigree Matrix")
})

test_that("pedmat invert_method options work", {
  tped <- tidyped(small_ped)

  Dinv_auto <- pedmat(tped, method = "Dinv", invert_method = "auto", sparse = FALSE)
  Dinv_general <- pedmat(tped, method = "Dinv", invert_method = "general", sparse = FALSE)

  expect_equal(as.matrix(Dinv_auto), as.matrix(Dinv_general), tolerance = 1e-6)

  expect_error(pedmat(tped, method = "Dinv", invert_method = "invalid"))
})

test_that("pedmat input validation works", {
  tped <- tidyped(small_ped)

  expect_error(pedmat(tped, method = c("A", "D")), "single method")
  expect_error(pedmat(tped, method = "invalid"), "Invalid method")
  expect_error(pedmat(tped, threads = -1), "non-negative integer")
  expect_error(pedmat(tped, threads = "bad"), "non-negative integer")
})

# ---------------------------------------------------------------------------
# vismat internal thresholds
# ---------------------------------------------------------------------------
# Centralised here so that documentation, code, and tests stay consistent.
# All values are the *inclusive upper bound* for the "lighter" behaviour,
# i.e. the condition that triggers the heavier/simpler fallback is "> MAX".

VISMAT_REORDER_MAX      <- 2000L   # N above this: skip hierarchical clustering
VISMAT_LABEL_MAX        <- 500L    # N above this: hide individual labels
VISMAT_GRID_MAX         <- 100L    # N above this: hide cell grid lines
VISMAT_EXPAND_MAX       <- 5000L   # original N above this: use compact representative view
VISMAT_BLOCK_AGG_MIN_N  <- 10000L  # aggregation switches to block algorithm when
VISMAT_BLOCK_AGG_MIN_G  <- 100L    #   n > _MIN_N  AND  n_grp > _MIN_G

#' Visualize Relationship Matrices
#'
#' @description
#' \code{vismat} provides visualization tools for relationship matrices (A, D, AA),
#' supporting individual-level heatmaps and relationship coefficient histograms.
#' This function is useful for exploring population genetic structure, identifying
#' inbred individuals, and analyzing kinship between families.
#'
#' @param mat A relationship matrix. Can be one of the following types:
#' \itemize{
#'   \item A \code{pedmat} object returned by \code{\link{pedmat}} — including
#'         compact matrices. When \code{by} is specified, group-level means are
#'         computed directly from the compact matrix (no full expansion needed).
#'         Without \code{by}, compact matrices are automatically expanded to
#'         full dimensions before plotting (see Details).
#'   \item A \code{\link{tidyped}} object (automatically calculates additive
#'         relationship matrix A).
#'   \item A named list containing matrices (preferring A, D, AA).
#'   \item A standard \code{matrix} or \code{Matrix} object.
#' }
#' \strong{Note}: Inverse matrices (Ainv, Dinv, AAinv) are not supported for 
#' visualization because their elements do not represent meaningful relationship
#' coefficients.
#'
#' @param ped Optional. A tidied pedigree object (\code{tidyped}), used for 
#' extracting labels or grouping information. Required when using the 
#' \code{by} parameter with a plain matrix input. If \code{mat} is a 
#' \code{pedmat} object, the pedigree is extracted automatically.
#'
#' @param type Character, type of visualization. Supported options:
#' \itemize{
#'   \item \code{"heatmap"}: Relationship matrix heatmap (default). Uses a 
#'         Nature Genetics style color palette (white-orange-red-dark red),
#'         with optional hierarchical clustering and group aggregation.
#'   \item \code{"histogram"}: Distribution histogram of relationship coefficients.
#'         Shows the frequency distribution of lower triangular elements (pairwise kinship).
#' }
#'
#' @param ids Character vector specifying individual IDs to display. Used to 
#' filter and display a submatrix of specific individuals. If \code{NULL} 
#' (default), all individuals are shown.
#'
#' @param reorder Logical. If \code{TRUE} (default), rows and columns are 
#' reordered using hierarchical clustering (Ward.D2 method) to bring closely 
#' related individuals together. Only affects heatmap visualization. 
#' Automatically skipped for large matrices (N > VISMAT_REORDER_MAX, default 2 000)
#' to improve performance.
#' 
#' \strong{Clustering principle}: Based on relationship profile distance (Euclidean
#' distance between rows). Full-sibs have nearly identical relationship profiles
#' with the whole population, so they cluster tightly together and appear as
#' contiguous blocks in the heatmap.
#'
#' @param by Optional. Column name in \code{ped} to group by (e.g., 
#' \code{"Family"}, \code{"Gen"}, \code{"Year"}). When grouping is enabled:
#' \itemize{
#'   \item Individual-level matrix is aggregated to a group-level matrix 
#'         (computing mean relationship coefficients between groups).
#'   \item For \code{"Family"} grouping, founders without family assignment are excluded.
#'   \item For other grouping columns, NA values are assigned to an \code{"Unknown"} group.
#' }
#' Useful for visualizing population structure in large pedigrees.
#'
#' @param grouping \code{[Deprecated]} Use \code{by} instead.
#'
#' @param labelcex Numeric. Manual control for font size of individual labels. 
#' If \code{NULL} (default), uses a dynamic font size that adjusts automatically 
#' based on matrix dimensions (range 0.2–0.7). Labels are hidden automatically
#' when N > VISMAT_LABEL_MAX (default 500).
#'
#' @param ... Additional arguments passed to the underlying plotting function:
#' \itemize{
#'   \item Heatmap uses \code{\link[lattice]{levelplot}}: can set \code{main},
#'         \code{xlab}, \code{ylab}, \code{col.regions}, \code{colorkey},
#'         \code{scales}, etc.
#'   \item Histogram uses \code{\link[lattice]{histogram}}: can set \code{main},
#'         \code{xlab}, \code{ylab}, \code{nint} (number of bins), etc.
#' }
#'
#' @details
#' \subsection{Compact Matrix Handling}{
#' When \code{mat} is a compact \code{pedmat} object (created with
#' \code{pedmat(..., compact = TRUE)}):
#' \itemize{
#'   \item \strong{With \code{by}}: Group-level mean relationships are computed
#'         algebraically from the K×K compact matrix, including a correction for
#'         sibling off-diagonal values. This avoids expanding to the full N×N
#'         matrix, making family-level or generation-level visualization feasible
#'         even for pedigrees with hundreds of thousands of individuals.
#'   \item \strong{Without \code{by}, N > VISMAT_EXPAND_MAX (5 000)}: The compact K\eqn{\times}K matrix
#'         is plotted directly using representative individuals. Labels show the
#'         number of individuals each representative stands for, e.g.,
#'         \code{"ID (\u00d7350)"}. This avoids memory-intensive full expansion.
#'   \item \strong{Without \code{by}, N \eqn{\le} 5 000}: The compact matrix is
#'         expanded via \code{\link{expand_pedmat}} to restore full dimensions.
#' }
#' }
#'
#' \subsection{Heatmap}{
#' \itemize{
#'   \item Uses a Nature Genetics style color palette (white to orange to red to dark red).
#'   \item Hierarchical clustering reordering (Ward.D2) is enabled by default.
#'   \item Grid lines shown when N \eqn{\le} VISMAT_GRID_MAX (100);
#'         labels shown when N \eqn{\le} VISMAT_LABEL_MAX (500).
#'   \item \code{mat[1,1]} is displayed at the top-left corner.
#' }
#' }
#'
#' \subsection{Histogram}{
#' \itemize{
#'   \item Shows the distribution of lower-triangular elements (pairwise kinship).
#'   \item X-axis: relationship coefficient values; Y-axis: frequency percentage.
#' }
#' }
#'
#' \subsection{Performance}{
#' The following automatic thresholds are defined as package-internal
#' constants (\code{VISMAT_*}) at the top of \code{R/vismat.R}:
#' \itemize{
#'   \item Passing a \code{tidyped} object together with \code{by} uses
#'         matrix-free additive relationship products to compute the grouped
#'         heatmap directly, without constructing the full individual-level
#'         relationship matrix.
#'   \item \code{VISMAT_EXPAND_MAX} (5 000): compact matrices with original
#'         N above this are shown in representative view instead of expanding.
#'   \item \code{VISMAT_REORDER_MAX} (2 000): hierarchical clustering is
#'         automatically skipped.
#'   \item \code{VISMAT_LABEL_MAX} (500): individual labels are hidden.
#'   \item \code{VISMAT_GRID_MAX} (100): cell grid lines are hidden.
#'   \item \code{by} grouping uses vectorized \code{rowsum()} algebra — suitable
#'         for large matrices.
#' }
#' }
#'
#' \subsection{Interpreting Relationship Coefficients}{
#' For the additive relationship matrix A:
#' \itemize{
#'   \item Diagonal elements = 1 + F (F = inbreeding coefficient).
#'   \item Off-diagonal elements = 2 × kinship coefficient.
#'   \item 0: unrelated; 0.25: half-sibs / grandparent–grandchild;
#'         0.5: full-sibs / parent–offspring; 1.0: same individual.
#' }
#' }
#'
#' @return Invisibly returns the \code{lattice} plot object. The plot is 
#' rendered on the current graphics device.
#'
#' @seealso 
#' \code{\link{pedmat}} for computing relationship matrices,
#' \code{\link{expand_pedmat}} for manually restoring compact matrix dimensions,
#' \code{\link{query_relationship}} for querying individual pairs,
#' \code{\link{tidyped}} for tidying pedigree data,
#' \code{\link{visped}} for visualizing pedigree structure graphs,
#' \code{\link[lattice]{levelplot}}, \code{\link[lattice]{histogram}}
#'
#' @examples
#' library(visPedigree)
#' data(small_ped)
#' ped <- tidyped(small_ped)
#'
#' # ============================================================
#' # Basic Usage
#' # ============================================================
#'
#' # Method 1: from tidyped object (auto-computes A)
#' vismat(ped)
#'
#' # Method 2: from pedmat object
#' A <- pedmat(ped)
#' vismat(A)
#'
#' # Method 3: from plain matrix
#' vismat(as.matrix(A))
#'
#' # ============================================================
#' # Compact Pedigree (auto-expanded before plotting)
#' # ============================================================
#'
#' # For pedigrees with large full-sib families, compute a compact matrix
#' # first for efficiency, then pass directly to vismat() — it automatically
#' # expands back to full dimensions.
#' A_compact <- pedmat(ped, compact = TRUE)
#' vismat(A_compact)   # prints: "Expanding compact matrix (N -> M individuals)"
#'
#' # For very large pedigrees, aggregate to a group-level view instead
#' vismat(A, ped = ped, by = "Gen",
#'        main = "Mean Relationship Between Generations")
#'
#' # ============================================================
#' # Heatmap Customization
#' # ============================================================
#'
#' # Custom title and axis labels
#' vismat(A, main = "Additive Relationship Matrix",
#'        xlab = "Individual", ylab = "Individual")
#'
#' # Preserve original pedigree order (no clustering)
#' vismat(A, reorder = FALSE)
#'
#' # Custom label font size
#' vismat(A, labelcex = 0.5)
#'
#' # Custom color palette (blue-white-red)
#' vismat(A, col.regions = colorRampPalette(c("blue", "white", "red"))(100))
#'
#' # ============================================================
#' # Display a Subset of Individuals
#' # ============================================================
#'
#' target_ids <- rownames(A)[1:8]
#' vismat(A, ids = target_ids)
#'
#' # ============================================================
#' # Histogram of Relationship Coefficients
#' # ============================================================
#'
#' vismat(A, type = "histogram")
#' vismat(A, type = "histogram", nint = 30)
#'
#' # ============================================================
#' # Group-level Aggregation
#' # ============================================================
#'
#' # Group by generation
#' vismat(A, ped = ped, by = "Gen",
#'        main = "Mean Relationship Between Generations")
#'
#' # Group by full-sib family (founders without a family are excluded)
#' vismat(A, ped = ped, by = "Family")
#'
#' # ============================================================
#' # Other Relationship Matrices
#' # ============================================================
#'
#' # Dominance relationship matrix
#' D <- pedmat(ped, method = "D")
#' vismat(D, main = "Dominance Relationship Matrix")
#' 
#' @export
#'
#' @importFrom grDevices colorRampPalette
#' @importFrom stats as.dist hclust
#' @importFrom lattice levelplot panel.levelplot panel.abline histogram
#' @importFrom data.table as.data.table
vismat <- function(mat, ped = NULL, type = "heatmap", ids = NULL, reorder = TRUE, by = NULL, grouping = NULL, labelcex = NULL, ...) {
  # Backward-compatible deprecation for 'grouping'
  if (!is.null(grouping)) {
    warning("'grouping' is deprecated in vismat(), use 'by' instead.", call. = FALSE)
    if (is.null(by)) by <- grouping
  }
  
  # Track original by label for axis labels (may be consumed early in compact path)
  by_label <- by
  grouping_main <- NULL

  # A grouped additive heatmap only needs W' A W, where W is the group
  # membership matrix. Compute that product directly from the pedigree rather
  # than materializing the individual-level A matrix.
  if (is_tidyped(mat) && !is.null(by)) {
    ped <- mat
    focal_ids <- if (!is.null(ids)) as.character(ids) else ped$Ind
    mat <- aggregate_pedigree_by_group(ped, focal_ids, by)
    grouping_main <- sprintf("Grouped Relationship Heatmap (%s)", by)
    by <- NULL
    ids <- NULL
  }
  
  # 0a. Extract ped from pedmat object if available
  is_pedmat <- inherits(mat, "pedmat") || !is.null(attr(mat, "pedmat_S4"))
  if (is_pedmat) {
    # Check for inverse matrices - not supported for visualization
    ci <- attr(mat, "call_info")
    if (!is.null(ci) && ci$method[1] %in% c("Ainv", "Dinv", "AAinv")) {
      stop(sprintf(
        paste0("vismat() does not support inverse matrices (%s). ",
               "Inverse matrix elements do not represent meaningful relationship ",
               "coefficients for visualization."),
        ci$method[1]
      ), call. = FALSE)
    }
    
    # Handle compact matrices
    if (isTRUE(ci$compact)) {
      if (!is.null(by)) {
        # Fast path: aggregate directly from K×K compact matrix → G×G
        if (is.null(ped)) ped <- ci$ped_original
        
        # If ids not specified, default to all individuals
        focal_ids <- if (!is.null(ids)) as.character(ids) else ci$ped_original$Ind
        
        agg_mat <- aggregate_compact_by_group(mat, focal_ids, by, ped)
        
        # Replace mat with the G×G result; disable downstream by/ids processing
        mat <- agg_mat
        grouping_main <- sprintf("Grouped Relationship Heatmap (%s)", by)
        by  <- NULL
        ids <- NULL
        
        # Strip pedmat class
        if (inherits(mat, "pedmat")) {
          class(mat) <- setdiff(class(mat), "pedmat")
        }
      } else {
        # No grouping: decide between expand vs compact representative view
        if (ci$n_original > VISMAT_EXPAND_MAX) {
          # Large pedigree: use compact K×K representative view directly
          message(sprintf(
            paste0("Using compact representative view (%d representatives ",
                   "for %d individuals). Use 'by' for group-level aggregation."),
            ci$n_compact, ci$n_original
          ))
          if (is.null(ped)) ped <- ci$ped_original
          cmap <- attr(mat, "compact_map")
          
          # Build informative labels: "ID (×n)" for representatives with siblings
          rep_rows <- cmap[cmap$IsRepresentative == TRUE, ]
          rep_rows <- rep_rows[match(rownames(mat), rep_rows$Ind), ]
          new_labels <- ifelse(
            rep_rows$FamilySize > 1,
            sprintf("%s (\u00d7%d)", rep_rows$Ind, rep_rows$FamilySize),
            rep_rows$Ind
          )
          
          # Strip pedmat class and apply labels
          if (inherits(mat, "pedmat")) {
            class(mat) <- setdiff(class(mat), "pedmat")
          }
          rownames(mat) <- new_labels
          colnames(mat) <- new_labels
          
          # Map ids to representative IDs if provided
          if (!is.null(ids)) {
            ids_char <- as.character(ids)
            rep_map <- cmap[cmap$Ind %chin% ids_char, ]
            rep_ids <- unique(rep_map$RepInd)
            rep_labels <- new_labels[match(rep_ids, rep_rows$Ind)]
            rep_labels <- rep_labels[!is.na(rep_labels)]
            if (length(rep_labels) == 0) {
              stop("None of the specified 'ids' were found in the compact matrix.")
            }
            ids <- rep_labels
          }
        } else {
          # Small enough to expand to full N×N
          message(sprintf(
            "Expanding compact matrix (%d -> %d individuals) for visualization.",
            ci$n_compact, ci$n_original
          ))
          if (is.null(ped)) ped <- ci$ped_original
          mat <- expand_pedmat(mat)
          # Strip pedmat class to get raw matrix
          if (inherits(mat, "pedmat")) {
            class(mat) <- setdiff(class(mat), "pedmat")
          }
        }
      }
    } else {
      if (is.null(ped)) ped <- attr(mat, "ped")
      # Strip pedmat class to get raw matrix
      if (inherits(mat, "pedmat")) {
        class(mat) <- setdiff(class(mat), "pedmat")
      }
    }
  }
  
  # 0b. If input is a tidyped object, calculate A first
  if (is_tidyped(mat)) {
    ped <- mat
    res <- pedmat(ped, method = "A")
    # Now res is a pure matrix with pedmat class
    if (is.null(ped)) {
      ped <- attr(res, "ped")
    }
    class(res) <- setdiff(class(res), "pedmat")
    mat <- res
  }

  # 1. Handle list input from pedmat
  if (is.list(mat) && !is.matrix(mat) && !inherits(mat, "Matrix")) {
    # Reject inverse matrices - primarily via list element names
    # Note: Due to S4 Matrix limitations, unnamed/custom-named lists may bypass this check
    inverse_methods <- c("Ainv", "Dinv", "AAinv")
    
    # Check list element names
    has_inverse_name <- any(names(mat) %in% inverse_methods)
    
    if (has_inverse_name) {
      stop("vismat() does not support inverse matrices (Ainv/Dinv/AAinv). ",
           "Inverse matrix elements do not represent meaningful relationship coefficients. ",
           "Please use pedmat() with method='A', 'D', or 'AA' instead.")
    }
    
    preferred <- c("A", "D", "AA")
    found <- intersect(preferred, names(mat))
    if (length(found) > 0) {
      target_name <- found[1]
      if (length(mat) > 1) {
        message(sprintf("Found multiple matrices in list. Using '%s'.", target_name))
      }
      mat <- mat[[target_name]]
    } else {
      is_mat <- sapply(mat, function(x) is.matrix(x) || inherits(x, "Matrix"))
      if (any(is_mat)) {
        mat <- mat[[which(is_mat)[1]]]
      } else {
        found_types <- sapply(mat, function(x) class(x)[1])
        stop(sprintf("Input list does not contain any valid matrices. Found: %s", paste(found_types, collapse = ", ")))
      }
    }
  }

  # 2. Validation
  if (!is.matrix(mat) && !inherits(mat, "Matrix")) {
    stop("Input 'mat' must be a matrix or a Matrix object.")
  }

  # 3. Subset by IDs if requested
  if (!is.null(ids)) {
    all_names <- rownames(mat)
    if (is.null(all_names)) {
      stop("Matrix has no rownames. Subset by 'ids' is not possible.")
    }
    valid_ids <- intersect(ids, all_names)
    if (length(valid_ids) == 0) {
      stop("None of the specified 'ids' were found in the matrix.")
    }
    mat <- mat[valid_ids, valid_ids, drop = FALSE]
  }

  # 3b. Reorder for clustering (if requested and type is heatmap)
  if (reorder && type == "heatmap") {
    n <- nrow(mat)
    if (n > VISMAT_REORDER_MAX) {
      warning(sprintf("Matrix too large for reordering (N > %d). Skipping clustering.", VISMAT_REORDER_MAX))
      reorder <- FALSE  # Mark as not reordered for downstream logic
    } else {
      # CRITICAL: We use dist() on the matrix rows (Relationship Profile Distance).
      # Siblings share almost identical relationship profiles with the whole population,
      # making their Euclidean distance much smaller than the Parent-Child distance.
      d_mat <- as.matrix(mat)
      dist_obj <- stats::dist(d_mat)
      hc <- stats::hclust(dist_obj, method = "ward.D2")
      mat <- mat[hc$order, hc$order, drop = FALSE]
    }
  }

  # 4. Handle Grouping (Aggregate)
  if (!is.null(by)) {
    if (is.null(ped)) {
      stop("'ped' must be provided when using 'by'.")
    }

    # Ensure ped is a data.table and includes the required column
    ped_dt <- data.table::as.data.table(ped)
    if (!by %in% names(ped_dt)) {
      stop(sprintf("Column '%s' not found in pedigree.", by))
    }

    # Match matrix IDs to groups
    # tidyped objects always use "Ind" as the ID column
    id_col <- "Ind"
    if (!id_col %in% names(ped_dt)) {
      # Fallback to the first column if "Ind" is missing (unlikely for tidyped)
      id_col <- names(ped_dt)[1]
    }
    mapping <- ped_dt[, .(id = as.character(get(id_col)), grp = get(by))]

    # Get group for each matrix row
    mat_ids <- rownames(mat)
    if (is.null(mat_ids)) stop("Matrix must have row names for grouping.")

    # Match matrix IDs to pedigree
    match_idx <- match(mat_ids, mapping$id)
    if (any(is.na(match_idx))) {
      warning("Some individuals in matrix not found in pedigree. They will be assigned to 'Unknown' group.")
    }
    
    # Extract groups (may contain NA if by column has NA values)
    mat_grps <- mapping[match_idx, grp]
    
    # Handle NA groups
    if (any(is.na(mat_grps))) {
      n_na <- sum(is.na(mat_grps))
      
      # For Family grouping, exclude NA individuals (founders without family)
      if (by == "Family") {
        na_idx <- which(is.na(mat_grps))
        na_ids <- mat_ids[na_idx]
        message(sprintf(
          "Note: Excluding %d founder(s) with no family assignment: %s%s",
          n_na, 
          paste(head(na_ids, 5), collapse = ", "),
          if (n_na > 5) sprintf(" (and %d more)", n_na - 5) else ""
        ))
        
        # Remove NA individuals from matrix and by groups
        mat <- mat[-na_idx, -na_idx, drop = FALSE]
        mat_grps <- mat_grps[-na_idx]
        mat_ids <- mat_ids[-na_idx]
      } else {
        # For other by columns, assign to 'Unknown' group
        message(sprintf("Note: %d individual(s) have NA in '%s' column. Assigning to 'Unknown' group.", 
                        n_na, by))
        mat_grps[is.na(mat_grps)] <- "Unknown"
      }
    }

    # Aggregate: Mean relationship between groups
    mat_grps <- as.character(mat_grps)
    grps_unique <- sort(unique(mat_grps))
    n_grp <- length(grps_unique)

    # Check if this is an inverse matrix (averaging inverse elements is usually not useful)
    # Note: Detection based on common naming patterns in row/col names or attributes
    mat_name <- if (!is.null(attr(mat, "method"))) attr(mat, "method") else ""
    is_inv <- grepl("inv", mat_name, ignore.case = TRUE) || 
              grepl("inv", paste(rownames(mat), collapse = ""), ignore.case = TRUE)
    if (is_inv) {
      warning("Aggregating an inverse relationship matrix. Mean values may not be biologically meaningful.")
    }

    message(sprintf("Aggregating %d individuals into %d groups based on '%s'...", nrow(mat), n_grp, by))

    # Optimized aggregation using matrix algebra instead of nested loops
    # This is O(n * n_grp) instead of O(n_grp^2 * avg_group_size^2)
    
    # Create group membership matrix (n x n_grp indicator matrix)
    grp_factor <- factor(mat_grps, levels = grps_unique)
    n <- nrow(mat)
    
    # Count members per group for averaging
    grp_sizes <- tabulate(grp_factor, nbins = n_grp)
    
    # For very large matrices with many groups, use block-wise aggregation
    # to avoid memory issues while being much faster than element-wise loops
    if (n > VISMAT_BLOCK_AGG_MIN_N && n_grp > VISMAT_BLOCK_AGG_MIN_G) {
      message("  Using memory-efficient block aggregation for large matrix...")
      
      # Pre-compute group indices once
      idx_list <- split(seq_len(n), grp_factor)
      
      # Initialize result matrix
      agg_mat <- matrix(0, nrow = n_grp, ncol = n_grp, 
                        dimnames = list(grps_unique, grps_unique))
      
      # Process in row blocks for memory efficiency
      # Aggregate row sums first, then column sums
      for (i in seq_len(n_grp)) {
        idx_i <- idx_list[[i]]
        n_i <- length(idx_i)
        
        # Extract rows for group i (sparse-friendly)
        rows_i <- mat[idx_i, , drop = FALSE]
        
        # Sum across columns for each target group
        for (j in i:n_grp) {
          idx_j <- idx_list[[j]]
          n_j <- length(idx_j)
          
          # Extract the subblock and compute sum (faster than mean on subblocks)
          sub_sum <- sum(rows_i[, idx_j, drop = FALSE])
          agg_mat[i, j] <- sub_sum / (n_i * n_j)
          if (i != j) agg_mat[j, i] <- agg_mat[i, j]
        }
      }
    } else {
      # For smaller matrices, use direct vectorized approach
      # Convert to dense if sparse (for small matrices this is fine)
      mat_dense <- as.matrix(mat)
      
      # Use rowsum for fast aggregation: first sum rows within groups
      # rowsum(mat, group) gives sum of rows for each group
      row_agg <- rowsum(mat_dense, grp_factor, reorder = TRUE)  # n_grp x n
      
      # Then sum columns within groups
      agg_sum <- rowsum(t(row_agg), grp_factor, reorder = TRUE)  # n_grp x n_grp
      agg_sum <- t(agg_sum)
      
      # Compute mean by dividing by group size products
      size_mat <- outer(grp_sizes, grp_sizes)
      agg_mat <- agg_sum / size_mat
      
      # Ensure symmetry and proper naming
      agg_mat <- (agg_mat + t(agg_mat)) / 2
      dimnames(agg_mat) <- list(grps_unique, grps_unique)
    }
    
    # Ensure all values are finite
    agg_mat[!is.finite(agg_mat)] <- 0
    
    mat <- agg_mat
    # Note: User's reorder setting is preserved for grouped matrices

    # Set default titles for later use in heatmap
    grouping_main <- sprintf("Grouped Relationship Heatmap (%s)", by)
  }

  # 5. Rendering
  if (type == "heatmap") {
    dots <- list(...)
    if (!"main" %in% names(dots)) {
      dots$main <- if(!is.null(grouping_main)) grouping_main else "Relationship Matrix Heatmap"
    }
    if (!"xlab" %in% names(dots)) dots$xlab <- if(!is.null(by_label)) by_label else (if(reorder) "Clustered Individuals" else "Individuals")
    if (!"ylab" %in% names(dots)) dots$ylab <- if(!is.null(by_label)) by_label else (if(reorder) "Clustered Individuals" else "Individuals")
    
    nature_genetics_palette <- grDevices::colorRampPalette(c(
      "#fff7ec", "#fee8c8", "#fdd49e", "#fdbb84", 
      "#fc8d59", "#ef6548", "#d7301f", "#b30000", "#7f0000"
    ))(100)
    if (!"col.regions" %in% names(dots)) {
      dots$col.regions <- nature_genetics_palette
    }

    # Add a color key (legend) by default
    if (!"colorkey" %in% names(dots)) {
      dots$colorkey <- TRUE
    }

    # IMPORTANT: Matrix::image() from the Matrix package handles dispatch in a way 
    # that often conflicts with lattice arguments for base matrices.
    # We use lattice::levelplot directly to ensure full grid (incl. 0s) and borders.
    mat_to_plot <- as.matrix(mat)
    
    # Final safety check for lattice
    if (!any(is.finite(mat_to_plot))) {
      stop("The relationship matrix contains no finite values to plot.")
    }
    
    # Set up scales for individual labels
    # We want Individual 1 at Top-Left, so Y-axis labels must be reversed.
    # Show labels up to VISMAT_LABEL_MAX with dynamic font scaling.
    if (nrow(mat) <= VISMAT_LABEL_MAX && !"scales" %in% names(dots)) {
      # Use manual labelcex if provided, otherwise dynamic font size
      use_cex <- if(!is.null(labelcex)) labelcex else max(0.2, 0.7 * (1 - (nrow(mat) - 20) / 600))
      
      dots$scales <- list(
        x = list(at = 1:nrow(mat), labels = rownames(mat), rot = 90, cex = use_cex),
        y = list(at = 1:ncol(mat), labels = rev(colnames(mat)), cex = use_cex)
      )
    }

    # Border/Grid logic: Use panel function to draw white grid lines.
    # Limit grid to VISMAT_GRID_MAX to avoid visual clutter.
    if (nrow(mat) <= VISMAT_GRID_MAX && !"panel" %in% names(dots)) {
      dots$panel <- function(x, y, z, ...) {
        lattice::panel.levelplot(x, y, z, ...)
        lattice::panel.abline(h = (0:nrow(mat)) + 0.5, v = (0:nrow(mat)) + 0.5, col = "white", lwd = 0.4)
      }
    }

    # Execute plotting using lattice::levelplot
    # Transpose and reverse rows to make mat[1,1] appear at Top-Left (1, n)
    p <- do.call(lattice::levelplot, c(list(x = t(mat_to_plot[rev(seq_len(nrow(mat_to_plot))), , drop = FALSE])), dots))
    
    print(p)
    return(invisible(p))
    
  } else if (type == "histogram") {
    # Histogram of relationship coefficients
    # Useful for checking the range of inbreeding/kinship
    dots <- list(...)
    
    # Convert to regular matrix if sparse (Matrix package)
    # This ensures lower.tri() works correctly
    if (inherits(mat, "Matrix")) {
      mat <- as.matrix(mat)
    }
    
    vals <- as.numeric(mat[lower.tri(mat)])
    
    if (!"main" %in% names(dots)) dots$main <- "Distribution of Relationship Coefficients (Kinship)"
    if (!"xlab" %in% names(dots)) dots$xlab <- "Relationship Value"
    if (!"ylab" %in% names(dots)) dots$ylab <- "Frequency (%)"
    
    p <- do.call(lattice::histogram, c(list(x = ~vals), dots))
    print(p)
    return(invisible(p))

  } else {
    stop(sprintf("Visualization type '%s' is not supported. Use 'heatmap' or 'histogram'.", type))
  }
}


# --------------------------------------------------------------------------
# Internal: compute G×G group means directly from a complete pedigree
# --------------------------------------------------------------------------
# For group membership matrix W, the raw group sums are W' A W. The
# matrix-free product avoids constructing the dense individual-level A matrix.
aggregate_pedigree_by_group <- function(ped, ids, by) {
  if (!by %in% names(ped)) {
    stop(sprintf("Column '%s' not found in pedigree.", by))
  }

  valid_ids <- intersect(as.character(ids), ped$Ind)
  if (length(valid_ids) == 0L) {
    stop("None of the specified 'ids' were found in the pedigree.")
  }

  focal <- data.table::as.data.table(ped)[
    Ind %chin% valid_ids,
    .(Ind, grp = get(by))
  ]
  focal[, grp := as.character(grp)]

  if (anyNA(focal$grp)) {
    n_na <- sum(is.na(focal$grp))
    if (by == "Family") {
      na_ids <- focal$Ind[is.na(focal$grp)]
      message(sprintf(
        "Note: Excluding %d founder(s) with no family assignment: %s%s",
        n_na,
        paste(utils::head(na_ids, 5L), collapse = ", "),
        if (n_na > 5L) sprintf(" (and %d more)", n_na - 5L) else ""
      ))
      focal <- focal[!is.na(grp)]
    } else {
      message(sprintf(
        "Note: %d individual(s) have NA in '%s' column. Assigning to 'Unknown' group.",
        n_na, by
      ))
      focal[is.na(grp), grp := "Unknown"]
    }
  }

  if (nrow(focal) == 0L) {
    stop("No individuals remain after excluding missing group assignments.")
  }

  groups <- sort(unique(focal$grp))
  group_factor <- factor(focal$grp, levels = groups)
  group_sizes <- tabulate(group_factor, nbins = length(groups))
  focal_rows <- match(focal$Ind, ped$Ind)

  message(sprintf(
    "Aggregating %d individuals into %d groups based on '%s'...",
    nrow(focal), length(groups), by
  ))

  f_res <- cpp_calculate_inbreeding(ped$SireNum, ped$DamNum)
  n <- nrow(ped)
  n_groups <- length(groups)
  bytes_per_column <- 2 * 8 * n
  batch_size <- max(
    1L,
    min(32L, floor((256 * 1024^2) / max(1, bytes_per_column)))
  )
  group_sum <- matrix(0, nrow = n_groups, ncol = n_groups)

  for (batch_start in seq.int(1L, n_groups, by = batch_size)) {
    columns <- batch_start:min(batch_start + batch_size - 1L, n_groups)
    rhs <- matrix(0, nrow = n, ncol = length(columns))
    local_columns <- match(as.integer(group_factor), columns)
    selected <- !is.na(local_columns)
    rhs[cbind(focal_rows[selected], local_columns[selected])] <- 1

    product <- cpp_multiply_A(
      ped$SireNum, ped$DamNum, f_res$dii, rhs
    )
    group_sum[, columns] <- rowsum(
      product[focal_rows, , drop = FALSE],
      group_factor,
      reorder = TRUE
    )
  }

  result <- group_sum / outer(group_sizes, group_sizes)
  result <- (result + t(result)) / 2
  result[!is.finite(result)] <- 0
  dimnames(result) <- list(groups, groups)
  result
}


# --------------------------------------------------------------------------
# Internal: compute G×G group-mean matrix directly from K×K compact matrix
# --------------------------------------------------------------------------
# This avoids expanding the compact matrix to N×N (which may be infeasible
# when N > ~50k) and instead computes the group-level aggregation
# algebraically.
#
# Algorithm
# ---------
# Let A_c be the K×K compact matrix and W an K×G weight matrix where
# W[k, g] = number of individuals in group g whose representative is k.
#
# Step 1  Raw aggregate:   S_raw = W^T %*% A_c %*% W
# Step 2  Sibling correction: for each representative k that stands for
#         n_k > 1 siblings, the off-diagonal sibling value (sib_val_k)
#         differs from the diagonal value A_c[k,k]. The correction is
#         accumulated per group pair.
# Step 3  Mean:            M[g,h] = S_corrected[g,h] / (n_g * n_h)
#
# @param mat       A compact pedmat object (dsCMatrix or similar).
# @param ids       Character vector of focal individual IDs.
# @param by        Character, name of column in ped to group by.
# @param ped       A tidyped data.table with at least Ind and the `by` column.
# @return A dense G×G matrix of group means, with group labels as dimnames.
aggregate_compact_by_group <- function(mat, ids, by, ped) {
  ci   <- attr(mat, "call_info")
  cmap <- data.table::copy(attr(mat, "compact_map"))
  primary_method <- ci$method[1]

  # Strip pedmat class for raw matrix ops
  A_c <- mat
  if (inherits(A_c, "pedmat")) {
    class(A_c) <- setdiff(class(A_c), "pedmat")
  }

  # --- Map focal ids to groups ---
  ped_dt <- data.table::as.data.table(ped)
  focal  <- ped_dt[ped_dt$Ind %chin% ids, .(Ind, grp = as.character(get(by)))]

  # Exclude NA groups (e.g. founders without family)
  na_mask <- is.na(focal$grp)
  if (any(na_mask)) {
    n_na <- sum(na_mask)
    na_ids <- focal$Ind[na_mask]
    if (by == "Family") {
      message(sprintf(
        "Note: Excluding %d founder(s) with no family assignment: %s%s",
        n_na, paste(head(na_ids, 5), collapse = ", "),
        if (n_na > 5) sprintf(" (and %d more)", n_na - 5) else ""
      ))
    } else {
      message(sprintf(
        "Note: %d individual(s) have NA in '%s' column. Excluding from aggregation.",
        n_na, by
      ))
    }
    focal <- focal[!na_mask]
  }

  # Attach representative index from compact_map
  focal <- merge(focal, cmap[, .(Ind, RepIndNum)], by = "Ind", all.x = TRUE)
  if (anyNA(focal$RepIndNum)) {
    stop("Some focal individuals are not present in the compact_map.")
  }

  grps_unique <- sort(unique(focal$grp))
  n_grp <- length(grps_unique)
  K <- nrow(A_c)

  message(sprintf("Aggregating %d individuals into %d groups based on '%s'...",
                  nrow(focal), n_grp, by))

  # --- Build K×G weight matrix W ---
  # W[k, g] = count of focal individuals with RepIndNum == k in group g
  grp_factor <- factor(focal$grp, levels = grps_unique)
  W <- matrix(0L, nrow = K, ncol = n_grp)
  for (i in seq_len(nrow(focal))) {
    W[focal$RepIndNum[i], as.integer(grp_factor[i])] <-
      W[focal$RepIndNum[i], as.integer(grp_factor[i])] + 1L
  }

  # Group sizes
  grp_sizes <- colSums(W)  # length n_grp

  # --- Step 1: raw aggregate S = W^T A_c W ---
  # For sparse A_c, A_c %*% W is efficient
  AcW <- A_c %*% W          # K × G (sparse × dense → dense)
  S   <- crossprod(W, AcW)   # G × G  (W^T (K×G)^T, but crossprod does W^T %*% AcW)

  # --- Step 2: sibling correction ---
  # For each representative k that maps >1 focal individuals, the compact
  # matrix stores the *self* relationship (diagonal) for both the diagonal
  # AND the off-diagonal of the sibling block. The true off-diagonal is
  # sib_val_k, computed from parent relationships.
  # Correction delta per group pair:
  #   g != h:  Δ_gh += W[k,g] * W[k,h] * (sib_val_k - A_c[k,k])
  #   g == h:  Δ_gg += W[k,g] * (W[k,g] - 1) * (sib_val_k - A_c[k,k])
  if (primary_method %in% c("A", "D", "AA")) {
    # A matrix for parent lookups
    A_mat <- if (primary_method == "A") A_c else attr(mat, "A_intermediate")

    if (!is.null(A_mat)) {
      # Find representatives with >1 focal individual mapped
      row_totals <- rowSums(W)
      reps_with_mult <- which(row_totals > 1)

      if (length(reps_with_mult) > 0) {
        for (k in reps_with_mult) {
          s_idx <- cmap$SireNum[cmap$RepIndNum == k][1]
          d_idx <- cmap$DamNum[cmap$RepIndNum == k][1]
          sib_val  <- calc_sib_offdiag(A_mat, s_idx, d_idx, primary_method)
          diag_val <- A_c[k, k]
          delta    <- sib_val - diag_val

          if (abs(delta) < .Machine$double.eps) next

          w_k <- W[k, ]  # length n_grp
          # Off-diagonal correction: w_k[g] * w_k[h] * delta for g != h
          # Diagonal correction:     w_k[g] * (w_k[g] - 1) * delta
          # Combined: outer(w_k, w_k) * delta, then subtract w_k * delta on diag
          # (because diagonal of outer = w_k^2, but we want w_k*(w_k-1))
          correction <- outer(w_k, w_k) * delta
          diag(correction) <- w_k * (w_k - 1) * delta
          S <- S + correction
        }
      }
    }
  }

  # --- Step 3: mean ---
  size_mat <- outer(grp_sizes, grp_sizes)
  agg_mat  <- as.matrix(S) / size_mat
  agg_mat  <- (agg_mat + t(agg_mat)) / 2  # enforce symmetry
  agg_mat[!is.finite(agg_mat)] <- 0
  dimnames(agg_mat) <- list(grps_unique, grps_unique)

  agg_mat
}

#' Calculate Generation Intervals
#'
#' Computes the generation intervals for the four gametic pathways:
#' Sire to Son (SS), Sire to Daughter (SD), Dam to Son (DS), and Dam to Daughter (DD).
#' The generation interval is defined as the age of the parents at the birth of their offspring.
#'
#' @param ped A \code{tidyped} object.
#' @param timevar Character. The name of the column containing the \strong{birth date}
#'   (or hatch date) of each individual. The column must be one of:
#'   \itemize{
#'     \item \code{Date} or \code{POSIXct} (recommended).
#'     \item A date string parseable by \code{as.POSIXct} (e.g., \code{"2020-03-15"}).
#'       Use \code{format} for non-ISO strings.
#'     \item A numeric year (e.g., \code{2020}). Automatically converted to
#'       \code{Date} (\code{"YYYY-07-01"}) with a message. For finer precision,
#'       convert to \code{Date} beforehand.
#'   }
#'   If \code{NULL}, auto-detects columns named \code{"BirthYear"}, \code{"Year"},
#'   \code{"BirthDate"}, or \code{"Date"}.
#' @param unit Character. Output time unit for the interval:
#'   \code{"year"} (default), \code{"month"}, \code{"day"}, or \code{"hour"}.
#' @param format Character. Optional format string for parsing \code{timevar}
#'   when it contains non-standard date strings (e.g., \code{"\%d/\%m/\%Y"}
#'   for \code{"15/03/2020"}).
#' @param cycle Numeric. Optional target (designed) length of one generation
#'   cycle expressed in \code{unit}s. When provided, an additional column
#'   \code{GenEquiv} is appended to the result, defined as:
#'   \deqn{GenEquiv_i = \frac{\bar{L}_i}{L_{cycle}}}
#'   where \eqn{\bar{L}_i} is the observed mean interval for pathway \eqn{i} and
#'   \eqn{L_{cycle}} is \code{cycle}. A value > 1 means the observed
#'   interval exceeds the target cycle (lower breeding efficiency).
#'   Example: for Pacific white shrimp with a 180-day target cycle, set
#'   \code{unit = "day", cycle = 180}.
#' @param by Character. Optional grouping column (e.g., \code{"Breed"}, \code{"Farm"}).
#'   If provided, intervals are calculated within each group.
#'
#' @return A \code{data.table} with columns:
#' \itemize{
#'   \item \code{Group}: Grouping level (if \code{by} is used).
#'   \item \code{Pathway}: One of \code{"SS"}, \code{"SD"}, \code{"DS"}, \code{"DD"},
#'     \code{"SO"}, \code{"DO"}, or \code{"Average"}.
#'     SS/SD/DS/DD require offspring sex; SO (Sire-Offspring) and DO (Dam-Offspring)
#'     are computed from all parent-offspring pairs regardless of offspring sex.
#'   \item \code{N}: Number of parent-offspring pairs used.
#'   \item \code{Mean}: Average generation interval in \code{unit}.
#'   \item \code{SD}: Standard deviation of the interval.
#'   \item \code{GenEquiv}: (Optional) Generation equivalents based on \code{cycle}.
#' }
#'
#' @details
#' Parent-offspring pairs with zero or negative intervals are excluded from
#' the calculation because they typically indicate data entry errors or
#' insufficient time resolution. If many zero intervals are expected (e.g.,
#' when using \code{unit = "year"} with annual spawners), consider using a
#' finer time unit such as \code{"month"} or \code{"day"}.
#'
#' Numeric year columns (e.g., \code{2020}) are automatically converted to
#' \code{Date} by appending \code{"-07-01"} (mid-year) as a reasonable default.
#' For more precise results, convert to \code{Date} before calling this function.
#'
#' @examples
#' \donttest{
#' # ---- Basic usage with package dataset (numeric Year auto-converted) ----
#' tped <- tidyped(big_family_size_ped)
#' gi <- pedgenint(tped, timevar = "Year")
#' gi
#'
#' # ---- Generation equivalents with cycle ----
#' gi2 <- pedgenint(tped, timevar = "Year", cycle = 2)
#' gi2
#' }
#'
#' @export
pedgenint <- function(ped, timevar = NULL, unit = c("year", "month", "day", "hour"),
                      format = NULL, cycle = NULL, by = NULL) {
  # Input validation
  ped <- ensure_complete_tidyped(ped, "pedgenint()")
  unit <- match.arg(unit)
  
  # Auto-detect timevar
  if (is.null(timevar)) {
    candidates <- c("BirthYear", "Year", "birth_year", "year", "BirthDate", "Date")
    match_col <- intersect(names(ped), candidates)
    if (length(match_col) > 0) {
      timevar <- match_col[1]
      message(sprintf("Using '%s' as time variable.", timevar))
    } else {
      stop("Please specify 'timevar' (e.g., 'BirthYear' or 'BirthDate').")
    }
  }
  
  if (!timevar %in% names(ped)) {
    stop(sprintf("Column '%s' not found in pedigree.", timevar))
  }
  
  # Check grouping column
  if (!is.null(by) && !by %in% names(ped)) {
    stop(sprintf("Grouping column '%s' not found.", by))
  }

  # Parse time to numeric axis
  numeric_time <- .parse_to_numeric_time(ped[[timevar]], unit = unit, format = format)
  
  # Prepare working data — include numeric parent IDs for fast integer lookup
  cols <- c("Ind", "Sire", "Dam", "Sex", "IndNum", "SireNum", "DamNum")
  if (!is.null(by)) cols <- c(cols, by)
  
  dt <- as.data.table(ped)[, ..cols]
  dt[, Time := numeric_time]
  
  # Build integer-indexed time vector instead of a named character vector.
  # time_vec[IndNum + 1L] = time for individual IndNum.
  # time_vec[1L] = NA  (sentinel for SireNum/DamNum = 0, i.e. unknown parent).
  # This replaces time_map <- setNames(dt$Time, dt$Ind) which requires a
  # 1M-entry character hash table and four O(N) hash lookups per call.
  n_ped <- nrow(ped)
  time_vec <- vector("double", n_ped + 1L)
  time_vec[1L] <- NA_real_
  time_vec[ped$IndNum + 1L] <- numeric_time   # IndNum 1-based → offset by 1
  
  calc_pathway <- function(parent_col, offspring_sex_label) {
    parent_num_col <- paste0(parent_col, "Num")  # "SireNum" or "DamNum"
    valid_offspring <- dt[!is.na(Sex) & Sex == offspring_sex_label]
    valid_pairs <- valid_offspring[get(parent_num_col) > 0L]
    
    if (nrow(valid_pairs) == 0) return(NULL)
    
    p_nums <- valid_pairs[[parent_num_col]]
    p_times <- time_vec[p_nums + 1L]  # integer array lookup — O(1) per element
    o_times <- valid_pairs$Time
    
    intervals <- o_times - p_times
    # Filter valid intervals (ignore zero/negative which might be data errors)
    valid_idx <- !is.na(intervals) & intervals > 0
    
    if (sum(valid_idx) == 0) return(NULL)
    
    path_code <- paste0(
      ifelse(parent_col == "Sire", "S", "D"),
      ifelse(offspring_sex_label == "male", "S", "D")
    )
    
    res <- data.table(
      Pathway = path_code,
      Interval = intervals[valid_idx]
    )
    
    if (!is.null(by)) {
      res[, Group := valid_pairs[[by]][valid_idx]]
    }
    
    return(res)
  }
  
  res_list <- list(
    calc_pathway("Sire", "male"),   # SS
    calc_pathway("Sire", "female"), # SD
    calc_pathway("Dam", "male"),    # DS
    calc_pathway("Dam", "female")   # DD
  )
  
  all_intervals <- rbindlist(res_list)
  
  # Compute Average from ALL parent-offspring pairs regardless of offspring sex.
  # This avoids severely underestimating N when most offspring lack Sex info.
  calc_all_pathway <- function(parent_col) {
    parent_num_col <- paste0(parent_col, "Num")
    valid_pairs <- dt[get(parent_num_col) > 0L]
    if (nrow(valid_pairs) == 0) return(NULL)
    p_nums <- valid_pairs[[parent_num_col]]
    p_times <- time_vec[p_nums + 1L]
    o_times <- valid_pairs$Time
    intervals <- o_times - p_times
    valid_idx <- !is.na(intervals) & intervals > 0
    if (sum(valid_idx) == 0) return(NULL)
    res <- data.table(
      Pathway = ifelse(parent_col == "Sire", "SO", "DO"),
      Interval = intervals[valid_idx]
    )
    if (!is.null(by)) {
      res[, Group := valid_pairs[[by]][valid_idx]]
    }
    return(res)
  }
  
  all_parent_intervals <- rbindlist(list(
    calc_all_pathway("Sire"),
    calc_all_pathway("Dam")
  ))
  
  if (nrow(all_intervals) == 0 && nrow(all_parent_intervals) == 0) {
    warning("No valid parent-offspring pairs found with birth dates.")
    res <- data.table(Pathway = character(0), N = integer(0), Mean = numeric(0), SD = numeric(0))
    attr(res, "unit") <- unit
    return(res)
  }
  
  agg_cols <- "Pathway"
  if (!is.null(by)) agg_cols <- c("Group", "Pathway")
  
  summary_dt <- if (nrow(all_intervals) > 0) {
    all_intervals[, .(
      N = .N,
      Mean = mean(Interval),
      SD = sd(Interval)
    ), by = agg_cols]
  } else {
    data.table(Pathway = character(0), N = integer(0), Mean = numeric(0), SD = numeric(0))
  }
  
  # SO/DO: Sire-Offspring and Dam-Offspring (sex-independent)
  so_do_agg_cols <- if (!is.null(by)) c("Group", "Pathway") else "Pathway"
  
  so_do_dt <- if (nrow(all_parent_intervals) > 0) {
    all_parent_intervals[, .(
      N = .N,
      Mean = mean(Interval),
      SD = sd(Interval)
    ), by = so_do_agg_cols]
  } else {
    data.table(Pathway = character(0), N = integer(0), Mean = numeric(0), SD = numeric(0))
  }
  
  # Average: all parent-offspring pairs combined
  overall_agg_cols <- if (!is.null(by)) "Group" else NULL
  
  overall_dt <- all_parent_intervals[, .(
    Pathway = "Average",
    N = .N,
    Mean = mean(Interval),
    SD = sd(Interval)
  ), by = overall_agg_cols]
  
  final_res <- rbind(summary_dt, so_do_dt, overall_dt, fill = TRUE)
  
  # Handle Cycle Length for Generation Equivalents
  if (!is.null(cycle) && is.numeric(cycle)) {
    final_res[, GenEquiv := Mean / cycle]
  }
  
  if (!is.null(by)) {
    setorderv(final_res, c("Group", "Pathway"))
  } else {
    setorder(final_res, Pathway)
  }
  
  attr(final_res, "unit") <- unit
  return(final_res[])
}

.parse_to_numeric_time <- function(x, unit = "year", format = NULL) {
  # Seconds per unit
  factors <- c(year = 365.25 * 86400, month = 30.4375 * 86400,
               day = 86400, hour = 3600)
  if (!unit %in% names(factors)) {
    stop("unit must be one of: ", paste(names(factors), collapse = ", "))
  }
  target_factor <- factors[unit]

  # --- Numeric input: assume integer years, auto-convert to Date ---
  if (is.numeric(x)) {
    message("Numeric time column detected. Converting to Date (YYYY-07-01). ",
            "For finer precision, convert to Date beforehand.")
    date_strings <- ifelse(is.na(x), NA_character_, paste0(as.integer(x), "-07-01"))
    x <- as.Date(date_strings)
  }

  # --- Date input ---
  if (inherits(x, "Date")) {
    seconds <- as.numeric(as.POSIXct(x, tz = "UTC"))
    return(seconds / target_factor)
  }

  # --- POSIXct input ---
  if (inherits(x, "POSIXct")) {
    return(as.numeric(x) / target_factor)
  }

  # --- Character input: parse date strings ---
  if (!is.character(x)) {
    stop("timevar must be Date, POSIXct, character, or numeric year. Got ",
         class(x)[1], ".")
  }

  u_x <- unique(x[!is.na(x)])
  if (length(u_x) == 0) return(rep(NA_real_, length(x)))

  parsed <- if (!is.null(format)) {
    as.POSIXct(u_x, format = format, tz = "UTC")
  } else {
    # Try ISO / common formats
    p <- suppressWarnings(as.POSIXct(u_x, tz = "UTC", optional = TRUE))
    if (all(is.na(p))) {
      p <- as.POSIXct(suppressWarnings(as.Date(u_x, optional = TRUE)), tz = "UTC")
    }
    if (all(is.na(p))) {
      p <- as.POSIXct(suppressWarnings(
        as.Date(u_x, format = "%Y/%m/%d", optional = TRUE)
      ), tz = "UTC")
    }
    # Fallback: bare year strings like "2020"
    if (all(is.na(p))) {
      num_try <- suppressWarnings(as.numeric(u_x))
      if (!all(is.na(num_try))) {
        p <- as.POSIXct(as.Date(paste0(as.integer(num_try), "-07-01")), tz = "UTC")
        message("Bare year strings detected. Converting to Date (YYYY-07-01). ",
                "For finer precision, convert to Date beforehand.")
      }
    }
    p
  }

  if (all(is.na(parsed))) {
    stop("Failed to parse time variable. Provide Date, POSIXct, or specify ",
         "'format' (e.g. '%Y-%m-%d').")
  }

  val_map <- setNames(as.numeric(parsed), u_x)
  numeric_seconds <- val_map[as.character(x)]

  return(numeric_seconds / target_factor)
}

#' Pedigree Subpopulations
#'
#' Summarizes pedigree subpopulations and group structure.
#'
#' When \code{by = NULL}, this function is a lightweight summary wrapper around
#' \code{\link{splitped}}, returning one row per disconnected pedigree component
#' plus an optional \code{"Isolated"} row for individuals with no known parents
#' and no offspring. When \code{by} is provided, it instead summarizes the pedigree
#' directly by the specified column (e.g. \code{"Gen"}, \code{"Year"}, \code{"Breed"}).
#'
#' @param ped A \code{tidyped} object.
#' @param by Character. The name of the column to group by. 
#'   If NULL, summarizes disconnected components via \code{\link{splitped}}.
#'
#' @return A \code{data.table} with columns:
#' \itemize{
#'   \item \code{Group}: Subpopulation label.
#'   \item \code{N}: Total individuals.
#'   \item \code{N_Sire}: Number of distinct sires.
#'   \item \code{N_Dam}: Number of distinct dams.
#'   \item \code{N_Founder}: Number of founders (parents unknown).
#' }
#'
#' @details
#' Use \code{pedsubpop()} when you want a compact analytical summary table.
#' Use \code{\link{splitped}} when you need the actual re-tidied sub-pedigree
#' objects for downstream plotting or analysis.
#'
#' @examples
#' tp <- tidyped(simple_ped)
#'
#' # Summarize disconnected pedigree components
#' pedsubpop(tp)
#'
#' # Summarize by an existing grouping variable
#' pedsubpop(tp, by = "Gen")
#' 
#' @export
pedsubpop <- function(ped, by = NULL) {
  ped <- ensure_tidyped(ped)
  
  if (is.null(by)) {
    splits <- splitped(ped)
    groups <- names(splits)
    n_grps <- length(splits)
    iso <- attr(splits, "isolated")
    n_iso <- length(iso)
    
    res_list <- vector("list", n_grps + (n_iso > 0))
    
    for (i in seq_along(splits)) {
      gped <- splits[[i]]
      res_list[[i]] <- list(
        Group = groups[i],
        N = nrow(gped),
        N_Sire = data.table::uniqueN(gped$Sire, na.rm = TRUE),
        N_Dam = data.table::uniqueN(gped$Dam, na.rm = TRUE),
        N_Founder = sum(is.na(gped$Sire) & is.na(gped$Dam))
      )
    }
    
    if (n_iso > 0) {
      res_list[[n_grps + 1]] <- list(
        Group = "Isolated",
        N = n_iso,
        N_Sire = 0,
        N_Dam = 0,
        N_Founder = n_iso
      )
    }
    
  } else {
    if (!by %in% names(ped)) stop(sprintf("Column '%s' not found.", by))
    
    res_list <- list(ped[, .(
      N = .N,
      N_Sire = data.table::uniqueN(Sire, na.rm = TRUE),
      N_Dam = data.table::uniqueN(Dam, na.rm = TRUE),
      N_Founder = sum(is.na(Sire) & is.na(Dam))
    ), by = by])
    
    setnames(res_list[[1]], by, "Group")
  }
  
  data.table::rbindlist(res_list)[]
}

#' Calculate Mean Relationship or Coancestry Within Groups
#'
#' Computes either the average pairwise additive genetic relationship
#' coefficients (\eqn{a_{ij}}) within cohorts, or the corrected population mean
#' coancestry used for pedigree-based diversity summaries.
#'
#' Let \eqn{g} be a zero-one indicator vector for the \eqn{N} selected
#' individuals and let \eqn{S = g^\prime A g}. When
#' \code{scale = "relationship"}, the returned value is
#' \deqn{\bar{a}_{off} =
#'   \frac{S - \sum_{i:g_i=1}(1 + F_i)}{N(N - 1)},}
#' the mean of the off-diagonal additive relationship coefficients. When
#' \code{scale = "coancestry"}, the returned value is
#' \deqn{\bar{C} = \frac{S}{2N^2},}
#' which includes both pairwise coancestry and diagonal self-coancestry. This
#' is equivalent to
#' \deqn{\bar{C} = \frac{N - 1}{N} \cdot
#'   \frac{\bar{a}_{off}}{2} + \frac{1 + \bar{F}}{2N}.}
#' The implementation computes \eqn{A g} directly from the pedigree and does
#' not construct the dense relationship matrix \eqn{A}.
#'
#' @param ped A \code{tidyped} object.
#' @param by Character. The column name to group by (e.g., "Year", "Breed", "Generation").
#' @param reference Character vector. An optional vector of reference individual IDs to calculate
#'   relationships for. If provided, only individuals matching these IDs in each group
#'   will be used. Default is NULL (use all individuals in the group).
#' @param compact Logical. Retained for backward compatibility. It is ignored
#'   because \code{pedrel()} now uses matrix-free pedigree products for all
#'   calculations.
#' @param scale Character. One of \code{"relationship"} or \code{"coancestry"}.
#'   \code{"relationship"} returns the pairwise off-diagonal mean additive
#'   relationship (current \code{pedrel()} behavior). \code{"coancestry"}
#'   returns the corrected population mean coancestry used for pedigree-based
#'   diversity calculations.
#'
#' @return A \code{data.table} with columns:
#' \itemize{
#'   \item A grouping identifier column, named after the \code{by} parameter (e.g., \code{Gen}, \code{Year}).
#'   \item \code{NTotal}: Total number of individuals in the group.
#'   \item \code{NUsed}: Number of individuals used in calculation (could be subset by reference).
#'   \item \code{MeanRel}: Present when \code{scale = "relationship"}; average of off-diagonal
#'     elements in the Additive Relationship (A) matrix for this group
#'     (\eqn{a_{ij} = 2f_{ij}}).
#'   \item \code{MeanCoan}: Present when \code{scale = "coancestry"}; diagonal-corrected
#'     population mean coancestry for this group.
#'   \item \code{Status}: \code{"ok"}, \code{"skipped"}, or \code{"failed"}.
#'   \item \code{Message}: Empty for successful groups; otherwise a diagnostic
#'     explaining why \code{NA} was returned.
#' }
#' 
#' @examples
#' \donttest{
#' library(data.table)
#' # Use the sample dataset and simulate a birth year
#' tp <- tidyped(small_ped)
#' tp$Year <- 2010 + tp$Gen
#' 
#' # Example 1: Calculate average relationship grouped by Generation (default)
#' rel_by_gen <- pedrel(tp, by = "Gen")
#' print(rel_by_gen)
#' 
#' # Example 2: Calculate average relationship grouped by Year
#' rel_by_year <- pedrel(tp, by = "Year")
#' print(rel_by_year)
#' 
#' # Example 3: Calculate corrected mean coancestry
#' coan_by_gen <- pedrel(tp, by = "Gen", scale = "coancestry")
#' print(coan_by_gen)
#' 
#' # Example 4: Filter calculations with a reference list in a chosen group
#' candidates <- c("N", "O", "P", "Q", "T", "U", "V", "X", "Y")
#' rel_subset <- pedrel(tp, by = "Gen", reference = candidates)
#' print(rel_subset)
#' }
#' 
#' @export
pedrel <- function(ped, by = "Gen", reference = NULL, compact = FALSE,
                   scale = c("relationship", "coancestry")) {
  ped <- ensure_complete_tidyped(ped, "pedrel()")
  scale <- match.arg(scale)
  if (!by %in% names(ped)) stop(sprintf("Column '%s' not found.", by))
  if (!isTRUE(compact) && !isFALSE(compact)) stop("'compact' must be TRUE or FALSE.")

  make_row <- function(group, n_total, n_used, mean, status = "ok", message = "") {
    data.table(Group = group, NTotal = n_total, NUsed = n_used,
               Mean = mean, Status = status, Message = message)
  }
  
  groups <- unique(ped[[by]])
  groups <- groups[!is.na(groups)]

  res_list <- vector("list", length(groups))
  group_specs <- vector("list", length(groups))

  for (group_no in seq_along(groups)) {
    g <- groups[group_no]
    row_idx <- which(ped[[by]] == g)
    sub_ped_full <- as.data.table(ped)[row_idx]
    n_total <- nrow(sub_ped_full)

    if (n_total < 2) {
      msg <- "Group has less than 2 individuals."
      res_list[[group_no]] <- make_row(
        g, n_total, n_total, NA_real_, "skipped", msg
      )
      next
    }

    if (!is.null(reference)) {
      sub_ped <- sub_ped_full[Ind %in% reference]
    } else {
      sub_ped <- sub_ped_full
    }
    n_used <- nrow(sub_ped)

    if (n_used < 2) {
      msg <- "Group has less than 2 individuals after applying 'reference'."
      res_list[[group_no]] <- make_row(
        g, n_total, n_used, NA_real_, "skipped", msg
      )
      next
    }

    group_specs[[group_no]] <- list(
      group = g,
      n_total = n_total,
      n_used = n_used,
      target_inds = sub_ped$Ind
    )
  }

  valid_groups <- which(!vapply(group_specs, is.null, logical(1)))

  if (length(valid_groups) > 0L) {
    all_targets <- unique(unlist(
      lapply(group_specs[valid_groups], `[[`, "target_inds"),
      use.names = FALSE
    ))

    traced <- tryCatch(
      suppressMessages(
        tidyped(ped, cand = all_targets, trace = "up", addnum = TRUE)
      ),
      error = identity
    )

    if (inherits(traced, "error")) {
      msg <- paste0("Unable to trace the selected individuals: ", conditionMessage(traced))
      for (group_no in valid_groups) {
        spec <- group_specs[[group_no]]
        res_list[[group_no]] <- make_row(
          spec$group, spec$n_total, spec$n_used, NA_real_, "failed", msg
        )
      }
    } else {
      f_res <- tryCatch(
        cpp_calculate_inbreeding(traced$SireNum, traced$DamNum),
        error = identity
      )

      if (inherits(f_res, "error")) {
        msg <- paste0(
          "Unable to calculate pedigree coefficients: ",
          conditionMessage(f_res)
        )
        for (group_no in valid_groups) {
          spec <- group_specs[[group_no]]
          res_list[[group_no]] <- make_row(
            spec$group, spec$n_total, spec$n_used, NA_real_, "failed", msg
          )
        }
      } else {
        target_indices <- lapply(valid_groups, function(group_no) {
          match(group_specs[[group_no]]$target_inds, traced$Ind)
        })
        missing_targets <- vapply(target_indices, anyNA, logical(1))

        if (any(missing_targets)) {
          for (position in which(missing_targets)) {
            group_no <- valid_groups[position]
            spec <- group_specs[[group_no]]
            res_list[[group_no]] <- make_row(
              spec$group, spec$n_total, spec$n_used, NA_real_, "failed",
              "One or more target individuals were not found in the traced pedigree."
            )
          }
        }

        calculable_positions <- which(!missing_targets)
        if (length(calculable_positions) > 0L) {
          # rhs and A-rhs coexist during each call. Keep their combined working
          # storage near 256 MiB and cap batches to avoid wide temporary matrices.
          bytes_per_column <- 2 * 8 * nrow(traced)
          batch_size <- max(
            1L,
            min(32L, floor((256 * 1024^2) / max(1, bytes_per_column)))
          )
          batch_starts <- seq.int(1L, length(calculable_positions), by = batch_size)

          for (batch_start in batch_starts) {
            batch_positions <- calculable_positions[
              batch_start:min(
                batch_start + batch_size - 1L,
                length(calculable_positions)
              )
            ]
            rhs <- matrix(
              0,
              nrow = nrow(traced),
              ncol = length(batch_positions)
            )
            for (column_no in seq_along(batch_positions)) {
              rhs[target_indices[[batch_positions[column_no]]], column_no] <- 1
            }

            product <- tryCatch(
              cpp_multiply_A(
                traced$SireNum, traced$DamNum, f_res$dii, rhs
              ),
              error = identity
            )

            if (inherits(product, "error")) {
              msg <- paste0(
                "Unable to apply the relationship matrix: ",
                conditionMessage(product)
              )
              for (position in batch_positions) {
                group_no <- valid_groups[position]
                spec <- group_specs[[group_no]]
                res_list[[group_no]] <- make_row(
                  spec$group, spec$n_total, spec$n_used,
                  NA_real_, "failed", msg
                )
              }
              next
            }

            for (column_no in seq_along(batch_positions)) {
              position <- batch_positions[column_no]
              group_no <- valid_groups[position]
              spec <- group_specs[[group_no]]
              idx <- target_indices[[position]]
              n_selected <- as.numeric(length(idx))
              total_sum <- sum(product[idx, column_no])

              if (scale == "relationship") {
                diagonal_sum <- n_selected + sum(f_res$f[idx])
                off_diagonal_sum <- total_sum - diagonal_sum
                rounding_scale <- max(1, abs(total_sum), abs(diagonal_sum))
                if (abs(off_diagonal_sum) <=
                    64 * .Machine$double.eps * rounding_scale) {
                  off_diagonal_sum <- 0
                }
                mean_value <- off_diagonal_sum /
                  (n_selected * (n_selected - 1))
              } else {
                mean_value <- total_sum /
                  (2 * n_selected * n_selected)
              }

              if (!is.finite(mean_value)) {
                res_list[[group_no]] <- make_row(
                  spec$group, spec$n_total, spec$n_used, NA_real_, "failed",
                  "Calculation returned a non-finite value."
                )
              } else {
                res_list[[group_no]] <- make_row(
                  spec$group, spec$n_total, spec$n_used, mean_value
                )
              }
            }
          }
        }
      }
    }
  }

  result <- rbindlist(res_list)
  setnames(result, "Group", by)
  setnames(result, "Mean", if (scale == "relationship") "MeanRel" else "MeanCoan")
  setorderv(result, cols = by)

  problem_rows <- result[result[["Status"]] != "ok"]
  if (nrow(problem_rows) > 0L) {
    warning(sprintf(
      paste0("pedrel(): %d of %d groups returned a non-ok status. ",
             "Inspect the 'Status' and 'Message' columns. First issue: %s = %s; %s"),
      nrow(problem_rows), nrow(result), by, as.character(problem_rows[[by]][1]), problem_rows$Message[1]
    ), call. = FALSE)
  }

  return(result[])
}

#' Calculate Equi-Generate Coefficient
#'
#' Estimates the number of distinct ancestral generations using the Equi-Generate Coefficient (ECG).
#' The ECG is calculated as 1/2 of the sum of the parents' ECG values plus 1.
#'
#' @param ped A \code{tidyped} object.
#'
#' @return A \code{data.table} with columns:
#' \itemize{
#'   \item \code{Ind}: Individual ID.
#'   \item \code{ECG}: Equi-Generate Coefficient.
#'   \item \code{FullGen}: Fully traced generations (depth of complete ancestry).
#'   \item \code{MaxGen}: Maximum ancestral generations (depth of deepest ancestor).
#' }
#'
#' @examples
#' tp <- tidyped(simple_ped)
#' ecg <- pedecg(tp)
#'
#' # ECG combines pedigree depth and completeness
#' head(ecg)
#'
#' # Individuals with deeper and more complete ancestry have larger ECG values
#' ecg[order(-ECG)][1:5]
#' 
#' @references
#' Boichard, D., Maignel, L., & Verrier, E. (1997). The value of using probabilities of gene origin to measure genetic variability in a population. Genetics Selection Evolution, 29(1), 5.
#'
#' @export
pedecg <- function(ped) {
  ped <- ensure_complete_tidyped(ped, "pedecg()")
  
  n <- nrow(ped)
  ecg <- numeric(n)
  full_gen <- numeric(n)
  max_gen_val <- numeric(n)
  
  sire_nums <- ped$SireNum
  dam_nums <- ped$DamNum
  
  gen_indices <- split(ped$IndNum, ped$Gen)
  gen_levels <- sort(as.integer(names(gen_indices)))
  
  for (g in gen_levels) {
    idx <- gen_indices[[as.character(g)]]
    k <- length(idx)
    
    sires <- sire_nums[idx]
    dams <- dam_nums[idx]
    
    sire_ecg <- numeric(k) 
    dam_ecg <- numeric(k)
    sire_full <- rep(-1, k)
    dam_full <- rep(-1, k)
    sire_max <- rep(-1, k)
    dam_max <- rep(-1, k)
    
    valid_s <- !is.na(sires) & sires > 0
    if (any(valid_s)) {
      parents <- sires[valid_s]
      sire_ecg[valid_s] <- ecg[parents]
      sire_full[valid_s] <- full_gen[parents]
      sire_max[valid_s] <- max_gen_val[parents]
    }
    
    valid_d <- !is.na(dams) & dams > 0
    if (any(valid_d)) {
      parents <- dams[valid_d]
      dam_ecg[valid_d] <- ecg[parents]
      dam_full[valid_d] <- full_gen[parents]
      dam_max[valid_d] <- max_gen_val[parents]
    }
    
    term_s <- ifelse(valid_s, sire_ecg + 1, 0)
    term_d <- ifelse(valid_d, dam_ecg + 1, 0)
    ecg[idx] <- 0.5 * (term_s + term_d)
    
    full_gen[idx] <- pmin(sire_full, dam_full) + 1
    max_gen_val[idx] <- pmax(sire_max, dam_max) + 1
  }
  
  res <- data.table(
    Ind = ped$Ind,
    ECG = ecg,
    FullGen = full_gen,
    MaxGen = max_gen_val
  )

  return(res[])
}

#' Pedigree Statistics
#'
#' Calculates comprehensive statistics for a pedigree, including population structure,
#' generation intervals, and ancestral depth.
#'
#' @param ped A \code{tidyped} object.
#' @param timevar Optional character. Name of the column containing the
#'   \strong{birth date} (or hatch date) of each individual.
#'   Accepted column formats:
#'   \itemize{
#'     \item \code{Date} or \code{POSIXct} (recommended).
#'     \item A date string parseable by \code{as.POSIXct}
#'       (e.g., \code{"2020-06-15"}).  Use \code{format} via \code{...}
#'       for non-ISO strings.
#'     \item A numeric year (e.g., \code{2020}).  Automatically converted
#'       to \code{Date} (\code{"YYYY-07-01"}) with a message.
#'   }
#'   If \code{NULL}, attempts auto-detection from common column names
#'   (\code{"BirthYear"}, \code{"Year"}, \code{"BirthDate"}, etc.).
#' @param unit Character. Time unit for reporting generation intervals:
#'   \code{"year"} (default), \code{"month"}, \code{"day"}, or \code{"hour"}.
#' @param cycle Numeric. Optional target generation cycle length in
#'   \code{unit}s. When provided, \code{gen_intervals} will include a
#'   \code{GenEquiv} column (observed Mean / cycle). See
#'   \code{\link{pedgenint}} for details.
#' @param ecg Logical. Whether to compute equivalent complete generations
#'   for each individual via \code{\link{pedecg}}. Default \code{TRUE}.
#' @param genint Logical. Whether to compute generation intervals via
#'   \code{\link{pedgenint}}. Requires a detectable \code{timevar} column.
#'   Default \code{TRUE}.
#' @param ... Additional arguments passed to \code{\link{pedgenint}},
#'   e.g., \code{format} for custom date parsing or \code{by} for grouping.
#'
#' @return An object of class \code{pedstats}, which is a list containing:
#' \itemize{
#'   \item \code{summary}: A \code{data.table} with one row summarising the
#'     whole pedigree.  Columns:
#'     \itemize{
#'       \item \code{N} — total number of individuals.
#'       \item \code{NSire} — number of unique sires.
#'       \item \code{NDam} — number of unique dams.
#'       \item \code{NFounder} — number of founder individuals
#'         (both parents unknown).
#'       \item \code{MaxGen} — maximum generation number.
#'     }
#'   \item \code{ecg}: A \code{data.table} with one row per individual
#'     (\code{NULL} if \code{ecg = FALSE}).  Columns:
#'     \itemize{
#'       \item \code{Ind} — individual identifier.
#'       \item \code{ECG} — equivalent complete generations.
#'       \item \code{FullGen} — number of fully known generations.
#'       \item \code{MaxGen} — maximum traceable generation depth.
#'     }
#'   \item \code{gen_intervals}: A \code{data.table} of generation intervals
#'     (\code{NULL} if no \code{timevar} is detected or
#'     \code{genint = FALSE}).  Columns:
#'     \itemize{
#'       \item \code{Pathway} — gametic pathway label.  Seven values:
#'         \code{"SS"} (sire to son), \code{"SD"} (sire to daughter),
#'         \code{"DS"} (dam to son), \code{"DD"} (dam to daughter) —
#'         require offspring sex; \code{"SO"} (sire to offspring) and
#'         \code{"DO"} (dam to offspring) — sex-independent; and
#'         \code{"Average"} — all parent-offspring pairs combined.
#'       \item \code{N} — number of parent-offspring pairs.
#'       \item \code{Mean} — mean generation interval.
#'       \item \code{SD} — standard deviation of the interval.
#'       \item \code{GenEquiv} — \code{Mean / cycle} (only present when
#'         \code{cycle} is supplied).
#'     }
#' }
#'
#' @examples
#' \donttest{
#' # ---- Without time variable ----
#' tp <- tidyped(simple_ped)
#' ps <- pedstats(tp)
#' ps$summary
#' ps$ecg
#'
#' # ---- With annual Year column (big_family_size_ped) ----
#' tp2 <- tidyped(big_family_size_ped)
#' ps2 <- pedstats(tp2, timevar = "Year")
#' ps2$summary
#' ps2$gen_intervals
#' }
#'
#' @export
pedstats <- function(ped, timevar = NULL, unit = "year", cycle = NULL, ecg = TRUE, genint = TRUE, ...) {
  ped <- if (isTRUE(ecg) || isTRUE(genint)) {
    ensure_complete_tidyped(ped, "pedstats()")
  } else {
    ensure_tidyped(ped)
  }
  
  summ <- data.table(
    N = nrow(ped),
    NSire = data.table::uniqueN(ped$Sire, na.rm = TRUE),
    NDam = data.table::uniqueN(ped$Dam, na.rm = TRUE),
    NFounder = sum(is.na(ped$Sire) & is.na(ped$Dam)),
    MaxGen = max(ped$Gen)
  )
  
  ecg_dt <- NULL
  if (ecg) {
    ecg_dt <- pedecg(ped)
  }
  
  gen_int <- NULL
  
  target_timevar <- timevar
  if (is.null(target_timevar)) {
    candidates <- c("BirthYear", "Year", "birth_year", "year", "BirthDate", "Date")
    match_col <- intersect(names(ped), candidates)
    if (length(match_col) > 0) target_timevar <- match_col[1]
  }
  
  if (genint && !is.null(target_timevar) && target_timevar %in% names(ped)) {
    gen_int <- tryCatch({
      pedgenint(ped, timevar = target_timevar, unit = unit, cycle = cycle, ...)
    }, error = function(e) {
      warning("Failed to calculate generation intervals: ", e$message)
      NULL
    })
  }
  
  res <- list(
    summary = summ,
    ecg = ecg_dt,
    gen_intervals = gen_int
  )
  class(res) <- "pedstats"
  res
}

#' Print Pedigree Statistics
#'
#' @param x A \code{pedstats} object.
#' @param ... Additional arguments.
#'
#' @export
print.pedstats <- function(x, ...) {
  cat("Pedigree Statistics\n")
  cat("===================\n")
  
  s <- x$summary
  cat(sprintf("Total Individuals: %d\n", s$N))
  cat(sprintf("Sires: %d | Dams: %d | Founders: %d\n", s$NSire, s$NDam, s$NFounder))
  cat(sprintf("Max Generation: %d\n", s$MaxGen))
  
  if (!is.null(x$gen_intervals)) {
    unit <- attr(x$gen_intervals, "unit")
    unit_label <- switch(unit, 
                        "year" = "Years", "month" = "Months", "day" = "Days", 
                        "hour" = "Hours", "gen" = "Units", "Years")
    cat(sprintf("\nGeneration Intervals (%s):\n", unit_label))
    print(x$gen_intervals)
  }
  
  if (!is.null(x$ecg)) {
    cat("\nAncestral Depth (ECG) Summary:\n")
    print(summary(x$ecg$ECG))
  }
  
  invisible(x)
}

#' Calculate Effective Population Size
#'
#' Calculates the effective population size (Ne) based on the rate of
#' coancestry, the rate of inbreeding, or demographic parent numbers.
#'
#' @param ped A \code{tidyped} object.
#' @param method Character. The method to compute Ne. One of \code{"coancestry"} (default), \code{"inbreeding"}, or \code{"demographic"}.
#' @param by Character. The name of the column used to group cohorts (e.g., "Year", "BirthYear").
#'   If NULL, calculates overall Ne for all individuals.
#' @param reference Character vector. Optional subset of individual IDs defining the reference cohort.
#'   If NULL, uses all individuals in the pedigree.
#' @param nsamples Integer. Number of individuals to randomly sample per cohort when using the \code{"coancestry"} method. Very large cohorts will be sampled down to this size to save memory and time (default: 1000).
#' @param ncores Integer. Number of cores for parallel processing. Currently only effective for \code{method = "coancestry"} (default: 1).
#' @param seed Integer or NULL. Random seed passed to \code{set.seed()} before sampling
#'   in the coancestry method, ensuring reproducible \eqn{N_e} estimates.
#'   Default is \code{NULL} (no fixed seed).
#'
#' @return A \code{data.table} with columns:
#' \itemize{
#'   \item \code{Cohort}: Cohort or grouping variable value.
#'   \item \code{N}: Number of individuals in the cohort.
#'   \item \code{Ne}: Calculated effective population size.
#'   \item \code{...}: Additional columns depending on the selected method (e.g., \code{NSampled}, \code{DeltaC}, \code{MeanF}, \code{DeltaF}, \code{Nm}, \code{Nf}).
#' }
#' 
#' @details
#' The effective population size can be calculated using one of three methods:
#' 
#' \itemize{
#'   \item \strong{"coancestry"} (Default): Based on the rate of coancestry (\eqn{f_{ij}}) between 
#'   pairs of individuals. In this context, \eqn{f_{ij}} is the probability that two alleles 
#'   randomly sampled from individuals $i$ and $j$ are identical by descent, which is equivalent 
#'   to half the additive genetic relationship (\eqn{f_{ij} = a_{ij} / 2}). 
#'   This method is generally more robust as it accounts for full genetic drift and 
#'   bottlenecks (Cervantes et al., 2011).
#'   \deqn{\Delta c_{ij} = 1 - (1 - c_{ij})^{1/(\frac{ECG_i + ECG_j}{2})}}
#'   \deqn{N_e = \frac{1}{2 \overline{\Delta c}}}
#'   To handle large populations, this method samples \code{nsamples} individuals per cohort 
#'   and computes the mean rate of coancestry among them. Sampled relationship
#'   coefficients are obtained from batched matrix-free \eqn{AX} products, so
#'   the full additive relationship matrix is not constructed.
#'   
#'   \item \strong{"inbreeding"}: Based on the individual rate of inbreeding ($F_i$) (Gutiérrez et al., 2008, 2009).
#'   \deqn{\Delta F_i = 1 - (1 - F_i)^{1/(ECG_i - 1)}}
#'   \deqn{N_e = \frac{1}{2 \overline{\Delta F}}}
#'   
#'   \item \strong{"demographic"}: Based on the demographic census of breeding males and females (Wright, 1931).
#'   \deqn{N_e = \frac{4 N_m N_f}{N_m + N_f}}
#'   Where \eqn{N_m} and \eqn{N_f} are the number of unique male and female parents contributing to the cohort.
#' }
#' 
#' @references
#' Cervantes, I., Goyache, F., Molina, A., Valera, M., & Gutiérrez, J. P. (2011). Estimation of effective population size from the rate of coancestry in pedigreed populations. \emph{Journal of Animal Breeding and Genetics}, 128(1), 56-63.
#' 
#' Gutiérrez, J. P., Cervantes, I., Molina, A., Valera, M., & Goyache, F. (2008). Individual increase in inbreeding allows estimating effective sizes from pedigrees. \emph{Genetics Selection Evolution}, 40(4), 359-370.
#' 
#' Gutiérrez, J. P., Cervantes, I., & Goyache, F. (2009). Improving the estimation of realized effective population sizes in farm animals. \emph{Journal of Animal Breeding and Genetics}, 126(4), 327-332.
#' 
#' Wright, S. (1931). Evolution in Mendelian populations. \emph{Genetics}, 16(2), 97-159.
#'
#' @examples
#' \donttest{
#' # Coancestry-based Ne (default) using a simple pedigree grouped by year
#' tp_simple <- tidyped(simple_ped)
#' tp_simple$BirthYear <- 2000 + tp_simple$Gen
#' ne_coan <- suppressMessages(pedne(tp_simple, by = "BirthYear", seed = 42L))
#' ne_coan
#'
#' # Inbreeding-based Ne using an inbred pedigree
#' tp_inbred <- tidyped(inbred_ped)
#' ne_inb <- suppressMessages(pedne(tp_inbred, method = "inbreeding", by = "Gen"))
#' ne_inb
#'
#' # Demographic Ne from the number of contributing sires and dams
#' ne_demo <- suppressMessages(pedne(tp_simple, method = "demographic", by = "BirthYear"))
#' ne_demo
#' }
#' 
#' @export
pedne <- function(ped, method = c("coancestry", "inbreeding", "demographic"),
                  by = NULL, reference = NULL, nsamples = 1000, ncores = 1,
                  seed = NULL) {
  
  ped <- if (match.arg(method) %in% c("coancestry", "inbreeding")) {
    ensure_complete_tidyped(ped, sprintf("pedne(method = \"%s\")", match.arg(method)))
  } else {
    ensure_tidyped(ped)
  }
  method <- match.arg(method)
  
  # Protect the original data from being modified by reference
  ped_dt <- data.table::copy(ped)
  
  # Handle 'by' parameter
  if (is.null(by)) {
    message("No 'by' parameter specified. Calculating Ne for the entire dataset. Grouping by a cohort or year variable is often more biologically meaningful.")
    by <- ".CohortAll"
    ped_dt[[by]] <- "All"
  } else {
    if (!by %in% names(ped_dt)) {
      stop(sprintf("Column '%s' not found in pedigree.", by))
    }
  }
  
  # Configure threads
  if (ncores > 1) {
    if (cpp_openmp_available()) {
      cpp_set_num_threads(ncores)
    } else {
      warning("OpenMP not available. Using single core.")
    }
  }

  # Pre-calculate mandatory genetic parameters on FULL pedigree before subsetting
  if (method %in% c("inbreeding", "coancestry")) {
    if (method == "inbreeding" && !"f" %in% names(ped_dt)) {
      ped_dt <- inbreed(ped_dt)
    }
    if (!"ECG" %in% names(ped_dt)) {
      ecg_dt <- pedecg(ped_dt)
      ped_dt <- merge(ped_dt, ecg_dt[, .(Ind, ECG)], by = "Ind", all.x = TRUE)
      # merge() sorts by Ind (alphabetically), breaking the IndNum == row-index
      # invariant that the fast-path BFS in tidyped() depends on.
      setorder(ped_dt, IndNum)
    }
  }
  
  # Filter reference population if specified
  if (!is.null(reference)) {
    if (!all(reference %in% ped_dt$Ind)) {
      missing <- reference[!reference %in% ped_dt$Ind]
      warning(sprintf("Reference IDs not found in pedigree: %s", paste(missing, collapse = ", ")))
    }
    ped_subset <- as.data.table(ped_dt)[Ind %in% reference]
  } else {
    ped_subset <- ped_dt
  }
  
  # Dispatch to specific calculator
  res <- switch(method,
    "inbreeding"  = calc_ne_inbreeding(ped_subset, by),
    "coancestry"  = {
      raw <- calc_ne_coancestry(
        ped_subset, ped_dt, by, nsamples,
        seed = seed, metrics = "ne"
      )
      # Strip fg/MeanCoan/NSampledCoan: exposed only via pediv(), not pedne()
      cols_to_drop <- intersect(names(raw), c("fg", "MeanCoan", "NSampledCoan"))
      if (length(cols_to_drop) > 0) raw[, (cols_to_drop) := NULL]
      raw
    },
    "demographic" = calc_ne_demographic(ped_subset, by)
  )

  return(res[])
}

calc_ne_inbreeding <- function(ped, by) {
  # Inbreeding coefficients and ECG should already be calculated on the full pedigree
  if (!"f" %in% names(ped) || !"ECG" %in% names(ped)) {
    stop("f or ECG is missing in subset tracking.")
  }
  
  ped_work <- ped[ECG >= 1]
  
  if (nrow(ped_work) == 0) {
    warning("No individuals with sufficient pedigree depth (ECG >= 1).")
    return(data.table(Cohort = numeric(0), N = integer(0), Ne = numeric(0)))
  }

  ped_work[, CohortLabel := get(by)]
  ped_work[, DeltaF_Ind := ifelse(ECG > 1, 1 - (1 - f)^(1 / (ECG - 1)), NA_real_)]
  
  # Filter valid DeltaF
  ped_valid <- ped_work[!is.na(DeltaF_Ind) & is.finite(DeltaF_Ind) & DeltaF_Ind >= 0]
  
  if (nrow(ped_valid) == 0) {
    warning("No valid DeltaF values calculated.")
    return(data.table(Cohort = numeric(0), N = integer(0), Ne = numeric(0)))
  }

  result <- ped_valid[, .(
    N = .N,
    MeanF = mean(f, na.rm = TRUE),
    DeltaF = mean(DeltaF_Ind, na.rm = TRUE),
    MeanECG = mean(ECG, na.rm = TRUE)
  ), by = .(Cohort = CohortLabel)]
  
  result[, Ne := ifelse(DeltaF > 0, 1 / (2 * DeltaF), NA_real_)]
  setorder(result, Cohort)
  
  return(result[])
}

calc_ne_demographic <- function(ped, by) {
  # Ensure numbers are present
  if (!all(c("Sire", "Dam", "Sex") %in% names(ped))) {
     stop("Pedigree must contain 'Sire', 'Dam', and 'Sex' columns for demographic Ne.")
  }

  ped <- copy(ped)
  ped[, CohortLabel := get(by)]
  
  # Need to find parents contributing to EACH cohort
  # Parent IDs are in Sire and Dam columns
  # We should count how many UNIQUE Sires and Dams produced offspring in this cohort
  
  result <- ped[, {
    sires <- unique(Sire[!is.na(Sire) & Sire != "" & Sire != "0"])
    dams <- unique(Dam[!is.na(Dam) & Dam != "" & Dam != "0"])
    
    nm <- length(sires)
    nf <- length(dams)
    
    ne_val <- if (nm > 0 && nf > 0) (4 * nm * nf) / (nm + nf) else NA_real_
    
    list(
      N = .N,
      Nm = nm,
      Nf = nf,
      Ne = ne_val
    )
  }, by = .(Cohort = CohortLabel)]
  
  setorder(result, Cohort)
  return(result[])
}

calc_sampled_relationship_stats <- function(sire, dam, dii, f, target_idx,
                                            ecg = NULL,
                                            need_delta = TRUE,
                                            need_mean = TRUE,
                                            batch_size = NULL) {
  target_idx <- unique(as.integer(target_idx))
  n <- length(sire)
  n_target <- length(target_idx)

  if (n_target < 2L) {
    return(list(
      DeltaC = NA_real_,
      MeanRelOff = NA_real_,
      MeanF = if (n_target == 1L) f[target_idx] else NA_real_
    ))
  }

  if (need_delta && is.null(ecg)) {
    stop("'ecg' is required when calculating DeltaC.", call. = FALSE)
  }

  mean_f <- if (need_mean) mean(f[target_idx], na.rm = TRUE) else NA_real_
  lower_sum <- 0
  delta_sum <- 0
  delta_pairs <- 0

  if (!need_delta && need_mean) {
    rhs <- matrix(0, nrow = n, ncol = 1L)
    rhs[target_idx, 1L] <- 1
    product <- cpp_multiply_A(sire, dam, dii, rhs)
    total_sum <- sum(product[target_idx, 1L])
    diagonal_sum <- n_target + sum(f[target_idx])
    off_diagonal_sum <- total_sum - diagonal_sum
    rounding_scale <- max(1, abs(total_sum), abs(diagonal_sum))
    if (abs(off_diagonal_sum) <=
        64 * .Machine$double.eps * rounding_scale) {
      off_diagonal_sum <- 0
    }

    return(list(
      DeltaC = NA_real_,
      MeanRelOff = off_diagonal_sum / (n_target * (n_target - 1)),
      MeanF = mean_f
    ))
  }

  if (is.null(batch_size)) {
    # rhs and A-rhs coexist during each call. Keep their combined working
    # storage near 256 MiB and cap batches to avoid wide temporary matrices.
    bytes_per_column <- 2 * 8 * n
    batch_size <- max(
      1L,
      min(32L, floor((256 * 1024^2) / max(1, bytes_per_column)))
    )
  }

  target_ecg <- if (need_delta) ecg[target_idx] else NULL
  batch_starts <- seq.int(1L, n_target, by = batch_size)

  for (batch_start in batch_starts) {
    columns <- batch_start:min(batch_start + batch_size - 1L, n_target)
    rhs <- matrix(0, nrow = n, ncol = length(columns))
    rhs[cbind(target_idx[columns], seq_along(columns))] <- 1
    product <- cpp_multiply_A(sire, dam, dii, rhs)
    target_product <- product[target_idx, , drop = FALSE]

    for (local_column in seq_along(columns)) {
      target_column <- columns[local_column]
      if (target_column <= 1L) next

      previous <- seq_len(target_column - 1L)
      relationships <- target_product[previous, local_column]

      if (need_mean) {
        lower_sum <- lower_sum + sum(relationships)
      }

      if (need_delta) {
        pair_ecg <- (target_ecg[previous] + target_ecg[target_column]) / 2
        valid <- pair_ecg > 0
        if (any(valid)) {
          coancestry <- relationships[valid] / 2
          delta_sum <- delta_sum + sum(
            1 - (1 - coancestry)^(1 / pair_ecg[valid])
          )
          delta_pairs <- delta_pairs + sum(valid)
        }
      }
    }
  }

  list(
    DeltaC = if (need_delta && delta_pairs > 0) {
      delta_sum / delta_pairs
    } else {
      NA_real_
    },
    MeanRelOff = if (need_mean) {
      2 * lower_sum / (n_target * (n_target - 1))
    } else {
      NA_real_
    },
    MeanF = mean_f
  )
}

calc_ne_coancestry <- function(ped_subset, ped_full, by, nsamples,
                               seed = NULL, max_trace = 25000L,
                               metrics = c("ne", "fg")) {
  # ECG is already merged into ped_subset from ped_full by the caller function.
  # Returns a stable extended schema for all metric selections. pedne() requests
  # only Ne and strips the fg columns; pediv() requests both metrics.

  metrics <- match.arg(metrics, c("ne", "fg"), several.ok = TRUE)
  need_ne <- "ne" %in% metrics
  need_fg <- "fg" %in% metrics
  ped_subset <- as.data.table(ped_subset)

  ped_subset[, CohortLabel := get(by)]
  cohorts <- sort(unique(ped_subset$CohortLabel))

  results_list <- list()

  for (coh in cohorts) {
    sub_ped <- ped_subset[CohortLabel == coh]
    n_total <- nrow(sub_ped)

    if (n_total == 0) next

    # Reproducible sampling
    if (!is.null(seed)) set.seed(seed)

    # Initial sampling
    if (n_total > nsamples) {
      sampled_inds <- sample(sub_ped$Ind, nsamples)
    } else {
      sampled_inds <- sub_ped$Ind
    }

    # Trace back ancestors using the FULL pedigree
    trace_ped <- tidyped(ped_full, cand = sampled_inds, trace = "up",
                         addgen = TRUE, addnum = TRUE)

    # Adaptive resampling: if traced pedigree exceeds memory limit, reduce sample
    if (nrow(trace_ped) > max_trace) {
      n_reduced <- max(2L, floor(length(sampled_inds) * max_trace / nrow(trace_ped)))
      if (!is.null(seed)) set.seed(seed)
      sampled_inds <- sample(sub_ped$Ind, min(n_reduced, n_total))
      trace_ped <- tidyped(ped_full, cand = sampled_inds, trace = "up",
                           addgen = TRUE, addnum = TRUE)
    }

    # Must be sorted by IndNum for correct C++ indexing (1..N)
    setorder(trace_ped, IndNum)

    # Re-identify target numeric IDs (1-based)
    target_nums <- trace_ped[Ind %in% sampled_inds, IndNum]

    # DeltaC needs ECG; fg-only consumers can skip this preprocessing entirely.
    if (need_ne) {
      if (!"ECG" %in% names(trace_ped)) {
        ecg_map <- ped_full[Ind %in% trace_ped$Ind, .(Ind, ECG)]
        trace_ped <- merge(trace_ped, ecg_map, by = "Ind", all.x = TRUE)
        setorder(trace_ped, IndNum)
      }
      trace_ped[is.na(ECG), ECG := 0]
    }

    avg_delta_c <- NA_real_
    ne_val <- NA_real_
    fg_val     <- NA_real_
    mean_coan  <- NA_real_

    tryCatch({
      f_res <- cpp_calculate_inbreeding(
        trace_ped$SireNum,
        trace_ped$DamNum
      )
      stats <- calc_sampled_relationship_stats(
        sire = trace_ped$SireNum,
        dam = trace_ped$DamNum,
        dii = f_res$dii,
        f = f_res$f,
        target_idx = target_nums,
        ecg = if (need_ne) trace_ped$ECG else NULL,
        need_delta = need_ne,
        need_mean = need_fg
      )

      if (need_ne) {
        avg_delta_c <- stats$DeltaC
        ne_val <- if (!is.na(avg_delta_c) && avg_delta_c > 0) {
          1 / (2 * avg_delta_c)
        } else {
          NA_real_
        }
      }

      if (need_fg) {
        # Diagonal-corrected mean coancestry:
        #   C_bar = (N-1)/N * rbar_off/2 + (1 + mean_F_s) / (2N)
        # where N is the full reference cohort size.
        mean_coan <- (n_total - 1) / n_total * stats$MeanRelOff / 2 +
                     (1 + stats$MeanF) / (2 * n_total)
        fg_val <- if (!is.na(mean_coan) && mean_coan > 0) {
          1 / (2 * mean_coan)
        } else {
          NA_real_
        }
      }
    }, error = function(e) {
      warning(paste(
        "Error in matrix-free coancestry calculation for cohort",
        coh, ":", e$message
      ))
    })

    results_list[[as.character(coh)]] <- data.table(
      Cohort       = coh,
      N            = n_total,
      NSampled     = length(sampled_inds),
      DeltaC       = avg_delta_c,
      Ne           = ne_val,
      MeanCoan     = mean_coan,
      fg           = fg_val,
      NSampledCoan = if (need_fg) length(sampled_inds) else NA_integer_
    )
  }

  result <- rbindlist(results_list)
  return(result[])
}

# Legacy wrapper/stub for documentation or compatibility if needed
# (none required as we replaced pedne in place) -> THIS LINE INTENTIONALLY LEFT BLANK
fill_phantoms <- function(ped) {
  missing_sire <- is.na(ped$Sire) & !is.na(ped$Dam)
  missing_dam <- !is.na(ped$Sire) & is.na(ped$Dam)
  
  if (!any(missing_sire) && !any(missing_dam)) {
    return(ped)
  }
  
  ped_new <- data.table::copy(ped)
  new_rows <- list()
  
  if (any(missing_sire)) {
    phantoms_sire <- paste0("*Phantom_Sire_", ped_new$Ind[missing_sire])
    ped_new$Sire[missing_sire] <- phantoms_sire
    new_rows[[1]] <- data.table::data.table(
      Ind = phantoms_sire,
      Sire = NA_character_,
      Dam = NA_character_
    )
  }
  
  if (any(missing_dam)) {
    phantoms_dam <- paste0("*Phantom_Dam_", ped_new$Ind[missing_dam])
    ped_new$Dam[missing_dam] <- phantoms_dam
    new_rows[[2]] <- data.table::data.table(
      Ind = phantoms_dam,
      Sire = NA_character_,
      Dam = NA_character_
    )
  }
  
  new_dt <- data.table::rbindlist(new_rows, fill = TRUE)
  extra_cols <- setdiff(names(ped_new), names(new_dt))
  if (length(extra_cols) > 0) {
    for (col in extra_cols) {
      data.table::set(new_dt, j = col, value = NA)
    }
  }
  
  ped_combined <- rbind(ped_new, new_dt, fill = TRUE)
  
  old_cols <- intersect(names(ped_combined), c("IndNum", "SireNum", "DamNum"))
  if (length(old_cols) > 0) {
    ped_combined[, (old_cols) := NULL]
  }
  
  return(tidyped(ped_combined, addnum = TRUE))
}

# Internal: Shannon entropy -> effective number (q=1 Hill number)
# p must be a probability vector (sums to 1); zeros are filtered out.
calc_shannon_effective <- function(p) {
  p <- p[p > 0]
  if (length(p) == 0L) return(NA_real_)
  exp(-sum(p * log(p)))
}

#' Calculate Founder and Ancestor Contributions
#'
#' Calculates genetic contributions from founders and influential ancestors.
#' Implements the gene dropping algorithm for founder contributions and Boichard's 
#' algorithm for ancestor contributions to estimate the effective number of founders ($f_e$) 
#' and ancestors ($f_a$).
#'
#' @param ped A \code{tidyped} object.
#' @param reference Character vector. Optional subset of individual IDs defining the reference population.
#'   If NULL, uses all individuals in the most recent generation.
#' @param mode Character. Type of contribution to calculate:
#' \itemize{
#'   \item \code{"founder"}: Founder contributions ($f_e$).
#'   \item \code{"ancestor"}: Ancestor contributions ($f_a$).
#'   \item \code{"both"}: Both founder and ancestor contributions.
#' }
#' @param top Integer. Number of top contributors to return. Default is 20.
#'
#' @return A list with class \code{pedcontrib} containing:
#' \itemize{
#'   \item \code{founders}: A \code{data.table} of founder contributions (if mode includes "founder", or "both").
#'   \item \code{ancestors}: A \code{data.table} of ancestor contributions (if mode includes "ancestor", or "both").
#'   \item \code{summary}: A \code{list} of summary statistics including:
#'     \itemize{
#'       \item \code{f_e}: Classical effective number of founders (\eqn{q=2}, Lacy 1989).
#'       \item \code{f_e_H}: Information-theoretic effective number of founders
#'         (\eqn{q=1}, Shannon entropy): \eqn{f_e^{(H)} = \exp(-\sum p_i \ln p_i)}.
#'       \item \code{f_a}: Classical effective number of ancestors (\eqn{q=2}, Boichard 1997).
#'       \item \code{f_a_H}: Information-theoretic effective number of ancestors
#'         (\eqn{q=1}): \eqn{f_a^{(H)} = \exp(-\sum q_k \ln q_k)}.
#'     }
#' }
#' 
#' Each contribution table contains:
#' \itemize{
#'   \item \code{Ind}: Individual ID.
#'   \item \code{Contrib}: Contribution to the reference population (0-1).
#'   \item \code{CumContrib}: Cumulative contribution.
#'   \item \code{Rank}: Rank by contribution.
#' }
#' 
#' @details
#' **Founder Contributions ($f_e$):**
#' Calculated by probabilistic gene flow from founders to the reference cohort. 
#' When individual ancestors with one unknown parent exist, "phantom" parents are temporarily injected correctly conserving the probability mass.
#' 
#' **Ancestor Contributions ($f_a$):**
#' Calculated using Boichard's iterative algorithm (1997), accounting for:
#' \itemize{
#'   \item Marginal genetic contribution of each ancestor
#'   \item Long-term contributions through multiple pathways
#' }
#' The parameter $f_a$ acts as a stringent metric since it identifies the bottlenecks of genetic variation in pedigrees. 
#' 
#' @examples
#' \donttest{
#' library(data.table)
#' # Load a sample pedigree
#' tp <- tidyped(small_ped)
#'
#' # Calculate both founder and ancestor contributions for reference population
#' ref_ids <- c("Z1", "Z2", "X", "Y")
#' contrib <- pedcontrib(tp, reference = ref_ids, mode = "both")
#'
#' # Print results including f_e, f_e(H), f_a, and f_a(H)
#' print(contrib)
#'
#' # Access Shannon-entropy effective numbers directly
#' contrib$summary$f_e_H   # Information-theoretic effective founders (q=1)
#' contrib$summary$f_e     # Classical effective founders (q=2)
#' contrib$summary$f_a_H   # Information-theoretic effective ancestors (q=1)
#' contrib$summary$f_a     # Classical effective ancestors (q=2)
#'
#' # Diversity ratio rho > 1 indicates long-tail founder value
#' contrib$summary$f_e_H / contrib$summary$f_e
#' }
#'
#' @references
#' Boichard, D., Maignel, L., & Verrier, É. (1997). The value of using probabilities 
#' of gene origin to measure genetic variability in a population. Genetics Selection 
#' Evolution, 29(1), 5-23.
#'
#' @export
pedcontrib <- function(ped, reference = NULL, mode = c("both", "founder", "ancestor"), top = 20) {
  ped <- ensure_complete_tidyped(ped, "pedcontrib()")
  
  # Inject phantom parents to conserve genetic contributions accurately
  ped <- fill_phantoms(ped)
  
  mode <- match.arg(mode)
  
  # Ensure integer indices exist
  if (!all(c("SireNum", "DamNum", "IndNum") %in% names(ped))) {
    ped <- tidyped(ped, addnum = TRUE)
  }
  
  # Define reference cohort
  if (is.null(reference)) {
    max_gen <- max(ped$Gen)
    reference <- ped[Gen == max_gen, Ind]
    message(sprintf("Using %d individuals from generation %d as reference population.",
                    length(reference), max_gen))
  } else {
    if (!all(reference %in% ped$Ind)) {
      missing <- reference[!reference %in% ped$Ind]
      warning(sprintf("Reference IDs not found in pedigree: %s", paste(head(missing, 5), collapse = ", ")))
    }
    reference <- reference[reference %in% ped$Ind]
  }
  
  if (length(reference) == 0) {
    stop("No valid reference individuals specified.")
  }
  
  n <- nrow(ped)
  n_ref <- length(reference)
  
  # Pre-compute integer arrays
  sire_num <- ped$SireNum   # 0 = unknown
  dam_num <- ped$DamNum     # 0 = unknown
  ind_ids <- ped$Ind
  
  # Map reference IDs to integer positions
  ind_to_pos <- setNames(seq_len(n), ind_ids)
  cohort_pos <- as.integer(ind_to_pos[reference])
  
  # Determine numeric mode for C++: 1=founder, 2=ancestor, 3=both
  cpp_mode <- switch(mode, founder = 1L, ancestor = 2L, both = 3L)
  
  # ---- Call Rcpp backend ----
  cpp_res <- cpp_pedcontrib(
    sire     = as.integer(sire_num),
    dam      = as.integer(dam_num),
    cohort_pos = cohort_pos,
    mode     = cpp_mode
  )
  
  result <- list()
  
  # ---- Founder Contributions ----
  fe2_sum <- NA_real_
  fe_H <- NA_real_
  n_founders_total <- 0L
  
  if (mode %in% c("founder", "both")) {
    message("Calculating founder contributions...")
    
    f_idx <- cpp_res$founder_idx
    f_val <- cpp_res$founder_contrib
    
    if (length(f_idx) == 0) {
      warning("No founders found (all individuals have at least one parent).")
      result$founders <- data.table(Ind = character(0), Contrib = numeric(0),
                                     CumContrib = numeric(0), Rank = integer(0))
    } else {
      founder_dt <- data.table(
        Ind = ind_ids[f_idx],
        Contrib = f_val
      )
      
      setorder(founder_dt, -Contrib)
      founder_dt[, CumContrib := cumsum(Contrib)]
      founder_dt[, Rank := seq_len(.N)]
      
      fe2_sum <- sum(founder_dt$Contrib^2, na.rm = TRUE)
      fe_H <- calc_shannon_effective(founder_dt$Contrib)
      n_founders_total <- nrow(founder_dt)
      
      result$founders <- if (nrow(founder_dt) > top) founder_dt[1:top] else founder_dt
    }
  }
  
  # ---- Ancestor Contributions ----
  fa2_sum <- NA_real_
  fa_H <- NA_real_
  n_ancestors_total <- 0L
  
  if (mode %in% c("ancestor", "both")) {
    message("Calculating ancestor contributions (Boichard's iterative algorithm)...")
    
    a_idx <- cpp_res$ancestor_idx
    a_val <- cpp_res$ancestor_contrib
    
    if (length(a_idx) == 0) {
      result$ancestors <- data.table(Ind = character(0), Contrib = numeric(0),
                                      CumContrib = numeric(0), Rank = integer(0))
    } else {
      ancestor_dt_full <- data.table(
        Ind = ind_ids[a_idx],
        Contrib = a_val
      )
      ancestor_dt_full[, CumContrib := cumsum(Contrib)]
      ancestor_dt_full[, Rank := seq_len(.N)]
      
      fa2_sum <- sum(ancestor_dt_full$Contrib^2, na.rm = TRUE)
      fa_H <- calc_shannon_effective(ancestor_dt_full$Contrib)
      n_ancestors_total <- nrow(ancestor_dt_full)
      
      result$ancestors <- if (nrow(ancestor_dt_full) > top) ancestor_dt_full[1:top] else ancestor_dt_full
    }
  }
  
  # ---- Summary Statistics ----
  summary_list <- list(n_ref = n_ref)
  
  if (!is.null(result$founders)) {
    summary_list$n_founder <- n_founders_total
    summary_list$n_founder_show <- nrow(result$founders)
    summary_list$f_e <- if (!is.na(fe2_sum) && fe2_sum > 0) 1 / fe2_sum else NA_real_
    summary_list$f_e_H <- fe_H
  }
  
  if (!is.null(result$ancestors)) {
    summary_list$n_ancestor <- n_ancestors_total
    summary_list$n_ancestor_show <- nrow(result$ancestors)
    summary_list$f_a <- if (!is.na(fa2_sum) && fa2_sum > 0) 1 / fa2_sum else NA_real_
    summary_list$f_a_H <- fa_H
  }
  
  result$summary <- summary_list
  class(result) <- "pedcontrib"
  
  return(result)
}

#' Calculate Ancestry Proportions
#'
#' Estimates the proportion of genes for each individual that originates from 
#' specific founder groups (e.g., breeds, source populations).
#'
#' @param ped A \code{tidyped} object.
#' @param foundervar Character. The name of the column containing founder-group
#'   labels (e.g., "Breed", "Origin").
#' @param target_labels Character vector. Specific founder-group labels to track.
#'   If NULL, all unique labels in \code{foundervar} among founders are used.
#'
#' @return A \code{data.table} with columns:
#' \itemize{
#'   \item \code{Ind}: Individual ID.
#'   \item One column per tracked label (named after each unique value in
#'     \code{foundervar} among founders, or as specified by
#'     \code{target_labels}).
#'     Each value gives the proportion of genes (0--1) originating from that
#'     founder group. Row sums across all label columns equal 1.
#' }
#'
#' @examples
#' \donttest{
#' library(data.table)
#' # Create dummy labels for founders
#' tp <- tidyped(small_ped)
#' tp_dated <- copy(tp)
#' founders <- tp_dated[is.na(Sire) & is.na(Dam), Ind]
#' # Assign 'LineA' and 'LineB'
#' tp_dated[Ind %in% founders[1:(length(founders)/2)], Origin := "LineA"]
#' tp_dated[is.na(Origin), Origin := "LineB"]
#' 
#' # Calculate ancestry proportions for all individuals
#' anc <- pedancestry(tp_dated, foundervar = "Origin")
#' print(tail(anc))
#' }
#' 
#' @export
pedancestry <- function(ped, foundervar, target_labels = NULL) {
  ped <- ensure_complete_tidyped(ped, "pedancestry()")
  if (!foundervar %in% names(ped)) stop(sprintf("Column '%s' not found.", foundervar))
  
  # Forward pass requires numbers
  if (!all(c("SireNum", "DamNum", "IndNum") %in% names(ped))) {
    ped_work <- tidyped(ped, addnum = TRUE)
  } else {
    ped_work <- copy(ped)
  }
  
  # Ensure sorted by Gen for correct forward propagation
  setorder(ped_work, Gen, IndNum)
  
  # Identify founder labels
  founders_idx <- which(is.na(ped_work$Sire) & is.na(ped_work$Dam))
  if (is.null(target_labels)) {
    target_labels <- unique(ped_work[[foundervar]][founders_idx])
    target_labels <- target_labels[!is.na(target_labels) & target_labels != ""]
  }
  
  if (length(target_labels) == 0) {
    stop("No valid founder-group labels found. Please check 'foundervar'.")
  }

  n <- nrow(ped_work)
  res_mat <- matrix(0, nrow = n, ncol = length(target_labels))
  colnames(res_mat) <- target_labels
  
  # Initialize founders with their respective labels using vectorized matrix indexing
  founder_labels <- ped_work[[foundervar]][founders_idx]
  valid_matches <- match(founder_labels, target_labels)
  valid_mask <- !is.na(valid_matches)
  
  if (any(valid_mask)) {
    res_mat[cbind(founders_idx[valid_mask], valid_matches[valid_mask])] <- 1.0
  }
  
  sire_nums <- ped_work$SireNum
  dam_nums <- ped_work$DamNum
  
  # Forward pass via Rcpp
  res_mat <- cpp_calculate_ancestry(
    sire_nums, 
    dam_nums, 
    res_mat
  )
  
  ans <- as.data.table(res_mat)
  colnames(ans) <- target_labels
  ans[, Ind := ped_work$Ind]
  setcolorder(ans, "Ind")
  return(ans[])
}

#' Summarize Inbreeding Levels
#'
#' Classifies individuals into inbreeding levels based on their inbreeding 
#' coefficients (F) according to standard or user-defined thresholds.
#'
#' @param ped A \code{tidyped} object.
#' @param breaks Numeric vector of strictly increasing positive upper bounds for
#'   inbreeding classes. Default is \code{c(0.0625, 0.125, 0.25)}, corresponding
#'   approximately to half-sib, avuncular/grandparent, and full-sib/parent-offspring
#'   mating thresholds. The class \code{"F = 0"} is always kept as a fixed first
#'   level. A final open-ended class \code{"F > max(breaks)"} is always appended
#'   automatically.
#' @param labels Optional character vector of interval labels. If \code{NULL},
#'   labels are generated automatically from \code{breaks}. When supplied, its
#'   length must equal \code{length(breaks)}, with each element naming the
#'   bounded interval \code{(breaks[i-1], breaks[i]]}. The open-ended tail
#'   class is always auto-generated and cannot be overridden.
#'
#' @return A \code{data.table} with 3 columns:
#' \describe{
#'   \item{\code{FClass}}{An ordered factor. By default it contains 5 levels:
#'   \code{"F = 0"}, \code{"0 < F <= 0.0625"}, \code{"0.0625 < F <= 0.125"},
#'   \code{"0.125 < F <= 0.25"}, and \code{"F > 0.25"}. The number of levels
#'   equals \code{length(breaks) + 2} (the fixed zero class plus one class per
#'   bounded interval plus the open-ended tail).}
#'   \item{\code{Count}}{Integer. Number of individuals in each class.}
#'   \item{\code{Percentage}}{Numeric. Percentage of individuals in each class.}
#' }
#'
#' @details
#' The default thresholds follow common pedigree interpretation rules:
#' \itemize{
#'   \item \code{F = 0.0625}: approximately the offspring of half-sib mating.
#'   \item \code{F = 0.125}: approximately the offspring of avuncular or grandparent-grandchild mating.
#'   \item \code{F = 0.25}: approximately the offspring of full-sib or parent-offspring mating.
#' }
#' Therefore, assigning \code{F = 0.25} to the class \code{"0.125 < F <= 0.25"}
#' is appropriate. If finer reporting is needed, supply custom \code{breaks},
#' for example to separate \code{0.25}, \code{0.375}, or \code{0.5}.
#'
#' @examples
#' tp <- tidyped(simple_ped, addnum = TRUE)
#' pedfclass(tp)
#'
#' # Finer custom classes (4 breaks, labels auto-generated)
#' pedfclass(tp, breaks = c(0.03125, 0.0625, 0.125, 0.25))
#'
#' # Custom labels aligned to breaks (3 labels for 3 breaks; tail is auto)
#' pedfclass(tp, labels = c("Low", "Moderate", "High"))
#'
#' \donttest{
#' tp_inbred <- tidyped(inbred_ped, addnum = TRUE)
#' pedfclass(tp_inbred)
#' }
#'
#' @export
pedfclass <- function(ped,
                      breaks = c(0.0625, 0.125, 0.25),
                      labels = NULL) {
  ped <- ensure_tidyped(ped)

  if (!is.numeric(breaks) || length(breaks) < 1L || any(!is.finite(breaks)) ||
      any(breaks <= 0) || is.unsorted(breaks, strictly = TRUE)) {
    stop("breaks must be a strictly increasing numeric vector of positive values")
  }

  # Build bounded-interval labels (one per break, no tail)
  if (is.null(labels)) {
    lower_bounds <- c(0, head(breaks, -1))
    labels <- ifelse(
      lower_bounds == 0,
      paste0("0 < F <= ", breaks),
      paste0(lower_bounds, " < F <= ", breaks)
    )
  } else {
    if (!is.character(labels) || length(labels) != length(breaks)) {
      stop("labels must be a character vector of length equal to length(breaks)")
    }
  }
  # The open-ended tail class is always auto-generated
  tail_label <- paste0("F > ", max(breaks))
  all_labels  <- c(labels, tail_label)
  
  if (!"f" %in% names(ped)) {
    ped <- ensure_complete_tidyped(ped, "pedfclass()")
    message("Calculating inbreeding coefficients...")
    ped <- inbreed(ped)
  }
  
  cut_breaks   <- c(-Inf, .Machine$double.eps, breaks, Inf)
  class_levels <- c("F = 0", all_labels)

  # Classify
  f_classes <- cut(ped$f, breaks = cut_breaks, labels = class_levels, include.lowest = TRUE)
  
  dt <- data.table(FClass = factor(f_classes, levels = class_levels, ordered = TRUE))
  summary_dt <- dt[, .(Count = .N), by = FClass]
  
  # Ensure all categories appear
  all_cats <- data.table(FClass = factor(class_levels, levels = class_levels, ordered = TRUE))
  summary_dt <- merge(all_cats, summary_dt, by = "FClass", all.x = TRUE)
  summary_dt[is.na(Count), Count := 0]
  summary_dt[, Count := as.integer(Count)]
  
  summary_dt[, Percentage := (Count / sum(Count)) * 100]
  setorder(summary_dt, FClass)
  
  return(summary_dt[])
}

#' Calculate Partial Inbreeding
#'
#' Decomposes individuals' inbreeding coefficients into marginal contributions from 
#' specific ancestors. This allows identifying which ancestors or lineages are 
#' responsible for the observed inbreeding.
#'
#' @param ped A \code{tidyped} object.
#' @param ancestors Character vector. IDs of ancestors to calculate partial 
#'   inbreeding for. If NULL, the top ancestors by marginal contribution are used.
#' @param top Integer. Number of top ancestors to include if \code{ancestors} is NULL.
#'
#' @return A \code{data.table} with the first column as \code{Ind} and subsequent 
#'   columns representing the partial inbreeding ($pF$) from each ancestor.
#'
#' @details
#' The sum of all partial inbreeding coefficients for an individual (including 
#' contributions from founders) equals $1 + f_i$, where $f_i$ is the total 
#' inbreeding coefficient. This function specifically isolates the terms in 
#' the Meuwissen & Luo (1992) decomposition that correspond to the selected ancestors.
#'
#' @references
#' Lacey, R. C. (1996). A formula for determining the partial inbreeding coefficient, 
#' \eqn{F_{ij}}. Journal of Heredity, 87(4), 337-339.
#' 
#' Meuwissen, T. H., & Luo, Z. (1992). Computing inbreeding coefficients in 
#' large populations. Genetics Selection Evolution, 24(4), 305-313.
#'
#' @examples
#' \donttest{
#' library(data.table)
#' tp <- tidyped(inbred_ped)
#' # Calculate partial inbreeding originating from specific ancestors
#' target_ancestors <- inbred_ped[is.na(Sire) & is.na(Dam), Ind]
#' pF <- pedpartial(tp, ancestors = target_ancestors)
#' print(tail(pF))
#' }
#' 
#' @export
pedpartial <- function(ped, ancestors = NULL, top = 20) {
  ped <- ensure_complete_tidyped(ped, "pedpartial()")
  
  # Need dii from inbreeding algorithm which requires integer indices
  if (!all(c("SireNum", "DamNum", "IndNum") %in% names(ped))) {
    ped_work <- tidyped(ped, addnum = TRUE)
  } else {
    ped_work <- copy(ped)
  }
  
  # Ensure sorted by IndNum for alignment with C++ output
  setorder(ped_work, IndNum)
  
  # Identify target ancestors
  if (is.null(ancestors)) {
    message("No ancestors specified. Identifying top contributors...")
    cont <- pedcontrib(ped, mode = "ancestor", top = top)
    ancestors <- cont$ancestors$Ind
    if (length(ancestors) == 0) stop("Could not identify top ancestors.")
  }
  
  # Map ancestor IDs to IndNum
  anc_nums <- ped_work[Ind %in% ancestors, IndNum]
  anc_ids <- ped_work[Ind %in% ancestors, Ind]
  
  if (length(anc_nums) == 0) stop("None of the specified ancestors found in pedigree.")
  
  message(sprintf("Calculating partial inbreeding for %d ancestors...", length(anc_nums)))
  
  # Calculate dii (Meuwissen & Luo engines)
  # We reuse the logic from cpp_calculate_inbreeding but only need dii
  res_f <- cpp_calculate_inbreeding(ped_work$SireNum, ped_work$DamNum)
  
  # Adjust dii for half-founders to match Gulisija & Crow (2007) tabular method.
  # For half-founders (one parent known, one unknown), the standard dii includes

  # 0.25 Mendelian sampling variance from the unknown parent that cannot be
  # attributed to any known ancestor. We subtract this to ensure that
  # sum(pF_j) over all founder ancestors j equals the total F_i.
  # Standard dii for half-founder: 0.75 - 0.25 * F_known
  # Adjusted dii for partial F:    0.50 - 0.25 * F_known
  dii_partial <- res_f$dii
  sire_vec <- ped_work$SireNum
  dam_vec <- ped_work$DamNum
  f_vec <- res_f$f
  half_founders <- which(xor(sire_vec > 0, dam_vec > 0))
  if (length(half_founders) > 0) {
    for (idx in half_founders) {
      known_parent <- if (sire_vec[idx] > 0) sire_vec[idx] else dam_vec[idx]
      dii_partial[idx] <- 0.5 - 0.25 * f_vec[known_parent]
    }
  }
  
  # Call C++ partial inbreeding engine
  pf_mat <- cpp_calculate_partial_inbreeding(
    ped_work$SireNum, 
    ped_work$DamNum, 
    dii_partial, 
    as.integer(anc_nums)
  )
  
  result <- as.data.table(pf_mat)
  colnames(result) <- anc_ids
  result[, Ind := ped_work$Ind]
  setcolorder(result, "Ind")
  
  return(result[])
}

#' Print Founder and Ancestor Contributions
#'
#' @param x A \code{pedcontrib} object.
#' @param ... Additional arguments.
#'
#' @export
print.pedcontrib <- function(x, ...) {
  cat("Founder and Ancestor Contributions\n")
  cat("===================================\n")
  
  s <- x$summary
  cat(sprintf("Reference population size: %d\n", s$n_ref))
  
  if (!is.null(x$founders)) {
    cat(sprintf("\nFounders: %d (reported top %d)\n", s$n_founder, s$n_founder_show))
    cat(sprintf("  f_e(H) = %.3f  |  f_e = %.3f\n",
                if (is.null(s$f_e_H) || is.na(s$f_e_H)) NA_real_ else s$f_e_H,
                s$f_e))
    cat("\nTop 10 Founder Contributions:\n")
    print(head(x$founders, 10))
  }
  
  if (!is.null(x$ancestors)) {
    cat(sprintf("\nAncestors: %d (reported top %d)\n", s$n_ancestor, s$n_ancestor_show))
    cat(sprintf("  f_a(H) = %.3f  |  f_a = %.3f\n",
                if (is.null(s$f_a_H) || is.na(s$f_a_H)) NA_real_ else s$f_a_H,
                s$f_a))
    cat("\nTop 10 Ancestor Contributions:\n")
    print(head(x$ancestors, 10))
  }
  
  invisible(x)
}

#' Calculate Genetic Diversity Indicators
#'
#' Combines founder/ancestor contributions ($f_e$, $f_a$) and effective population
#' size estimates (Ne) from three methods into a single summary object.
#'
#' @param ped A \code{tidyped} object.
#' @param reference Character vector. Optional subset of individual IDs defining the reference population.
#'   If NULL, uses all individuals in the most recent generation.
#' @param top Integer. Number of top contributors to return in founder/ancestor tables. Default is 20.
#' @param nsamples Integer. Number of individuals sampled per cohort for the coancestry
#'   Ne method and for \eqn{f_g} estimation. Very large cohorts are sampled down to this
#'   size to control memory usage (default: 1000).
#' @param ncores Integer. Number of cores for parallel processing in the coancestry method. Default is 1.
#' @param seed Integer or NULL. Random seed passed to \code{set.seed()} before sampling
#'   in the coancestry method, ensuring reproducible \eqn{f_g} and \eqn{N_e} estimates.
#'   Default is \code{NULL} (no fixed seed).
#'
#' @return A list with class \code{pediv} containing:
#' \itemize{
#'   \item \code{summary}: A single-row \code{data.table} with columns
#'     \code{NRef}, \code{NFounder}, \code{feH}, \code{fe}, \code{NAncestor},
#'     \code{faH}, \code{fa}, \code{fafe}, \code{fg}, \code{MeanCoan},
#'     \code{GeneDiv}, \code{NSampledCoan}, \code{NeCoancestry}, \code{NeInbreeding},
#'     \code{NeDemographic}.
#'     Here \code{feH} and \code{faH} are the Shannon-entropy (\eqn{q=1})
#'     effective numbers of founders and ancestors, respectively.
#'     \code{GeneDiv} is the pedigree-based retained genetic diversity,
#'     computed as \eqn{1 - \bar{C}}, where \eqn{\bar{C}} is the
#'     diagonal-corrected population mean coancestry (\code{MeanCoan}).
#'   \item \code{founders}: A \code{data.table} of top founder contributions.
#'   \item \code{ancestors}: A \code{data.table} of top ancestor contributions.
#' }
#'
#' @details
#' Internally calls \code{\link{pedcontrib}} for \eqn{f_e} and \eqn{f_a}.
#' The coancestry method is called via the internal \code{calc_ne_coancestry()}
#' function directly so that \eqn{f_g} and the Ne estimate can be obtained from
#' the same traced pedigree and the same batched matrix-free relationship
#' products without duplication.
#' The inbreeding and demographic Ne methods are obtained via \code{\link{pedne}}.
#' All calculations use the same \code{reference} population.
#' If any method fails (e.g., insufficient pedigree depth), its value is \code{NA}
#' rather than stopping execution.
#'
#' \eqn{f_g} (founder genome equivalents, Caballero & Toro 2000) is estimated from
#' the diagonal-corrected mean coancestry of the reference population:
#' \deqn{\hat{\bar{C}} = \frac{N-1}{N} \cdot \frac{\bar{a}_{off}}{2} + \frac{1 + \bar{F}_s}{2N}}
#' \deqn{f_g = \frac{1}{2 \hat{\bar{C}}}}
#' where \eqn{N} is the full reference cohort size, \eqn{\bar{a}_{off}} is the
#' off-diagonal mean relationship among sampled individuals, and \eqn{\bar{F}_s}
#' is their mean inbreeding coefficient.
#'
#' @examples
#' \donttest{
#' tp <- tidyped(small_ped)
#' div <- pediv(tp, reference = c("Z1", "Z2", "X", "Y"), seed = 42L)
#' print(div)
#'
#' # Access Shannon effective numbers from summary
#' div$summary$feH   # Shannon effective founders (q=1)
#' div$summary$faH   # Shannon effective ancestors (q=1)
#'
#' # Founder diversity profile: NFounder >= feH >= fe
#' with(div$summary, c(NFounder = NFounder, feH = feH, fe = fe))
#' }
#'
#' @seealso \code{\link{pedcontrib}}, \code{\link{pedne}}, \code{\link{pedstats}}
#' @export
pediv <- function(ped, reference = NULL, top = 20, nsamples = 1000, ncores = 1,
                  seed = NULL) {
  ped <- ensure_complete_tidyped(ped, "pediv()")

  # ---- Founder and ancestor contributions ----
  message("Calculating founder and ancestor contributions...")
  contrib <- pedcontrib(ped, reference = reference, mode = "both", top = top)
  s <- contrib$summary

  # ---- Prepare pedigree with ECG (mirrors pedne() preprocessing) ----
  ped_dt <- data.table::copy(ped)
  if (!"ECG" %in% names(ped_dt)) {
    ecg_dt <- pedecg(ped_dt)
    ped_dt <- merge(ped_dt, ecg_dt[, .(Ind, ECG)], by = "Ind", all.x = TRUE)
    # merge() sorts by Ind (alphabetically), breaking the IndNum == row-index
    # invariant that the fast-path BFS in tidyped() depends on.
    setorder(ped_dt, IndNum)
  }
  by_all <- ".CohortAll"
  ped_dt[[by_all]] <- "All"

  if (!is.null(reference)) {
    ped_subset <- as.data.table(ped_dt)[Ind %in% reference]
  } else {
    ped_subset <- ped_dt
  }

  # ---- Ne (coancestry) + fg: call internal function directly ----
  # pedne(method="coancestry") strips fg from its public output; bypass it here
  # so both Ne and fg are computed from the same single traced pedigree.
  message("Calculating Ne (coancestry) and fg...")
  raw_coan <- tryCatch(
    calc_ne_coancestry(
      ped_subset, ped_dt, by_all, nsamples,
      seed = seed, metrics = c("ne", "fg")
    ),
    error = function(e) NULL
  )

  # ---- Ne: inbreeding and demographic (via public pedne API) ----
  message("Calculating Ne (inbreeding)...")
  ne_i <- tryCatch(
    suppressMessages(pedne(ped, reference = reference, by = NULL,
                           method = "inbreeding")),
    error = function(e) NULL
  )

  message("Calculating Ne (demographic)...")
  ne_d <- tryCatch(
    suppressMessages(pedne(ped, reference = reference, by = NULL,
                           method = "demographic")),
    error = function(e) NULL
  )

  extract_ne <- function(dt) {
    if (is.null(dt) || nrow(dt) == 0) NA_real_ else dt$Ne[1]
  }

  # Extract fg and MeanCoan from raw coancestry result
  fg_val      <- if (!is.null(raw_coan) && nrow(raw_coan) > 0 && "fg" %in% names(raw_coan))
                   raw_coan$fg[1] else NA_real_
  mean_coan   <- if (!is.null(raw_coan) && nrow(raw_coan) > 0 && "MeanCoan" %in% names(raw_coan))
                   raw_coan$MeanCoan[1] else NA_real_
  n_samp_coan <- if (!is.null(raw_coan) && nrow(raw_coan) > 0 && "NSampledCoan" %in% names(raw_coan))
                   as.integer(raw_coan$NSampledCoan[1]) else NA_integer_

  # ---- Assemble summary ----
  summary_dt <- data.table::data.table(
    NRef          = s$n_ref,
    NFounder      = s$n_founder,
    feH           = if (!is.null(s$f_e_H)) s$f_e_H else NA_real_,
    fe            = s$f_e,
    NAncestor     = s$n_ancestor,
    faH           = if (!is.null(s$f_a_H)) s$f_a_H else NA_real_,
    fa            = s$f_a,
    fafe          = if (!is.na(s$f_a) && !is.na(s$f_e) && s$f_e > 0)
                      round(s$f_a / s$f_e, 6) else NA_real_,
    fg            = fg_val,
    MeanCoan      = mean_coan,
    GeneDiv       = if (!is.na(mean_coan)) pmax(0, pmin(1, 1 - mean_coan)) else NA_real_,
    NSampledCoan  = n_samp_coan,
    NeCoancestry  = extract_ne(raw_coan),
    NeInbreeding  = extract_ne(ne_i),
    NeDemographic = extract_ne(ne_d)
  )

  result <- list(
    summary   = summary_dt,
    founders  = contrib$founders,
    ancestors = contrib$ancestors
  )
  class(result) <- "pediv"
  return(result)
}

#' Print Genetic Diversity Summary
#'
#' @param x A \code{pediv} object.
#' @param ... Additional arguments.
#'
#' @export
print.pediv <- function(x, ...) {
  cat("Genetic Diversity Summary\n")
  cat("=========================\n")

  s <- x$summary

  cat(sprintf("Reference population size : %d\n", s$NRef))

  cat("\n-- Founder / Ancestor Contributions --\n")
  cat(sprintf("Founders  (total) : %d    fe(H) = %.3f    fe = %.3f\n",
              s$NFounder, s$feH, s$fe))
  cat(sprintf("Ancestors (total) : %d    fa(H) = %.3f    fa = %.3f\n",
              s$NAncestor, s$faH, s$fa))
  cat(sprintf("fa/fe ratio       : %.4f\n", s$fafe))
  if (!is.null(s$fg) && !is.na(s$fg)) {
    cat(sprintf("Founder genomes   : fg = %.3f  (MeanCoan = %.6f, NSampled = %d)\n",
                s$fg, s$MeanCoan, s$NSampledCoan))
    if (!is.null(s$GeneDiv) && !is.na(s$GeneDiv)) {
      cat(sprintf("Gene diversity    : GeneDiv = %.4f  (= 1 - MeanCoan)\n", s$GeneDiv))
    }
    # Hierarchy fg <= fa <= fe <= NFounder holds only when all metrics share
    # the same reference population. Display it only when the order is intact.
    if (!is.na(s$fa) && !is.na(s$fe) && s$fg <= s$fa + 1e-6 && s$fa <= s$fe + 1e-6) {
      cat(sprintf("Hierarchy: fg <= fa <= fe <= NFounder  =  %.2f <= %.3f <= %.3f <= %d\n",
                  s$fg, s$fa, s$fe, s$NFounder))
    }
  }

  cat("\n-- Effective Population Size (Ne) --\n")
  fmt_ne <- function(v) if (is.na(v)) "     NA" else sprintf("%7.1f", v)
  cat(sprintf("  Coancestry  : %s\n", fmt_ne(s$NeCoancestry)))
  cat(sprintf("  Inbreeding  : %s\n", fmt_ne(s$NeInbreeding)))
  cat(sprintf("  Demographic : %s\n", fmt_ne(s$NeDemographic)))

  if (!is.null(x$founders) && nrow(x$founders) > 0) {
    cat("\nTop Founder Contributions (top 5):\n")
    print(head(x$founders, 5))
  }

  if (!is.null(x$ancestors) && nrow(x$ancestors) > 0) {
    cat("\nTop Ancestor Contributions (top 5):\n")
    print(head(x$ancestors, 5))
  }

  invisible(x)
}

library(testthat)
library(visPedigree)
library(data.table)

# ===========================================================================
# Test Suite for pedcontrib()
# Purpose: Capture exact numerical outputs of the current (pure-R) implementation
#          so that any future Rcpp optimisation must produce identical results.
# ===========================================================================

# ---- Fixture 1: 10-individual "theory" pedigree (瓶颈模型) ----
# Hand-derived expected values from the manuscript
make_theory_ped <- function() {
  ped_dt <- data.table(
    Ind  = c("F1", "F2", "F3", "F4", "S1", "D1", "R1", "R2", "R3", "R4"),
    Sire = c(NA,   NA,   NA,   NA,  "F1", "F3", "S1", "S1", "S1", "S1"),
    Dam  = c(NA,   NA,   NA,   NA,  "F2", "F4", "D1", "D1", "D1", "F4")
  )
  suppressMessages(tidyped(ped_dt))
}

# ---- Fixture 2: small_ped (visPedigree built-in dataset) ----
make_small_ped <- function() {
  suppressMessages(tidyped(small_ped))
}

# ---- Fixture 3: two founders with one offspring, candidates are founders ----
make_founders_as_cand_ped <- function() {
  ped_dt <- data.table(
    Ind  = c("A", "B", "C"),
    Sire = c(NA, NA, "A"),
    Dam  = c(NA, NA, "B")
  )
  suppressMessages(tidyped(ped_dt))
}

# ---- Fixture 4: single offspring from two founders ----
make_minimal_ped <- function() {
  ped_dt <- data.table(
    Ind  = c("P1", "P2", "O1"),
    Sire = c(NA,  NA,  "P1"),
    Dam  = c(NA,  NA,  "P2")
  )
  suppressMessages(tidyped(ped_dt))
}

# ---- Fixture 5: single candidate (edge case) ----
make_single_cand_ped <- function() {
  ped_dt <- data.table(
    Ind  = c("A", "B", "C", "D"),
    Sire = c(NA,  NA, "A", "A"),
    Dam  = c(NA,  NA, "B", "B")
  )
  suppressMessages(tidyped(ped_dt))
}

# ---- Fixture 6: deeper chain (4 generations) ----
make_deep_chain_ped <- function() {
  ped_dt <- data.table(
    Ind  = c("G1M", "G1F", "G2", "G3M", "G3F", "G4"),
    Sire = c(NA,    NA,   "G1M", "G2",  NA,   "G3M"),
    Dam  = c(NA,    NA,   "G1F", NA,    NA,   "G3F")
  )
  suppressMessages(tidyped(ped_dt))
}


# ==========================================================================
# Group 1: Founder contributions (p_i) and f_e
# ==========================================================================

test_that("Theory pedigree: founder contributions match hand-derived values", {
  tp <- make_theory_ped()
  cand <- c("R1", "R2", "R3", "R4")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "founder", top = 100))

  # Extract founder table and sort by name for deterministic comparison
  fd <- res$founders
  setkey(fd, Ind)

  expect_equal(fd["F1", Contrib], 0.25,     tolerance = 1e-10)
  expect_equal(fd["F2", Contrib], 0.25,     tolerance = 1e-10)
  expect_equal(fd["F3", Contrib], 0.1875,   tolerance = 1e-10)
  expect_equal(fd["F4", Contrib], 0.3125,   tolerance = 1e-10)
})

test_that("Theory pedigree: f_e matches hand-derived value (3.878788)", {
  tp <- make_theory_ped()
  cand <- c("R1", "R2", "R3", "R4")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "both", top = 100))

  expected_fe <- 1 / (0.25^2 + 0.25^2 + 0.1875^2 + 0.3125^2)
  expect_equal(res$summary$f_e, expected_fe, tolerance = 1e-6)
})

test_that("Theory pedigree: founder contributions sum to 1.0", {
  tp <- make_theory_ped()
  cand <- c("R1", "R2", "R3", "R4")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "founder", top = 100))

  expect_equal(sum(res$founders$Contrib), 1.0, tolerance = 1e-10)
})


# ==========================================================================
# Group 2: Ancestor marginal contributions (q_i) and f_a
# ==========================================================================

test_that("Theory pedigree: ancestor marginal contributions match hand-derived values", {
  tp <- make_theory_ped()
  cand <- c("R1", "R2", "R3", "R4")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "ancestor", top = 100))

  anc <- res$ancestors
  # By rank order: S1(0.5), D1(0.375), F4(0.125)
  expect_equal(anc$Contrib[1], 0.500, tolerance = 1e-10)
  expect_equal(anc$Contrib[2], 0.375, tolerance = 1e-10)
  expect_equal(anc$Contrib[3], 0.125, tolerance = 1e-10)
  expect_equal(anc$Ind[1], "S1")
  expect_equal(anc$Ind[2], "D1")
  expect_equal(anc$Ind[3], "F4")
})

test_that("Theory pedigree: f_a matches hand-derived value (2.461538)", {
  tp <- make_theory_ped()
  cand <- c("R1", "R2", "R3", "R4")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "both", top = 100))

  expected_fa <- 1 / (0.5^2 + 0.375^2 + 0.125^2)
  expect_equal(res$summary$f_a, expected_fa, tolerance = 1e-6)
})

test_that("Theory pedigree: ancestor contributions sum to 1.0", {
  tp <- make_theory_ped()
  cand <- c("R1", "R2", "R3", "R4")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "ancestor", top = 100))

  expect_equal(sum(res$ancestors$Contrib), 1.0, tolerance = 1e-10)
})


# ==========================================================================
# Group 3: small_ped dataset (complex real-world pedigree)
# ==========================================================================

test_that("small_ped: f_e and f_a match known values", {
  tp <- make_small_ped()
  cand <- c("Z1", "Z2", "Y", "X")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "both", top = 100))

  # These values are from the current pure-R implementation runs
  expect_equal(res$summary$f_e, 6.585209, tolerance = 1e-4)
  expect_equal(res$summary$f_a, 2.666667, tolerance = 1e-4)
})

test_that("small_ped: founder contributions sum to 1.0", {
  tp <- make_small_ped()
  cand <- c("Z1", "Z2", "Y", "X")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "founder", top = 100))

  expect_equal(sum(res$founders$Contrib), 1.0, tolerance = 1e-10)
})

test_that("small_ped: ancestor contributions sum to 1.0", {
  tp <- make_small_ped()
  cand <- c("Z1", "Z2", "Y", "X")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "ancestor", top = 100))

  expect_equal(sum(res$ancestors$Contrib), 1.0, tolerance = 1e-10)
})

test_that("small_ped: top ancestor is N with q = 0.50", {
  tp <- make_small_ped()
  cand <- c("Z1", "Z2", "Y", "X")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "ancestor", top = 100))

  expect_equal(res$ancestors$Ind[1], "X")
  expect_equal(res$ancestors$Contrib[1], 0.50, tolerance = 1e-10)
})

test_that("small_ped: ancestor ranking order is correct", {
  tp <- make_small_ped()
  cand <- c("Z1", "Z2", "Y", "X")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "ancestor", top = 100))

  # Expected order: N(0.50), V(0.25), R(0.125), K(0.125), W(0.125), J1(0.0625), O(0.03125)
  expect_equal(res$ancestors$Ind[1], "X")
  expect_equal(res$ancestors$Ind[2], "N")
  expect_equal(res$ancestors$Contrib[2], 0.25, tolerance = 1e-10)
})


# ==========================================================================
# Group 4: f_e and f_a invariance under `top` parameter
# ==========================================================================

test_that("f_e and f_a are invariant to the top parameter", {
  tp <- make_theory_ped()
  cand <- c("R1", "R2", "R3", "R4")

  res_full <- suppressMessages(pedcontrib(tp, reference = cand, mode = "both", top = 100))
  res_top1 <- suppressMessages(pedcontrib(tp, reference = cand, mode = "both", top = 1))

  expect_equal(res_full$summary$f_e, res_top1$summary$f_e)
  expect_equal(res_full$summary$f_a, res_top1$summary$f_a)
})


# ==========================================================================
# Group 5: Edge cases
# ==========================================================================

test_that("Single candidate: founder and ancestor contributions work", {
  tp <- make_single_cand_ped()
  res <- suppressMessages(pedcontrib(tp, reference = "C", mode = "both", top = 100))

  # Single individual C has parents A and B
  # Founder contributions: A = 0.5, B = 0.5
  fd <- res$founders
  setkey(fd, Ind)
  expect_equal(fd["A", Contrib], 0.5, tolerance = 1e-10)
  expect_equal(fd["B", Contrib], 0.5, tolerance = 1e-10)
  expect_equal(sum(fd$Contrib), 1.0, tolerance = 1e-10)

  # Ancestor contributions should also sum to 1.0
  expect_equal(sum(res$ancestors$Contrib), 1.0, tolerance = 1e-10)
})

test_that("Minimal pedigree (1 offspring from 2 founders): correct values", {
  tp <- make_minimal_ped()
  res <- suppressMessages(pedcontrib(tp, reference = "O1", mode = "both", top = 100))

  # f_e: 1/(0.5^2 + 0.5^2) = 2.0
  expect_equal(res$summary$f_e, 2.0, tolerance = 1e-10)

  # Founders: P1=0.5, P2=0.5
  fd <- res$founders
  setkey(fd, Ind)
  expect_equal(fd["P1", Contrib], 0.5, tolerance = 1e-10)
  expect_equal(fd["P2", Contrib], 0.5, tolerance = 1e-10)
})

test_that("Deep chain: contributions propagate correctly through 4 generations", {
  tp <- make_deep_chain_ped()
  res <- suppressMessages(pedcontrib(tp, reference = "G4", mode = "both", top = 100))

  # G4's parents are G3M and G3F

  # G3F is a founder -> contributes 0.5 to G4
  # G3M's parent is G2 (sire side only)
  # G2's parents are G1M and G1F
  # So G1M -> G2 -> G3M -> G4: contribution = 0.5^3 = 0.125
  # G1F -> G2 -> G3M -> G4: contribution = 0.5^3 = 0.125
  # G3M has no dam, so half of G3M is "missing" => goes to phantom founder
  # G2 -> G3M: 0.5 (sire contribution to G3M), then G3M -> G4: 0.5
  # So from G2: 0.25 to G4

  fd <- res$founders
  setkey(fd, Ind)
  expect_equal(fd["G3F", Contrib], 0.5, tolerance = 1e-10)

  # G3M has no dam => 25% of G4's genes go to a phantom founder
  # Real founders account for only 0.75 of total variation
  expect_equal(sum(fd$Contrib), 1.0, tolerance = 1e-10)

  # Ancestor contributions also reflect the same incomplete pedigree
  expect_true(sum(res$ancestors$Contrib) <= 1.0 + 1e-10)
})

test_that("Founders as candidates: they are their own founders, no ancestors above", {
  tp <- make_founders_as_cand_ped()
  # Use the two founders A and B as candidates
  res <- suppressMessages(pedcontrib(tp, reference = c("A", "B"), mode = "both", top = 100))

  # Founder contributions: A = 0.5, B = 0.5
  expect_equal(nrow(res$founders), 2)
  expect_equal(sum(res$founders$Contrib), 1.0, tolerance = 1e-10)

  # f_e = 1 / (0.5^2 + 0.5^2) = 2.0
  expect_equal(res$summary$f_e, 2.0, tolerance = 1e-10)

  # No ancestors should be found (founders have no parents)
  expect_equal(nrow(res$ancestors), 2)
})


# ==========================================================================
# Group 6: mode parameter behaviour
# ==========================================================================

test_that("mode='founder' does not produce ancestors", {
  tp <- make_theory_ped()
  cand <- c("R1", "R2", "R3", "R4")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "founder"))

  expect_true(!is.null(res$founders))
  expect_true(is.null(res$ancestors))
})

test_that("mode='ancestor' does not produce founders", {
  tp <- make_theory_ped()
  cand <- c("R1", "R2", "R3", "R4")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "ancestor"))

  expect_true(is.null(res$founders))
  expect_true(!is.null(res$ancestors))
})

test_that("mode='both' produces both founders and ancestors", {
  tp <- make_theory_ped()
  cand <- c("R1", "R2", "R3", "R4")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "both"))

  expect_true(!is.null(res$founders))
  expect_true(!is.null(res$ancestors))
})


# ==========================================================================
# Group 7: Return structure and class
# ==========================================================================

test_that("pedcontrib returns correct structure", {
  tp <- make_theory_ped()
  cand <- c("R1", "R2", "R3", "R4")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "both"))

  expect_s3_class(res, "pedcontrib")
  expect_true(is.list(res))
  expect_true(all(c("founders", "ancestors", "summary") %in% names(res)))

  # founders table columns
  expect_true(all(c("Ind", "Contrib", "CumContrib", "Rank") %in% names(res$founders)))

  # ancestors table columns
  expect_true(all(c("Ind", "Contrib", "CumContrib", "Rank") %in% names(res$ancestors)))

  # summary fields
  expect_true(all(c("n_ref", "n_founder", "f_e",
                     "n_ancestor", "f_a") %in% names(res$summary)))
})

test_that("CumContrib is monotonically non-decreasing", {
  tp <- make_small_ped()
  cand <- c("Z1", "Z2", "Y", "X")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "both", top = 100))

  expect_true(all(diff(res$founders$CumContrib) >= -1e-15))
  expect_true(all(diff(res$ancestors$CumContrib) >= -1e-15))
})


# ==========================================================================
# Group 8: Error handling
# ==========================================================================

test_that("pedcontrib rejects non-tidyped input", {
  expect_error(pedcontrib(data.frame(a = 1)), "tidyped")
})

test_that("pedcontrib errors on empty candidate set", {
  tp <- make_theory_ped()
  expect_error(
    suppressMessages(suppressWarnings(pedcontrib(tp, reference = "NONEXISTENT", mode = "both"))),
    "No valid reference"
  )
})


# ==========================================================================
# Group 9: Relationship between f_e and f_a
# ==========================================================================

test_that("f_a <= f_e (bottleneck effect)", {
  tp <- make_theory_ped()
  cand <- c("R1", "R2", "R3", "R4")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "both", top = 100))

  expect_true(res$summary$f_a <= res$summary$f_e)
})

test_that("small_ped: f_a <= f_e", {
  tp <- make_small_ped()
  cand <- c("Z1", "Z2", "Y", "X")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "both", top = 100))

  expect_true(res$summary$f_a <= res$summary$f_e)
})


# ==========================================================================
# Group 10: Shannon entropy effective numbers f_e(H) and f_a(H)
# ==========================================================================

test_that("Theory pedigree: f_e_H matches hand-derived value", {
  tp <- make_theory_ped()
  cand <- c("R1", "R2", "R3", "R4")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "founder", top = 100))

  p <- c(0.25, 0.25, 0.1875, 0.3125)
  expected_fe_H <- exp(-sum(p * log(p)))
  expect_equal(res$summary$f_e_H, expected_fe_H, tolerance = 1e-6)
})

test_that("Theory pedigree: f_e_H >= f_e (Hill monotonicity)", {
  tp <- make_theory_ped()
  cand <- c("R1", "R2", "R3", "R4")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "founder", top = 100))

  expect_true(res$summary$f_e_H >= res$summary$f_e)
})

test_that("Theory pedigree: f_a_H >= f_a", {
  tp <- make_theory_ped()
  cand <- c("R1", "R2", "R3", "R4")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "both", top = 100))

  expect_true(res$summary$f_a_H >= res$summary$f_a)
})

test_that("Uniform contributions: f_e_H == f_e == Nf", {
  # Two founders each contributing exactly 0.5
  tp <- make_founders_as_cand_ped()
  res <- suppressMessages(pedcontrib(tp, reference = "C", mode = "founder", top = 100))

  # Both parents contribute equally -> uniform
  expect_equal(res$summary$f_e_H, res$summary$f_e, tolerance = 1e-6)
  expect_equal(res$summary$f_e_H, res$summary$n_founder, tolerance = 1e-6)
})

test_that("small_ped: f_e_H >= f_e", {
  tp <- make_small_ped()
  cand <- c("Z1", "Z2", "Y", "X")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "both", top = 100))

  expect_true(res$summary$f_e_H >= res$summary$f_e)
  expect_true(res$summary$f_a_H >= res$summary$f_a)
})

test_that("f_e_H is NA when no founders found", {
  tp <- make_theory_ped()
  cand <- c("R1", "R2", "R3", "R4")
  res <- suppressMessages(pedcontrib(tp, reference = cand, mode = "ancestor", top = 100))

  # When mode = "ancestor", founders is NULL, f_e_H should not exist

  expect_null(res$summary$f_e_H)
  # But f_a_H should exist
  expect_true(!is.null(res$summary$f_a_H) && !is.na(res$summary$f_a_H))
})

library(testthat)
library(visPedigree)
library(data.table)

# Create a minimal testing dataset
test_ped_dt <- data.table::data.table(
  Ind = c("A", "B", "C", "D", "E", "F"),
  Sire = c(NA, NA, "A", "A", "C", "C"),
  Dam = c(NA, NA, "B", "B", "D", "D"),
  Sex = c("male", "female", "male", "female", "male", "female"),
  Gen = c(1, 1, 2, 2, 3, 3),
  Group = c("G1", "G1", "G2", "G2", "G3", "G3")
)
# C and D are full siblings (F _ offspring = 0.25). 
# E and F are offspring of full siblings.

test_ped <- suppressMessages(tidyped(test_ped_dt, reference = c("E", "F")))

test_that("pedrel calculates relation correctly and handles boundaries", {
  # N < 2 case for G1 (after subsetting by Gen, if one was dropped, or just test N=1)
  dt_n1 <- data.table::data.table(Ind = c("A", "B", "C"), Sire = c(NA, NA, "A"), Dam = c(NA, NA, "B"), Sex = c("male", "female", "male"), Gen = c(1, 1, 2))
  ped_n1 <- suppressMessages(tidyped(dt_n1))
  
  res_n1 <- suppressWarnings(pedrel(ped_n1, by = "Gen"))
  expect_equal(res_n1[Gen == 2, NUsed], 1)
  expect_true(is.na(res_n1[Gen == 2, MeanRel]))
  expect_equal(res_n1[Gen == 2, Status], "skipped")
  expect_match(res_n1[Gen == 2, Message], "less than 2 individuals")
  
  # For test_ped, group by Gen
  rel <- pedrel(test_ped, by = "Gen")
  
  # Gen 1: A, B (Unrelated) -> 0
  expect_equal(rel[Gen == 1, MeanRel], 0)
  
  # Gen 2: C, D (full sibs). a_ij = 0.5. Mean of off-diagonals = 0.5
  expect_equal(rel[Gen == 2, MeanRel], 0.5)
  
  # Test sample size behavior (less than 2 individuals after filtering)
  rel_sample <- suppressWarnings(pedrel(test_ped, by = "Gen", reference = c("A", "C")))
  expect_true(all(is.na(rel_sample$MeanRel)))
  expect_true(all(rel_sample$Status == "skipped"))
  
  # Test with a different by parameter (e.g. BirthYear)
  test_ped$YearGroup <- c(1, 1, 2, 2, 3, 3) 
  rel_year <- pedrel(test_ped, by = "YearGroup")
  expect_true("YearGroup" %in% names(rel_year))
  expect_false("Group" %in% names(rel_year))
  expect_equal(rel_year[YearGroup == 2, MeanRel], 0.5) # C, D in YearGroup 2
  
  # Test reference parameter filtering with another by variable
  # If reference is only C, D, E, then YearGroup 3 should only have E left -> NUsed=1, NA MeanRel
  rel_year_ref <- suppressWarnings(pedrel(test_ped, by = "YearGroup", reference = c("C", "D", "E")))
  expect_equal(rel_year_ref[YearGroup == 2, NUsed], 2)
  expect_equal(rel_year_ref[YearGroup == 2, MeanRel], 0.5)
  expect_equal(rel_year_ref[YearGroup == 3, NUsed], 1)
  expect_true(is.na(rel_year_ref[YearGroup == 3, MeanRel]))
  expect_equal(rel_year_ref[YearGroup == 3, Status], "skipped")

  coan <- pedrel(test_ped, by = "Gen", scale = "coancestry")
  expect_true("MeanCoan" %in% names(coan))
  expect_false("MeanRel" %in% names(coan))
  expect_equal(coan[Gen == 1, MeanCoan], 0.25)
  expect_equal(coan[Gen == 2, MeanCoan], 0.375)
  expect_equal(coan[Gen == 3, MeanCoan], 0.5)
})

test_that("pedrel captures deep inbreeding and correct ancestor tracing", {
  # Deep inbreeding: Full-sib mating (Gen 3), then their offspring full-sib mating (Gen 4)
  dt_deep <- data.table::data.table(
    Ind  = c("A", "B", "C", "D", "E", "F", "G", "H"),
    Sire = c(NA, NA, "A", "A", "C", "C", "E", "E"),
    Dam  = c(NA, NA, "B", "B", "D", "D", "F", "F"),
    Gen  = c(1, 1, 2, 2, 3, 3, 4, 4)
  )
  tp_deep <- tidyped(dt_deep)
  rel_deep <- pedrel(tp_deep, by = "Gen")
  
  # Gen 1: Unrelated founders
  expect_equal(rel_deep[Gen == 1, MeanRel], 0.0)
  
  # Gen 2: C, D are full sibs (0.5)
  expect_equal(rel_deep[Gen == 2, MeanRel], 0.5)
  
  # Gen 3: E, F's parents are C, D (full sibs). a_EF = 0.5 + 0.5 * f_CD = 0.5 + 0.5 * 0.5 = 0.75
  # Error check: if ancestor tracing failed, this would be 0.5
  expect_equal(rel_deep[Gen == 3, MeanRel], 0.75)
  
  # Gen 4: G, H's parents are E, F. a_GH = 0.5 + 0.5 * f_EF = 0.5 + 0.5 * 1.0 (since F_E=0.25+0.25*0.5=0.375? No, a_EF=0.75)
  # Actually a_GH = 0.5 + 0.5 * a_EF = 0.5 + 0.5 * 0.75 = 0.875? 
  # Wait, manual calibration: 
  # A_mat[G, H] = 0.5 * (A[E, G] + A[F, G]) = 0.5 * (0.5*(A[E,E]+A[E,F]) + 0.5*(A[F,E]+A[F,F]))
  # A[E,E] = 1 + f_CD = 1 + 0.25 = 1.25. A[E,F] = 0.75.
  # A_mat[G,H] = 0.5 * (0.5*(1.25+0.75) + 0.5*(0.75+1.25)) = 0.5 * (1.0 + 1.0) = 1.0
  expect_equal(rel_deep[Gen == 4, MeanRel], 1.0)

  coan_deep <- pedrel(tp_deep, by = "Gen", scale = "coancestry")
  expect_equal(coan_deep[Gen == 1, MeanCoan], 0.25)
  expect_equal(coan_deep[Gen == 2, MeanCoan], 0.375)
  expect_equal(coan_deep[Gen == 3, MeanCoan], 0.5)
  expect_equal(coan_deep[Gen == 4, MeanCoan], 0.59375)
})

test_that("pedrel retains compact as a backward-compatible no-op", {
  baseline <- pedrel(test_ped, by = "Gen")
  legacy_compact <- pedrel(test_ped, by = "Gen", compact = TRUE)

  expect_equal(legacy_compact, baseline)
})

test_that("pedgenint computes Average from all parent-offspring pairs", {
  test_ped$BirthYear <- c(2000, 2001, 2005, 2006, 2010, 2012)
  # SS: C-A (5), E-C (5) -> mean=5

  # SD: D-A (6), F-C (7) -> mean=6.5
  # DS: C-B (4), E-D (4) -> mean=4
  # DD: D-B (5), F-D (6) -> mean=5.5
  # Average from ALL 8 pairs: (5+6+5+7+4+5+4+6)/8 = 5.25
  # Note: numeric years auto-converted to Date (YYYY-07-01); leap years cause
  # tiny deviations from exact integers, so we use tolerance = 0.01.
  
  suppressMessages(
    genint_res <- pedgenint(test_ped, timevar = "BirthYear", unit = "year")
  )
  
  avg_res <- genint_res[Pathway == "Average"]
  expect_equal(avg_res$Mean, 5.25, tolerance = 0.01)
  expect_equal(avg_res$N, 8L)
  expect_true(!is.na(avg_res$SD))
  
  # Sex-specific pathways should still work
  expect_equal(genint_res[Pathway == "SS", Mean], 5, tolerance = 0.01)
  expect_equal(genint_res[Pathway == "SD", Mean], 6.5, tolerance = 0.01)
  expect_equal(genint_res[Pathway == "DS", Mean], 4, tolerance = 0.01)
  expect_equal(genint_res[Pathway == "DD", Mean], 5.5, tolerance = 0.01)
  
  # SO/DO: sex-independent pathways
  # SO (Sire→Offspring): A→C(5), A→D(6), C→E(5), C→F(7) -> N=4, Mean=5.75
  so_res <- genint_res[Pathway == "SO"]
  expect_equal(so_res$N, 4L)
  expect_equal(so_res$Mean, 5.75, tolerance = 0.01)
  
  # DO (Dam→Offspring): B→C(4), B→D(5), D→E(4), D→F(6) -> N=4, Mean=4.75
  do_res <- genint_res[Pathway == "DO"]
  expect_equal(do_res$N, 4L)
  expect_equal(do_res$Mean, 4.75, tolerance = 0.01)
  
  # All 7 pathways present
  expect_equal(sort(genint_res$Pathway),
               c("Average", "DD", "DO", "DS", "SD", "SO", "SS"))
})

test_that("pedgenint works with Date and character date inputs", {
  # Date input — exact control, no auto-conversion message
  test_ped$BirthDate <- as.Date(c(
    "2000-03-15", "2001-06-20",  # A, B
    "2005-03-15", "2006-06-20",  # C, D
    "2010-03-15", "2012-06-20"   # E, F
  ))
  genint_date <- suppressMessages(
    pedgenint(test_ped, timevar = "BirthDate", unit = "day")
  )
  expect_true(all(genint_date$Mean > 0))
  expect_true("Average" %in% genint_date$Pathway)

  # Character date input
  test_ped$BirthStr <- as.character(test_ped$BirthDate)
  genint_str <- suppressMessages(
    pedgenint(test_ped, timevar = "BirthStr", unit = "month")
  )
  expect_true(all(genint_str$Mean > 0))

  # unit = "hour"
  genint_hr <- suppressMessages(
    pedgenint(test_ped, timevar = "BirthDate", unit = "hour")
  )
  expect_true(all(genint_hr$Mean > 0))
})

test_that("pedcontrib f_e and f_a compute on full sets despite top cutoff", {
  cont_all <- suppressMessages(pedcontrib(test_ped, mode = "both", top = 100))
  cont_top1 <- suppressMessages(pedcontrib(test_ped, mode = "both", top = 1))
  
  # f_e and f_a should be identical because they're based on full arrays before truncation
  expect_equal(cont_all$summary$f_e, cont_top1$summary$f_e)
  expect_equal(cont_all$summary$f_a, cont_top1$summary$f_a)
  
  # Ensure reported reflects top
  expect_equal(cont_top1$summary$n_founder_show, 1)
  expect_equal(cont_all$summary$n_founder_show, length(cont_all$founders$Ind))
  expect_true(cont_top1$summary$n_founder > 1) 
})

test_that("pedpartial and pedancestry run without addnum=TRUE initially", {
  # Create tidyped without IndNum (addnum = FALSE)
  ped_nonum <- suppressMessages(tidyped(test_ped_dt, addnum = FALSE))
  
  expect_false("IndNum" %in% names(ped_nonum))
  
  # Should not crash and automatically handle missing nums
  part_res <- pedpartial(ped_nonum, ancestors = c("A", "B"), top = 2)
  expect_true(all(c("A", "B") %in% names(part_res)))
  expect_equal(nrow(part_res), nrow(ped_nonum))
  
  # pedancestry with labels
  ped_nonum$Label <- ifelse(ped_nonum$Ind %in% c("A", "B"), "Base", NA)
  anc_res <- pedancestry(ped_nonum, foundervar = "Label")
  expect_true("Base" %in% names(anc_res))
  # All non-founders descend completely from A and B, so Base should be 1.0 for all
  expect_true(all(anc_res$Base == 1.0))
})

# --- Additional tests for pedne, pedecg, pedsubpop, pedfclass ---

test_that("pedecg computes equivalent complete generations", {
  ecg_res <- pedecg(test_ped)
  expect_true(all(c("ECG", "FullGen", "MaxGen") %in% names(ecg_res)))
  expect_equal(nrow(ecg_res), nrow(test_ped))
  # Founders A, B have ECG = 0
  expect_equal(ecg_res[Ind == "A", ECG], 0)
  expect_equal(ecg_res[Ind == "B", ECG], 0)
  # C has 2 known parents -> ECG = 1
  expect_equal(ecg_res[Ind == "C", ECG], 1)
  # E has parents C, D who each have ECG=1 -> ECG = 1 + (1+1)/2 = 2
  expect_equal(ecg_res[Ind == "E", ECG], 2)
})

test_that("pedne computes effective population size by cohort", {
  test_ped$BirthYear <- c(2000, 2000, 2005, 2005, 2010, 2010)
  res <- suppressMessages(pedne(test_ped, by = "BirthYear"))
  expect_true(all(c("Cohort", "N", "DeltaC", "Ne") %in% names(res)))
  # Only cohort 2010 has ECG > 1 (founders/gen1 filtered out)
  expect_equal(nrow(res), 3)
  expect_equal(res$Cohort, c(2000, 2005, 2010))
  expect_true(any(res$DeltaC >= 0, na.rm=TRUE))
  expect_true(any(res$Ne > 0 & is.finite(res$Ne)))
})

test_that("pedsubpop splits pedigree by grouping variable", {
  res_gen <- pedsubpop(test_ped, by = "Gen")
  expect_true(is.data.table(res_gen))
  expect_equal(nrow(res_gen), length(unique(test_ped$Gen)))
  expect_true(all(c("Group", "N") %in% names(res_gen)))
  # Gen 1 (Group=1) has 2 individuals
  expect_equal(res_gen[Group == 1, N], 2)
  expect_equal(res_gen[Group == 2, N], 2)
})

test_that("pedsubpop summarizes disconnected groups and isolated individuals", {
  ped_multi_dt <- data.table::data.table(
    Ind = c("A", "B", "C", "D", "E", "F", "ISO"),
    Sire = c(NA, NA, "A", NA, NA, "D", NA),
    Dam = c(NA, NA, "B", NA, NA, "E", NA),
    Sex = c("male", "female", "male", "male", "female", "male", NA_character_)
  )
  ped_multi <- suppressMessages(tidyped(ped_multi_dt))

  res_null <- pedsubpop(ped_multi)

  expect_true(is.data.table(res_null))
  expect_equal(nrow(res_null), 3)
  expect_true(all(c("GP1", "GP2", "Isolated") %in% res_null$Group))

  expect_equal(res_null[Group == "GP1", N], 3)
  expect_equal(res_null[Group == "GP1", N_Founder], 2)

  expect_equal(res_null[Group == "GP2", N], 3)
  expect_equal(res_null[Group == "GP2", N_Founder], 2)

  expect_equal(res_null[Group == "Isolated", N], 1)
  expect_equal(res_null[Group == "Isolated", N_Sire], 0)
  expect_equal(res_null[Group == "Isolated", N_Dam], 0)
  expect_equal(res_null[Group == "Isolated", N_Founder], 1)
})

test_that("pedfclass classifies inbreeding levels", {
  res <- pedfclass(test_ped)
  expect_true(is.data.table(res))
  expect_true("FClass" %in% names(res))
  expect_s3_class(res$FClass, "ordered")
  # A, B, C, D have f=0 (4 individuals); E, F have f=0.25 (2 individuals)
  expect_equal(sum(res$Count), nrow(test_ped))
  expect_equal(res[FClass == "F = 0", Count], 4)
  expect_equal(res[FClass == "0.125 < F <= 0.25", Count], 2)
})

test_that("pedfclass handles threshold boundaries correctly", {
  threshold_ped <- data.table(
    Ind = c("A", "B", "C", "D", "E"),
    Sire = NA_character_,
    Dam = NA_character_,
    Sex = c("male", "female", "male", "female", "male"),
    f = c(0, 0.0625, 0.125, 0.25, 0.250001)
  )
  threshold_ped <- new_tidyped(threshold_ped)

  res <- pedfclass(threshold_ped)

  expect_equal(res[FClass == "F = 0", Count], 1L)
  expect_equal(res[FClass == "0 < F <= 0.0625", Count], 1L)
  expect_equal(res[FClass == "0.0625 < F <= 0.125", Count], 1L)
  expect_equal(res[FClass == "0.125 < F <= 0.25", Count], 1L)
  expect_equal(res[FClass == "F > 0.25", Count], 1L)
})

test_that("pedfclass supports custom breaks", {
  threshold_ped <- data.table(
    Ind = c("A", "B", "C", "D", "E", "F", "G"),
    Sire = NA_character_,
    Dam = NA_character_,
    Sex = c("male", "female", "male", "female", "male", "female", "male"),
    f = c(0, 0.03125, 0.0625, 0.125, 0.25, 0.5, 0.6)
  )
  threshold_ped <- new_tidyped(threshold_ped)

  res <- pedfclass(threshold_ped, breaks = c(0.03125, 0.0625, 0.125, 0.25, 0.5))

  # labels auto-generated: 5 bounded + 1 tail = 6 positives, plus "F = 0" = 7 levels
  expect_equal(
    as.character(res$FClass),
    c("F = 0", "0 < F <= 0.03125", "0.03125 < F <= 0.0625",
      "0.0625 < F <= 0.125", "0.125 < F <= 0.25", "0.25 < F <= 0.5", "F > 0.5")
  )
  expect_equal(res[FClass == "0 < F <= 0.03125",      Count], 1L)
  expect_equal(res[FClass == "0.03125 < F <= 0.0625", Count], 1L)
  expect_equal(res[FClass == "0.0625 < F <= 0.125",   Count], 1L)
  expect_equal(res[FClass == "0.125 < F <= 0.25",     Count], 1L)
  expect_equal(res[FClass == "0.25 < F <= 0.5",       Count], 1L)
  expect_equal(res[FClass == "F > 0.5",               Count], 1L)
})

test_that("pedfclass respects custom labels aligned to breaks", {
  threshold_ped <- data.table(
    Ind = c("A", "B", "C", "D", "E"),
    Sire = NA_character_,
    Dam = NA_character_,
    Sex = c("male", "female", "male", "female", "male"),
    f = c(0, 0.0625, 0.125, 0.25, 0.4)
  )
  threshold_ped <- new_tidyped(threshold_ped)

  res <- pedfclass(
    threshold_ped,
    breaks = c(0.0625, 0.125, 0.25),
    labels = c("Low", "Moderate", "High")
  )
  # tail auto-generated as "F > 0.25"
  expect_equal(as.character(res$FClass), c("F = 0", "Low", "Moderate", "High", "F > 0.25"))
  expect_equal(res[FClass == "Low",       Count], 1L)
  expect_equal(res[FClass == "High",      Count], 1L)
  expect_equal(res[FClass == "F > 0.25",  Count], 1L)

  # Wrong length should error
  expect_error(
    pedfclass(threshold_ped, breaks = c(0.0625, 0.125, 0.25), labels = c("A", "B")),
    "length equal to length\\(breaks\\)"
  )
})

test_that("pedstats returns correct structure without timevar", {
  # Use a pedigree without any time column to test no-timevar path
  ped_no_time <- data.table::data.table(
    ID = c("A", "B", "C", "D"),
    Sire = c(NA, NA, "A", "A"),
    Dam = c(NA, NA, "B", "B"),
    Sex = c("male", "female", "male", "female")
  )
  tp <- suppressMessages(tidyped(ped_no_time))
  ps <- suppressMessages(pedstats(tp))

  # class
  expect_s3_class(ps, "pedstats")

  # $summary columns and types
  expect_true(is.data.table(ps$summary))
  expect_named(ps$summary, c("N", "NSire", "NDam", "NFounder", "MaxGen"))
  expect_equal(ps$summary$N, nrow(tp))
  expect_true(ps$summary$NFounder > 0)

  # $ecg columns
  expect_true(is.data.table(ps$ecg))
  expect_named(ps$ecg, c("Ind", "ECG", "FullGen", "MaxGen"))
  expect_equal(nrow(ps$ecg), nrow(tp))

  # no timevar -> gen_intervals is NULL
  expect_null(ps$gen_intervals)
})

test_that("pedstats returns gen_intervals with timevar", {
  tp2 <- suppressMessages(tidyped(big_family_size_ped))
  ps2 <- pedstats(tp2, timevar = "Year")

  # $gen_intervals columns
  expect_true(is.data.table(ps2$gen_intervals))
  expect_true(all(c("Pathway", "N", "Mean", "SD") %in% names(ps2$gen_intervals)))
  expect_true("Average" %in% ps2$gen_intervals$Pathway)

  # ecg = FALSE suppresses ecg
  ps_no_ecg <- pedstats(tp2, timevar = "Year", ecg = FALSE)
  expect_null(ps_no_ecg$ecg)

  # genint = FALSE suppresses gen_intervals
  ps_no_gi <- pedstats(tp2, timevar = "Year", genint = FALSE)
  expect_null(ps_no_gi$gen_intervals)
})

test_that("pedancestry proportions sum to 1 for multi-line admixture", {
  ped_df <- data.table::data.table(
    Ind = c("A","B","C","D", "E","F","G"),
    Sire = c(NA, NA, NA, NA, "A","C","E"),
    Dam =  c(NA, NA, NA, NA, "B","D","F"),
    Sex = c("male","female","male","female", "male","female","male")
  )
  ped_df$Line <- c("Line1", "Line1", "Line2", "Line2", NA, NA, NA)

  tp <- suppressMessages(tidyped(ped_df))
  res <- pedancestry(tp, foundervar = "Line")

  # Calculate sum of all label columns for each individual
  res[, Sum := rowSums(.SD, na.rm = TRUE), .SDcols = c("Line1", "Line2")]
  
  # Check specific admixture for the final hybrid (G)
  expect_equal(res[Ind == "G", Line1], 0.5)
  expect_equal(res[Ind == "G", Line2], 0.5)
  
  # All proportions must sum strictly to 1.0 (with floating point tolerance)
  expect_true(all(abs(res$Sum - 1.0) < 1e-12))
})

test_that("pedrel results match between compact and non-compact modes", {
  # Use a medium subset to explicitly test folding performance equivalence
  # 500 rows is large enough to contain multi-generational sib-groups for compression
  ped_sub <- head(visPedigree::big_family_size_ped, 500)
  tp <- suppressMessages(tidyped(ped_sub))

  rel_comp <- pedrel(tp, compact = TRUE)
  rel_full <- pedrel(tp, compact = FALSE)

  expect_equal(rel_comp, rel_full)

  coan_comp <- pedrel(tp, compact = TRUE, scale = "coancestry")
  coan_full <- pedrel(tp, compact = FALSE, scale = "coancestry")

  expect_equal(coan_comp, coan_full)
})

library(testthat)
library(visPedigree)
library(data.table)

# ===========================================================================
# Test Suite for pediv()
# ===========================================================================

# ---- Fixtures (reuse theory pedigree from test-pedcontrib.R) ----
make_theory_ped <- function() {
  ped_dt <- data.table(
    Ind  = c("F1", "F2", "F3", "F4", "S1", "D1", "R1", "R2", "R3", "R4"),
    Sire = c(NA,   NA,   NA,   NA,  "F1", "F3", "S1", "S1", "S1", "S1"),
    Dam  = c(NA,   NA,   NA,   NA,  "F2", "F4", "D1", "D1", "D1", "F4")
  )
  suppressMessages(tidyped(ped_dt))
}

make_small_ped <- function() {
  suppressMessages(tidyped(small_ped))
}


# ==========================================================================
# Group 1: Return structure
# ==========================================================================

test_that("pediv returns correct class and list structure", {
  tp <- make_theory_ped()
  div <- suppressMessages(pediv(tp, reference = c("R1", "R2", "R3", "R4")))

  expect_s3_class(div, "pediv")
  expect_true(is.list(div))
  expect_true(all(c("summary", "founders", "ancestors") %in% names(div)))
})

test_that("pediv summary is a single-row data.table with expected columns", {
  tp <- make_theory_ped()
  div <- suppressMessages(pediv(tp, reference = c("R1", "R2", "R3", "R4")))

  expected_cols <- c("NRef", "NFounder", "feH", "fe",
                     "NAncestor", "faH", "fa", "fafe",
                     "NeCoancestry", "NeInbreeding", "NeDemographic")
  expect_true(all(expected_cols %in% names(div$summary)))
  expect_equal(nrow(div$summary), 1L)
})


# ==========================================================================
# Group 2: f_e and f_a values match pedcontrib directly
# ==========================================================================

test_that("pediv f_e matches pedcontrib for theory pedigree", {
  tp <- make_theory_ped()
  ref <- c("R1", "R2", "R3", "R4")

  div  <- suppressMessages(pediv(tp, reference = ref))
  cont <- suppressMessages(pedcontrib(tp, reference = ref, mode = "both"))

  expect_equal(div$summary$fe, cont$summary$f_e, tolerance = 1e-10)
})

test_that("pediv f_a matches pedcontrib for theory pedigree", {
  tp <- make_theory_ped()
  ref <- c("R1", "R2", "R3", "R4")

  div  <- suppressMessages(pediv(tp, reference = ref))
  cont <- suppressMessages(pedcontrib(tp, reference = ref, mode = "both"))

  expect_equal(div$summary$fa, cont$summary$f_a, tolerance = 1e-10)
})

test_that("pediv f_e and f_a match known values for theory pedigree", {
  tp  <- make_theory_ped()
  div <- suppressMessages(pediv(tp, reference = c("R1", "R2", "R3", "R4")))

  expect_equal(div$summary$fe, 1 / (0.25^2 + 0.25^2 + 0.1875^2 + 0.3125^2),
               tolerance = 1e-6)
  expect_equal(div$summary$fa, 1 / (0.5^2 + 0.375^2 + 0.125^2),
               tolerance = 1e-6)
})

test_that("pediv small_ped: f_e and f_a match known values", {
  tp  <- make_small_ped()
  div <- suppressMessages(pediv(tp, reference = c("Z1", "Z2", "Y", "X")))

  expect_equal(div$summary$fe, 6.585209, tolerance = 1e-4)
  expect_equal(div$summary$fa, 2.666667, tolerance = 1e-4)
})


# ==========================================================================
# Group 3: f_a_fe_ratio
# ==========================================================================

test_that("fafe equals fa / fe", {
  tp  <- make_theory_ped()
  div <- suppressMessages(pediv(tp, reference = c("R1", "R2", "R3", "R4")))

  expect_equal(div$summary$fafe,
               div$summary$fa / div$summary$fe,
               tolerance = 1e-6)
})

test_that("fafe is between 0 and 1", {
  tp  <- make_small_ped()
  div <- suppressMessages(pediv(tp, reference = c("Z1", "Z2", "Y", "X")))

  expect_true(div$summary$fafe > 0)
  expect_true(div$summary$fafe <= 1)
})


# ==========================================================================
# Group 4: Ne values are numeric (or NA, never an error)
# ==========================================================================

test_that("Ne columns are numeric (not character or list)", {
  tp  <- make_theory_ped()
  div <- suppressMessages(pediv(tp, reference = c("R1", "R2", "R3", "R4")))

  expect_true(is.numeric(div$summary$NeCoancestry))
  expect_true(is.numeric(div$summary$NeInbreeding))
  expect_true(is.numeric(div$summary$NeDemographic))
})

test_that("Ne_coancestry is positive when computable", {
  tp  <- make_small_ped()
  div <- suppressMessages(pediv(tp, reference = c("Z1", "Z2", "Y", "X")))

  # coancestry method should produce a positive Ne for a real pedigree
  if (!is.na(div$summary$NeCoancestry)) {
    expect_true(div$summary$NeCoancestry > 0)
  }
})


# ==========================================================================
# Group 5: n_ref propagates correctly
# ==========================================================================

test_that("n_ref reflects length of reference argument", {
  tp  <- make_theory_ped()
  ref <- c("R1", "R2", "R3", "R4")
  div <- suppressMessages(pediv(tp, reference = ref))

  expect_equal(div$summary$NRef, length(ref))
})


# ==========================================================================
# Group 6: founders / ancestors tables are propagated from pedcontrib
# ==========================================================================

test_that("founders table matches pedcontrib", {
  tp  <- make_theory_ped()
  ref <- c("R1", "R2", "R3", "R4")

  div  <- suppressMessages(pediv(tp, reference = ref, top = 100))
  cont <- suppressMessages(pedcontrib(tp, reference = ref, mode = "both", top = 100))

  expect_equal(div$founders, cont$founders)
})

test_that("ancestors table matches pedcontrib", {
  tp  <- make_theory_ped()
  ref <- c("R1", "R2", "R3", "R4")

  div  <- suppressMessages(pediv(tp, reference = ref, top = 100))
  cont <- suppressMessages(pedcontrib(tp, reference = ref, mode = "both", top = 100))

  expect_equal(div$ancestors, cont$ancestors)
})


# ==========================================================================
# Group 7: Input validation
# ==========================================================================

test_that("pediv rejects non-tidyped input", {
  expect_error(pediv(data.frame(a = 1)), "tidyped")
})

test_that("pediv feH matches pedcontrib f_e_H", {
  tp  <- make_theory_ped()
  ref <- c("R1", "R2", "R3", "R4")
  div  <- suppressMessages(pediv(tp, reference = ref))
  cont <- suppressMessages(pedcontrib(tp, reference = ref, mode = "both"))

  expect_equal(div$summary$feH, cont$summary$f_e_H, tolerance = 1e-10)
})

test_that("pediv faH matches pedcontrib f_a_H", {
  tp  <- make_theory_ped()
  ref <- c("R1", "R2", "R3", "R4")
  div  <- suppressMessages(pediv(tp, reference = ref))
  cont <- suppressMessages(pedcontrib(tp, reference = ref, mode = "both"))

  expect_equal(div$summary$faH, cont$summary$f_a_H, tolerance = 1e-10)
})

test_that("print.pediv runs without error", {
  tp  <- make_theory_ped()
  div <- suppressMessages(pediv(tp, reference = c("R1", "R2", "R3", "R4")))
  expect_output(print(div), "Genetic Diversity Summary")
  expect_output(print(div), "fe\\(H\\)")
  expect_output(print(div), "fa\\(H\\)")
  expect_output(print(div), "Ne")
})


# ==========================================================================
# Group 8: GeneDiv
# ==========================================================================

test_that("GeneDiv column is present in summary", {
  tp  <- make_theory_ped()
  div <- suppressMessages(pediv(tp, reference = c("R1", "R2", "R3", "R4")))
  expect_true("GeneDiv" %in% names(div$summary))
})

test_that("GeneDiv equals 1 - MeanCoan", {
  tp  <- make_small_ped()
  div <- suppressMessages(pediv(tp, reference = c("Z1", "Z2", "Y", "X")))
  if (!is.na(div$summary$MeanCoan) && !is.na(div$summary$GeneDiv)) {
    expect_equal(div$summary$GeneDiv, 1 - div$summary$MeanCoan, tolerance = 1e-12)
  }
})

test_that("GeneDiv is in [0, 1]", {
  tp  <- make_small_ped()
  div <- suppressMessages(pediv(tp, reference = c("Z1", "Z2", "Y", "X")))
  if (!is.na(div$summary$GeneDiv)) {
    expect_true(div$summary$GeneDiv >= 0)
    expect_true(div$summary$GeneDiv <= 1)
  }
})

test_that("print.pediv displays GeneDiv", {
  tp  <- make_small_ped()
  div <- suppressMessages(pediv(tp, reference = c("Z1", "Z2", "Y", "X")))
  expect_output(print(div), "GeneDiv")
})

test_that("matrix-free sampled statistics match an explicit relationship matrix", {
  ped <- suppressMessages(tidyped(simple_ped))
  ecg <- pedecg(ped)
  ped <- merge(ped, ecg[, .(Ind, ECG)], by = "Ind", all.x = TRUE)
  data.table::setorder(ped, IndNum)

  target_idx <- which(ped$Gen == 2L)
  f_res <- cpp_calculate_inbreeding(ped$SireNum, ped$DamNum)
  actual <- calc_sampled_relationship_stats(
    sire = ped$SireNum,
    dam = ped$DamNum,
    dii = f_res$dii,
    f = f_res$f,
    target_idx = target_idx,
    ecg = ped$ECG,
    need_delta = TRUE,
    need_mean = TRUE,
    batch_size = 3L
  )

  A <- as.matrix(pedmat(ped, method = "A", sparse = FALSE))
  A_target <- A[target_idx, target_idx, drop = FALSE]
  pair_mask <- lower.tri(A_target)
  pair_ecg <- outer(ped$ECG[target_idx], ped$ECG[target_idx], "+") / 2
  valid <- pair_mask & pair_ecg > 0
  expected_delta <- mean(
    1 - (1 - A_target[valid] / 2)^(1 / pair_ecg[valid])
  )

  expect_equal(actual$DeltaC, expected_delta, tolerance = 1e-12)
  expect_equal(actual$MeanRelOff, mean(A_target[pair_mask]), tolerance = 1e-12)
  expect_equal(actual$MeanF, mean(diag(A_target) - 1), tolerance = 1e-12)
})

test_that("coancestry metric selection avoids unnecessary calculations", {
  ped <- suppressMessages(tidyped(simple_ped))
  ecg <- pedecg(ped)
  ped <- merge(ped, ecg[, .(Ind, ECG)], by = "Ind", all.x = TRUE)
  data.table::setorder(ped, IndNum)
  ped$Cohort <- ped$Gen

  ne_only <- suppressWarnings(calc_ne_coancestry(
    ped, ped, "Cohort", nsamples = 1000, seed = 42L, metrics = "ne"
  ))
  fg_only <- suppressWarnings(calc_ne_coancestry(
    ped, ped, "Cohort", nsamples = 1000, seed = 42L, metrics = "fg"
  ))
  both <- suppressWarnings(calc_ne_coancestry(
    ped, ped, "Cohort", nsamples = 1000, seed = 42L,
    metrics = c("ne", "fg")
  ))

  expect_equal(ne_only$DeltaC, both$DeltaC, tolerance = 1e-12)
  expect_equal(ne_only$Ne, both$Ne, tolerance = 1e-12)
  expect_true(all(is.na(ne_only$MeanCoan)))
  expect_true(all(is.na(ne_only$fg)))
  expect_true(all(is.na(ne_only$NSampledCoan)))

  expect_equal(fg_only$MeanCoan, both$MeanCoan, tolerance = 1e-12)
  expect_equal(fg_only$fg, both$fg, tolerance = 1e-12)
  expect_true(all(is.na(fg_only$DeltaC)))
  expect_true(all(is.na(fg_only$Ne)))
})

test_that("single sampled individual preserves missing pair statistics", {
  ped <- suppressMessages(tidyped(simple_ped))
  f_res <- cpp_calculate_inbreeding(ped$SireNum, ped$DamNum)
  target_idx <- which.max(ped$Gen)

  actual <- calc_sampled_relationship_stats(
    sire = ped$SireNum,
    dam = ped$DamNum,
    dii = f_res$dii,
    f = f_res$f,
    target_idx = target_idx,
    ecg = rep(1, nrow(ped))
  )

  expect_true(is.na(actual$DeltaC))
  expect_true(is.na(actual$MeanRelOff))
  expect_equal(actual$MeanF, f_res$f[target_idx])
})

#' Resolve generation-label text size
#' @keywords internal
resolve_visped_genlabcex <- function(node_size, genlabcex = NULL) {
  if (!is.null(genlabcex)) {
    return(genlabcex)
  }
  max(0.5, min(1.5, max(node_size, 1.5) * 0.3))
}

#' Render pedigree graph using Two-Pass strategy
#' @importFrom igraph V E plot.igraph vertex_attr edge_attr
#' @importFrom utils modifyList
#' @importFrom stats median
#' @keywords internal
plot_ped_igraph <- function(g, l, node_size, gen_info = NULL, genlab = FALSE,
                            genlabcex = NULL, ...) {
  genlab_margin <- 0.05
  genlab_color <- "#888888"

  # ============================================================================
  # SCALING LOGIC FOR RENDERING
  # ============================================================================
  scaling_ref <- max(node_size, 1.5)
  
  # Access attributes as lists/vectors ONCE to avoid repeated V(g)/E(g) lookups
  # Using vertex_attr/edge_attr is generally more stable for large graphs
  v_attrs <- igraph::vertex_attr(g)
  e_attrs <- igraph::edge_attr(g)
  
  # Scale frame width
  v_frame_width_base <- if (!is.null(v_attrs$frame.width)) v_attrs$frame.width else 0.2
  lwd_scale <- if (scaling_ref < 2.5) scaling_ref / 2.5 else 1
  v_frame_width <- v_frame_width_base * lwd_scale
  
  # Scale edge width
  e_width_base <- if (!is.null(e_attrs$width)) e_attrs$width else 1
  e_width <- pmax(e_width_base * (scaling_ref * 0.15), 0.001)
  
  # Scale arrows
  e_arrow_size_base <- if (!is.null(e_attrs$arrow.size)) e_attrs$arrow.size else 1
  calc_arrow_size <- e_arrow_size_base * (scaling_ref * 0.005)
  
  e_arrow_size <- calc_arrow_size
  e_arrow_width <- if (!is.null(e_attrs$arrow.width)) e_attrs$arrow.width else 1
  e_arrow_mode <- if (!is.null(e_attrs$arrow.mode)) e_attrs$arrow.mode else 2
  
  if (median(calc_arrow_size, na.rm = TRUE) < 0.005) {
    e_arrow_size <- 0
    e_arrow_width <- 0
    e_arrow_mode <- 0
  } else {
    e_arrow_width <- e_arrow_width * (scaling_ref * 0.015)
  }
  
  user_args <- list(...)
  margin <- max(node_size / 100, 0.02)
  genlab_space <- if (genlab && !is.null(gen_info) && nrow(gen_info) > 0) genlab_margin else 0
  
  # PASS 1: Draw EDGES ONLY
  # Instead of modifying g, we pass visual attributes directly to plot.igraph
  plot_args_edges <- list(
    x = g, rescale = FALSE,
    xlim = c(-margin - genlab_space, 1 + margin), ylim = c(1 + margin, -margin),
    layout = l, asp = 0, add = FALSE,
    # Edge styles
    edge.width = e_width,
    edge.color = if (!is.null(e_attrs$color)) e_attrs$color else "#333333",
    edge.curved = if (!is.null(e_attrs$curved)) e_attrs$curved else 0.1,
    edge.arrow.size = e_arrow_size,
    edge.arrow.width = e_arrow_width,
    edge.arrow.mode = e_arrow_mode,
    # Hide nodes
    vertex.size = 0,
    vertex.size2 = 0,
    vertex.label = NA,
    vertex.color = NA,
    vertex.frame.color = NA,
    vertex.frame.width = 0
  )
  
  if (length(user_args) > 0) {
    # Filter out user vertex args from edge pass
    safe_user_args <- user_args[!grepl("^(vertex|label)\\.", names(user_args))]
    plot_args_edges <- utils::modifyList(plot_args_edges, safe_user_args)
    # Re-enforce node hiding for pass 1
    plot_args_edges$vertex.size <- 0
    plot_args_edges$vertex.size2 <- 0
    plot_args_edges$vertex.label <- NA
  }
  
  suppressWarnings(do.call(igraph::plot.igraph, plot_args_edges))
  
  # PASS 2: Draw NODES AND LABELS ONLY
  # We reuse the same graph object but hide edges
  plot_args_nodes <- list(
    x = g, rescale = FALSE,
    xlim = c(-margin - genlab_space, 1 + margin), ylim = c(1 + margin, -margin),
    layout = l, asp = 0, add = TRUE,
    # Node styles
    vertex.size = if (!is.null(v_attrs$size)) v_attrs$size else 15,
    vertex.size2 = if (!is.null(v_attrs$size2)) v_attrs$size2 else 15,
    vertex.shape = if (!is.null(v_attrs$shape)) v_attrs$shape else "circle",
    vertex.color = if (!is.null(v_attrs$color)) v_attrs$color else VISPED_COLORS[["unknown"]],
    vertex.frame.color = if (!is.null(v_attrs$frame.color)) v_attrs$frame.color else VISPED_COLORS[["unknown_frame"]],
    vertex.frame.width = v_frame_width,
    vertex.label = if (!is.null(v_attrs$label)) v_attrs$label else v_attrs$name,
    vertex.label.cex = if (!is.null(v_attrs$label.cex)) v_attrs$label.cex else 1,
    vertex.label.color = if (!is.null(v_attrs$label.color)) v_attrs$label.color else "#0d0312",
    vertex.label.font = if (!is.null(v_attrs$label.font)) v_attrs$label.font else 1,
    # Hide edges
    edge.color = "#FFFFFF00",
    edge.width = 0,
    edge.arrow.size = 0,
    edge.arrow.mode = 0
  )
  
  if (length(user_args) > 0) {
    plot_args_nodes <- utils::modifyList(plot_args_nodes, user_args)
  }
  
  suppressWarnings(do.call(igraph::plot.igraph, plot_args_nodes))

  # PASS 3: Draw generation labels on the left margin
  if (genlab && !is.null(gen_info) && nrow(gen_info) > 0) {
    gen_label_cex <- resolve_visped_genlabcex(node_size, genlabcex)
    gen_label_x <- -margin - genlab_space * 0.5
    gen_labels <- if ("label" %in% names(gen_info)) {
      gen_info$label
    } else {
      paste0("G", gen_info$gen)
    }
    for (i in seq_len(nrow(gen_info))) {
      graphics::text(
        x = gen_label_x,
        y = gen_info$y[i],
        labels = gen_labels[i],
        cex = gen_label_cex,
        col = genlab_color,
        font = 2,
        adj = c(0.5, 0.5)
      )
    }
  }
}

#' Test if an object is a tidyped
#'
#' @param x An object to test.
#' @return Logical scalar.
#' @export
is_tidyped <- function(x) {
  inherits(x, "tidyped")
}

#' Build a unified metadata list for a tidyped object
#'
#' @param selfing Logical: whether selfing/monoecious mode was used.
#' @param bisexual_parents Character vector of IDs that appear as both sire and dam.
#' @param genmethod Character: generation assignment method ("top" or "bottom").
#' @return A named list of pedigree metadata.
#' @keywords internal
build_ped_meta <- function(selfing = FALSE,
                           bisexual_parents = character(0),
                           genmethod = "top") {
  list(
    selfing           = selfing,
    bisexual_parents  = bisexual_parents,
    genmethod         = genmethod
  )
}

#' Access pedigree metadata from a tidyped object
#'
#' @param x A tidyped object.
#' @return The \code{ped_meta} list, or \code{NULL} if not set.
#' @export
pedmeta <- function(x) {
  if (!is_tidyped(x)) stop("x is not a tidyped object.", call. = FALSE)
  attr(x, "ped_meta")
}

#' Check whether a tidyped object contains inbreeding coefficients
#'
#' @param x A tidyped object.
#' @return Logical scalar.
#' @export
has_inbreeding <- function(x) {
  is_tidyped(x) && "f" %in% names(x)
}

#' Check whether a tidyped object contains candidate flags
#'
#' @param x A tidyped object.
#' @return Logical scalar.
#' @export
has_candidates <- function(x) {
  is_tidyped(x) && "Cand" %in% names(x)
}

#' Internal constructor for tidyped class
#' @param x A data.table object
#' @return A tidyped object
#' @keywords internal
new_tidyped <- function(x) {
  stopifnot(inherits(x, "data.table"))
  if (!is_tidyped(x)) {
    data.table::setattr(x, "class", c("tidyped", class(x)))
  }
  # x[] clears data.table's invisible flag set by := and set*() operations.
  # [.tidyped uses setattr (in-place), so no copy is created.
  x[]
}

#' Check whether all referenced parents are present
#'
#' @param x A pedigree-like object.
#' @return Logical scalar.
#' @keywords internal
is_complete_pedigree <- function(x) {
  if (!is.data.frame(x)) return(FALSE)

  core_cols <- c("Ind", "Sire", "Dam")
  if (!all(core_cols %in% names(x))) return(FALSE)

  sires_ok <- is.na(x$Sire) | x$Sire %in% x$Ind
  dams_ok  <- is.na(x$Dam)  | x$Dam %in% x$Ind

  all(sires_ok) && all(dams_ok)
}

#' Error on row-truncated pedigree subsets
#'
#' @param ped A pedigree-like object.
#' @param fun Character scalar. Calling function name for the error message.
#' @return The input object, invisibly.
#' @keywords internal
require_complete_pedigree <- function(ped, fun) {
  if (is_complete_pedigree(ped)) return(invisible(ped))

  stop(
    fun, " requires a structurally complete pedigree. ",
    "This input appears to be a row-truncated subset with missing parent records.\n",
    "Compute on the full pedigree first, or extract a valid sub-pedigree with ",
    "`tidyped(tp, cand = ids, trace = \"up\")`.",
    call. = FALSE
  )
}

#' Internal helper to ensure ped is a tidyped object
#'
#' If the object has lost its \code{tidyped} class (e.g., after \code{merge()},
#' \code{rbind()}, or dplyr operations) but still contains the required columns,
#' the class is automatically restored with an informational message. Otherwise,
#' an error is raised guiding the user to call \code{tidyped()} or
#' \code{as_tidyped()}.
#'
#' @param ped An object expected to be a tidyped.
#' @return A valid tidyped object.
#' @keywords internal
ensure_tidyped <- function(ped) {
  if (is_tidyped(ped)) return(ped)

  if (!is.data.frame(ped)) {
    stop("'ped' must be a tidyped object. Run `tp <- tidyped(ped)` first.",
         call. = FALSE)
  }

  core <- c("Ind", "Sire", "Dam", "Sex", "Gen", "IndNum", "SireNum", "DamNum")
  missing <- setdiff(core, names(ped))

  if (length(missing) > 0) {
    stop("'ped' must be a tidyped object. Run `tp <- tidyped(ped)` first.",
         call. = FALSE)
  }

  if (!inherits(ped, "data.table")) {
    ped <- data.table::as.data.table(ped)
  }

  # Validate that IndNum indices are still consistent with Ind/Sire/Dam.
  # After rbind() or row subsetting, IndNum may be duplicated or no longer
  # match the row position, which would cause silent errors in Rcpp routines.
  if (anyDuplicated(ped$IndNum) > 0L ||
      !identical(ped$IndNum, seq_len(nrow(ped)))) {
    # Rebuild numeric indices from current row order
    ped[, IndNum := .I]
    ped[, SireNum := match(Sire, Ind, nomatch = 0L)]
    ped[, DamNum  := match(Dam, Ind, nomatch = 0L)]
  }

  # Structure is intact, only the class label was dropped
  message(
    "Note: 'ped' lost its tidyped class ",
    "(common after merge/rbind/dplyr). Restoring automatically."
  )
  new_tidyped(ped)
}

#' Internal helper to ensure ped is a complete tidyped object
#'
#' Like \code{ensure_tidyped()}, but also rejects row-truncated pedigree
#' subsets whose referenced parents are no longer present.
#'
#' @param ped An object expected to be a complete tidyped pedigree.
#' @param fun Character scalar. Calling function name for the error message.
#' @return A valid, structurally complete tidyped object.
#' @keywords internal
ensure_complete_tidyped <- function(ped, fun) {
  if (is_tidyped(ped)) {
    require_complete_pedigree(ped, fun)
    return(ped)
  }

  if (!is.data.frame(ped)) {
    stop("'ped' must be a tidyped object. Run `tp <- tidyped(ped)` first.",
         call. = FALSE)
  }

  core <- c("Ind", "Sire", "Dam", "Sex", "Gen", "IndNum", "SireNum", "DamNum")
  missing <- setdiff(core, names(ped))

  if (length(missing) > 0) {
    stop("'ped' must be a tidyped object. Run `tp <- tidyped(ped)` first.",
         call. = FALSE)
  }

  if (!inherits(ped, "data.table")) {
    ped <- data.table::as.data.table(ped)
  }

  require_complete_pedigree(ped, fun)

  if (anyDuplicated(ped$IndNum) > 0L ||
      !identical(ped$IndNum, seq_len(nrow(ped)))) {
    ped[, IndNum := .I]
    ped[, SireNum := match(Sire, Ind, nomatch = 0L)]
    ped[, DamNum  := match(Dam, Ind, nomatch = 0L)]
  }

  message(
    "Note: 'ped' lost its tidyped class ",
    "(common after merge/rbind/dplyr). Restoring automatically."
  )
  new_tidyped(ped)
}

#' Restore the tidyped class to a manipulated pedigree
#'
#' Rapidly restores the \code{tidyped} class to a \code{data.table} or
#' \code{data.frame} that was previously processed by \code{\link{tidyped}()}
#' but lost its class attributes due to data manipulation.
#'
#' This is a lightweight operation that only checks for the required columns
#' and re-attaches the class---it does \strong{not} re-run the full pedigree
#' sorting, generation inference, or loop detection.
#'
#' @param x A \code{data.table} or \code{data.frame} that was previously a
#'   tidyped object. It must still contain the core columns: \code{Ind},
#'   \code{Sire}, \code{Dam}, \code{Sex}, \code{Gen}, \code{IndNum},
#'   \code{SireNum}, \code{DamNum}.
#'
#' @return A \code{tidyped} object.
#'
#' @details
#' This helper is intended for objects that still contain the core pedigree
#' columns and numeric indices, but no longer inherit from \code{tidyped}.
#' A common reproducible case is \code{rbind()} on two \code{tidyped}
#' fragments, which typically returns a plain \code{data.table}. Converting a
#' \code{tidyped} object to a plain \code{data.frame} and then subsetting it
#' also drops the class.
#'
#' Some operations, such as \code{merge()} or certain dplyr workflows, may or
#' may not preserve the \code{tidyped} class depending on the versions of
#' \pkg{data.table}, \pkg{dplyr}, and the exact method dispatch path used in
#' the current R session. Therefore, \code{as_tidyped()} should be viewed as a
#' safe recovery helper rather than something only needed after one specific
#' verb.
#'
#' Typical class-loss scenarios include:
#' \itemize{
#'   \item \code{rbind(tped1, tped2)} --- often returns plain \code{data.table}
#'   \item \code{as.data.frame(tped)[rows, ]} --- returns plain \code{data.frame}
#'   \item manual class removal or serialization / import workflows
#' }
#' After such operations, downstream analysis functions (e.g.,
#' \code{\link{pedstats}}, \code{\link{pedne}}) will either error or
#' automatically restore the class. You can also call \code{as_tidyped()}
#' explicitly to restore the class yourself.
#'
#' @examples
#' library(visPedigree)
#' tp <- tidyped(simple_ped)
#' class(tp)
#' # [1] "tidyped"    "data.table" "data.frame"
#'
#' # Simulate class loss via rbind()
#' tp2 <- rbind(tp[1:5], tp[6:10])
#' class(tp2)
#' # [1] "data.table" "data.frame"
#'
#' # Restore the class
#' tp3 <- as_tidyped(tp2)
#' class(tp3)
#' # [1] "tidyped"    "data.table" "data.frame"
#'
#' # It can also restore from a plain data.frame if core columns are intact
#' tp_df <- as.data.frame(tp)
#' tp4 <- tp_df[tp_df$Gen > 1, ]
#' class(tp4)
#' # [1] "data.frame"
#'
#' tp5 <- as_tidyped(tp4)
#' class(tp5)
#' # [1] "tidyped"    "data.table" "data.frame"
#'
#' @seealso \code{\link{tidyped}}, \code{\link{new_tidyped}}
#' @export
as_tidyped <- function(x) {
  if (is_tidyped(x)) return(x)

  if (!is.data.frame(x)) {
    stop("Cannot coerce to tidyped: input is not a data.frame or data.table.",
         call. = FALSE)
  }

  if (!inherits(x, "data.table")) {
    x <- data.table::as.data.table(x)
  }

  core <- c("Ind", "Sire", "Dam", "Sex", "Gen", "IndNum", "SireNum", "DamNum")
  missing <- setdiff(core, names(x))

  if (length(missing) > 0) {
    stop(
      "Cannot restore tidyped class. Missing columns: ",
      paste(missing, collapse = ", "),
      ". Run `tidyped()` on raw data instead.",
      call. = FALSE
    )
  }

  # Rebuild numeric indices if they are inconsistent (e.g. after rbind/subset)
  if (anyDuplicated(x$IndNum) > 0L ||
      !identical(x$IndNum, seq_len(nrow(x)))) {
    x[, IndNum := .I]
    x[, SireNum := match(Sire, Ind, nomatch = 0L)]
    x[, DamNum  := match(Dam, Ind, nomatch = 0L)]
  }

  new_tidyped(x)
}

#' Internal validator for tidyped class
#'
#' Validates a tidyped object. If the object has lost its class but retains the
#' required columns, it is automatically restored via \code{ensure_tidyped()}.
#' Fatal structural problems (e.g., missing core columns) raise an error.
#'
#' @param x A tidyped object
#' @return The (possibly restored) object if valid, otherwise an error
#' @keywords internal
validate_tidyped <- function(x) {
  # Attempt auto-recovery if class was lost

  if (!is_tidyped(x)) {
    x <- tryCatch(
      suppressMessages(ensure_tidyped(x)),
      error = function(e) {
        stop("Object must be of class 'tidyped'. ", conditionMessage(e),
             call. = FALSE)
      }
    )
  }

  needed <- c("Ind", "Sire", "Dam", "Sex")
  missing_cols <- setdiff(needed, names(x))
  
  if (length(missing_cols) > 0) {
    stop(
      "The pedigree object is missing required columns: ",
      paste(missing_cols, collapse = ", "),
      ". Please ensure it has been processed by tidyped().",
      call. = FALSE
    )
  }
  
  x
}

#' Subset a tidyped object
#'
#' Intercepts \code{data.table}'s \code{[} method for \code{tidyped} objects.
#' After subsetting, the method checks whether the result is still a valid
#' pedigree (all referenced parents still present). If so, \code{IndNum},
#' \code{SireNum}, and \code{DamNum} are rebuilt and the \code{tidyped} class
#' is preserved. If the pedigree becomes structurally incomplete (missing parent
#' records), the result is degraded to a plain \code{data.table} with a warning.
#' Column-only selections (missing core columns) also return a plain
#' \code{data.table}.
#'
#' @param x A \code{tidyped} object.
#' @param ... Arguments passed to the \code{data.table} \code{[} method.
#' @return A \code{tidyped} object if the result is still a complete pedigree,
#'   otherwise a plain \code{data.table}.
#' @export
`[.tidyped` <- function(x, ...) {
  # Save metadata before dispatching to data.table
  meta <- attr(x, "ped_meta")
  cl <- class(x)

  sc <- sys.call()

  # Detect := (modify-in-place) operations.
  # := may appear in any argument position, e.g.:
  #   x[, col := val]              -> sc[[3]] is `:=`(col, val)
  #   x[cond, col := val]          -> sc[[4]] is `:=`(col, val)
  #   x[, `:=`(col1=v1, col2=v2)] -> sc[[3]] is `:=`(...)
  has_assign <- FALSE
  for (k in seq_along(sc)[-c(1L, 2L)]) {
    is_assign_k <- tryCatch({
      arg <- sc[[k]]
      is.call(arg) && identical(arg[[1L]], quote(`:=`))
    }, error = function(e) FALSE)
    if (isTRUE(is_assign_k)) {
      has_assign <- TRUE
      break
    }
  }

  # Use data.table::setattr() to strip the class IN PLACE (no copy!).
  # This is critical: class(x) <- ... triggers copy-on-modify, which breaks

  # := operations because the caller's variable still points to the original.
  data.table::setattr(x, "class", setdiff(cl, "tidyped"))
  # Ensure class is always restored, even if an error occurs
  on.exit(data.table::setattr(x, "class", cl), add = TRUE)

  # Re-dispatch the original call with proper NSE support.
  # NextMethod() breaks data.table's substitute() mechanism, so we
  # reconstruct the call and evaluate in a helper environment that
  # chains to the caller's frame (for user-defined variables in i/j).
  sc[[1L]] <- quote(`[`)
  sc[[2L]] <- quote(x)
  env <- list2env(list(x = x), parent = parent.frame())
  result <- eval(sc, envir = env)

  # Restore class on the original object (in-place)
  data.table::setattr(x, "class", cl)

  if (has_assign) {
    # := modifies x by reference. x IS the caller's object (no copy).
    # Class already restored above. Return invisibly.
    return(invisible(x))
  }

  # If result is not a data.table (e.g., single column vector), return as-is
  if (!is.data.table(result)) return(result)

  # Check if core columns are still present (column subset may have removed them)
  core_cols <- c("Ind", "Sire", "Dam")
  if (!all(core_cols %in% names(result))) return(result)

  # Check pedigree completeness: all non-NA parents must still be in Ind
  if (is_complete_pedigree(result)) {
    # Pedigree is still complete -> rebuild IndNum and restore class.
    # Use data.table::set() to avoid recursive [.tidyped dispatch.
    if ("IndNum" %in% names(result)) {
      data.table::set(result, j = "IndNum",  value = seq_len(nrow(result)))
      data.table::set(result, j = "SireNum",
                      value = match(result$Sire, result$Ind, nomatch = 0L))
      data.table::set(result, j = "DamNum",
                      value = match(result$Dam, result$Ind, nomatch = 0L))
    }
    data.table::setattr(result, "ped_meta", meta)
    data.table::setattr(result, "class", cl)
  } else {
    # Pedigree is incomplete -> degrade to plain data.table
    warning(
      "Subsetting removed parent records. ",
      "Result is a plain data.table, not a tidyped.\n",
      "Use tidyped(tp, cand = ids, trace = \"up\") ",
      "to extract a valid sub-pedigree.",
      call. = FALSE
    )
  }

  result
}

#' Print method for tidyped pedigree
#' @param x A tidyped object
#' @param ... Additional arguments passed to the data.table print method
#' @return The input object, invisibly.
#' @export
print.tidyped <- function(x, ...) {
  cat("Tidy Pedigree Object\n")
  # Temporarily remove tidyped class to use data.table's print method
  cl <- class(x)
  class(x) <- setdiff(cl, "tidyped")
  print(x, ...)
  # Restore the class
  class(x) <- cl
  invisible(x)
}

#' Plot a tidy pedigree
#'
#' @param x A \code{tidyped} object.
#' @param ... Additional arguments passed to \code{\link{visped}}.
#' @return Invisibly returns a list of graph data from \code{\link{visped}} (node/edge
#'   data and layout components) used to render the pedigree; the primary result is
#'   the plot drawn on the current device.
#' @export
plot.tidyped <- function(x, ...) {
  validate_tidyped(x)
  visped(x, ...)
}

#' Summary method for tidyped objects
#'
#' @param object A tidyped object.
#' @param ... Additional arguments (ignored).
#' @return A summary.tidyped object (list) containing:
#' \itemize{
#'   \item \code{n_ind}: Total number of individuals.
#'   \item \code{n_male}, \code{n_female}, \code{n_unknown_sex}: Sex composition counts.
#'   \item \code{n_founders}: Number of individuals with no known parents.
#'   \item \code{n_both_parents}: Count of individuals with complete parentage.
#'   \item \code{max_gen}, \code{gen_dist}: (Optional) Maximum generation and its distribution.
#'   \item \code{n_families}, \code{family_sizes}, \code{top_families}: (Optional) Family statistics.
#'   \item \code{f_stats}, \code{n_inbred}: (Optional) Inbreeding coefficient statistics.
#'   \item \code{n_cand}, \code{cand_f_stats}: (Optional) Candidate-specific statistics.
#' }
#' @export
summary.tidyped <- function(object, ...) {
  x <- validate_tidyped(object)
  res <- list()
  
  # Basic counts
  res$n_ind <- nrow(x)
  res$n_male <- sum(x$Sex == "male", na.rm = TRUE)
  res$n_female <- sum(x$Sex == "female", na.rm = TRUE)
  res$n_monoecious <- sum(x$Sex == "monoecious", na.rm = TRUE)
  res$n_unknown_sex <- sum(is.na(x$Sex))
  
  # Founders: Individuals with both parents unknown (NA)
  res$n_founders <- sum(is.na(x$Sire) & is.na(x$Dam))
  
  # Pedigree completeness
  res$n_both_parents <- sum(!is.na(x$Sire) & !is.na(x$Dam))
  res$n_sire_only <- sum(!is.na(x$Sire) & is.na(x$Dam))
  res$n_dam_only <- sum(is.na(x$Sire) & !is.na(x$Dam))
  
  # Isolated: Gen == 0
  if ("Gen" %in% names(x)) {
    res$n_isolated <- sum(x$Gen == 0)
    res$max_gen <- max(x$Gen, na.rm = TRUE)
    
    # Generation distribution
    gen_table <- table(x$Gen)
    res$gen_dist <- as.data.frame(gen_table)
    names(res$gen_dist) <- c("Generation", "Count")
  } else {
    res$n_isolated <- NA
    res$max_gen <- NA
    res$gen_dist <- NULL
  }
  
  # Candidates
  if ("Cand" %in% names(x)) {
    res$n_cand <- sum(x$Cand, na.rm = TRUE)
  } else {
    res$n_cand <- NA
  }

  # Offspring statistics
  # Count offspring for each individual
  if (!is.null(x$Sire) && !is.null(x$Dam)) {
    sire_counts <- table(x$Sire[!is.na(x$Sire)])
    dam_counts <- table(x$Dam[!is.na(x$Dam)])
    
    # Individuals with offspring (as sire)
    res$n_sires <- length(sire_counts)
    res$max_offspring_sire <- if(length(sire_counts) > 0) max(sire_counts) else 0
    res$mean_offspring_sire <- if(length(sire_counts) > 0) mean(sire_counts) else 0
    
    # Individuals with offspring (as dam)
    res$n_dams <- length(dam_counts)
    res$max_offspring_dam <- if(length(dam_counts) > 0) max(dam_counts) else 0
    res$mean_offspring_dam <- if(length(dam_counts) > 0) mean(dam_counts) else 0
    
    # Combined: any individual with offspring
    parents <- unique(c(names(sire_counts), names(dam_counts)))
    res$n_parents <- length(parents)
  }

  # Family information
  if ("Family" %in% names(x)) {
    # Count non-NA families
    families <- x$Family[!is.na(x$Family)]
    if (length(families) > 0) {
      family_table <- table(families)
      res$n_families <- length(family_table)
      res$family_sizes <- as.numeric(family_table)
      res$max_family_size <- max(res$family_sizes)
      res$mean_family_size <- mean(res$family_sizes)
      
      # Top 5 largest families
      top5 <- head(sort(family_table, decreasing = TRUE), 5)
      res$top_families <- data.frame(
        Family = names(top5),
        Size = as.numeric(top5),
        stringsAsFactors = FALSE
      )
    } else {
      res$n_families <- 0
      res$top_families <- NULL
    }
  }

  # Inbreeding coefficients statistics
  if ("f" %in% names(x)) {
    has_f <- any(!is.na(x$f))
    if (has_f) {
      res$f_stats <- list(
        min = min(x$f, na.rm = TRUE),
        max = max(x$f, na.rm = TRUE),
        mean = mean(x$f, na.rm = TRUE)
      )
    } else {
      res$f_stats <- list(
        min = NA_real_,
        max = NA_real_,
        mean = NA_real_
      )
    }
    
    # Count inbred individuals (f > 0)
    res$n_inbred <- sum(x$f > 0, na.rm = TRUE)

    # Stats for candidates if present
    if (!is.na(res$n_cand) && res$n_cand > 0) {
      cand_f <- x[x$Cand == TRUE, f]
      if (any(!is.na(cand_f))) {
        res$cand_f_stats <- list(
          min = min(cand_f, na.rm = TRUE),
          max = max(cand_f, na.rm = TRUE),
          mean = mean(cand_f, na.rm = TRUE)
        )
        res$n_cand_inbred <- sum(cand_f > 0, na.rm = TRUE)
      }
    }
  }

  class(res) <- "summary.tidyped"
  res
}

#' Print method for summary.tidyped
#'
#' @param x A summary.tidyped object.
#' @param ... Additional arguments (ignored).
#' @return The input object, invisibly.
#' @export
print.summary.tidyped <- function(x, ...) {
  cat("Pedigree Summary\n")
  cat("================\n\n")
  
  # Basic information
  cat("Total Individuals: ", x$n_ind, "\n")
  pct_male <- if(x$n_ind > 0) 100 * x$n_male / x$n_ind else 0
  pct_female <- if(x$n_ind > 0) 100 * x$n_female / x$n_ind else 0
  
  cat("  - Males:   ", sprintf("%d (%.1f%%)", x$n_male, pct_male), "\n")
  cat("  - Females: ", sprintf("%d (%.1f%%)", x$n_female, pct_female), "\n")
  
  if (!is.null(x$n_monoecious) && x$n_monoecious > 0) {
    pct_monoecious <- if(x$n_ind > 0) 100 * x$n_monoecious / x$n_ind else 0
    cat("  - Monoecious: ", sprintf("%d (%.1f%%)", x$n_monoecious, pct_monoecious), "\n")
  }
  
  if (x$n_unknown_sex > 0) {
    pct_unknown <- if(x$n_ind > 0) 100 * x$n_unknown_sex / x$n_ind else 0
    cat("  - Unknown: ", sprintf("%d (%.1f%%)", x$n_unknown_sex, pct_unknown), "\n")
  }
  cat("\n")
  
  # Pedigree structure
  cat("Pedigree Structure:\n")
  cat("  - Founders (no parents):  ", x$n_founders, "\n")
  cat("  - Both parents known:     ", x$n_both_parents, "\n")
  if (x$n_sire_only > 0) {
    cat("  - Sire only known:        ", x$n_sire_only, "\n")
  }
  if (x$n_dam_only > 0) {
    cat("  - Dam only known:         ", x$n_dam_only, "\n")
  }
  
  if (!is.na(x$n_isolated) && x$n_isolated > 0) {
    cat("  - Isolated (Gen 0):       ", x$n_isolated, "\n")
  }
  cat("\n")
  
  # Generation information
  if (!is.na(x$max_gen)) {
    cat("Generation:\n")
    cat("  - Maximum: ", x$max_gen, "\n")
    if (!is.null(x$gen_dist) && nrow(x$gen_dist) <= 10) {
      cat("  - Distribution:\n")
      for (i in seq_len(nrow(x$gen_dist))) {
        cat(sprintf("      Gen %s: %d individuals\n", 
                    x$gen_dist[i, 1], x$gen_dist[i, 2]))
      }
    } else if (!is.null(x$gen_dist)) {
      cat(sprintf("  - Distribution: %d generations (use str() for details)\n", 
                  nrow(x$gen_dist)))
    }
    cat("\n")
  }
  
  # Reproduction information
  if (!is.null(x$n_parents)) {
    cat("Reproduction:\n")
    cat("  - Individuals with offspring: ", x$n_parents, "\n")
    cat("  - Sires: ", x$n_sires, 
        sprintf(" (Mean=%.1f, Max=%d offspring)\n", 
                x$mean_offspring_sire, x$max_offspring_sire))
    cat("  - Dams:  ", x$n_dams, 
        sprintf(" (Mean=%.1f, Max=%d offspring)\n", 
                x$mean_offspring_dam, x$max_offspring_dam))
    cat("\n")
  }
  
  # Family information
  if (!is.null(x$n_families) && x$n_families > 0) {
    cat("Full-sibling Families:\n")
    cat("  - Number of families:     ", x$n_families, "\n")
    cat("  - Mean family size:       ", sprintf("%.2f\n", x$mean_family_size))
    cat("  - Maximum family size:    ", x$max_family_size, "\n")
    
    if (!is.null(x$top_families) && nrow(x$top_families) > 0) {
      cat("  - Top families by size:\n")
      for (i in seq_len(nrow(x$top_families))) {
        cat(sprintf("      %s: %d\n", 
                    x$top_families$Family[i], 
                    x$top_families$Size[i]))
      }
    }
    cat("\n")
  }
  
  # Candidates
  if (!is.na(x$n_cand)) {
    cat("Candidates Traced: ", x$n_cand, "\n\n")
  }

  # Inbreeding information
  if (!is.null(x$f_stats)) {
    cat("Inbreeding Coefficients:\n")
    if (all(is.na(unlist(x$f_stats)))) {
      cat("  - (Calculated column present, but all values are NA)\n")
    } else {
      cat("  - All individuals:\n")
      cat(sprintf("      Mean = %.4f, Min = %.4f, Max = %.4f\n",
                  x$f_stats$mean, x$f_stats$min, x$f_stats$max))
      if (!is.null(x$n_inbred)) {
        pct_inbred <- if(x$n_ind > 0) 100 * x$n_inbred / x$n_ind else 0
        cat(sprintf("      Inbred (f > 0): %d (%.1f%%)\n",
                    x$n_inbred, pct_inbred))
      }
      
      if (!is.null(x$cand_f_stats)) {
        cat("  - Candidates:\n")
        cat(sprintf("      Mean = %.4f, Min = %.4f, Max = %.4f\n",
                    x$cand_f_stats$mean, x$cand_f_stats$min, x$cand_f_stats$max))
        if (!is.null(x$n_cand_inbred)) {
          pct_cand_inbred <- if(x$n_cand > 0) 100 * x$n_cand_inbred / x$n_cand else 0
          cat(sprintf("      Inbred (f > 0): %d (%.1f%%)\n",
                      x$n_cand_inbred, pct_cand_inbred))
        }
      }
    }
    cat("\n")
  }
  
  cat("================\n")
  invisible(x)
}

# test-as-tidyped.R
# Tests for as_tidyped(), ensure_tidyped(), and S3 class recovery

test_that("as_tidyped restores class after rbind", {
  tp <- tidyped(simple_ped)
  tp2 <- rbind(tp[1:5], tp[6:10])

  expect_false(inherits(tp2, "tidyped"))
  expect_true(inherits(tp2, "data.table"))

  tp3 <- as_tidyped(tp2)
  expect_s3_class(tp3, "tidyped")
  expect_s3_class(tp3, "data.table")
})

test_that("as_tidyped restores class from plain data.frame", {
  tp <- tidyped(simple_ped)
  tp_df <- as.data.frame(tp)
  tp_sub <- tp_df[tp_df$Gen > 1, ]

  expect_false(inherits(tp_sub, "tidyped"))
  expect_false(inherits(tp_sub, "data.table"))

  tp_restored <- as_tidyped(tp_sub)
  expect_s3_class(tp_restored, "tidyped")
  expect_s3_class(tp_restored, "data.table")
})

test_that("as_tidyped rebuilds IndNum after rbind with overlapping indices", {

  tp <- tidyped(simple_ped)
  # rbind two slices that have overlapping IndNum values
  tp2 <- rbind(tp[1:5], tp[6:10])
  tp3 <- as_tidyped(tp2)

  # IndNum must be sequential 1:nrow

  expect_equal(tp3$IndNum, seq_len(nrow(tp3)))

  # SireNum/DamNum must reference valid IndNum or be 0
  valid_range <- c(0L, tp3$IndNum)
  expect_true(all(tp3$SireNum %in% valid_range))
  expect_true(all(tp3$DamNum %in% valid_range))
})

test_that("as_tidyped rebuilds IndNum from data.frame subset", {
  tp <- tidyped(simple_ped)
  tp_df <- as.data.frame(tp)
  tp_sub <- tp_df[tp_df$Gen > 1, ]
  tp_restored <- as_tidyped(tp_sub)

  expect_equal(tp_restored$IndNum, seq_len(nrow(tp_restored)))
  expect_true(all(tp_restored$SireNum >= 0L))
  expect_true(all(tp_restored$DamNum >= 0L))

  # SireNum should correctly map to Ind in restored object
  for (i in seq_len(nrow(tp_restored))) {
    if (tp_restored$SireNum[i] > 0L) {
      expect_equal(
        tp_restored$Ind[tp_restored$SireNum[i]],
        tp_restored$Sire[i]
      )
    }
    if (tp_restored$DamNum[i] > 0L) {
      expect_equal(
        tp_restored$Ind[tp_restored$DamNum[i]],
        tp_restored$Dam[i]
      )
    }
  }
})

test_that("as_tidyped is a no-op on valid tidyped", {
  tp <- tidyped(simple_ped)
  tp2 <- as_tidyped(tp)

  expect_identical(tp, tp2)
})

test_that("as_tidyped errors on missing core columns", {
  df <- data.frame(Ind = c("A", "B"), Sire = c(NA, "A"))

  expect_error(as_tidyped(df), "Missing columns")
})

test_that("as_tidyped errors on non-data.frame input", {
  expect_error(as_tidyped(1:10), "not a data.frame")
  expect_error(as_tidyped("abc"), "not a data.frame")
})

test_that("ensure_tidyped auto-recovers with message", {
  tp <- tidyped(simple_ped)
  tp2 <- rbind(tp[1:5], tp[6:10])

  expect_message(
    tp3 <- ensure_tidyped(tp2),
    "Restoring automatically"
  )
  expect_s3_class(tp3, "tidyped")
  expect_equal(tp3$IndNum, seq_len(nrow(tp3)))
})

test_that("ensure_tidyped is silent on valid tidyped", {
  tp <- tidyped(simple_ped)
  expect_silent(tp2 <- ensure_tidyped(tp))
  expect_identical(tp, tp2)
})

test_that("ensure_tidyped errors on non-data.frame", {
  expect_error(ensure_tidyped(1:10), "tidyped object")
})

test_that("ensure_tidyped errors on missing core columns", {
  df <- data.frame(Ind = c("A", "B"), Sire = c(NA, "A"))
  expect_error(ensure_tidyped(df), "tidyped object")
})

test_that("analysis functions work after class loss via rbind", {
  tp <- tidyped(simple_ped)
  tp2 <- rbind(tp[1:10], tp[11:20])

  # pedstats should auto-recover and work
  expect_message(
    stats <- pedstats(tp2),
    "Restoring automatically"
  )
  expect_s3_class(stats, "pedstats")
  expect_true(is_tidyped(tp2))

  # pedecg should then work on the already-restored object
  expect_silent(ecg <- pedecg(tp2))
  expect_true(is.data.table(ecg))
  expect_true("ECG" %in% names(ecg))
})

test_that("splitped works after class loss", {
  tp <- tidyped(small_ped)
  tp2 <- data.table::as.data.table(tp)
  class(tp2) <- c("data.table", "data.frame")  # strip tidyped class

  expect_message(
    result <- splitped(tp2),
    "Restoring automatically"
  )
  expect_s3_class(result, "splitped")
})

test_that("inbreed works after class loss", {
  tp <- tidyped(simple_ped)
  tp2 <- data.table::as.data.table(tp)
  class(tp2) <- c("data.table", "data.frame")  # strip tidyped class

  expect_message(
    result <- inbreed(tp2),
    "Restoring automatically"
  )
  expect_s3_class(result, "tidyped")
  expect_true("f" %in% names(result))
})

library(testthat)
library(data.table)

test_that("[.tidyped degrades incomplete row subsets with warning", {
  ped <- data.table(
    Ind = c("A", "B", "C", "D"),
    Sire = c(NA, NA, "A", "C"),
    Dam = c(NA, NA, "B", "B"),
    Year = c(2000, 2000, 2005, 2006)
  )

  tp <- tidyped(ped)

  expect_warning(
    sub_dt <- tp[Year > 2005],
    "Subsetting removed parent records"
  )

  expect_identical(class(sub_dt), c("data.table", "data.frame"))
  expect_false(is_tidyped(sub_dt))
})

test_that("[.tidyped preserves complete pedigree subsets and rebuilds indices", {
  ped <- data.table(
    Ind = c("A", "B", "C", "D", "E", "F"),
    Sire = c(NA, NA, "A", NA, NA, "D"),
    Dam = c(NA, NA, "B", NA, NA, "E")
  )

  tp <- tidyped(ped)
  sub_tp <- tp[Ind %in% c("A", "B", "C")]

  expect_true(is_tidyped(sub_tp))
  expect_identical(sub_tp$IndNum, 1:3)
  expect_equal(sub_tp[Ind == "C", SireNum], sub_tp[Ind == "A", IndNum])
  expect_equal(sub_tp[Ind == "C", DamNum], sub_tp[Ind == "B", IndNum])
})

test_that(":= keeps tidyped class and metadata", {
  tp <- tidyped(simple_ped)

  tp[, phenotype := seq_len(.N)]

  expect_true(is_tidyped(tp))
  expect_true("phenotype" %in% names(tp))
  expect_identical(pedmeta(tp)$genmethod, "top")
})

test_that("tidyped fast path matches full tracing result", {
  tp_master <- tidyped(simple_ped)

  res_raw <- tidyped(simple_ped, cand = "J5X804", trace = "up", tracegen = 2)
  res_fast <- tidyped(tp_master, cand = "J5X804", trace = "up", tracegen = 2)

  expect_true(is_tidyped(res_fast))
  expect_true(has_candidates(res_fast))
  expect_equal(as.list(res_fast), as.list(res_raw))
  expect_equal(pedmeta(res_fast), pedmeta(res_raw))
})

test_that("fast path: cand works on tidyped created with addnum=FALSE (regression)", {
  # Regression: fast-path previously used IndNum which is absent when addnum=FALSE,
  # causing "None of the specified candidates were found" even for valid IDs.
  tp_no <- tidyped(simple_ped, addnum = FALSE, addgen = FALSE)
  expect_false("IndNum" %in% names(tp_no))

  # Must not error and must find the candidate
  res <- tidyped(tp_no, cand = "J5X804", trace = "up")
  expect_true(is_tidyped(res))
  expect_true(has_candidates(res))
  expect_true("J5X804" %in% res$Ind)
  expect_true(any(res$Cand))

  # addnum=FALSE: output must not contain integer index columns
  res_no <- tidyped(tp_no, cand = "J5X804", addnum = FALSE, addgen = FALSE)
  expect_false("IndNum"  %in% names(res_no))
  expect_false("SireNum" %in% names(res_no))
  expect_false("DamNum"  %in% names(res_no))

  # addnum=TRUE: output must contain integer index columns
  res_yes <- tidyped(tp_no, cand = "J5X804", addnum = TRUE)
  expect_true("IndNum"  %in% names(res_yes))
  expect_true("SireNum" %in% names(res_yes))
  expect_true("DamNum"  %in% names(res_yes))

  # Results must match a full-path trace from the raw pedigree
  res_ref <- tidyped(simple_ped, cand = "J5X804", trace = "up")
  expect_equal(sort(res_yes$Ind), sort(res_ref$Ind))
})

test_that("state accessors report tidyped contents correctly", {
  tp <- tidyped(simple_ped)
  tp_f <- inbreed(tp)
  tp_c <- tidyped(tp, cand = "J5X804", trace = "up")

  expect_true(is_tidyped(tp))
  expect_false(has_inbreeding(tp))
  expect_true(has_inbreeding(tp_f))
  expect_false(has_candidates(tp))
  expect_true(has_candidates(tp_c))
})

test_that("splitped returns objects while pedsubpop returns summary table", {
  ped <- data.table(
    Ind = c("A", "B", "C", "D", "E", "F", "G"),
    Sire = c(NA, NA, "A", "A", NA, NA, "E"),
    Dam = c(NA, NA, "B", "B", NA, NA, "F")
  )

  tp <- tidyped(ped)
  splits <- splitped(tp)
  stats <- pedsubpop(tp)

  expect_s3_class(splits, "splitped")
  expect_true(all(vapply(splits, is_tidyped, logical(1))))
  expect_s3_class(stats, "data.table")
  expect_false("subpop" %in% names(stats))
  expect_true(all(c("Group", "N", "N_Sire", "N_Dam", "N_Founder") %in% names(stats)))
})

test_that("completeness-sensitive analyses error on truncated subsets", {
  tp <- tidyped(simple_ped)

  expect_warning(
    tp_sub <- tp[Gen > 2],
    "Subsetting removed parent records"
  )

  expect_false(is_tidyped(tp_sub))

  expect_error(
    inbreed(tp_sub),
    "structurally complete pedigree"
  )

  expect_error(
    pedecg(tp_sub),
    "structurally complete pedigree"
  )

  expect_error(
    pedmat(tp_sub, method = "f"),
    "structurally complete pedigree"
  )
})

library(testthat)
library(data.table)

test_that("summary.tidyped returns expected core fields", {
  data(simple_ped, package = "visPedigree")
  ped <- tidyped(simple_ped)
  s <- summary(ped)

  expect_s3_class(s, "summary.tidyped")
  expect_equal(s$n_ind, nrow(ped))
  expect_equal(s$n_male + s$n_female + s$n_unknown_sex, s$n_ind)
  expect_true(is.numeric(s$max_gen))
  expect_true(!is.null(s$gen_dist))
  expect_true(is.numeric(s$n_families))
  expect_true(is.numeric(s$n_parents))
})

test_that("summary.tidyped handles addgen=FALSE", {
  data(simple_ped, package = "visPedigree")
  ped <- tidyped(simple_ped, addgen = FALSE)
  s <- summary(ped)

  expect_true(is.na(s$max_gen))
  expect_true(is.na(s$n_isolated))
  expect_null(s$gen_dist)
})

test_that("summary.tidyped handles all-NA inbreeding column", {
  data(simple_ped, package = "visPedigree")
  ped <- tidyped(simple_ped)
  data.table::setalloccol(ped)
  data.table::set(ped, j = "f", value = rep(NA_real_, nrow(ped)))

  s <- summary(ped)
  expect_equal(s$n_inbred, 0)
  expect_true(all(is.na(unlist(s$f_stats))))
})

test_that("summary.tidyped validates input object", {
  bad_class <- data.table(Ind = "A", Sire = NA_character_, Dam = NA_character_, Sex = "male")
  expect_error(summary.tidyped(bad_class), "Object must be of class 'tidyped'")

  bad_cols <- data.table(Ind = "A", Sire = NA_character_, Dam = NA_character_)
  class(bad_cols) <- c("tidyped", class(bad_cols))
  expect_error(summary(bad_cols), "missing required columns: Sex")
})

#' Visualize Pedigree Statistics (internal)
#'
#' Internal plotting backend for \code{\link{plot.pedstats}}.
#' Users should call \code{plot(stats, ...)} instead of this function directly.
#'
#' @param x A \code{pedstats} object returned by \code{\link{pedstats}}.
#' @param type Character. The type of plot to generate:
#' \itemize{
#'   \item \code{"genint"}: Bar chart of mean generation intervals.
#'   \item \code{"ecg"}: Histogram of ancestral depth metrics (ECG, FullGen, or MaxGen).
#' }
#' @param metric Character. Specific metric to plot when \code{type = "ecg"}.
#' Options: \code{"ECG"} (default), \code{"FullGen"}, \code{"MaxGen"}.
#' @param ... Additional arguments passed to \code{\link[lattice]{barchart}} or 
#' \code{\link[lattice]{histogram}}.
#'
#' @return A lattice plot object.
#'
#' @seealso \code{\link{pedstats}}, \code{\link{plot.pedstats}}
#' 
#' @keywords internal
vispstat <- function(x, type = c("genint", "ecg"), metric = "ECG", ...) {
  if (!inherits(x, "pedstats")) {
    stop("x must be a pedstats object")
  }
  
  type <- match.arg(type)
  
  if (type == "genint") {
    if (is.null(x$gen_intervals)) {
      stop("No generation interval data found in pedstats object. ",
           "Ensure the pedigree has a time variable (e.g., 'Year' or 'BirthYear').")
    }
    
    dt <- data.table::copy(x$gen_intervals)
    # Keep only the four classic gametic pathways
    dt <- dt[Pathway %in% c("SS", "SD", "DS", "DD")]
    
    if (nrow(dt) == 0) {
      stop("No pathway-specific generation intervals found.")
    }
    
    # Order pathways consistently
    dt[, Pathway := factor(Pathway, levels = c("SS", "SD", "DS", "DD"))]
    
    # Get Unit for Label
    unit <- attr(x$gen_intervals, "unit")
    unit_label <- switch(unit, 
                        "year" = "Years", "month" = "Months", "day" = "Days", 
                        "hour" = "Hours", "Years")
    
    # Basic formula
    f <- Mean ~ Pathway
    if ("Group" %in% names(dt)) {
      f <- Mean ~ Pathway | Group
    }
    
    p <- lattice::barchart(
      f, 
      data = dt,
      ylab = sprintf("Generation Interval (%s)", unit_label),
      xlab = "Pathway",
      main = "Generation Intervals by Pathway",
      col = "steelblue",
      border = "transparent",
      scales = list(x = list(rot = 0)),
      panel = function(x, y, ...) {
        lattice::panel.barchart(x, y, ...)
        lattice::panel.grid(h = -1, v = 0, col = "gray90", lty = "dotted")
      },
      ...
    )
    
    return(p)
    
  } else if (type == "ecg") {
    if (is.null(x$ecg)) {
      stop("No ECG data found in pedstats object.")
    }
    
    valid_metrics <- c("ECG", "FullGen", "MaxGen")
    if (!metric %in% valid_metrics) {
      stop("Invalid metric. Choose from: ", paste(valid_metrics, collapse = ", "))
    }
    
    # Check if metric column exists
    if (!metric %in% names(x$ecg)) {
      stop(sprintf("Metric '%s' not found in ECG data.", metric))
    }
    
    # Dynamic formula
    f <- as.formula(paste("~", metric))
    
    # Determine appropriate number of bins based on data range
    metric_data <- x$ecg[[metric]]
    n_bins <- min(30, max(10, ceiling(sqrt(length(metric_data)))))
    
    # For integer metrics (FullGen, MaxGen), use integer breaks
    if (metric %in% c("FullGen", "MaxGen")) {
      max_val <- max(metric_data, na.rm = TRUE)
      breaks <- seq(0, max_val + 1, by = 1)
      n_bins <- NULL
    } else {
      breaks <- NULL
    }
    
    p <- lattice::histogram(
      f, 
      data = x$ecg,
      xlab = metric,
      main = paste("Distribution of", metric),
      col = "lightgray",
      border = "white",
      breaks = breaks,
      nint = n_bins,
      type = "count",
      ...
    )
    
    return(p)
  }
}

#' @rdname vispstat
#' @export
plot.pedstats <- function(x, ...) {
  vispstat(x, ...)
}

#' Convert pedigree to igraph structure
#' @import data.table
#' @keywords internal
ped2igraph <- function(ped, compact = FALSE, highlight = NULL, trace = FALSE,
                       showf = FALSE, labelvar = NULL, shapeby = "sex") {
  if (nrow(ped) == 0) {
    return(list(
      node = data.table(id = integer(), nodetype = character(), gen = integer(), layer = numeric(), label = character()),
      edge = data.table(from = integer(), to = integer())
    ))
  }
  
  # 1. Inject missing parents (for subsetted pedigrees)
  ped_new <- inject_missing_parents(ped)
  
  # 2. Resolve highlight IDs and relatives
  highlight_info <- get_highlight_ids(ped_new, highlight, trace)
  h_ids <- highlight_info$all_ids
  
  # 3. Prepare initial node table
  ped_node <- prepare_initial_nodes(ped_new, labelvar = labelvar)
  
  # 4. Compact pedigree (if requested)
  ped_node <- compact_pedigree(ped_node, compact, h_ids)
  
  # 5. Generate edges and virtual nodes
  graph_struct <- generate_graph_structure(ped_node, h_ids)
  ped_node <- graph_struct$node
  ped_edge <- graph_struct$edge
  
  # 6. Apply styles (colors, shapes, highlighting)
  ped_node <- apply_node_styles(ped_node, highlight_info, shapeby)
  
  # 7. Finalize graph (reindex IDs, set edge colors)
  result <- finalize_graph(ped_node, ped_edge, highlight_info, trace, showf)
  
  return(result)
}

#' Inject missing parents for subsetted pedigrees
#' @param ped A data.table containing pedigree info.
#' @keywords internal
inject_missing_parents <- function(ped) {
  ped_new <- copy(ped)
  # Ensure keys for fast matching
  setkey(ped_new, IndNum)
  all_ind_nums <- ped_new$IndNum
  
  # Find missing Sires
  s_nums <- ped_new$SireNum
  missing_sire_nums <- unique(s_nums[s_nums != 0 & !is.na(s_nums) & !(s_nums %in% all_ind_nums)])
  
  # Find missing Dams (compute early to detect monoecious)
  d_nums <- ped_new$DamNum
  missing_dam_nums <- unique(d_nums[d_nums != 0 & !is.na(d_nums) & !(d_nums %in% all_ind_nums)])
  
  # Detect monoecious: individuals that are both missing sires and missing dams
  monoecious_nums <- intersect(missing_sire_nums, missing_dam_nums)
  
  if (length(missing_sire_nums) > 0) {
    sire_info <- unique(ped_new[SireNum %in% missing_sire_nums, .(SireNum, Sire)])
    # Determine sex: monoecious if also a dam, otherwise male
    sire_sex <- ifelse(sire_info$SireNum %in% monoecious_nums, "monoecious", "male")
    new_sires <- data.table(
      Ind = sire_info$Sire,
      Sire = NA_character_,
      Dam = NA_character_,
      Sex = sire_sex,
      Gen = min(ped_new$Gen, na.rm=TRUE) - 1,
      IndNum = sire_info$SireNum,
      SireNum = 0L,
      DamNum = 0L
    )
    extra_cols <- setdiff(names(ped_new), names(new_sires))
    if (length(extra_cols) > 0) new_sires[, (extra_cols) := NA]
    ped_new <- rbind(ped_new, new_sires, fill = TRUE)
    setkey(ped_new, IndNum)
    all_ind_nums <- ped_new$IndNum
  }

  # Find missing Dams (exclude monoecious already added as sires)
  missing_dam_nums_only <- setdiff(missing_dam_nums, monoecious_nums)
  
  if (length(missing_dam_nums_only) > 0) {
    dam_info <- unique(ped_new[DamNum %in% missing_dam_nums_only, .(DamNum, Dam)])
    new_dams <- data.table(
      Ind = dam_info$Dam,
      Sire = NA_character_,
      Dam = NA_character_,
      Sex = "female",
      Gen = min(ped_new$Gen, na.rm=TRUE) - 1,
      IndNum = dam_info$DamNum,
      SireNum = 0L,
      DamNum = 0L
    )
    extra_cols <- setdiff(names(ped_new), names(new_dams))
    if (length(extra_cols) > 0) new_dams[, (extra_cols) := NA]
    ped_new <- rbind(ped_new, new_dams, fill = TRUE)
  }
  
  return(ped_new[])
}

#' Prepare initial node table for igraph conversion
#' @param ped A data.table containing pedigree info.
#' @keywords internal
prepare_initial_nodes <- function(ped, labelvar = NULL) {
  cols <- c("IndNum", "Ind", "SireNum", "DamNum", "Sire", "Dam", "Sex", "Gen")
  if ("Cand" %in% colnames(ped)) cols <- c(cols, "Cand")
  if ("f" %in% colnames(ped)) cols <- c(cols, "f")
  
  ped_node <- ped[, ..cols]
  if (!is.null(labelvar)) {
    set(ped_node, j = "DisplayLabel", value = as.character(ped[[labelvar]]))
  }
  setnames(ped_node, 
           old = c("IndNum", "Ind", "SireNum", "DamNum", "Sire", "Dam", "Sex", "Gen"),
           new = c("id", "label", "sirenum", "damnum", "sirelabel", "damlabel", "sex", "gen"))
  
  if ("Cand" %in% colnames(ped)) setnames(ped_node, "Cand", "cand")
  
  max_id <- max(ped_node$id, na.rm = TRUE)
  familylabel = NULL
  ped_node[!(is.na(sirelabel) & is.na(damlabel)),
           familylabel := paste(sirelabel, damlabel, sep = "x")]
  
  ped_node[!is.na(familylabel), 
           familynum := .GRP + max_id, 
           by = familylabel]
  ped_node[is.na(familynum), familynum := 0]
  
  nodetype = NULL
  ped_node[, nodetype := "real"]
  
  return(ped_node[])
}

#' Compact pedigree by merging full siblings
#' @param ped_node A data.table of nodes.
#' @param compact Logical, whether to compact.
#' @param h_ids Highlighted IDs to exempt from compaction.
#' @keywords internal
compact_pedigree <- function(ped_node, compact, h_ids) {
  if (!compact) return(ped_node)
  
  # Identify parents (individuals who appear in sirelabel or damlabel)
  # Efficiently find unique parents
  parents <- unique(c(ped_node$sirelabel, ped_node$damlabel))
  parents <- parents[!is.na(parents)]
  
  # Candidates for compaction: NOT parents, HAS a family, and NOT highlighted
  # Using %in% on unique labels is usually fast in data.table
  ped_node_1 <- ped_node[!(label %in% parents) & !is.na(familylabel)]
  
  if (!is.null(h_ids) && length(h_ids) > 0) {
    ped_node_1 <- ped_node_1[!(label %in% h_ids)]
  }
  
  if (nrow(ped_node_1) > 0) {
    familysize <- NULL
    # Efficiently count family sizes for eligible nodes
    ped_node_1[, familysize := .N, by = .(familylabel)]
    fullsib_id_DT <- ped_node_1[familysize >= 2]
    
    if (nrow(fullsib_id_DT) > 0) {
      # Take one representative from each family
      compact_family <- unique(fullsib_id_DT, by = c("familylabel"))
      compact_family[, `:=`(
        label = paste0("FS\u00d7", familysize),
        nodetype = "compact",
        sex = NA_character_
      )]
      
      next_id <- max(ped_node$id, na.rm = TRUE)
      compact_family[, id := seq.int(next_id + 1, length.out = .N)]
      
      # Mapping from old individual IDs to new compact family node IDs
      map_dt <- fullsib_id_DT[, .(from_id = id, familylabel)]
      map_dt <- compact_family[map_dt, on = .(familylabel), .(from_id, to_id = id, to_label = label)]
      
      # Remove individuals that were compacted
      ped_node <- ped_node[!(id %in% fullsib_id_DT$id)]
      
      # Update children of these individuals to point to the compact node instead
      if (nrow(map_dt) > 0) {
        # Note: In practice, these individuals are NOT parents (vetted above),
        # but we keep this for robustness if the logic ever changes.
        ped_node[map_dt, on = .(sirenum = from_id), `:=`(sirenum = to_id, sirelabel = to_label)]
        ped_node[map_dt, on = .(damnum = from_id), `:=`(damnum = to_id, damlabel = to_label)]
      }
      
      ped_node <- rbind(ped_node, compact_family, fill = TRUE)
      
      # Re-calculate family numbers to include new compact nodes
      max_id <- max(ped_node$id, na.rm = TRUE)
      ped_node[!is.na(familylabel), 
               familynum := .GRP + max_id, 
               by = familylabel]
      ped_node[is.na(familynum), familynum := 0]
    }
  }
  return(ped_node[])
}

#' Generate edges and virtual family nodes
#' @param ped_node A data.table of nodes.
#' @param h_ids Highlighted IDs.
#' @keywords internal
generate_graph_structure <- function(ped_node, h_ids) {
  # Individual → family edges
  edge_ind <- ped_node[familynum > 0, .(from = id, to = familynum, role = "ind")]

  # Family → sire edges (dedup siblings in same family)
  edge_sire <- unique(
    ped_node[sirenum > 0 & familynum > 0, .(from = familynum, to = sirenum, role = "sire")]
  )

  # Family → dam edges (dedup siblings in same family)
  edge_dam <- unique(
    ped_node[damnum > 0 & familynum > 0, .(from = familynum, to = damnum, role = "dam")]
  )

  # Detect selfing: same (from, to) appearing in both sire and dam roles
  edge_parents <- rbind(edge_sire, edge_dam)
  dup_ft <- edge_parents[, .N, by = .(from, to)]
  selfing_ft <- dup_ft[N > 1]
  if (nrow(selfing_ft) > 0) {
    edge_parents[selfing_ft, on = .(from, to), role := "selfing"]
    edge_parents <- unique(edge_parents, by = c("from", "to"))
  }

  ped_edge <- rbind(edge_ind, edge_parents)
  ped_edge <- ped_edge[order(from, to)]

  width = arrow.size = arrow.width = color = curved = NULL
  edge_default_color <- if (length(h_ids) > 0) "#3333334D" else "#333333"
  ped_edge[, ":="(width = 1, arrow.size = 1, arrow.width = 1, arrow.mode = 2, 
                  color = edge_default_color, curved = 0.10)]
  
  virtual_nodes <- unique(ped_node[familynum > 0, .(
    id = familynum,
    familylabel,
    label = familylabel,
    sirenum,
    damnum,
    sirelabel,
    damlabel,
    gen,
    familynum
  )])
  
  ped_node <- rbind(ped_node, virtual_nodes, fill = TRUE)
  ped_node[is.na(nodetype), nodetype := "virtual"]
  
  layer = NULL
  ped_node[nodetype %in% c("real", "compact"), layer := 2 * gen - 1]
  ped_node[nodetype %in% c("virtual"), layer := 2 * (gen - 1)]
  
  list(node = ped_node, edge = ped_edge)
}

test_that("highlight edges are correct for shared parents", {
  # Test case: J5X804's ancestor J0Z938 also has another child J1H419
  # The edge to J1H419's family should NOT be highlighted
  
  tidy_ped <- tidyped(simple_ped, genmethod = "bottom")
  graph_data <- visPedigree:::ped2igraph(
    tidy_ped, 
    compact = FALSE, 
    highlight = "J5X804", 
    trace = "up", 
    showf = FALSE
  )
  
  # Find J0Z938 node
  j0z938_node <- graph_data$node[label == "J0Z938"]
  expect_equal(nrow(j0z938_node), 1)
  expect_true(j0z938_node$highlighted)
  
  # Find edges from virtual family nodes to J0Z938
  edges_to_j0z938 <- graph_data$edge[to == j0z938_node$id & from > max(graph_data$node[nodetype %in% c("real", "compact")]$id)]
  
  # Should have 2 edges: one from highlighted family (J0Z938xJ0Z843), one from non-highlighted family (J0Z938xJ0Z167)
  expect_equal(nrow(edges_to_j0z938), 2)
  
  # Edge from highlighted family should still match parent node color
  highlighted_edge <- edges_to_j0z938[from_highlighted == TRUE]
  expect_equal(nrow(highlighted_edge), 1)
  expect_equal(highlighted_edge$color, j0z938_node$color)
  
  # Edge from non-highlighted family should match parent node color with fading
  non_highlighted_edge <- edges_to_j0z938[from_highlighted == FALSE]
  expect_equal(nrow(non_highlighted_edge), 1)
  expect_true(grepl(paste0("^", j0z938_node$color, "4D$"), non_highlighted_edge$color))
})

test_that("highlight edges are correct for shared children", {
  # Test case: J0Z475 and J0C612 have two children: J1Y339 (in path) and J1F266 (not in path)
  # Only the edge from J1Y339 to the family virtual node should be highlighted
  
  tidy_ped <- tidyped(simple_ped, genmethod = "bottom")
  graph_data <- visPedigree:::ped2igraph(
    tidy_ped, 
    compact = FALSE, 
    highlight = "J4Y326", 
    trace = "up", 
    showf = FALSE
  )
  
  # Find the family virtual node
  family_node <- graph_data$node[familylabel == "J0Z475xJ0C612" & nodetype == "virtual"]
  expect_equal(nrow(family_node), 1)
  expect_true(family_node$highlighted)
  
  # Find children
  j1y339_node <- graph_data$node[label == "J1Y339"]
  j1f266_node <- graph_data$node[label == "J1F266"]
  expect_true(j1y339_node$highlighted)
  expect_false(j1f266_node$highlighted)
  
  # Find edges from children to family node
  edges_to_family <- graph_data$edge[to == family_node$id]
  expect_equal(nrow(edges_to_family), 2)
  
  # Edge from highlighted child (J1Y339) should be solid
  edge_from_j1y339 <- edges_to_family[from == j1y339_node$id]
  expect_equal(edge_from_j1y339$color, "#333333")
  
  # Edge from non-highlighted child (J1F266) should be faded
  edge_from_j1f266 <- edges_to_family[from == j1f266_node$id]
  expect_equal(edge_from_j1f266$color, "#3333334D")
})

test_that("edges work correctly without highlighting", {
  tidy_ped <- tidyped(simple_ped, genmethod = "bottom")
  graph_data <- visPedigree:::ped2igraph(
    tidy_ped, 
    compact = FALSE, 
    highlight = NULL, 
    trace = FALSE, 
    showf = FALSE
  )
  
  # Without highlighting, edges from virtual nodes should use parent colors
  real_max <- max(graph_data$node[nodetype %in% c("real", "compact")]$id)
  virtual_edges <- graph_data$edge[from > real_max]
  
  # All virtual edges should have color matching their target node's color (fill color)
  expect_true(all(!is.na(virtual_edges$color)))
  expect_true(all(virtual_edges$color %in% graph_data$node$color))
})

test_that("highlight edges work correctly with trace down", {
  # Test descendant tracing: edges should follow the same rules
  tidy_ped <- tidyped(simple_ped, genmethod = "bottom")
  graph_data <- visPedigree:::ped2igraph(
    tidy_ped, 
    compact = FALSE, 
    highlight = "J1J576", 
    trace = "down", 
    showf = FALSE
  )
  
  # J1J576 should be highlighted
  j1j576_node <- graph_data$node[label == "J1J576"]
  expect_true(j1j576_node$highlighted)
  
  # Edge from J1J576 to its family should be highlighted
  j1j576_family_id <- j1j576_node$familynum
  edge_to_family <- graph_data$edge[from == j1j576_node$id & to == j1j576_family_id]
  expect_equal(nrow(edge_to_family), 1)
  expect_equal(edge_to_family$color, "#3333334D")
  
  # J1J576's parents should NOT be highlighted (trace = "down" only traces descendants)
  j0z938_node <- graph_data$node[label == "J0Z938"]
  j0z843_node <- graph_data$node[label == "J0Z843"]
  expect_false(j0z938_node$highlighted)
  expect_false(j0z843_node$highlighted)
  
  # Edges from family to parents should match parent node colors
  family_node <- graph_data$node[id == j1j576_family_id]
  edges_from_family <- graph_data$edge[from == j1j576_family_id]
  parent_colors <- graph_data$node[id %in% edges_from_family$to, color]
  expect_true(all(edges_from_family$color %in% parent_colors))
})

test_that("highlight edges work correctly with trace all", {
  # Test both ancestor and descendant tracing
  tidy_ped <- tidyped(simple_ped, genmethod = "bottom")
  graph_data <- visPedigree:::ped2igraph(
    tidy_ped, 
    compact = FALSE, 
    highlight = "J3Y620", 
    trace = "all", 
    showf = FALSE
  )
  
  # Focal individual should be highlighted
  focal_node <- graph_data$node[label == "J3Y620"]
  expect_true(focal_node$highlighted)
  
  # Ancestor (J2C161) should be highlighted
  ancestor_node <- graph_data$node[label == "J2C161"]
  expect_true(ancestor_node$highlighted)
  
  # Descendant (J4Y326) should be highlighted
  descendant_node <- graph_data$node[label == "J4Y326"]
  expect_true(descendant_node$highlighted)
  
  # Edges involving highlighted nodes should follow the rules
  # Individual to family: only if individual is highlighted
  edge_focal_to_family <- graph_data$edge[from == focal_node$id & to == focal_node$familynum]
  if (nrow(edge_focal_to_family) > 0) {
    expect_equal(edge_focal_to_family$color, "#333333")
  }
})

test_that("highlight edges work correctly with multiple individuals without trace", {
  # Test multiple highlighted individuals sharing a parent
  # When trace = FALSE, NO edges should be highlighted (only individuals are marked)
  tidy_ped <- tidyped(simple_ped, genmethod = "bottom")
  graph_data <- visPedigree:::ped2igraph(
    tidy_ped, 
    compact = FALSE, 
    highlight = c("J1J576", "J1H419"), 
    trace = FALSE, 
    showf = FALSE
  )
  
  # Both individuals should be highlighted
  j1j576_node <- graph_data$node[label == "J1J576"]
  j1h419_node <- graph_data$node[label == "J1H419"]
  expect_true(j1j576_node$highlighted)
  expect_true(j1h419_node$highlighted)
  
  # Their shared parent J0Z938 should NOT be highlighted (no trace)
  j0z938_node <- graph_data$node[label == "J0Z938"]
  expect_false(j0z938_node$highlighted)
  
  # Both family virtual nodes should be highlighted
  family1 <- graph_data$node[id == j1j576_node$familynum]
  family2 <- graph_data$node[id == j1h419_node$familynum]
  expect_true(family1$highlighted)
  expect_true(family2$highlighted)
  
  # WITHOUT trace, edges from individuals to their families should be FADED
  # (No relationship path is being traced, so edges remain faded)
  edge1 <- graph_data$edge[from == j1j576_node$id & to == j1j576_node$familynum]
  edge2 <- graph_data$edge[from == j1h419_node$id & to == j1h419_node$familynum]
  expect_equal(edge1$color, "#3333334D")
  expect_equal(edge2$color, "#3333334D")
  
  # Edges from families to shared parent should match parent node colors
  real_max <- max(graph_data$node[nodetype %in% c("real", "compact")]$id)
  edges_to_parent <- graph_data$edge[to == j0z938_node$id & from > real_max]
  # Faded edges should have 4D suffix
  expect_true(all(grepl(paste0("^", j0z938_node$color, "(4D)?$"), edges_to_parent$color)))
})

test_that("highlight edges work correctly with compact mode", {
  # Test that compact mode doesn't break edge highlighting logic
  tidy_ped <- tidyped(simple_ped, genmethod = "bottom")
  graph_data <- visPedigree:::ped2igraph(
    tidy_ped, 
    compact = TRUE, 
    highlight = "J5X804", 
    trace = "up", 
    showf = FALSE
  )
  
  # Basic edge highlighting rules should still apply
  real_max <- max(graph_data$node[nodetype %in% c("real", "compact")]$id)
  
  # All edges should be either highlighted (#333333), faded (#3333334D), or match parent colors (with optional 4D)
  all_edges <- graph_data$edge
  node_colors <- graph_data$node$color
  node_colors_faded <- paste0(node_colors, "4D")
  valid_colors <- unique(c("#333333", "#3333334D", node_colors, node_colors_faded))
  expect_true(all(all_edges$color %in% valid_colors))
  
  # Edges from real/compact nodes should respect from_highlighted
  individual_edges <- all_edges[from <= real_max]
  if (nrow(individual_edges) > 0) {
    highlighted_ind_edges <- individual_edges[from_highlighted == TRUE]
    if (nrow(highlighted_ind_edges) > 0) {
      expect_true(all(highlighted_ind_edges$color == "#333333"))
    }
    
    faded_ind_edges <- individual_edges[from_highlighted == FALSE]
    if (nrow(faded_ind_edges) > 0) {
      expect_true(all(faded_ind_edges$color == "#3333334D"))
    }
  }
  
  # Edges from virtual nodes should follow parent colors (with fading for non-highlighted families)
  virtual_edges <- all_edges[from > real_max]
  if (nrow(virtual_edges) > 0) {
    parent_colors <- graph_data$node[id %in% virtual_edges$to, color]
    parent_colors_faded <- paste0(parent_colors, "4D")
    valid_parent_colors <- unique(c(parent_colors, parent_colors_faded))
    expect_true(all(virtual_edges$color %in% valid_parent_colors))
  }
})

#' Normalize custom labels before pedigree tidying
#' @keywords internal
prepare_visped_labels <- function(ped, labelvar) {
  if (is.null(labelvar)) {
    return(list(ped = ped, label_column = NULL))
  }

  if (!is.character(labelvar)) {
    stop(
      "'labelvar' must be NULL, a single non-empty column name, ",
      "or a character vector with one label per pedigree row."
    )
  }

  if (length(labelvar) == 1L) {
    if (is.na(labelvar) || labelvar == "") {
      stop(
        "'labelvar' must be NULL, a single non-empty column name, ",
        "or a character vector with one label per pedigree row."
      )
    }
    return(list(ped = ped, label_column = labelvar))
  }

  if (!is.data.frame(ped)) {
    stop("A 'labelvar' vector requires 'ped' to be a data.frame or data.table.")
  }

  if (length(labelvar) != nrow(ped)) {
    stop(
      sprintf(
        "A 'labelvar' vector must have one label per pedigree row (%d), not %d.",
        nrow(ped),
        length(labelvar)
      )
    )
  }

  ped <- copy(ped)
  label_column <- ".visped_label"
  while (label_column %in% names(ped)) {
    label_column <- paste0(label_column, "_")
  }
  set(ped, j = label_column, value = labelvar)

  list(ped = ped, label_column = label_column)
}

#' Normalize generation-label display settings
#' @keywords internal
prepare_visped_genlab <- function(genlab) {
  if (is.logical(genlab) && length(genlab) == 1L && !is.na(genlab)) {
    return(list(show = genlab, labels = NULL))
  }

  if (is.character(genlab) && length(genlab) > 0L) {
    if (anyNA(genlab) || any(!nzchar(genlab))) {
      stop("Custom 'genlab' labels must not be NA or empty strings.")
    }

    label_names <- names(genlab)
    if (!is.null(label_names) && any(!is.na(label_names) & nzchar(label_names))) {
      stop("Custom 'genlab' labels must be an unnamed character vector.")
    }

    return(list(show = TRUE, labels = unname(genlab)))
  }

  stop(
    "'genlab' must be TRUE, FALSE, or a non-empty unnamed character vector."
  )
}

#' Attach display labels to the final plotted generation layers
#' @keywords internal
prepare_visped_geninfo <- function(gen_info, labels = NULL) {
  gen_info <- copy(gen_info)

  if (!is.null(labels) && length(labels) != nrow(gen_info)) {
    stop(sprintf(
      "Custom 'genlab' labels must have one label for each displayed generation (%d), not %d.",
      nrow(gen_info),
      length(labels)
    ))
  }

  gen_info[, label := if (is.null(labels)) paste0("G", gen) else labels]
  return(gen_info[])
}

#' Visualize a tidy pedigree
#'
#' \code{visped} function draws a graph of a full or compact pedigree.
#'
#' This function takes a pedigree tidied by the \code{\link{tidyped}} function and outputs a hierarchical graph for all individuals in the pedigree. The graph can be shown on the default graphic device or saved as a PDF or SVG file. The vector output is legible and avoids overlapping labels. It is especially useful when the number of individuals is large and individual labels are long.
#'
#' Rendering is performed using a Two-Pass strategy: edges are drawn first to ensure center-to-center connectivity, followed by nodes and labels. This ensures perfect visual alignment in high-resolution vector outputs. The function also supports real-time ancestry and descendant highlighting.
#'
#' This function can draw the graph of a very large pedigree (> 10,000 individuals per generation) by compacting full-sib individuals. It is highly effective for aquatic animal pedigrees, which usually include many full-sib families per generation in nucleus breeding populations. The outline of a pedigree without individual labels is still shown if the width of a pedigree graph exceeds the maximum width (500 inches) of the PDF file.
#'
#' By default, \code{shapeby = "sex"} encodes individual sex with node shape:
#' females are circles, males are squares, individuals of unknown sex are
#' diamonds, and monoecious individuals are hexagons. Set
#' \code{shapeby = "role"} to use the legacy role-based scheme, where circles
#' represent individual records and green-grey rectangles represent compact
#' full-sib family summaries. Compact family summaries remain green-grey
#' rectangles in both modes. In \code{shapeby = "role"} mode, real individuals
#' remain circles, but their fill colors still encode sex. Dark sky blue
#' indicates males, dark goldenrod indicates females, teal indicates monoecious
#' individuals, and neutral grey indicates unknown sex. Purple borders indicate
#' highlighted individuals rather than sex.
#'
#' @param ped A \code{tidyped} object (which inherits from \code{data.table}). It is recommended that the pedigree is tidied and pruned by candidates using the \code{\link{tidyped}} function with the non-null parameter \code{cand}.
#' @param compact A logical value indicating whether terminal, non-parent
#'   full-sib individuals will be replaced by one family display node. For
#'   example, 100 eligible full siblings are shown as \code{"FS×100"} when
#'   \code{compact = TRUE}. Highlighted individuals are not compacted, and the
#'   underlying pedigree is unchanged. The default value is FALSE.
#' @param outline A logical value indicating whether shapes without labels will be shown. A graph of the pedigree without individual labels is shown when setting \code{outline = TRUE}. This is useful for viewing the pedigree outline and identifying immigrant individuals in each generation when the graph width exceeds the maximum PDF width (500 inches). The default value is FALSE.
#' @param cex NULL or a numeric value changing the size of individual labels shown in the graph. \emph{cex} is an abbreviation for 'character expansion factor'. The \code{visped} function will attempt to estimate (\code{cex=NULL}) the appropriate cex value and report it in the messages. Based on the reported cex from a previous run, this parameter should be increased if labels are wider than their shapes in the PDF; conversely, it should be decreased if labels are narrower than their shapes. The default value is NULL.
#' @param showgraph A logical value indicating whether a plot will be shown in the default graphic device (e.g., the Plots panel in RStudio). This is useful for quick viewing without opening a PDF file. However, the graph on the default device may not be legible (e.g., overlapping labels or aliasing lines) due to size restrictions. It is recommended to set \code{showgraph = FALSE} for large pedigrees. The default value is TRUE.
#' @param file NULL or a character value specifying whether the pedigree graph will be saved as a vector file. Files ending in '.svg' are written with the SVG device; all other file names use PDF output. The vector output is legible and avoids overlapping labels even with many individuals or long labels. It is recommended to save the pedigree graph as a vector file. The default value is NULL.
#' @param highlight NULL, a character vector of individual IDs, or a list specifying individuals to highlight. If a character vector is provided, individuals will be highlighted with a purple border while preserving their sex-based fill color. If a list is provided, it should contain:
#' \itemize{
#'   \item \code{ids}: (required) character vector of individual IDs to highlight.
#'   \item \code{frame.color}: (optional) hex color for the border of focal individuals.
#'   \item \code{color}: (optional) hex color for the fill of focal individuals.
#'   \item \code{rel.frame.color}: (optional) hex color for the border of relatives (used when \code{trace} is not NULL).
#'   \item \code{rel.color}: (optional) hex color for the fill of relatives (used when \code{trace} is not NULL).
#' }
#' For example: \code{c("A", "B")} or \code{list(ids = c("A", "B"), frame.color = "#9c27b0")}. The function will check if the specified individuals exist in the pedigree and issue a warning for any missing IDs. The default value is NULL.
#' @param trace A logical value or a character string. If TRUE, all ancestors and descendants of the individuals specified in \code{highlight} will be highlighted. If a character string, it specifies the tracing direction: "\strong{up}" (ancestors), "\strong{down}" (descendants), or "\strong{all}" (union of ancestors and descendants). This is useful for focusing on specific families within a large pedigree. The default value is FALSE.
#' @param showf A logical value indicating whether inbreeding coefficients will be shown in the graph. If \code{showf = TRUE} and the column \strong{f} is missing, \code{visped()} will try to compute it automatically with \code{\link{inbreed}} on a structurally complete pedigree. If automatic computation is not possible, a warning is issued and labels are drawn without \strong{f}. The default value is FALSE.
#' @param labelvar NULL, a single non-empty character value naming a column in
#'   \code{ped}, or a character vector with one displayed label per input
#'   pedigree row. Row-aligned vectors are attached before any internal
#'   tidying or reordering. Compact full-sib family nodes still show family
#'   size when \code{compact = TRUE}. Missing or empty values fall back to
#'   individual IDs. The default value is NULL.
#' @param shapeby Character string controlling what node shapes encode.
#'   \code{"sex"} (default) uses circles for females, squares for males,
#'   diamonds for unknown sex, and hexagons for monoecious individuals.
#'   \code{"role"} distinguishes individual records from compact family
#'   summaries by shape while keeping sex-based fill colors for real
#'   individuals. Compact family summaries are rectangles in both modes.
#' @param pagewidth A numeric value specifying the width of the PDF file in inches. This controls the horizontal scaling of the layout. The default value is 200.
#' @param symbolsize A numeric value specifying the scaling factor for node size relative to the label size. Values greater than 1 increase the node size (adding padding around the label), while values less than 1 decrease it. This is useful for fine-tuning the whitespace and legibility of dense graphs. The default value is 1.
#' @param maxiter An integer specifying the maximum number of iterations for the Sugiyama layout algorithm to minimize edge crossings. Higher values (e.g., 2000 or 5000) may result in fewer crossed lines for complex pedigrees but will increase computation time. The default value is 1000.
#' @param genlab FALSE to omit generation labels; TRUE to draw the default
#'   labels G1, G2, ...; or an unnamed, non-empty character vector of custom
#'   labels. Custom labels are assigned from top to bottom to the final
#'   displayed generation layers and must contain exactly one label per layer.
#'   The default value is FALSE.
#' @param genlabcex NULL or a numeric value controlling the size of displayed
#'   generation labels. If \code{NULL}, \code{visped()} uses an automatic size
#'   based on node scaling. Set a larger value to keep generation labels
#'   readable in deep pedigrees. The default value is NULL.
#' @param ... Additional arguments passed to \code{\link[igraph:plot.igraph]{plot.igraph}}.
#' @return The function mainly produces a plot on the current graphics device and/or a vector file. It invisibly returns a list containing the graph object, layout coordinates, and node sizes.
#'
#' @seealso
#' \code{\link{tidyped}} for tidying pedigree data (required input)
#' \code{\link{vismat}} for visualizing relationship matrices as heatmaps
#' \code{\link{pedmat}} for computing relationship matrices
#' \code{\link{splitped}} for splitting pedigree into connected components
#' \code{\link[igraph:plot.igraph]{plot.igraph}} underlying plotting function
#'
#' @note Isolated individuals (those with no parents and no progeny, assigned Gen 0) are automatically filtered out and not shown in the plot. A message will be issued if any such individuals are removed.
#'
#' @examples
#' library(visPedigree)
#' library(data.table)
#' # Drawing a simple pedigree
#' simple_ped_tidy <- tidyped(simple_ped)
#' visped(simple_ped_tidy, 
#'        cex=0.25, 
#'        symbolsize=5.5)
#'
#' # Use application-specific labels for the displayed generations.
#' # Custom labels are assigned from top to bottom.
#' generation_labels <- paste("Generation", sort(unique(simple_ped_tidy$Gen)))
#' visped(simple_ped_tidy,
#'        genlab = generation_labels,
#'        cex = 0.25,
#'        symbolsize = 5.5)
#'
#' # Highlighting an individual and its ancestors and descendants
#' visped(simple_ped_tidy, 
#'        highlight = "J5X804", 
#'        trace = "all", 
#'        cex=0.25, 
#'        symbolsize=5.5)
#'
#' # Showing inbreeding coefficients in the graph
#' simple_ped_tidy_inbreed <- tidyped(simple_ped, inbreed = TRUE)
#' visped(simple_ped_tidy_inbreed,
#'        showf = TRUE, 
#'        cex=0.25, 
#'        symbolsize=5.5)
#'
#' # visped() will automatically compute inbreeding coefficients if 'f' is missing
#' visped(simple_ped_tidy,
#'        showf = TRUE,
#'        cex=0.25,
#'        symbolsize=5.5)
#'
#' # Use the role-based legacy symbol scheme
#' visped(simple_ped_tidy,
#'        shapeby = "role",
#'        cex=0.25,
#'        symbolsize=5.5)
#'
#' # Adjusting page width and symbol size for better layout
#' # Increase pagewidth to spread nodes horizontally in the pdf file
#' # Increase symbolsize for more padding around individual labels
#' visped(simple_ped_tidy, 
#'        cex=0.25, 
#'        symbolsize=5.5, 
#'        pagewidth = 100, 
#'        file = tempfile(fileext = ".pdf"))
#'
#' # Saving pedigree as SVG with larger labels and tighter symbols
#' visped(simple_ped_tidy,
#'        cex=0.8,
#'        symbolsize=1,
#'        pagewidth = 100,
#'        file = tempfile(fileext = ".svg"))
#'
#' # Highlighting multiple individuals with custom colors
#' visped(simple_ped_tidy,
#'        highlight = list(ids = c("J3Y620", "J1X971"),
#'                         frame.color = "#4caf50",
#'                         color = "#81c784"),
#'        cex=0.25,
#'        symbolsize=5.5)
#'
#' # Handling large pedigrees: Saving to PDF is recommended for legibility
#' # The 'trace' and 'tracegen' parameters in tidyped() help prune the graph
#' cand_labels <- big_family_size_ped[(Year == 2007) & (substr(Ind,1,2) == "G8"), Ind]
#' \donttest{
#' big_ped_tidy <- tidyped(big_family_size_ped, 
#'                         cand = cand_labels, 
#'                         trace = "up", 
#'                         tracegen = 2)
#' # Use compact = TRUE for large families
#' visped(big_ped_tidy, 
#'        compact = TRUE, 
#'        cex=0.08, 
#'        symbolsize=5.5, 
#'        file = tempfile(fileext = ".pdf"))
#'
#' # Use outline = TRUE if individual labels are not required
#' visped(big_ped_tidy, 
#'        compact = TRUE, 
#'        outline = TRUE, 
#'        file = tempfile(fileext = ".pdf"))
#' }
#'
#' @import data.table
#' @import igraph
#' @importFrom grDevices pdf svg dev.off dev.cur
#' @importFrom graphics strwidth
#' @export
visped <- function(
  ped,
  compact = FALSE,
  outline = FALSE,
  cex = NULL,
  showgraph = TRUE,
  file = NULL,
  highlight = NULL,
  trace = FALSE,
  showf = FALSE,
  labelvar = NULL,
  shapeby = c("sex", "role"),
  pagewidth = 200,
  symbolsize = 1,
  maxiter = 1000,
  genlab = FALSE,
  genlabcex = NULL,
  ...
) {
  label_info <- prepare_visped_labels(ped, labelvar)
  ped <- label_info$ped
  label_column <- label_info$label_column
  genlab_info <- prepare_visped_genlab(genlab)

  # Automatically convert raw data to tidyped object if needed.
  # If the object already looks like a tidyped pedigree but only lost its class,
  # restore/validate it instead of rebuilding from raw input.
  tidyped_core <- c("Ind", "Sire", "Dam", "Sex", "Gen", "IndNum", "SireNum", "DamNum")
  looks_tidyped <- is.data.frame(ped) && all(tidyped_core %in% colnames(ped))

  if (!is_tidyped(ped) && !looks_tidyped) {
    ped <- tidyped(ped, addgen = TRUE, addnum = TRUE)
  }

  ped <- validate_tidyped(ped)
  shapeby <- match.arg(shapeby)

  if (!isTRUE(compact) && !isFALSE(compact)) {
    stop("'compact' must be TRUE or FALSE.")
  }

  if (!isTRUE(outline) && !isFALSE(outline)) {
    stop("'outline' must be TRUE or FALSE.")
  }

  if (!isTRUE(showgraph) && !isFALSE(showgraph)) {
    stop("'showgraph' must be TRUE or FALSE.")
  }

  if (!isTRUE(showf) && !isFALSE(showf)) {
    stop("'showf' must be TRUE or FALSE.")
  }

  if (!is.null(label_column) && !label_column %in% names(ped)) {
    stop(sprintf("Column '%s' specified by 'labelvar' was not found in the pedigree.", label_column))
  }

  if (
    !is.null(cex) &&
      (!is.numeric(cex) || length(cex) != 1 || is.na(cex) || cex <= 0)
  ) {
    stop("'cex' must be NULL or a single positive number.")
  }

  if (
    !is.null(file) && (!is.character(file) || length(file) != 1 || is.na(file))
  ) {
    stop("'file' must be NULL or a single character string.")
  }

  # 1. Ensure at least one output is selected (validated after flags are checked)
  if (!showgraph && is.null(file)) {
    stop(
      "Both 'showgraph' and 'file' are disabled. No output will be generated."
    )
  }

  if (!is.null(highlight) && !is.character(highlight) && !is.list(highlight)) {
    stop("'highlight' must be NULL, a character vector, or a list.")
  }

  if (!isFALSE(trace)) {
    if (isTRUE(trace)) {
      trace <- "all"
    } else if (
      !is.character(trace) ||
        length(trace) != 1 ||
        is.na(trace) ||
        !trace %in% c("up", "down", "all")
    ) {
      stop("'trace' must be TRUE, FALSE, 'up', 'down', or 'all'.")
    }
  }

  if (
    !is.numeric(pagewidth) ||
      length(pagewidth) != 1 ||
      is.na(pagewidth) ||
      pagewidth <= 0
  ) {
    stop("'pagewidth' must be a single positive number.")
  }

  if (
    !is.numeric(symbolsize) ||
      length(symbolsize) != 1 ||
      is.na(symbolsize) ||
      symbolsize <= 0
  ) {
    stop("'symbolsize' must be a single positive number.")
  }

  if (
    !is.numeric(maxiter) ||
      length(maxiter) != 1 ||
      is.na(maxiter) ||
      maxiter <= 0
  ) {
    stop("'maxiter' must be a single positive integer.")
  }
  maxiter <- as.integer(maxiter)

  if (
    !is.null(genlabcex) &&
      (!is.numeric(genlabcex) || length(genlabcex) != 1 || is.na(genlabcex) || genlabcex <= 0)
  ) {
    stop("'genlabcex' must be NULL or a single positive number.")
  }

  # 2. Sanitize highlight inputs
  if (!is.null(highlight)) {
    if (is.character(highlight)) {
      highlight <- highlight[!is.na(highlight) & highlight != ""]
    } else if (is.list(highlight) && !is.null(highlight[["ids"]])) {
      highlight[["ids"]] <- highlight[["ids"]][
        !is.na(highlight[["ids"]]) & highlight[["ids"]] != ""
      ]
    }
    # If sanitization left it empty, set to NULL
    if (is.character(highlight) && length(highlight) == 0) {
      highlight <- NULL
    }
    if (is.list(highlight) && length(highlight[["ids"]]) == 0) highlight <- NULL
  }

  if (showf && !has_inbreeding(ped)) {
    if (is_complete_pedigree(ped)) {
      ped <- inbreed(ped)
      message(
        "Note: 'showf = TRUE' requested but 'f' column was missing. ",
        "Calculated inbreeding coefficients automatically."
      )
    } else {
      warning(
        "Inbreeding coefficients ('f' column) not found and cannot be ",
        "computed automatically because the pedigree is structurally incomplete. ",
        "Run tidyped(..., inbreed = TRUE) on a complete pedigree first, ",
        "or extract a valid sub-pedigree with tidyped(tp, cand = ids, trace = \"up\")."
      )
      showf <- FALSE
    }
  }

  # Prepare graph data
  # Use a temporary PDF device to calculate text metrics if file output is requested
  # This ensures strwidth/strheight match the target PDF device (e.g. font metrics)
  if (!is.null(file)) {
    tmp_pdf <- tempfile()
    # Pass font family if provided in ... to ensure accurate cex estimation
    dots <- list(...)
    pdf_args <- list(file = tmp_pdf)
    if ("family" %in% names(dots)) {
      pdf_args$family <- dots$family
    }
    do.call(pdf, pdf_args)

    on.exit(
      {
        if (file.exists(tmp_pdf)) unlink(tmp_pdf)
      },
      add = TRUE
    )

    # Calculate graph data using temp device for font metrics
    graph_data <- tryCatch(
      {
        prepare_ped_graph(
          ped = ped,
          compact = compact,
          outline = outline,
          cex = cex,
          highlight = highlight,
          trace = trace,
          showf = showf,
          labelvar = label_column,
          shapeby = shapeby,
          pagewidth = pagewidth,
          symbolsize = symbolsize,
          maxiter = maxiter,
          ...
        )
      },
      finally = {
        if (dev.cur() > 1) dev.off()
      }
    )
  } else {
    graph_data <- prepare_ped_graph(
      ped = ped,
      compact = compact,
      outline = outline,
      cex = cex,
      highlight = highlight,
      trace = trace,
      showf = showf,
      labelvar = label_column,
      shapeby = shapeby,
      pagewidth = pagewidth,
      symbolsize = symbolsize,
      maxiter = maxiter,
      ...
    )
  }

  g <- graph_data$g
  l <- graph_data$layout
  node_size <- graph_data$node_size
  if (genlab_info$show) {
    graph_data$gen_info <- prepare_visped_geninfo(
      graph_data$gen_info,
      genlab_info$labels
    )
  }

  #===Draw the pedigree================================================================
  if (showgraph) {
    plot_ped_igraph(
      g,
      l,
      node_size,
      gen_info = graph_data$gen_info,
      genlab = genlab_info$show,
      genlabcex = genlabcex,
      ...
    )
  }

  if (!is.null(file)) {
    file_ext <- tolower(tools::file_ext(file))
    output_device <- if (identical(file_ext, "svg")) grDevices::svg else grDevices::pdf
    output_device(
      file = file,
      width = graph_data$canvas_width,
      height = graph_data$canvas_height
    )
    # Ensure the device is closed even if plotting fails.
    on.exit(if (dev.cur() > 1) dev.off(), add = TRUE)
    plot_ped_igraph(
      g,
      l,
      node_size,
      gen_info = graph_data$gen_info,
      genlab = genlab_info$show,
      genlabcex = genlabcex,
      ...
    )

    # Correct path normalization for the message
    saved_path <- tryCatch(
      normalizePath(file, mustWork = FALSE),
      error = function(e) file
    )
    message(paste("Pedigree saved to: ", saved_path, sep = ""))
  }

  if ((showgraph || !is.null(file)) && !outline) {
    current_cex <- if (is.null(cex)) graph_data$best_cex else cex
    current_genlabcex <- if (
      genlab_info$show &&
        !is.null(graph_data$gen_info) &&
        nrow(graph_data$gen_info) > 0L
    ) {
      resolve_visped_genlabcex(node_size, genlabcex)
    } else {
      NULL
    }
    message(paste(
      "Label cex: ",
      current_cex,
      ". Symbol size: ",
      symbolsize,
      if (!is.null(current_genlabcex)) {
        paste0(". Generation label cex: ", current_genlabcex)
      },
      if (!is.null(current_genlabcex)) {
        ". Adjust 'cex', 'symbolsize', and 'genlabcex' if labels are too large or small."
      } else {
        ". Adjust 'cex' and 'symbolsize' if labels are too large or small."
      },
      sep = ""
    ))
  }

  if ((showgraph || !is.null(file)) && is.null(file)) {
    message("Tip: Use 'file' to save as a legible vector PDF or SVG.")
  }

  if (showf && "f" %in% colnames(ped) && any(ped$f == 0, na.rm = TRUE)) {
    message("Note: Inbreeding coefficients of 0 are not shown in the graph.")
  }

  graph_data$gen_info <- NULL
  invisible(graph_data)
}

test_that("visped handles isolated nodes in layout correctly", {
  skip_if_not_installed("visPedigree")
  
  # This specific test case (small_ped with bottom sorting + specific highlight)
  # created a graph where some nodes existed in the node table but not in the edge list
  # causing a mismatch when assigning x coordinates back to the node table.
  # The fix was to ensure graph_from_data_frame includes 'vertices' argument.
  
  tped_bottom <- tidyped(small_ped, genmethod = "bottom")
  expect_error(visped(tped_bottom, highlight = "Z1"), NA)
  
  # Test with another method to be sure
  tped_top <- tidyped(small_ped, genmethod = "top")
  expect_error(visped(tped_top, highlight = "Z1"), NA)
  
  # Expanded test cases with other datasets
  
  # 1. simple_ped
  tped_simple_bottom <- tidyped(simple_ped, genmethod = "bottom")
  expect_error(visped(tped_simple_bottom), NA)
  
  # 2. deep_ped (larger dataset)
  tped_deep <- tidyped(deep_ped, genmethod = "bottom")
  # Taking a subset via cand to test layout logic without overwhelming plot
  tped_deep_sub <- tidyped(deep_ped, cand="K110550H", trace="up", tracegen=3, genmethod="bottom")
  expect_error(visped(tped_deep_sub), NA)
  
  # 3. loop_ped (with loops - although tidyped cleans/warns, visped should layout cleanly whatever tidyped produces)
  # loop_ped causes error in tidyped by default, skip full fail test here, assume clean input
  
  # 4. Edge case: very small layout - requires at least some parentage to be built otherwise tidyped stops
  # Create a minimal valid pedigree
  mini_ped <- data.frame(Ind=c("A", "B"), Sire=c(NA, "A"), Dam=c(NA, NA))
  tmini <- tidyped(mini_ped)
  expect_error(visped(tmini), NA)
})

#' A large pedigree with big family sizes
#'
#' A dataset containing a pedigree with many full-sib individuals per family.
#'
#' @format A data.table with 8 columns:
#' \describe{
#'   \item{Ind}{Individual ID}
#'   \item{Sire}{Sire ID}
#'   \item{Dam}{Dam ID}
#'   \item{Sex}{Sex of the individual}
#'   \item{Year}{Year of birth}
#'   \item{IndNum}{Numeric ID for individual}
#'   \item{SireNum}{Numeric ID for sire}
#'   \item{DamNum}{Numeric ID for dam}
#' }
"big_family_size_ped"

#' A deep pedigree
#'
#' A dataset containing a pedigree with many generations.
#'
#' @format A data.table with 4 columns:
#' \describe{
#'   \item{Ind}{Individual ID}
#'   \item{Sire}{Sire ID}
#'   \item{Dam}{Dam ID}
#'   \item{Sex}{Sex of the individual}
#' }
"deep_ped"

#' A pedigree with loops
#'
#' A dataset containing a pedigree with circular mating loops.
#'
#' @format A data.table with 3 columns:
#' \describe{
#'   \item{Ind}{Individual ID}
#'   \item{Sire}{Sire ID}
#'   \item{Dam}{Dam ID}
#' }
"loop_ped"

#' A simple pedigree
#'
#' A small dataset containing a simple pedigree for demonstration.
#'
#' @format A data.table with 4 columns:
#' \describe{
#'   \item{Ind}{Individual ID}
#'   \item{Sire}{Sire ID}
#'   \item{Dam}{Dam ID}
#'   \item{Sex}{Sex of the individual}
#' }
"simple_ped"

#' A small pedigree
#'
#' A small dataset containing a pedigree with some missing parents.
#'
#' @format A data.frame with 3 columns:
#' \describe{
#'   \item{Ind}{Individual ID}
#'   \item{Sire}{Sire ID}
#'   \item{Dam}{Dam ID}
#' }
"small_ped"

#' A highly inbred pedigree
#'
#' A simulated pedigree designed to demonstrate high levels of inbreeding and partial inbreeding decomposition.
#' Contains full-sib mating and backcrossing.
#'
#' @format A data.table with 5 columns:
#' \describe{
#'   \item{Ind}{Individual ID}
#'   \item{Sire}{Sire ID}
#'   \item{Dam}{Dam ID}
#'   \item{Sex}{Sex of the individual}
#'   \item{Gen}{Generation number}
#' }
"inbred_ped"

#' A pedigree with half founders
#'
#' A dataset from ENDOG containing individuals with a single missing parent (half founders).
#' Useful for testing genetic algorithms correctly conserving probability mass for missing lineages.
#'
#' @format A data.frame with 4 columns:
#' \describe{
#'   \item{Ind}{Individual ID}
#'   \item{Sire}{Sire ID}
#'   \item{Dam}{Dam ID}
#'   \item{Sex}{Sex of the individual}
#' }
"half_founder_ped"

#' A complex pedigree
#'
#' A dataset containing a large, complex pedigree covering about 100 generations,
#' useful for testing the performance and accuracy of partial inbreeding and similar calculations.
#'
#' @format A data.table with a standard pedigree structure.
"complex_ped"

#' Internal layout engine for pedigree visualization
#' @import data.table
#' @importFrom igraph graph_from_data_frame layout_with_sugiyama norm_coords V
#' @importFrom graphics strwidth
#' @keywords internal
prepare_ped_graph <- function(ped, compact = FALSE, outline = FALSE, cex = NULL, 
                              highlight = NULL, trace = FALSE, showf = FALSE,
                              labelvar = NULL, shapeby = "sex", pagewidth = 200,
                              symbolsize = 1, maxiter = 1000, ...) {
  ped_new <- copy(ped)

  # Check and tidyped if necessary
  if (!"Gen" %in% colnames(ped_new) || !all(c("IndNum", "SireNum", "DamNum") %in% colnames(ped_new))) {
    ped_new <- tidyped(ped_new, addgen = TRUE, addnum = TRUE)
  }
  
  # Remove isolated individuals early to optimize performance and simplify graph conversion
  if ("Gen" %in% colnames(ped_new)) {
    n_iso <- sum(ped_new$Gen == 0, na.rm = TRUE)
    if (n_iso > 0) {
      message(sprintf("Note: Removed %d isolated individuals (no parents, no progeny) from the plot.", n_iso))
      ped_new <- ped_new[Gen != 0]
    }
  }

  # Reserved digits
  fixed_digits <- 7
  old_digits <- getOption("digits")
  options(digits = 20)
  on.exit(options(digits = old_digits), add = TRUE)

  ped_igraph_data <- ped2igraph(
    ped_new,
    compact = compact,
    highlight = highlight,
    trace = trace,
    showf = showf,
    labelvar = labelvar,
    shapeby = shapeby
  )
  ped_igraph <- ped_igraph_data
  real_node <- ped_igraph$node[nodetype %in% c("real", "compact")]
  
  if (nrow(real_node) == 0) {
    stop("Pedigree contains no individuals to plot.")
  }

  gen_node_num <- real_node[, .N, by = gen]
  gen_max_size <- max(gen_node_num$N, na.rm = TRUE)

  pdf_max_width_soft <- pagewidth
  s_size <- symbolsize
  cexs <- seq(from = 0.1, to = 1, by = 0.05)
  best_cex <- 0
  max_strwidth_label <- real_node[which.max(strwidth(real_node$label, cex = 1, units = "inches")), label]
  
  for (i in length(cexs):1) {
    label_max_width <- max(strwidth(max_strwidth_label, cex = cexs[i], units = "inches"), na.rm = TRUE)
    if (gen_max_size <= 16 & label_max_width < 0.8) {
      label_max_width <- 0.8
    }
    if ((label_max_width * s_size * gen_max_size) < pdf_max_width_soft) {
      best_cex <- cexs[i] * 0.65
      break
    }
  }

  if (!outline & best_cex == 0 & is.null(cex)) {
    stop(sprintf("Too many individuals per generation; cannot fit within %.1f inches. Use compact=TRUE, outline=TRUE, or increase pagewidth.", pdf_max_width_soft))
  }

  if (!outline && !is.null(cex)) {
    label_max_width <- max(strwidth(max_strwidth_label, cex = cex, units = "inches"), na.rm = TRUE)
    if (gen_max_size <= 16 && label_max_width < 0.8) {
      label_max_width <- 0.8
    }
    if ((label_max_width * s_size * gen_max_size) >= pdf_max_width_soft) {
      warning(sprintf("Provided cex likely exceeds PDF width (%.1f inches); labels may clip. Try compact=TRUE, outline=TRUE, or increase pagewidth.", pdf_max_width_soft), call. = FALSE)
    }
  }

  hgap <- round(1 / gen_max_size, 8)
  gen_num <- max(real_node$gen, na.rm = TRUE)
  max_layer <- max(ped_igraph$node$layer, na.rm = TRUE)
  
  # Create a minimal graph for layout calculation to save memory and time
  g_layout <- graph_from_data_frame(ped_igraph$edge[, .(from, to)], directed = TRUE, vertices = ped_igraph$node[, .(id)])
  # match names to get layers - V(g_layout)$name exists by default from graph_from_data_frame
  layer_idx <- ped_igraph$node$layer[match(v_names <- V(g_layout)$name, as.character(ped_igraph$node$id))]
  layer_idx <- max_layer - layer_idx + 1
  
  l <- layout_with_sugiyama(g_layout, layers = layer_idx, hgap = hgap, maxiter = maxiter, attributes = "all")$layout
  l <- norm_coords(l, xmin = 0, xmax = 1, ymin = 0, ymax = 1)
  
  # Map coordinates back to ped_igraph$node based on names
  # The layout l corresponds to V(g_layout)
  layout_dt <- data.table(id = as.integer(v_names), x = l[, 1], y = l[, 2])
  ped_igraph$node[layout_dt, on = "id", `:=`(x = i.x, y = i.y)]

  real_node <- ped_igraph$node[nodetype %in% c("real", "compact")]
  for (i in 1:gen_num) {
    v_rank <- rank(real_node[gen == i, x], na.last = TRUE, ties.method = "first")
    x_sorted <- round(sort(real_node[gen == i, x]), fixed_digits)
    x_new <- repeloverlap(x_sorted)
    real_node[gen == i, x := x_new[v_rank]]
  }
  ped_igraph$node[nodetype %in% c("real", "compact")] <- real_node

  if (gen_max_size >= 2) {
    min_gap <- 0.8 / gen_max_size
    fam_gap <- min_gap * 1.2
    
    for (i in seq_len(gen_num)) {
      current_nodes <- real_node[gen == i]
      if (nrow(current_nodes) == 0) next
      
      if (any(is.na(current_nodes$familylabel))) {
        current_nodes[is.na(familylabel), familylabel := paste0("u_", id)]
      }
      
      parent_coords <- real_node[, .(Ind, px = x)]
      current_nodes[parent_coords, on = .(sirelabel = Ind), sire_x := i.px]
      current_nodes[parent_coords, on = .(damlabel = Ind), dam_x := i.px]
      current_nodes[, parent_mid := rowMeans(.SD, na.rm = TRUE), .SDcols = c("sire_x", "dam_x")]
      current_nodes[is.na(parent_mid), parent_mid := x]
      
      fam_stats <- current_nodes[, .(
        N = .N,
        target_x = mean(parent_mid, na.rm=TRUE),
        original_x_mean = mean(x, na.rm=TRUE)
      ), by = familylabel]
      
      fam_stats[, half_width := ((N - 1) * min_gap) / 2]
      setorder(fam_stats, target_x, original_x_mean)
      
      n_fams <- nrow(fam_stats)
      if (n_fams > 0) {
        fam_stats[, fwd_pos := target_x]
        fam_stats[, bwd_pos := target_x]
        
        if (n_fams > 1) {
          for (k in 2:n_fams) {
            req_dist <- fam_stats$half_width[k-1] + fam_gap + fam_stats$half_width[k]
            if (fam_stats$fwd_pos[k] < fam_stats$fwd_pos[k-1] + req_dist) {
              fam_stats$fwd_pos[k] <- fam_stats$fwd_pos[k-1] + req_dist
            }
          }
          for (k in (n_fams-1):1) {
            req_dist <- fam_stats$half_width[k] + fam_gap + fam_stats$half_width[k+1]
            if (fam_stats$bwd_pos[k] > fam_stats$bwd_pos[k+1] - req_dist) {
              fam_stats$bwd_pos[k] <- fam_stats$bwd_pos[k+1] - req_dist
            }
          }
        }
        
        fam_stats[, final_center := (fwd_pos + bwd_pos) / 2]
        current_nodes[fam_stats, on = "familylabel", fam_center := i.final_center]
        setorder(current_nodes, familylabel, x) 
        current_nodes[, sib_offset := (seq_len(.N) - 1 - (.N - 1)/2) * min_gap, by = familylabel]
        real_node[current_nodes, on = "id", x := i.fam_center + i.sib_offset]
      }
    }
    
    rx <- range(real_node$x, na.rm = TRUE)
    if (diff(rx) > 1e-9) {
      real_node[, x := (x - rx[1]) / diff(rx)]
    } else {
      real_node[, x := 0.5]
    }
    ped_igraph$node[nodetype %in% c("real", "compact")] <- real_node
  }

  virtual_node <- ped_igraph$node[nodetype %in% c("virtual")]
  if (nrow(virtual_node) > 0) {
    real_family_min_x <- real_node[, .(minx = min(x, na.rm = TRUE)), by = .(gen, familylabel)]
    virtual_node[real_family_min_x, x := i.minx, on = .(gen, familylabel)]
  }
  ped_igraph$node[nodetype %in% c("virtual")] <- virtual_node

  # Update layout matrix with adjusted x coordinates for correct canvas width calculation
  l[, 1] <- ped_igraph$node[, x]

  s_size <- if (!is.null(symbolsize)) symbolsize else 1
  calc_node_width <- label_max_width * s_size
  node_width_s <- label_max_width * s_size
  
  if (!outline) {
    if (gen_max_size >= 2) {
      x_stats_gen <- real_node[, .(.N, range = max(x, na.rm = TRUE) - min(x, na.rm = TRUE)), by = gen]
      x_stats_gen[range > 0 & N > 1, ":="(meanspace = range / (N - 1))]
      min_node_space <- min(x_stats_gen[meanspace > 0, meanspace], na.rm = TRUE)
      f <- if (max(x_stats_gen$N) <= 16) 3 * round(node_width_s / min_node_space, 8) else round(node_width_s / min_node_space, 8)
    } else {
      f <- 1
    }
    canvas_width_s <- max(f * l[, 1], na.rm = TRUE) - min(f * l[, 1], na.rm = TRUE) + 6 * node_width_s
  } else {
    canvas_width_s <- min(label_max_width * gen_max_size, pdf_max_width_soft)
    node_width_s <- 0.0001
    if (0.0001 * gen_max_size > pdf_max_width_soft) stop("Outline view unavailable.")
  }

  canvas_width_s <- max(10, min(pdf_max_width_soft, canvas_width_s))
  min_layer_height <- if (outline) 0.1 else 0.8
  effective_layer_height <- max(calc_node_width, min_layer_height)
  canvas_height <- gen_num * effective_layer_height + 3 * effective_layer_height
  canvas_height <- max(8, canvas_height, canvas_width_s * 0.618)

  # Prepare final attributes in data.table BEFORE creating the graph
  ped_igraph$node[, `:=`(size = 0.001, size2 = 0.001, label.cex = 1.0)]
  
  node_size_val <- round(node_width_s * 100 / canvas_width_s, 8)
  current_cex <- if (is.null(cex)) best_cex else cex
  
  if (outline) {
    # Outline mode: icons are tiny, no labels except highlighted
    ped_igraph$node[nodetype %in% c("real", "compact"), `:=`(
      size = 0.0001,
      size2 = 0.0001
    )]
    is_h <- !is.na(ped_igraph$node$highlighted) & ped_igraph$node$highlighted
    ped_igraph$node[!is_h, label := ""]
    
    if (any(is_h)) {
      highlight_node_size <- round(label_max_width * 100 / canvas_width_s, 8)
      ped_igraph$node[is_h & nodetype %in% c("real", "compact"), `:=`(
        size = highlight_node_size,
        size2 = highlight_node_size,
        label.cex = if (is.null(cex)) (if (best_cex > 0) best_cex else 0.6) else cex
      )]
    }
  } else {
    ped_igraph$node[nodetype %in% c("real", "compact"), `:=`(
      size = node_size_val,
      size2 = node_size_val,
      label.cex = current_cex
    )]
  }

  # Compute generation info for labels (using final layout coordinates)
  gen_info <- ped_igraph$node[nodetype %in% c("real", "compact"),
                              .(y = mean(y, na.rm = TRUE), N = .N), by = gen]
  setorder(gen_info, gen)
  
  # Final graph creation with all attributes
  # Ensure the order of columns in edge and node are correct for igraph
  g <- graph_from_data_frame(ped_igraph$edge, directed = TRUE, ped_igraph$node)
  max_node_size <- max(ped_igraph$node$size, na.rm = TRUE)

  return(list(
    g = g,
    layout = as.matrix(ped_igraph$node[, .(x, y)]),
    canvas_width = canvas_width_s,
    canvas_height = canvas_height,
    node_size = max_node_size,
    best_cex = best_cex,
    gen_info = gen_info
  ))
}

#' Repel overlapping nodes on the x-axis
#' @param x A numeric vector of x positions.
#' @return A numeric vector of unique x positions.
#' @keywords internal
repeloverlap <- function(x) {
  n <- length(x)
  if (n <= 1 || anyDuplicated(x) == 0) return(x)
  
  x_dt <- as.data.table(x)
  x_counts <- x_dt[, .N, by = x]
  dup_info <- x_counts[N > 1]
  if (nrow(dup_info) == 0 || nrow(x_counts) == 1) return(x)
  
  setorder(dup_info, x)
  unique_pos <- sort(x_counts[, x])
  n_unique <- length(unique_pos)
  n_dup <- nrow(dup_info)
  
  max_new_vals <- sum(dup_info$N) - n_dup
  result <- numeric(n_unique + max_new_vals)
  result[seq_len(n_unique)] <- unique_pos
  result_idx <- n_unique
  
  for (i in seq_len(n_dup)) {
    dup_val <- dup_info$x[i]
    n_copies <- dup_info$N[i]
    idx <- match(dup_val, unique_pos)
    
    if (i == n_dup - 1 && n_dup >= 2) {
      next_dup_val <- dup_info$x[i + 1]
      next_idx <- match(next_dup_val, unique_pos)
      if (next_idx == idx + 1) {
        next_n_copies <- dup_info$N[i + 1]
        total_copies <- n_copies + next_n_copies - 1
        gap <- (next_dup_val - dup_val) / total_copies
        n_new1 <- n_copies - 1
        n_new2 <- next_n_copies - 1
        result[(result_idx + 1):(result_idx + n_new1)] <- dup_val + gap * seq_len(n_new1)
        result[(result_idx + n_new1 + 1):(result_idx + n_new1 + n_new2)] <- 
          dup_val + gap * seq(n_copies, total_copies - 1)
        result_idx <- result_idx + n_new1 + n_new2
        break
      }
    }
    
    n_new <- n_copies - 1
    if (n_new > 0) {
      if (idx < n_unique) {
        next_pos <- unique_pos[idx + 1]
        gap <- (next_pos - dup_val) / n_copies
        result[(result_idx + 1):(result_idx + n_new)] <- dup_val + gap * seq_len(n_new)
      } else if (idx > 1) {
        prev_pos <- unique_pos[idx - 1]
        gap <- (dup_val - prev_pos) / n_copies
        result[(result_idx + 1):(result_idx + n_new)] <- prev_pos + gap * seq_len(n_new)
      }
      result_idx <- result_idx + n_new
    }
  }
  return(result[seq_len(result_idx)])
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

cpp_set_num_threads <- function(threads) {
    .Call(`_visPedigree_cpp_set_num_threads`, threads)
}

cpp_openmp_available <- function() {
    .Call(`_visPedigree_cpp_openmp_available`)
}

cpp_calculate_inbreeding <- function(sire, dam) {
    .Call(`_visPedigree_cpp_calculate_inbreeding`, sire, dam)
}

cpp_build_ainv_triplets <- function(sire, dam, dii) {
    .Call(`_visPedigree_cpp_build_ainv_triplets`, sire, dam, dii)
}

cpp_calculate_A <- function(sire, dam) {
    .Call(`_visPedigree_cpp_calculate_A`, sire, dam)
}

cpp_calculate_D <- function(sire, dam, A) {
    .Call(`_visPedigree_cpp_calculate_D`, sire, dam, A)
}

cpp_calculate_AA <- function(A) {
    .Call(`_visPedigree_cpp_calculate_AA`, A)
}

cpp_invert_dense <- function(M) {
    .Call(`_visPedigree_cpp_invert_dense`, M)
}

cpp_invert_sympd <- function(M) {
    .Call(`_visPedigree_cpp_invert_sympd`, M)
}

cpp_invert_auto <- function(M) {
    .Call(`_visPedigree_cpp_invert_auto`, M)
}

cpp_multiply_A <- function(sire, dam, dii, rhs) {
    .Call(`_visPedigree_cpp_multiply_A`, sire, dam, dii, rhs)
}

cpp_multiply_Ainv <- function(sire, dam, dii, rhs) {
    .Call(`_visPedigree_cpp_multiply_Ainv`, sire, dam, dii, rhs)
}

cpp_assign_generations_top <- function(sire, dam, topo_order) {
    .Call(`_visPedigree_cpp_assign_generations_top`, sire, dam, topo_order)
}

cpp_calculate_partial_inbreeding <- function(sire, dam, dii, ancestors) {
    .Call(`_visPedigree_cpp_calculate_partial_inbreeding`, sire, dam, dii, ancestors)
}

cpp_assign_generations_bottom <- function(sire, dam, topo_order) {
    .Call(`_visPedigree_cpp_assign_generations_bottom`, sire, dam, topo_order)
}

cpp_pedcontrib <- function(sire, dam, cohort_pos, mode) {
    .Call(`_visPedigree_cpp_pedcontrib`, sire, dam, cohort_pos, mode)
}

cpp_calculate_ancestry <- function(sire, dam, res_mat) {
    .Call(`_visPedigree_cpp_calculate_ancestry`, sire, dam, res_mat)
}

cpp_trace_ancestors <- function(sire, dam, targets, max_gen = 0L) {
    .Call(`_visPedigree_cpp_trace_ancestors`, sire, dam, targets, max_gen)
}

cpp_trace_descendants <- function(sire, dam, targets, max_gen = 0L) {
    .Call(`_visPedigree_cpp_trace_descendants`, sire, dam, targets, max_gen)
}

cpp_topo_order <- function(sire, dam) {
    .Call(`_visPedigree_cpp_topo_order`, sire, dam)
}

#' @importFrom Rcpp sourceCpp
#' @importFrom stats as.formula sd
#' @useDynLib visPedigree, .registration = TRUE
if (getRversion() >= "2.15.1") {
  utils::globalVariables(
    c(
      ".",
      "..new_names",
      "..ped_column_name_new",
      "Cand",
      "Dam",
      "DamNum",
      "FamilyLabel",
      "Gen",
      "Ind",
      "IndNum",
      "MaxTraceGen",
      "N",
      "Sex",
      "Sire",
      "SireNum",
      "TraceDirec",
      "TraceGen",
      "TraceGenInd",
      "TraceGenInterval",
      "damlabel",
      "damnum",
      "f",
      "familylabel",
      "familynum",
      "from",
      "gen",
      "i.f_val",
      "i.minx",
      "id",
      "label",
      "layer",
      "nodetype",
      "sex",
      "sirelabel",
      "sirenum",
      "to",
      "tonodecolor",
      "x",
      "meanspace",
      "range",
      "familylabelsex",
      "familysize",
      "highlighted",
      "color",
      "frame.color",
      "label.color",
      "frame.width",
      "label.font",
      "to_highlighted",
      "from_highlighted",
      "compact_id",
      "compact_label",
      "from_id",
      "to_id",
      "to_label",
      "sib_offset",
      "fam_center",
      "final_center",
      "bwd_pos",
      "fwd_pos",
      "half_width",
      "target_x",
      "original_x_mean",
      "parent_mid",
      "dam_x",
      "sire_x",
      "px",
      "i.final_center",
      "i.px",
      "DamH",
      "Height",
      "MaxH",
      "SireH",
      "i.fam_center",
      "i.sib_offset",
      "i.x",
      "i.y",
      "size",
      "size2",
      "y",
      "..cols",
      "..new_names_edge",
      "..new_names_node",
      "family_key",
      "family_label",
      "family_size",
      "FamilySize",
      "FamilyID",
      "RepInd",
      "RepIndNum",
      "IsRepresentative",
      "IsCompacted",
      "IsParent",
      "SizeCategory",
      "i.Ind",
      "i.IndNum",
      "FamilyLabel",
      "OldIndNum",
      "OldRepIndNum",
      "NewNum",
      "OldNum",
      "NewSireNum",
      "NewDamNum",
      "NewRepIndNum",
      "NewIndNum",
      "OrigRepInd"
    )
  )
}

utils::globalVariables(c('i.SireNum', 'i.DamNum', 'i.Sire', 'i.Dam', 'i.IsParent', 'sort_key', 'NCompressed', 'Family', 'grp', 'NewGen', 'Limit', 'CohortLabel', 'DeltaF_Ind', 'ECG', 'monoecious_nums', 'role',
  '.orig_order_visped', 'Cohort', 'Contrib', 'Count', 'CumContrib', 'DeltaF', 'FClass', 'GenEquiv', 'Group', 'Interval', 'Mean', 'Ne', 'Pathway', 'Percentage', 'Rank', 'Time', 'child', 'child_label', 'family_id', 'from_label', 'i.label', 'rep_dam', 'rep_sire', 'Fe', 'Fa', 'Fg', 'LnFe', 'LnFa', 'LnFg', 'LnFaFe', 'LnFgFa', 'fe', 'fa', 'fg', 'lnfe', 'lnfa', 'lnfg', 'lnfafe', 'lnfgfa', 'TimeStep', 'Value', 'Metric'))

#' Calculate Information-Theoretic Diversity Half-Life
#'
#' @description
#' Calculates the diversity half-life (\eqn{T_{1/2}}) of a pedigree across
#' time points using a Renyi-2 entropy cascade framework.  The total loss rate of
#' genetic diversity is partitioned into three additive components:
#' \itemize{
#'   \item \eqn{\lambda_e}: foundation bottleneck (unequal founder
#'     contributions).
#'   \item \eqn{\lambda_b}: breeding bottleneck (overuse of key ancestors).
#'   \item \eqn{\lambda_d}: genetic drift (random sampling loss).
#' }
#'
#' The function rolls over time points defined by \code{timevar}, computing
#' \eqn{f_e} and \eqn{f_a} (via \code{\link{pedcontrib}}) and \eqn{f_g}
#' (via the internal matrix-free coancestry engine) for each time point. Only
#' the \eqn{f_g} statistic is requested; no redundant Ne calculations are
#' performed.
#'
#' @param ped A \code{tidyped} object.
#' @param timevar Character.
#'   Column name in \code{ped} that defines time points
#'   (e.g. \code{"Gen"}, \code{"Year"}).  Default: \code{"Gen"}.
#' @param at Optional vector of values selecting which time points to include
#'   (e.g., \code{2:4}, \code{2010:2020}, or non-consecutive \code{c(2015, 2018, 2022)}).
#'   Values must match entries in the \code{timevar} column.  Non-numeric values
#'   are accepted but the OLS time axis will fall back to sequential indices.
#'   If \code{NULL} (default), all non-NA unique values in \code{timevar} are used.
#' @param nsamples Integer.
#'   Sample size per time point for coancestry estimation
#'   (passed to the internal coancestry engine).  Default: 1000.
#' @param ncores Integer.
#'   Number of OpenMP threads for C++ backends.  Default: 1.
#' @param seed Integer or \code{NULL}.
#'   Random seed for reproducible coancestry sampling.
#'   Default: \code{NULL}.
#'
#' @return A list of class \code{pedhalflife} with two \code{data.table}
#'   components:
#'   \describe{
#'     \item{\code{timeseries}}{Per-time-point tracking with columns
#'       \code{Time} (time-point label from \code{timevar}), \code{NRef},
#'       \code{fe}, \code{fa}, \code{fg} and their log transformations
#'       (\code{lnfe}, \code{lnfa}, \code{lnfg}, \code{lnfafe},
#'       \code{lnfgfa}), plus \code{TimeStep} (numeric OLS time axis).}
#'     \item{\code{decay}}{Single-row table with \code{lambda_e},
#'       \code{lambda_b}, \code{lambda_d}, \code{lambda_total}, and
#'       \code{thalf}.}
#'   }
#'
#' @details
#' The mathematical identity underlying the cascade is:
#' \deqn{\ln f_g = \ln f_e + \ln(f_a / f_e) + \ln(f_g / f_a)}
#' Taking the negative time-slope of each term gives the \eqn{\lambda}
#' components, which sum exactly by linearity of OLS:
#' \deqn{\lambda_{total} = \lambda_e + \lambda_b + \lambda_d}
#'
#' \eqn{T_{1/2} = \ln 2 / \lambda_{total}} is the number of time-units
#' (time points, years, generations) for diversity to halve.
#'
#' @examples
#' \donttest{
#' library(visPedigree)
#' data(simple_ped)
#' tp <- tidyped(simple_ped)
#'
#' # 1. Calculate half-life using all available generations
#' hl <- pedhalflife(tp, timevar = "Gen")
#' print(hl)
#'
#' # 2. View the underlying log-linear decay plot
#' plot(hl, type = "log")
#'
#' # 3. Calculate half-life for a specific time window (e.g., Generations 2 to 4)
#' hl_subset <- pedhalflife(tp, timevar = "Gen", at = c(2, 3, 4))
#' print(hl_subset)
#' }
#'
#' @seealso \code{\link{pediv}}, \code{\link{pedcontrib}}, \code{\link{tidyped}}
#' @export
pedhalflife <- function(ped, timevar = "Gen", at = NULL, nsamples = 1000,
                        ncores = 1, seed = NULL) {
  ped <- ensure_complete_tidyped(ped, "pedhalflife()")

  if (!timevar %in% names(ped)) {
    stop(sprintf("Column '%s' not found in pedigree.", timevar))
  }

  # ---- Determine time points ----
  if (is.null(at)) {
    at <- sort(unique(ped[[timevar]]))
    at <- at[!is.na(at)]
  } else {
    at <- sort(unique(at))
  }

  if (length(at) < 2) {
    stop("At least two distinct time points are required to calculate half-life rates.")
  }

  ped_dt <- data.table::copy(ped)

  # Strip tidyped class from the working copy to avoid strict subsetting warnings
  data.table::setattr(ped_dt, "class", c("data.table", "data.frame"))

  # Configure OpenMP threads once (used by C++ coancestry backend)
  if (ncores > 1 && cpp_openmp_available()) {
    cpp_set_num_threads(ncores)
  }

  by_all <- ".CohortAll"
  ped_dt[[by_all]] <- "All"

  ts_list <- vector("list", length(at))

  message(sprintf("Calculating diversity across %d time points...", length(at)))
  for (i in seq_along(at)) {
    c_val <- at[i]
    ref_ids <- ped_dt[get(timevar) == c_val, Ind]

    if (length(ref_ids) == 0) next

    # ---- fe / fa via pedcontrib (public API, C++ backend) ----
    contrib <- suppressMessages(
      pedcontrib(ped, reference = ref_ids, mode = "both", top = 20)
    )
    s <- contrib$summary

    # ---- fg via internal coancestry engine (avoids redundant Ne calcs) ----
    fg_val <- NA_real_
    tryCatch({
      ped_subset <- as.data.table(ped_dt)[Ind %in% ref_ids]
      raw <- calc_ne_coancestry(
        ped_subset, ped_dt, by_all, nsamples,
        seed = seed, metrics = "fg"
      )
      if (!is.null(raw) && nrow(raw) > 0 && "fg" %in% names(raw)) {
        fg_val <- raw$fg[1]
      }
    }, error = function(e) {
      warning(sprintf("fg calculation failed for time point %s: %s",
                      as.character(c_val), e$message))
    })

    ts_list[[i]] <- data.table::data.table(
      Time   = c_val,
      NRef   = s$n_ref,
      fe     = s$f_e,
      fa     = s$f_a,
      fg     = fg_val
    )
  }

  ts_dt <- data.table::rbindlist(ts_list)

  if (nrow(ts_dt) < 2) {
    stop("Not enough valid time points to compute decay rates.")
  }

  # ---- Guard: require positive finite fe/fa/fg for log ----
  valid <- !is.na(ts_dt$fe) & !is.na(ts_dt$fa) & !is.na(ts_dt$fg) &
    ts_dt$fe > 0 & ts_dt$fa > 0 & ts_dt$fg > 0

  n_dropped <- sum(!valid)
  if (n_dropped > 0) {
    warning(sprintf(
      "%d time point(s) with NA or non-positive fe/fa/fg removed from decay analysis.",
      n_dropped
    ))
  }

  ts_valid <- ts_dt[valid]

  if (nrow(ts_valid) < 2) {
    stop("Not enough time points with valid positive fe/fa/fg to compute decay rates.")
  }

  # ---- Logarithmic metrics ----
  ts_valid[, lnfe := log(fe)]
  ts_valid[, lnfa := log(fa)]
  ts_valid[, lnfg := log(fg)]
  ts_valid[, lnfafe := log(fa / fe)]
  ts_valid[, lnfgfa := log(fg / fa)]

  # ---- Time axis ----
  t_vals <- suppressWarnings(as.numeric(ts_valid$Time))
  if (any(is.na(t_vals))) {
    warning("Time variable could not be parsed as numeric. Using sequential indices.")
    t_vals <- seq_len(nrow(ts_valid))
  }
  ts_valid[, TimeStep := t_vals]

  # ---- Decay rates (negative OLS slope of log vs time) ----
  calc_lambda <- function(y, t) {
    ok <- is.finite(y) & is.finite(t)
    if (sum(ok) < 2) return(NA_real_)
    y <- y[ok]; t <- t[ok]
    if (stats::sd(t) == 0) return(NA_real_)
    fit <- stats::lm(y ~ t)
    -stats::coef(fit)[[2]]
  }

  lambda_e     <- calc_lambda(ts_valid$lnfe,   ts_valid$TimeStep)
  lambda_b     <- calc_lambda(ts_valid$lnfafe, ts_valid$TimeStep)
  lambda_d     <- calc_lambda(ts_valid$lnfgfa, ts_valid$TimeStep)
  lambda_total <- calc_lambda(ts_valid$lnfg,   ts_valid$TimeStep)

  thalf <- if (!is.na(lambda_total) && lambda_total > 0) {
    log(2) / lambda_total
  } else {
    NA_real_
  }

  decay_dt <- data.table::data.table(
    LambdaE     = lambda_e,
    LambdaB     = lambda_b,
    LambdaD     = lambda_d,
    LambdaTotal = lambda_total,
    THalf       = thalf
  )

  result <- list(
    timeseries = ts_valid[],
    decay      = decay_dt[],
    timevar    = timevar
  )

  class(result) <- "pedhalflife"
  return(result)
}


# ---- S3 methods --------------------------------------------------------

#' @rdname pedhalflife
#' @param x A \code{pedhalflife} object.
#' @param ... Additional arguments (ignored).
#' @export
print.pedhalflife <- function(x, ...) {
  d <- x$decay

  fmt <- function(val) {
    if (is.na(val)) "        NA" else sprintf("%10.6f", val)
  }

  cat("Information-Theoretic Diversity Half-Life\n")
  cat("-----------------------------------------\n")
  cat(sprintf("Total Loss Rate (lambda_total): %s\n", fmt(d$LambdaTotal)))
  cat(sprintf("  Foundation  (lambda_e)      : %s\n", fmt(d$LambdaE)))
  cat(sprintf("  Bottleneck  (lambda_b)      : %s\n", fmt(d$LambdaB)))
  cat(sprintf("  Drift       (lambda_d)      : %s\n", fmt(d$LambdaD)))
  cat("-----------------------------------------\n")

  tv <- if (!is.null(x$timevar)) x$timevar else "time steps"
  if (!is.na(d$THalf) && d$THalf > 0) {
    cat(sprintf("Diversity Half-life (T_1/2)   : %10.2f (%s)\n", d$THalf, tv))
  } else {
    cat("Diversity Half-life (T_1/2)   : NA\n")
  }
  cat(sprintf("\nTimeseries: %d time points\n", nrow(x$timeseries)))
  invisible(x)
}


#' @rdname pedhalflife
#' @param type Character.
#'   \code{"log"} for log-transformed values;
#'   \code{"raw"} for \eqn{f_e}, \eqn{f_a}, \eqn{f_g}.
#' @export
plot.pedhalflife <- function(x, type = c("log", "raw"), ...) {
  type <- match.arg(type)

  if (!requireNamespace("lattice", quietly = TRUE)) {
    stop("Package 'lattice' is required for plotting.")
  }

  dt <- x$timeseries

  thalf <- x$decay$THalf
  tv    <- if (!is.null(x$timevar)) x$timevar else "Time"

  if (type == "log") {
    plot_data <- data.frame(
      Time   = rep(dt$TimeStep, 3),
      Value  = c(dt$lnfe, dt$lnfa, dt$lnfg),
      Metric = factor(rep(c("ln(fe)", "ln(fa)", "ln(fg)"), each = nrow(dt)),
                       levels = c("ln(fe)", "ln(fa)", "ln(fg)"))
    )

    # Pre-fit OLS on lnfg for the panel closure
    ols_fit <- stats::lm(lnfg ~ TimeStep, data = dt)

    lattice::xyplot(
      Value ~ Time,
      groups = Metric,
      data   = plot_data,
      type   = "b",
      auto.key = list(columns = 3),
      xlab   = tv,
      ylab   = "Log Diversity",
      main   = "Information-Theoretic Diversity Decay",
      panel  = function(...) {
        lattice::panel.xyplot(...)
        # OLS regression line for ln(fg) total decay
        lattice::panel.abline(ols_fit, lty = 2, col = "black", lwd = 1.5)
        # Half-life vertical reference
        if (is.finite(thalf)) {
          lattice::panel.abline(v = thalf, lty = 3, col = "grey40")
          lattice::panel.text(
            x = thalf, y = max(plot_data$Value, na.rm = TRUE),
            labels = sprintf("T1/2 = %.1f", thalf),
            pos = 4, cex = 0.8, col = "grey30"
          )
        }
      },
      ...
    )

  } else {
    plot_data <- data.frame(
      Time   = rep(dt$TimeStep, 3),
      Value  = c(dt$fe, dt$fa, dt$fg),
      Metric = factor(rep(c("fe", "fa", "fg"), each = nrow(dt)),
                       levels = c("fe", "fa", "fg"))
    )

    lattice::xyplot(
      Value ~ Time,
      groups = Metric,
      data   = plot_data,
      type   = "b",
      auto.key = list(columns = 3),
      xlab   = tv,
      ylab   = "Equivalent Numbers",
      main   = "Diversity Loss over Time",
      ...
    )
  }
}

test_that("pedhalflife returns correct structure", {
  data(simple_ped)
  tped <- suppressMessages(tidyped(simple_ped))

  res <- suppressMessages(
    suppressWarnings(pedhalflife(tped, timevar = "Gen"))
  )

  # Class and top-level components

  expect_s3_class(res, "pedhalflife")
  expect_true("timeseries" %in% names(res))
  expect_true("decay" %in% names(res))

  # Timeseries columns
  ts <- res$timeseries
  expect_s3_class(ts, "data.table")
  expect_true(all(c("Time", "NRef", "fe", "fa", "fg",
                     "lnfe", "lnfa", "lnfg", "lnfafe", "lnfgfa",
                     "TimeStep") %in% names(ts)))
  expect_true(nrow(ts) >= 2)

  # Decay columns
  d <- res$decay
  expect_s3_class(d, "data.table")
  expect_true(all(c("LambdaE", "LambdaB", "LambdaD",
                     "LambdaTotal", "THalf") %in% names(d)))
  expect_equal(nrow(d), 1L)
})


test_that("lambda decomposition sums correctly (OLS linearity)", {
  data(simple_ped)
  tped <- suppressMessages(tidyped(simple_ped))

  res <- suppressMessages(
    suppressWarnings(pedhalflife(tped, timevar = "Gen"))
  )
  d <- res$decay

  # By linearity of least squares, lambda_total == lambda_e + lambda_b + lambda_d
  if (!is.na(d$LambdaTotal)) {
    expect_equal(d$LambdaTotal,
                 d$LambdaE + d$LambdaB + d$LambdaD,
                 tolerance = 1e-10)
  }

  # THalf identity
  if (!is.na(d$THalf) && d$LambdaTotal > 0) {
    expect_equal(d$THalf, log(2) / d$LambdaTotal, tolerance = 1e-10)
  }
})


test_that("log columns are consistent with raw values", {
  data(simple_ped)
  tped <- suppressMessages(tidyped(simple_ped))

  res <- suppressMessages(
    suppressWarnings(pedhalflife(tped, timevar = "Gen"))
  )
  ts <- res$timeseries

  expect_equal(ts$lnfe, log(ts$fe), tolerance = 1e-12)
  expect_equal(ts$lnfa, log(ts$fa), tolerance = 1e-12)
  expect_equal(ts$lnfg, log(ts$fg), tolerance = 1e-12)
  expect_equal(ts$lnfafe, log(ts$fa / ts$fe), tolerance = 1e-12)
  expect_equal(ts$lnfgfa, log(ts$fg / ts$fa), tolerance = 1e-12)
})


test_that("pedhalflife rejects invalid inputs", {
  data(simple_ped)
  tped <- suppressMessages(tidyped(simple_ped))

  # Missing column
  expect_error(pedhalflife(tped, timevar = "NoSuchCol"),
               "Column 'NoSuchCol' not found")

  # Single time point
  tped2 <- data.table::copy(tped)
  tped2$Year <- 2000
  expect_error(suppressMessages(pedhalflife(tped2, timevar = "Year")),
               "At least two distinct")
})


test_that("print and plot methods work without error", {
  data(simple_ped)
  tped <- suppressMessages(tidyped(simple_ped))

  res <- suppressMessages(
    suppressWarnings(pedhalflife(tped, timevar = "Gen"))
  )

  # print returns invisible x
  out <- capture.output(ret <- print(res))
  expect_identical(ret, res)
  expect_true(length(out) > 0)

  # plot returns a lattice trellis object
  p <- plot(res, type = "log")
  expect_s3_class(p, "trellis")

  p2 <- plot(res, type = "raw")
  expect_s3_class(p2, "trellis")
})

#' Split Pedigree into Disconnected Groups
#'
#' @description
#' Detects and splits a tidyped object into disconnected groups (connected components).
#' Uses igraph to efficiently find groups of individuals that have no genetic
#' relationships with each other. Isolated individuals (Gen = 0, those with no
#' parents and no offspring) are excluded from group splitting and stored separately.
#'
#' @param ped A tidyped object created by \code{\link{tidyped}}.
#'
#' @return A list of class "splitped" containing:
#'   \item{GP1, GP2, ...}{tidyped objects for each disconnected group (with at least
#'     2 individuals), with renumbered IndNum, SireNum, DamNum}
#'
#'   The returned object has the following attributes:
#'   \item{n_groups}{Number of disconnected groups found (excluding isolated individuals)}
#'   \item{sizes}{Named vector of group sizes}
#'   \item{total}{Total number of individuals in groups (excluding isolated)}
#'   \item{isolated}{Character vector of isolated individual IDs (Gen = 0)}
#'   \item{n_isolated}{Number of isolated individuals}
#'
#' @details
#' This function identifies connected components in the pedigree graph where
#' edges represent parent-offspring relationships. Two individuals are in the
#' same group if they share any ancestry (direct or indirect).
#'
#' Isolated individuals (Gen = 0 in tidyped output) are those who:
#' \itemize{
#'   \item Have no known parents (Sire and Dam are both NA)
#'   \item Are not parents of any other individual in the pedigree
#' }
#' These isolated individuals are excluded from splitting and stored in the
#' \code{isolated} attribute. Each resulting group contains at least 2 individuals
#' (at least one parent-offspring relationship).
#'
#' The function always returns a list, even if there is only one group (i.e.,
#' the pedigree is fully connected). Groups are sorted by size in descending
#' order.
#'
#' Each group in the result is a valid tidyped object with:
#' \itemize{
#'   \item Renumbered IndNum (1 to n for each group)
#'   \item Updated SireNum and DamNum referencing the new IndNum
#'   \item Recalculated Gen (generation) based on the group's structure
#' }
#'
#' @examples
#' # Load example data
#' library(visPedigree)
#' data(small_ped)
#'
#' # First tidy the pedigree
#' tped <- tidyped(small_ped)
#'
#' # Split into groups
#' result <- splitped(tped)
#' print(result)
#'
#' # Access individual groups (each is a tidyped object)
#' result$GP1
#'
#' # Check isolated individuals
#' attr(result, "isolated")
#'
#' @seealso \code{\link{tidyped}} for pedigree tidying
#'
#' @export
splitped <- function(ped) {
  # Input validation: ensure tidyped class (auto-recover if possible)
  ped <- ensure_tidyped(ped)

  # Copy to avoid modifying original
  ped <- data.table::copy(ped)

  # Separate isolated individuals (Gen = 0)
  # These have no parents AND are not parents of anyone
  isolated_idx <- ped$Gen == 0L
  isolated_ids <- ped$Ind[isolated_idx]
  n_isolated <- sum(isolated_idx)

  # Filter to non-isolated individuals only
  ped_valid <- ped[!isolated_idx, ]
  n <- nrow(ped_valid)

  # Handle edge case: no valid individuals (all isolated)
  if (n == 0) {
    result <- list()
    class(result) <- c("splitped", "list")
    attr(result, "n_groups") <- 0L
    attr(result, "sizes") <- integer(0)
    attr(result, "total") <- 0L
    attr(result, "isolated") <- isolated_ids
    attr(result, "n_isolated") <- n_isolated
    return(result)
  }

  # Handle edge case: single non-isolated individual
  # (This shouldn't happen in practice since a valid pedigree needs at least 2)
  if (n == 1) {
    group_ped <- retidy_subset(ped_valid)
    result <- list(GP1 = group_ped)
    class(result) <- c("splitped", "list")
    attr(result, "n_groups") <- 1L
    attr(result, "sizes") <- c(GP1 = 1L)
    attr(result, "total") <- 1L
    attr(result, "isolated") <- isolated_ids
    attr(result, "n_isolated") <- n_isolated
    return(result)
  }

  # Find connected components using numeric indices
  group_ids <- find_components(ped_valid)

  # Build result from group assignments
  build_splitped_result(ped_valid, group_ids, n, isolated_ids, n_isolated)
}

# Find connected components using numeric indices
find_components <- function(ped) {
  n <- nrow(ped)
  ind_nums <- ped$IndNum
  sire_nums <- ped$SireNum
  dam_nums <- ped$DamNum

  # Build edges using numeric indices directly
  # In tidyped, SireNum=0 and DamNum=0 means unknown parent
  # All parents are guaranteed to be in the pedigree

  edges_from <- integer(0)
  edges_to <- integer(0)

  # Add sire-offspring edges
  valid_sire <- sire_nums > 0
  if (any(valid_sire)) {
    edges_from <- c(edges_from, ind_nums[valid_sire])
    edges_to <- c(edges_to, sire_nums[valid_sire])
  }

  # Add dam-offspring edges
  valid_dam <- dam_nums > 0
  if (any(valid_dam)) {
    edges_from <- c(edges_from, ind_nums[valid_dam])
    edges_to <- c(edges_to, dam_nums[valid_dam])
  }

  # Create graph and find components
  if (length(edges_from) == 0) {
    # No edges: each individual is its own group
    return(seq_len(n))
  }

  # Build edge matrix
  edge_matrix <- cbind(edges_from, edges_to)

  # Create graph with numeric vertices
  g <- igraph::graph_from_edgelist(edge_matrix, directed = FALSE)

  # Add isolated vertices if any
  max_vertex <- max(ind_nums)
  if (igraph::vcount(g) < max_vertex) {
    g <- igraph::add_vertices(g, max_vertex - igraph::vcount(g))
  }

  # Find connected components
  components <- igraph::components(g)

  # Map back to original order
  components$membership[ind_nums]
}

# Re-tidy a subset to be a valid standalone tidyped
# Uses tidyped() to properly recalculate Gen, IndNum, SireNum, DamNum
retidy_subset <- function(ped) {
  if (nrow(ped) == 0) return(ped)

  # Extract only the basic pedigree columns (Ind, Sire, Dam, Sex)
  # and call tidyped() to properly re-process the subset
  basic_ped <- data.table::data.table(
    Ind = ped$Ind,
    Sire = ped$Sire,
    Dam = ped$Dam
  )

  # Add Sex column if present
  if ("Sex" %in% names(ped)) {
    basic_ped$Sex <- ped$Sex
  }

  # Check if all parents are missing (founders only group)
  # tidyped() does not accept pedigrees where all parents are NA
  all_founders <- all(is.na(basic_ped$Sire) & is.na(basic_ped$Dam))

  if (all_founders) {
    # Build a minimal tidyped manually for founders-only groups
    n <- nrow(basic_ped)
    result <- data.table::data.table(
      Ind = basic_ped$Ind,
      Sire = rep(NA_character_, n),
      Dam = rep(NA_character_, n),
      Gen = rep(1L, n),
      Sex = if ("Sex" %in% names(basic_ped)) basic_ped$Sex else rep(NA_character_, n),
      IndNum = seq_len(n),
      SireNum = rep(0L, n),
      DamNum = rep(0L, n)
    )
    result <- new_tidyped(result)
    meta <- attr(ped, "ped_meta")
    if (!is.null(meta)) data.table::setattr(result, "ped_meta", meta)
    return(result)
  }

  # Propagate selfing attribute if present in the original pedigree
  meta <- attr(ped, "ped_meta")
  selfing_val <- if (!is.null(meta)) isTRUE(meta$selfing) else FALSE
  
  # Call tidyped to properly recalculate Gen, IndNum, SireNum, DamNum
  tidyped(basic_ped, selfing = selfing_val)
}

# Build result from group membership vector
build_splitped_result <- function(ped, group_ids, n, isolated_ids, n_isolated) {
  # Get unique groups and their sizes
  group_table <- table(group_ids)
  group_order <- order(group_table, decreasing = TRUE)
  unique_groups <- as.integer(names(group_table)[group_order])

  n_groups <- length(unique_groups)

  # Create result list
  result <- vector("list", n_groups)
  sizes <- integer(n_groups)

  for (i in seq_len(n_groups)) {
    grp <- unique_groups[i]
    idx <- which(group_ids == grp)
    group_ped <- ped[idx, ]

    # Re-tidy the subset using tidyped() to properly recalculate Gen, IndNum, etc.
    group_ped <- retidy_subset(group_ped)

    result[[i]] <- group_ped
    sizes[i] <- length(idx)
  }

  # Name the groups
  names(result) <- paste0("GP", seq_len(n_groups))
  names(sizes) <- names(result)

  # Set class and attributes
  class(result) <- c("splitped", "list")
  attr(result, "n_groups") <- n_groups
  attr(result, "sizes") <- sizes
  attr(result, "total") <- n
  attr(result, "isolated") <- isolated_ids
  attr(result, "n_isolated") <- n_isolated

  result
}

#' @export
print.splitped <- function(x, ...) {
  n_groups <- attr(x, "n_groups")
  sizes <- attr(x, "sizes")
  total <- attr(x, "total")
  n_isolated <- attr(x, "n_isolated")

  cat("Pedigree Split Result\n")
  cat("======================\n")
  cat("Total individuals in groups:", total, "\n")
  cat("Isolated individuals (Gen=0):", n_isolated, "\n")
  cat("Number of groups: ", n_groups, "\n")

  if (n_groups > 0) {
    cat("\nGroup sizes:\n")
    for (i in seq_len(min(n_groups, 10))) {
      cat(sprintf("  %s: %d individuals (%.1f%%)\n",
                  names(sizes)[i], sizes[i], 100 * sizes[i] / total))
    }
    if (n_groups > 10) {
      cat(sprintf("  ... and %d more groups\n", n_groups - 10))
    }
  }

  invisible(x)
}

#' @export
summary.splitped <- function(object, ...) {
  n_groups <- attr(object, "n_groups")
  sizes <- attr(object, "sizes")
  total <- attr(object, "total")
  n_isolated <- attr(object, "n_isolated")

  cat("Summary of Pedigree Split\n")
  cat("=========================\n")
  cat("Total individuals in groups:", total, "\n")
  cat("Isolated individuals (Gen=0):", n_isolated, "\n")
  cat("Grand total:", total + n_isolated, "\n")
  cat("Number of groups: ", n_groups, "\n")

  if (n_groups > 0) {
    cat("\nSize statistics:\n")
    cat("  Min:    ", min(sizes), "\n")
    cat("  Max:    ", max(sizes), "\n")
    cat("  Mean:   ", round(mean(sizes), 1), "\n")
    cat("  Median: ", median(sizes), "\n")

    if (n_groups > 1) {
      cat("\nConnectivity: Pedigree contains disconnected groups\n")
    } else {
      cat("\nConnectivity: Pedigree is fully connected\n")
    }
  }

  invisible(object)
}

# Tests for selfing support (Issue #10: plant breeding monoecious pedigrees)

test_that("selfing parameter validation", {
  skip_if_not_installed("visPedigree")
  
  ped <- data.frame(
    Ind = c("A", "B", "C"),
    Sire = c(NA, NA, "A"),
    Dam = c(NA, NA, "B")
  )
  
  # selfing must be logical

  expect_error(tidyped(ped, selfing = 1), "only is assigned using TRUE or FALSE")
  expect_error(tidyped(ped, selfing = "yes"), "only is assigned using TRUE or FALSE")
  
  # selfing = FALSE is default and works
  res <- tidyped(ped, selfing = FALSE)
  expect_s3_class(res, "tidyped")
})

test_that("selfing = TRUE allows same individual as both Sire and Dam", {
  skip_if_not_installed("visPedigree")
  
  # Simple selfing: A selfs to produce C
  ped <- data.frame(
    Ind = c("A", "B", "C"),
    Sire = c(NA, NA, "A"),
    Dam = c(NA, NA, "A"),  # A is both Sire and Dam
    stringsAsFactors = FALSE
  )
  
  # Should error without selfing
  expect_error(tidyped(ped), "Sex conflict detected")
  
  # Should succeed with selfing = TRUE
  expect_message(tidyped(ped, selfing = TRUE), "Selfing mode")
  res <- suppressMessages(tidyped(ped, selfing = TRUE))
  
  expect_s3_class(res, "tidyped")
  expect_equal(res[Ind == "A", Sex], "monoecious")
  expect_true(isTRUE(attr(res, "ped_meta")$selfing))
  expect_equal(attr(res, "ped_meta")$bisexual_parents, "A")
})

test_that("selfing with multiple monoecious individuals", {
  skip_if_not_installed("visPedigree")
  
  # Plant pedigree: P1 and P2 are both monoecious
  ped <- data.frame(
    Ind = c("P1", "P2", "C1", "C2", "C3"),
    Sire = c(NA, NA, "P1", "P2", "P1"),
    Dam = c(NA, NA, "P2", "P1", "P1"),  # P1 and P2 both appear as Sire and Dam
    stringsAsFactors = FALSE
  )
  
  res <- suppressMessages(tidyped(ped, selfing = TRUE))
  
  expect_equal(res[Ind == "P1", Sex], "monoecious")
  expect_equal(res[Ind == "P2", Sex], "monoecious")
  expect_equal(sort(attr(res, "ped_meta")$bisexual_parents), c("P1", "P2"))
})

test_that("selfing: non-monoecious individuals get normal sex inference", {
  skip_if_not_installed("visPedigree")
  
  # Mixed pedigree: P1 is monoecious, M1 is only a Sire, F1 is only a Dam
  ped <- data.frame(
    Ind = c("P1", "M1", "F1", "C1", "C2", "C3"),
    Sire = c(NA, NA, NA, "P1", "M1", "P1"),
    Dam = c(NA, NA, NA, "P1", "F1", "F1"),
    stringsAsFactors = FALSE
  )
  
  res <- suppressMessages(tidyped(ped, selfing = TRUE))
  
  expect_equal(res[Ind == "P1", Sex], "monoecious")
  expect_equal(res[Ind == "M1", Sex], "male")
  expect_equal(res[Ind == "F1", Sex], "female")
})

test_that("selfing: explicit sex annotation conflict for monoecious", {
  skip_if_not_installed("visPedigree")
  
  # A is both Sire and Dam but explicitly annotated as "male" -> should error
  ped <- data.frame(
    Ind = c("A", "B", "C"),
    Sire = c(NA, NA, "A"),
    Dam = c(NA, NA, "A"),
    Sex = c("male", NA, NA),
    stringsAsFactors = FALSE
  )
  
  expect_error(
    tidyped(ped, selfing = TRUE),
    "Sex annotation conflicts for monoecious"
  )
})

test_that("selfing: explicit monoecious annotation accepted", {
  skip_if_not_installed("visPedigree")
  
  # A is both Sire and Dam, explicitly annotated as "monoecious" -> should work
  ped <- data.frame(
    Ind = c("A", "B", "C"),
    Sire = c(NA, NA, "A"),
    Dam = c(NA, NA, "A"),
    Sex = c("monoecious", NA, NA),
    stringsAsFactors = FALSE
  )
  
  res <- suppressMessages(tidyped(ped, selfing = TRUE))
  expect_equal(res[Ind == "A", Sex], "monoecious")
})

test_that("selfing: summary statistics count monoecious correctly", {
  skip_if_not_installed("visPedigree")
  
  ped <- data.frame(
    Ind = c("P1", "P2", "M1", "C1", "C2"),
    Sire = c(NA, NA, NA, "P1", "M1"),
    Dam = c(NA, NA, NA, "P1", "P2"),
    stringsAsFactors = FALSE
  )
  
  res <- suppressMessages(tidyped(ped, selfing = TRUE))
  s <- summary(res)
  
  expect_equal(s$n_monoecious, 1)  # P1
  expect_equal(s$n_male, 1)        # M1
  expect_equal(s$n_female, 1)      # P2 (only Dam)
  
  # n_parents should deduplicate P1 (which is both sire and dam)
  # Sires: P1, M1 -> 2 unique sires
  # Dams: P1, P2 -> 2 unique dams
  # Parents (union): P1, M1, P2 -> 3 unique parents
  expect_equal(s$n_parents, 3)
})

test_that("selfing: true self-fertilization (Sire == Dam for offspring)", {
  skip_if_not_installed("visPedigree")
  
  # Self-fertilization: C1 has same Sire and Dam (A selfs)
  ped <- data.frame(
    Ind = c("A", "C1", "C2"),
    Sire = c(NA, "A", "A"),
    Dam = c(NA, "A", "A"),
    stringsAsFactors = FALSE
  )
  
  res <- suppressMessages(tidyped(ped, selfing = TRUE))
  
  expect_equal(res[Ind == "A", Sex], "monoecious")
  expect_equal(res[Ind == "C1", Sire], "A")
  expect_equal(res[Ind == "C1", Dam], "A")
  
  # Generation check: A is founder (Gen 1), C1/C2 are offspring (Gen 2)
  expect_equal(res[Ind == "A", Gen], 1L)
  expect_equal(res[Ind == "C1", Gen], 2L)
})

test_that("selfing: inbreeding coefficient computation works", {
  skip_if_not_installed("visPedigree")
  
  # Self-fertilization should produce inbred offspring (f = 0.5)
  ped <- data.frame(
    Ind = c("A", "C1"),
    Sire = c(NA, "A"),
    Dam = c(NA, "A"),
    stringsAsFactors = FALSE
  )
  
  res <- suppressMessages(tidyped(ped, selfing = TRUE, inbreed = TRUE))
  
  expect_equal(res[Ind == "A", f], 0, tolerance = 1e-10)
  expect_equal(res[Ind == "C1", f], 0.5, tolerance = 1e-10)
})

test_that("selfing: multi-generation plant pedigree", {
  skip_if_not_installed("visPedigree")
  
  # Simulate a realistic plant breeding scenario:
  # P1 x P2 -> F1
  # F1 selfs -> F2_1, F2_2
  # F2_1 selfs -> F3
  ped <- data.frame(
    Ind =  c("P1", "P2", "F1", "F2_1", "F2_2", "F3"),
    Sire = c(NA,   NA,   "P1", "F1",   "F1",   "F2_1"),
    Dam =  c(NA,   NA,   "P2", "F1",   "F1",   "F2_1"),
    stringsAsFactors = FALSE
  )
  
  res <- suppressMessages(tidyped(ped, selfing = TRUE, inbreed = TRUE))
  
  # F1 is monoecious (appears as both Sire and Dam)
  expect_equal(res[Ind == "F1", Sex], "monoecious")
  # F2_1 is monoecious (appears as both Sire and Dam for F3)
  expect_equal(res[Ind == "F2_1", Sex], "monoecious")
  # P1 is only sire -> male
  expect_equal(res[Ind == "P1", Sex], "male")
  # P2 is only dam -> female
  expect_equal(res[Ind == "P2", Sex], "female")
  
  # Generation structure
  expect_equal(res[Ind == "P1", Gen], 1L)
  expect_equal(res[Ind == "F1", Gen], 2L)
  expect_equal(res[Ind == "F2_1", Gen], 3L)
  expect_equal(res[Ind == "F3", Gen], 4L)
  
  # Inbreeding: F2_1 and F2_2 from selfing F1, f = 0.5
  expect_equal(res[Ind == "F2_1", f], 0.5, tolerance = 1e-10)
  # F3 from selfing F2_1 (which itself has f=0.5), f = 0.75
  expect_equal(res[Ind == "F3", f], 0.75, tolerance = 1e-10)
})

test_that("selfing: splitped propagates selfing attribute", {
  skip_if_not_installed("visPedigree")
  
  # Two disconnected groups, one with selfing
  ped <- data.frame(
    Ind =  c("A", "C1", "X", "Y", "Z"),
    Sire = c(NA,  "A",  NA,  "X", NA),
    Dam =  c(NA,  "A",  NA,  NA,  NA),
    stringsAsFactors = FALSE
  )
  
  res <- suppressMessages(tidyped(ped, selfing = TRUE))
  expect_true(isTRUE(attr(res, "ped_meta")$selfing))
  
  # splitped should propagate selfing
  groups <- splitped(res)
  
  # Each group should be a valid tidyped
  for (gp in groups) {
    expect_s3_class(gp, "tidyped")
  }
})

test_that("selfing: selfing hint in error message when selfing = FALSE", {
  skip_if_not_installed("visPedigree")
  
  ped <- data.frame(
    Ind = c("A", "C"),
    Sire = c(NA, "A"),
    Dam = c(NA, "A"),
    stringsAsFactors = FALSE
  )
  
  expect_error(
    tidyped(ped, selfing = FALSE),
    "selfing = TRUE"
  )
})

test_that("splitped works correctly on connected pedigree", {
  ped <- data.frame(
    Ind = c("A", "B", "C", "D"),
    Sire = c(NA, NA, "A", "A"),
    Dam = c(NA, NA, "B", "B")
  )
  tped <- tidyped(ped)
  
  res <- splitped(tped)
  
  expect_s3_class(res, "splitped")
  expect_equal(attr(res, "n_groups"), 1)
  expect_equal(attr(res, "total"), 4)
  expect_equal(attr(res, "n_isolated"), 0)
  expect_equal(length(res), 1)
  expect_s3_class(res$GP1, "tidyped")
  expect_equal(nrow(res$GP1), 4)
})

test_that("splitped works correctly on disconnected pedigree", {
  ped <- data.frame(
    Ind = c("A", "B", "C", "D", "E", "F", "G"),
    Sire = c(NA, NA, "A", "A", NA, NA, "E"),
    Dam = c(NA, NA, "B", "B", NA, NA, "F")
  )
  tped <- tidyped(ped)
  
  res <- splitped(tped)
  
  expect_s3_class(res, "splitped")
  expect_equal(attr(res, "n_groups"), 2)
  expect_equal(attr(res, "total"), 7)
  expect_equal(attr(res, "n_isolated"), 0)
  expect_equal(length(res), 2)
  
  # Groups are sorted by size descending
  expect_equal(nrow(res$GP1), 4)
  expect_equal(nrow(res$GP2), 3)
  
  expect_true(all(res$GP1$Ind %in% c("A", "B", "C", "D")))
  expect_true(all(res$GP2$Ind %in% c("E", "F", "G")))
})

test_that("splitped handles isolated individuals", {
  ped <- data.frame(
    Ind = c("A", "B", "C", "D", "E", "F"),
    Sire = c(NA, NA, "A", NA, NA, NA),
    Dam = c(NA, NA, "B", NA, NA, NA)
  )
  tped <- tidyped(ped)
  
  res <- splitped(tped)
  
  expect_s3_class(res, "splitped")
  expect_equal(attr(res, "n_groups"), 1)
  expect_equal(attr(res, "total"), 3)
  expect_equal(attr(res, "n_isolated"), 3)
  expect_equal(sort(attr(res, "isolated")), c("D", "E", "F"))
  expect_equal(nrow(res$GP1), 3)
})

test_that("splitped handles all isolated individuals", {
  # Cannot construct an empty tidyped with tidyped() because "All parents are missing"
  # So we manually construct a mimic tidyped object for testing splitped() logic
  
  ped <- data.table::data.table(
    Ind = c("A", "B", "C"),
    Sire = rep(NA_character_, 3),
    Dam = rep(NA_character_, 3),
    Sex = rep(NA_character_, 3),
    Gen = rep(0L, 3), # Isolated individuals have Gen=0
    IndNum = 1:3,
    SireNum = rep(0L, 3),
    DamNum = rep(0L, 3)
  )
  class(ped) <- c("tidyped", "data.table", "data.frame")
  
  res <- splitped(ped)
  
  expect_s3_class(res, "splitped")
  expect_equal(attr(res, "n_groups"), 0)
  expect_equal(attr(res, "total"), 0)
  expect_equal(attr(res, "n_isolated"), 3)
  expect_equal(length(res), 0)
})

test_that("splitped input validation", {
  expect_error(splitped(data.frame(Ind="A", Sire=NA, Dam=NA)), "tidyped object")
})

test_that("print and summary methods for splitped", {
  ped <- data.frame(
    Ind = c("A", "B", "C", "D", "E", "F", "G", "H"),
    Sire = c(NA, NA, "A", "A", NA, NA, "E", NA),
    Dam = c(NA, NA, "B", "B", NA, NA, "F", NA)
  )
  tped <- tidyped(ped)
  res <- splitped(tped)
  
  expect_output(print(res), "Pedigree Split Result")
  expect_output(print(res), "Total individuals in groups: 7")
  expect_output(print(res), "Isolated individuals \\(Gen=0\\): 1")
  expect_output(print(res), "Number of groups:  2")
  
  expect_output(summary(res), "Summary of Pedigree Split")
  expect_output(summary(res), "Grand total: 8")
  expect_output(summary(res), "Connectivity: Pedigree contains disconnected groups")
})

#' Tidy and prepare a pedigree
#'
#' This function standardizes pedigree records, checks for duplicate IDs and
#' incompatible parental roles, detects pedigree loops, injects missing
#' founders, assigns generation numbers, sorts the pedigree, and optionally
#' traces the pedigree of specified candidates. If the \code{cand} parameter
#' contains individual IDs, only those individuals and their ancestors or
#' descendants are retained. Tracing direction and the number of generations
#' can be specified using the \code{trace} and \code{tracegen} parameters.
#'
#' Compared to the legacy version, this function reports cyclic pedigrees more
#' clearly and uses a mixed implementation. There are two candidate-tracing
#' paths: when the input is a raw pedigree, \code{igraph} is used for loop
#' checking, candidate tracing, and topological sorting; when the input is an
#' already validated \code{tidyped} object and \code{cand} is supplied,
#' tracing and topological sorting use integer-indexed C++ routines. Generation
#' assignment can be performed using either a top-down approach (default,
#' aligning founders at the top) or a bottom-up approach (aligning terminal
#' nodes at the bottom).
#'
#' @param ped A data.table or data frame containing the pedigree. The first three columns must be \strong{individual}, \strong{sire}, and \strong{dam} IDs. Additional columns, such as sex or generation, can be included. Column names can be customized, but their order must remain unchanged. Individual IDs should not be coded as "", " ", "0", "*", or "NA"; otherwise, they will be removed. Missing parents should be denoted by "NA", "0", or "*". Spaces and empty strings ("") are also treated as missing parents but are not recommended.
#' @param cand A character vector of individual IDs, or NULL. If provided, only the candidates and their ancestors/descendants are retained.
#' @param trace A character value specifying the tracing direction: "\strong{up}", "\strong{down}", or "\strong{all}". "up" traces ancestors; "down" traces descendants; "all" traces the union of ancestors and descendants. This parameter is only used if \code{cand} is not NULL. Default is "up".
#' @param tracegen An integer specifying the number of generations to trace. This parameter is only used if \code{trace} is not NULL. If NULL or 0, all available generations are traced.
#' @param addgen A logical value indicating whether to generate generation numbers. Default is TRUE, which adds a \strong{Gen} column to the output.
#' @param addnum A logical value indicating whether to generate a numeric pedigree. Default is TRUE, which adds \strong{IndNum}, \strong{SireNum}, and \strong{DamNum} columns to the output.
#' @param inbreed A logical value indicating whether to calculate inbreeding coefficients. Default is FALSE. If TRUE, an \strong{f} column is added to the output. This uses the same optimized engine as \code{pedmat(..., method = "f")}.
#' @param selfing A logical value indicating whether to allow the same individual to appear as both sire and dam. This is common in plant breeding (monoecious species) where the same plant can serve as both male and female parent. If TRUE, individuals appearing in both the Sire and Dam columns will be assigned Sex = "monoecious" instead of triggering an error. Default is FALSE.
#' @param genmethod A character value specifying the generation assignment method: "\strong{top}" or "\strong{bottom}". "top" (top-aligned) assigns generations from parents to offspring, starting founders at Gen 1. "bottom" (bottom-aligned) assigns generations from offspring to parents, aligning terminal nodes at the bottom. Default is "top".
#' @param ... Additional arguments passed to \code{\link{inbreed}}.
#' 
#' @return A \code{tidyped} object (which inherits from \code{data.table}). Individual, sire, and dam ID columns are renamed to \strong{Ind}, \strong{Sire}, and \strong{Dam}. Missing parents are replaced with \strong{NA}. The \strong{Sex} column contains \code{"male"}, \code{"female"}, \code{"monoecious"}, or \code{NA}. The \strong{Cand} column is included if \code{cand} is not NULL. The \strong{Gen} column is included if \code{addgen} is TRUE. The \strong{IndNum}, \strong{SireNum}, and \strong{DamNum} columns are included if \code{addnum} is TRUE. The \strong{Family} and \strong{FamilySize} columns identify full-sibling families (for example, \code{"AxB"} for offspring of sire \code{A} and dam \code{B}). The \strong{f} column is included if \code{inbreed} is TRUE.
#' 
#' @seealso 
#' \code{\link{summary.tidyped}} for summarizing tidyped objects
#' \code{\link{visped}} for visualizing pedigree structure
#' \code{\link{pedmat}} for computing relationship matrices
#' \code{\link{vismat}} for visualizing relationship matrices
#' \code{\link{splitped}} for splitting pedigree into connected components
#' \code{\link{inbreed}} for calculating inbreeding coefficients
#'
#' @examples
#' library(visPedigree)
#' library(data.table)
#'
#' # Tidy a simple pedigree
#' tidy_ped <- tidyped(simple_ped)
#' head(tidy_ped)
#'
#' # Trace ancestors of a specific individual within 2 generations
#' tidy_ped_tracegen <- tidyped(simple_ped, cand = "J5X804", trace = "up", tracegen = 2)
#' head(tidy_ped_tracegen)
#' 
#' # Trace both ancestors and descendants for multiple candidates
#' # This is highly optimized and works quickly even on 100k+ individuals
#' cand_list <- c("J5X804", "J3Y620")
#' tidy_ped_all <- tidyped(simple_ped, cand = cand_list, trace = "all")
#' 
#' # Check for loops (will error if loops exist)
#' try(tidyped(loop_ped))
#' 
#' # Example with a large pedigree: extract 2 generations of ancestors for 2007 candidates
#' cand_2007 <- big_family_size_ped[Year == 2007, Ind]
#' \donttest{
#' tidy_big <- tidyped(big_family_size_ped, cand = cand_2007, trace = "up", tracegen = 2)
#' summary(tidy_big)
#' }
#'
#' @import data.table
#' @importFrom igraph graph_from_data_frame is_dag topo_sort components subcomponent distances ego degree as_adj_list vcount add_vertices neighbors shortest_paths which_loop ends graph_from_edgelist neighborhood
#' @importFrom stats setNames
#' @importFrom utils head
#' @export
tidyped <- function(ped,
                          cand = NULL,
                          trace = "up",
                          tracegen = NULL,
                          addgen = TRUE,
                          addnum = TRUE,
                          inbreed = FALSE,
                          selfing = FALSE,
                          genmethod = "top",
                          ...) {
  # 1. Parameter Validation
  if (missing(ped) || is.null(ped)) {
    stop("The ped parameter cannot be NULL or missing")
  }
  if ((is.data.frame(ped) || is.matrix(ped)) && NROW(ped) == 0) {
    stop("The ped parameter cannot be empty")
  }
  
  # Boolean checks
  for (arg in c("addgen", "addnum", "inbreed", "selfing")) {
    val <- get(arg)
    if (!is.logical(val) || length(val) != 1 || is.na(val)) {
      stop(sprintf("The %s parameter only is assigned using TRUE or FALSE", arg))
    }
  }

  valid_trace <- c("up", "down", "all")
  if (length(trace) != 1 || !trace %in% valid_trace) {
    stop(sprintf("The trace parameter must be one of: %s", paste(valid_trace, collapse = ", ")))
  }

  valid_genmethod <- c("top", "bottom")
  if (length(genmethod) != 1 || !genmethod %in% valid_genmethod) {
    stop(sprintf("The genmethod parameter must be one of: %s", paste(valid_genmethod, collapse = ", ")))
  }
  
  if (!is.null(tracegen)) {
    if (!is.numeric(tracegen) || length(tracegen) != 1 || is.na(tracegen) || !is.finite(tracegen)) {
      stop("The tracegen parameter must be a single numeric value")
    }
    tracegen <- as.integer(tracegen)
  }

  # ---------- B2 Fast Path ----------
  # When input is already a tidyped and cand is requested, skip the full
  # validation / cycle-detection / sex-inference pipeline.  The existing
  # object has already passed all of those checks.
  if (is_tidyped(ped) && !is.null(cand)) {
    ped_dt <- data.table::copy(ped)
    meta   <- attr(ped, "ped_meta")
    bisexual_parents <- if (!is.null(meta)) meta$bisexual_parents else character(0)
    selfing_val      <- if (!is.null(meta)) isTRUE(meta$selfing) else selfing
    genmethod_val    <- if (!is.null(meta) && !is.null(meta$genmethod)) meta$genmethod else genmethod

    # Strip class so internal := ops don't hit [.tidyped
    data.table::setattr(ped_dt, "class", setdiff(class(ped_dt), "tidyped"))

    # Ensure numeric parent indices exist for BFS — the input tidyped may have
    # been created with addnum = FALSE, so IndNum/SireNum/DamNum may be absent.
    had_num_cols <- "IndNum" %in% names(ped_dt)
    if (!had_num_cols) {
      ped_dt[, IndNum  := .I]
      ped_dt[, SireNum := match(Sire, Ind, nomatch = 0L)]
      ped_dt[, DamNum  := match(Dam,  Ind, nomatch = 0L)]
    }

    # --- C++ BFS tracing (replaces igraph build + trace_ped_candidates) ---
    cand_clean <- as.character(cand[!is.na(cand) & cand != "" & cand != " "])
    if (length(cand_clean) == 0) stop("The cand parameter contains no valid individual IDs.")
    cand_nums <- ped_dt$IndNum[match(cand_clean, ped_dt$Ind)]
    valid_mask <- !is.na(cand_nums)
    if (!any(valid_mask)) stop("None of the specified candidates were found in the pedigree.")
    if (sum(!valid_mask) > 0) {
      missing_cands <- cand_clean[!valid_mask]
      warning(sprintf("The following %d candidates were not found in the pedigree: %s",
                      length(missing_cands), paste(head(missing_cands, 5), collapse = ", ")))
    }
    cand_nums <- cand_nums[valid_mask]

    tracegen_int <- if (is.null(tracegen) || tracegen <= 0L) 0L else as.integer(tracegen)
    sn_full <- ped_dt$SireNum
    dn_full <- ped_dt$DamNum

    subset_idx <- if (trace == "all") {
      sort(unique(c(
        cpp_trace_ancestors(sn_full, dn_full, cand_nums, tracegen_int),
        cpp_trace_descendants(sn_full, dn_full, cand_nums, tracegen_int)
      )))
    } else if (trace == "up") {
      cpp_trace_ancestors(sn_full, dn_full, cand_nums, tracegen_int)
    } else {
      cpp_trace_descendants(sn_full, dn_full, cand_nums, tracegen_int)
    }

    ped_dt <- ped_dt[subset_idx]
    # NA-out parents dropped from the subset
    keep_inds <- ped_dt$Ind
    ped_dt[!is.na(Sire) & !(Sire %in% keep_inds), Sire := NA_character_]
    ped_dt[!is.na(Dam)  & !(Dam  %in% keep_inds), Dam  := NA_character_]

    # Rebuild integer indices for the subset (needed for C++ topo / gen assignment)
    new_sn <- match(ped_dt$Sire, ped_dt$Ind, nomatch = 0L)
    new_dn <- match(ped_dt$Dam,  ped_dt$Ind, nomatch = 0L)

    # --- C++ topological sort (replaces igraph build + topo_sort) ---
    topo <- cpp_topo_order(new_sn, new_dn)

    # Family / FamilySize
    ped_dt[, Family := ifelse(
      !is.na(Sire) & !is.na(Dam), paste0(Sire, "x", Dam), NA_character_
    )]
    ped_dt[, FamilySize := .N, by = Family]
    ped_dt[is.na(Family), FamilySize := 1L]

    if (addgen) {
      gen_vec <- if (genmethod_val == "top") {
        cpp_assign_generations_top(new_sn, new_dn, topo)
      } else {
        cpp_assign_generations_bottom(new_sn, new_dn, topo)
      }
      # Isolated nodes (no parents AND no children in the subset): set Gen = 0
      n_sub <- nrow(ped_dt)
      is_ref <- logical(n_sub)
      if (any(new_sn > 0L)) is_ref[new_sn[new_sn > 0L]] <- TRUE
      if (any(new_dn > 0L)) is_ref[new_dn[new_dn > 0L]] <- TRUE
      iso_mask <- (new_sn == 0L) & (new_dn == 0L) & !is_ref
      if (any(iso_mask)) gen_vec[iso_mask] <- 0L
      ped_dt[, Gen := gen_vec]

      if (genmethod_val == "bottom") {
        align_bottom_generations(ped_dt)
      }
    }

    # Sort
    if (addgen) {
      setorder(ped_dt, Gen, Ind)
    } else {
      ped_dt <- ped_dt[topo]
    }

    # Numeric IDs
    if (addnum) {
      ped_dt[, IndNum  := .I]
      ped_dt[, SireNum := match(Sire, Ind, nomatch = 0L)]
      ped_dt[, DamNum  := match(Dam,  Ind, nomatch = 0L)]
    } else {
      # Remove index cols that were added temporarily (only if they were absent
      # in the original input).
      if (!had_num_cols) {
        drop_cols <- intersect(c("IndNum", "SireNum", "DamNum"), names(ped_dt))
        if (length(drop_cols) > 0L) ped_dt[, (drop_cols) := NULL]
      }
    }

    ped_dt[, Cand := Ind %in% cand]

    ped_dt <- new_tidyped(ped_dt)
    data.table::setattr(ped_dt, "ped_meta", build_ped_meta(
      selfing          = selfing_val,
      bisexual_parents = bisexual_parents,
      genmethod        = genmethod_val
    ))

    if (inbreed) ped_dt <- inbreed(ped_dt, ...)
    return(ped_dt)
  }
  # ---------- End Fast Path ----------

  # 2. Data Preparation
  res_prep <- validate_and_prepare_ped(ped, selfing = selfing)
  ped_dt <- res_prep$ped_dt
  bisexual_parents <- res_prep$bisexual_parents
  
  # 3. Build Initial Graph
  # We need the graph for cycle detection and tracing
  graph_res <- build_ped_graph(ped_dt)
  g <- graph_res$g
  
  # 4. Cycle Detection
  check_ped_loops(g)
  
  # 5. Trace Candidates (if needed)
  if (!is.null(cand)) {
    ped_dt <- trace_ped_candidates(g, ped_dt, cand, trace, tracegen)
    
    # Rebuild graph for the subsetted pedigree
    # This is necessary because generation assignment depends on the structure of the subset
    graph_res <- build_ped_graph(ped_dt)
    g <- graph_res$g
  }
  
  # 6. Generation Assignment
  # We need topo_order for sorting even if addgen=FALSE
  topo_order <- as.integer(topo_sort(g))
  
  # 6.5. Pre-calculate Family (moved from step 8.5) to assist generation assignment
  # Create family identifier based on parent combination
  ped_dt[, Family := ifelse(
    !is.na(Sire) & !is.na(Dam),
    paste0(Sire, "x", Dam),
    NA_character_
  )]
  
  # Calculate family sizes
  ped_dt[, FamilySize := .N, by = Family]
  # Individuals without both parents have FamilySize = 1
  ped_dt[is.na(Family), FamilySize := 1L]
  
  if (addgen) {
    ped_dt <- assign_ped_generations(g, ped_dt, topo_order, genmethod)
    
    # Optimization for bottom-up method:
    # 1. Align full siblings to the same generation (the highest/minimum Gen among them)
    # This prevents leaves (individuals with no progeny) from dropping to the bottom
    # when they have siblings who are parents of deep lineages.
    if (genmethod == "bottom") {
      align_bottom_generations(ped_dt)
    }
  }
  
  # 7. Sex Inference and Check
  ped_dt <- infer_and_check_sex(ped_dt, selfing = selfing)
  
  # 8. Sorting and Numeric IDs
  if (addgen) {
    setorder(ped_dt, Gen, Ind)
  } else {
    # If no Gen, preserve topo sort order
    sorted_inds <- V(g)$name[topo_order]
    ped_dt <- ped_dt[match(sorted_inds, Ind)]
  }
  
  if (addnum) {
    ped_dt[, IndNum := .I]
    ped_dt[, SireNum := match(Sire, Ind, nomatch = 0)]
    ped_dt[, DamNum := match(Dam, Ind, nomatch = 0)]
  }
  
  if (!is.null(cand)) {
    ped_dt[, Cand := Ind %in% cand]
  }

  # 9. Final S3 and Inbreeding
  ped_dt <- new_tidyped(ped_dt)
  data.table::setattr(ped_dt, "ped_meta", build_ped_meta(
    selfing          = selfing,
    bisexual_parents = bisexual_parents,
    genmethod        = genmethod
  ))
  
  if (inbreed) {
    ped_dt <- inbreed(ped_dt, ...)
  }
  
  return(ped_dt)
}

# ==============================================================================
# Internal Helper Functions
# ==============================================================================

#' Validate and Prepare Pedigree Data
#' @noRd
validate_and_prepare_ped <- function(ped, selfing = FALSE) {
  ped_dt <- if (is.data.table(ped)) copy(ped) else as.data.table(ped)
  
  # Standardize primary column names
  setnames(ped_dt, 1:3, c("Ind", "Sire", "Dam"))
  
  # Ensure character types for IDs
  ped_dt[, c("Ind", "Sire", "Dam") := lapply(.SD, as.character), .SDcols = c("Ind", "Sire", "Dam")]
  
  # Remove records with missing Ind
  if (any(ped_dt$Ind %in% c("", " ", "0", "*", "NA", NA))) {
    warning("Missing values in 'Ind' column; records discarded.")
    ped_dt <- ped_dt[!Ind %in% c("", " ", "0", "*", "NA", NA)]
  }
  
  # Normalize missing parents as NA
  ped_dt[Sire %in% c("", " ", "0", "*", "NA"), Sire := NA_character_]
  ped_dt[Dam %in% c("", " ", "0", "*", "NA"), Dam := NA_character_]
  
  if (all(is.na(ped_dt$Sire)) && all(is.na(ped_dt$Dam))) {
    stop("All parents are missing; no pedigree can be built.")
  }

  # Duplicate checks
  if (anyDuplicated(ped_dt, by = c("Ind", "Sire", "Dam")) > 0) {
    n_duped <- sum(duplicated(ped_dt, by = c("Ind", "Sire", "Dam")))
    warning(sprintf("Removed %d records with duplicate Ind/Sire/Dam IDs.", n_duped))
    ped_dt <- unique(ped_dt, by = c("Ind", "Sire", "Dam"))
  }
  
  if (anyDuplicated(ped_dt, by = "Ind") > 0) {
    stop("Fatal error: Duplicate Ind IDs with different parents found!")
  }

  # Sex conflict check: same individual appears as both sire and dam
  sires <- unique(ped_dt[!is.na(Sire), Sire])
  dams <- unique(ped_dt[!is.na(Dam), Dam])
  bisexual_parents <- sort(intersect(sires, dams))
  
  if (length(bisexual_parents) > 0 && !selfing) {
    stop(sprintf(
      paste0("Sex conflict detected: The following individual(s) appear as both Sire and Dam: %s. ",
             "This is biologically impossible. Please check and correct the pedigree data. ",
             "If this is a plant pedigree with monoecious species, set selfing = TRUE."),
      paste(bisexual_parents, collapse = ", ")
    ), call. = FALSE)
  }
  
  if (length(bisexual_parents) > 0 && selfing) {
    message(sprintf(
      "Selfing mode: %d individual(s) appear as both Sire and Dam: %s. These will be assigned Sex = 'monoecious'.",
      length(bisexual_parents),
      paste(head(bisexual_parents, 10), collapse = ", ")
    ))
  }

  # Add missing founders
  all_parents <- unique(c(sires, dams))
  missing_founders <- setdiff(all_parents, ped_dt$Ind)
  
  if (length(missing_founders) > 0) {
    founder_dt <- data.table(Ind = missing_founders, Sire = NA_character_, Dam = NA_character_)
    ped_dt <- rbind(founder_dt, ped_dt, fill = TRUE)
  }
  
  list(ped_dt = ped_dt, bisexual_parents = bisexual_parents)
}

#' Build igraph Object from Pedigree
#' @noRd
build_ped_graph <- function(ped_dt) {
  node_names <- ped_dt$Ind
  
  # Use match() instead of named vector lookup for speed
  s_from <- match(ped_dt$Sire[!is.na(ped_dt$Sire)], node_names)
  s_to <- which(!is.na(ped_dt$Sire))
  
  d_from <- match(ped_dt$Dam[!is.na(ped_dt$Dam)], node_names)
  d_to <- which(!is.na(ped_dt$Dam))
  
  edge_mat <- matrix(c(s_from, d_from, s_to, d_to), ncol = 2)
  
  # Using graph_from_edgelist is faster usually
  g <- graph_from_edgelist(edge_mat, directed = TRUE)
  
  # Ensure all vertices exist (even isolated ones)
  num_needed <- length(node_names)
  if (vcount(g) < num_needed) {
    g <- add_vertices(g, num_needed - vcount(g))
  }
  
  V(g)$name <- node_names
  
  list(g = g)
}

#' Check for Cycles/Loops in Pedigree
#' @noRd
check_ped_loops <- function(g) {
  if (!is_dag(g)) {
    # 1. Check for self-loops (A -> A)
    # which_loop returns TRUE for edges that are self-loops
    loop_edges <- which(which_loop(g))
    self_loops <- character()
    if (length(loop_edges) > 0) {
      # ends() returns the vertex IDs of edge endpoints
      self_loops <- unique(V(g)[ends(g, loop_edges)[, 1]]$name)
    }
    
    # 2. Check for larger cycles via SCC
    scc <- components(g, mode = "strong")
    cyclic_nodes_idx <- which(scc$csize > 1)
    
    cycle_reports <- character()
    
    if (length(self_loops) > 0) {
      cycle_reports <- c(cycle_reports, paste(self_loops, "->", self_loops, "(self-loop)"))
    }
    
    if (length(cyclic_nodes_idx) > 0) {
      scc_reports <- sapply(cyclic_nodes_idx, function(i) {
        nodes_in_scc <- names(which(scc$membership == i))
        v_start <- nodes_in_scc[1]
        successors <- intersect(names(neighbors(g, v_start, mode = "out")), nodes_in_scc)
        
        if (length(successors) > 0) {
          v_next <- successors[1]
          p <- shortest_paths(g, from = v_next, to = v_start, mode = "out")$vpath[[1]]
          path_names <- c(v_start, names(p))
          paste(path_names, collapse = " -> ")
        } else {
          paste(nodes_in_scc, collapse = ", ")
        }
      })
      cycle_reports <- c(cycle_reports, scc_reports)
    }
    
    # Fallback
    if (length(cycle_reports) == 0) {
      cycle_reports <- "Complex cycle structure detected (try checking strong components)."
    }
    
    stop(paste("Pedigree error! Pedigree loops detected:\n", 
               paste(cycle_reports, collapse = "\n")), call. = FALSE)
  }
}

#' Trace Candidates and Subset Pedigree
#' @noRd
trace_ped_candidates <- function(g, ped_dt, cand, trace, tracegen) {
  # Drop NAs and empty strings from input
  cand_clean <- as.character(cand[!is.na(cand) & cand != "" & cand != " "])
  
  if (length(cand_clean) == 0) {
    stop("The cand parameter contains no valid individual IDs.")
  }

  valid_cand <- cand_clean[cand_clean %in% ped_dt$Ind]
  if (length(valid_cand) == 0) {
    stop("None of the specified candidates were found in the pedigree.")
  }

  if (length(valid_cand) < length(cand_clean)) {
    missing_cands <- setdiff(cand_clean, valid_cand)
    warning(sprintf("The following %d candidates were not found in the pedigree: %s", 
                    length(missing_cands), 
                    paste(head(missing_cands, 5), collapse = ", ")))
  }
  
  # Determine recursion depth
  # order must be numeric and >= 0 for neighborhood()
  order <- if (is.null(tracegen) || tracegen <= 0) vcount(g) else as.integer(tracegen)
  
  # For mode "all", we want the union of ancestors (in) and descendants (out).
  # We should NOT use igraph's mode="all" which is undirected search (connected component).
  if (trace == "all") {
    rel_in <- neighborhood(g, order = order, nodes = valid_cand, mode = "in")
    rel_out <- neighborhood(g, order = order, nodes = valid_cand, mode = "out")
    # Efficiently flatten and get unique IDs from both directions
    keep_indices <- unique(c(
      unlist(lapply(rel_in, as.integer)),
      unlist(lapply(rel_out, as.integer))
    ))
  } else {
    m <- if (trace == "up") "in" else "out"
    reaching_nodes_list <- neighborhood(g, order = order, nodes = valid_cand, mode = m)
    # Efficiently flatten and get unique IDs
    # lapply(list, as.integer) converts vertex sequences to integer IDs
    keep_indices <- unique(unlist(lapply(reaching_nodes_list, as.integer)))
  }
  
  keep_inds <- V(g)$name[keep_indices]
  
  ped_dt <- ped_dt[Ind %in% keep_inds]
  # Update parents not in set after tracing
  ped_dt[!(Sire %in% keep_inds), Sire := NA_character_]
  ped_dt[!(Dam %in% keep_inds), Dam := NA_character_]
  
  return(ped_dt[])
}

#' Assigns individual generation numbers based on topological sorting and parentage.
#'
#' @param g An igraph object representing the pedigree.
#' @param ped_dt A data.table containing the pedigree.
#' @param topo_order Integer vector specifying the topological order of vertices.
#' @param genmethod Character, "top" or "bottom".
#' @return A data.table with a \strong{Gen} column added.
#' @noRd
assign_ped_generations <- function(g, ped_dt, topo_order, genmethod) {
  # Get numeric parents for C++ (1-based, 0 for missing)
  # Pedigree is already complete (missing founders added)
  inds <- ped_dt$Ind
  sire_num <- match(ped_dt$Sire, inds, nomatch = 0)
  dam_num <- match(ped_dt$Dam, inds, nomatch = 0)
  
  if (genmethod == "top") {
    gen_vec <- cpp_assign_generations_top(sire_num, dam_num, as.integer(topo_order))
  } else {
    gen_vec <- cpp_assign_generations_bottom(sire_num, dam_num, as.integer(topo_order))
  }
  
  # Isolated nodes check
  in_deg <- degree(g, mode = "in")
  out_deg <- degree(g, mode = "out")
  iso_mask <- (in_deg == 0) & (out_deg == 0)
  if (any(iso_mask)) {
    # Match iso_mask (which is in graph order) to ped_dt order
    # topo_order is a permutation of 1:vcount(g)
    # The C++ gen_vec is in the same order as sire_num/dam_num, which is ped_dt order
    # BUT wait, the graph g was built from ped_dt. So V(g)$name is ped_dt$Ind.
    # So vertex i in the graph corresponds to row i in ped_dt.
    gen_vec[iso_mask] <- 0L
  }
  
  ped_dt[, Gen := gen_vec]
  return(ped_dt[])
}

#' Infer and Check Sex of Individuals
#' @noRd
infer_and_check_sex <- function(ped_dt, selfing = FALSE) {
  if (!("Sex" %in% names(ped_dt))) {
    ped_dt[, Sex := NA_character_]
  } else {
    ped_dt[, Sex := tolower(as.character(Sex))]
    ped_dt[Sex %in% c("", " ", "na"), Sex := NA_character_]
    # Normalize common abbreviations to canonical values
    ped_dt[Sex %in% c("m", "1", "male"),   Sex := "male"]
    ped_dt[Sex %in% c("f", "2", "female"), Sex := "female"]
  }
  
  sires <- unique(ped_dt[!is.na(Sire), Sire])
  dams <- unique(ped_dt[!is.na(Dam), Dam])
  
  # Identify monoecious individuals (appear as both Sire and Dam)
  monoecious_ids <- intersect(sires, dams)
  
  if (selfing && length(monoecious_ids) > 0) {
    # In selfing mode: monoecious individuals can be both Sire and Dam
    # Check for conflicts with explicit sex annotation (only non-monoecious annotations)
    sex_conflicts <- ped_dt[
      !(Ind %in% monoecious_ids) & 
      ((Ind %in% sires & Sex == "female") | (Ind %in% dams & Sex == "male")), Ind]
    if (length(sex_conflicts) > 0) {
      stop(sprintf(
        paste0("Sex annotation conflicts detected for: %s. ",
               "These individuals have explicit sex that contradicts their role as Sire/Dam. ",
               "Please check and correct the pedigree data."),
        paste(sex_conflicts, collapse = ", ")
      ), call. = FALSE)
    }
    
    # Check that monoecious individuals are not explicitly annotated as male or female
    mono_sex_conflicts <- ped_dt[
      Ind %in% monoecious_ids & Sex %in% c("male", "female"), Ind]
    if (length(mono_sex_conflicts) > 0) {
      stop(sprintf(
        paste0("Sex annotation conflicts for monoecious individuals: %s. ",
               "These individuals appear as both Sire and Dam but have explicit male/female annotation. ",
               "Remove or change their Sex annotation to 'monoecious' or NA."),
        paste(mono_sex_conflicts, collapse = ", ")
      ), call. = FALSE)
    }
    
    # Assign monoecious sex
    ped_dt[Ind %in% monoecious_ids & (is.na(Sex) | Sex != "monoecious"), Sex := "monoecious"]
    
    # Infer sex for remaining individuals (excluding monoecious)
    sires_only <- setdiff(sires, monoecious_ids)
    dams_only <- setdiff(dams, monoecious_ids)
    ped_dt[is.na(Sex) & (Ind %in% sires_only), Sex := "male"]
    ped_dt[is.na(Sex) & (Ind %in% dams_only), Sex := "female"]
  } else {
    # Standard mode: no selfing allowed
    # Sex conflicts with explicit sex annotation
    sex_conflicts <- ped_dt[(Ind %in% sires & Sex == "female") | (Ind %in% dams & Sex == "male"), Ind]
    if (length(sex_conflicts) > 0) {
      stop(sprintf(
        paste0("Sex annotation conflicts detected for: %s. ",
               "These individuals have explicit sex that contradicts their role as Sire/Dam. ",
               "Please check and correct the pedigree data."),
        paste(sex_conflicts, collapse = ", ")
      ), call. = FALSE)
    }
    
    # Infer sex from roles
    ped_dt[is.na(Sex) & (Ind %in% sires), Sex := "male"]
    ped_dt[is.na(Sex) & (Ind %in% dams), Sex := "female"]
  }
  
  return(ped_dt[])
}

#' Align generations for bottom-up method
#'
#' Aligns full siblings and mates to consistent generations after
#' bottom-up generation assignment.  Called by both the main path and
#' the fast path of \code{tidyped()}.
#' @param ped_dt A data.table with Ind, Sire, Dam, Family, Gen columns.
#' @return ped_dt modified in-place (Gen column updated).
#' @noRd
align_bottom_generations <- function(ped_dt) {
  # 1. Sibling alignment: all full-sibs share the minimum Gen
  ped_dt[!is.na(Family), Gen := min(Gen), by = Family]

  # 2. Mate alignment: iteratively push mismatched mates to the same Gen
  iter <- 0L
  max_iter <- 10L
  repeat {
    iter <- iter + 1L
    sire_gen <- ped_dt$Gen[match(ped_dt$Sire, ped_dt$Ind)]
    dam_gen  <- ped_dt$Gen[match(ped_dt$Dam,  ped_dt$Ind)]

    valid_mask <- !is.na(sire_gen) & !is.na(dam_gen) & (sire_gen != dam_gen)
    if (!any(valid_mask) || iter > max_iter) break

    target_gen <- pmax(sire_gen[valid_mask], dam_gen[valid_mask])
    upd_dt <- data.table(
      Ind    = c(ped_dt$Sire[valid_mask], ped_dt$Dam[valid_mask]),
      NewGen = c(target_gen, target_gen)
    )
    upd_dt <- upd_dt[, .(NewGen = max(NewGen)), by = Ind]

    # Constraint: NewGen must be < min(child Gen)
    inds_vec <- upd_dt$Ind
    relevant_kids <- ped_dt[Sire %in% inds_vec | Dam %in% inds_vec,
                            .(Sire, Dam, Gen)]
    limits <- rbind(
      relevant_kids[Sire %in% inds_vec, .(Ind = Sire, Limit = Gen)],
      relevant_kids[Dam  %in% inds_vec, .(Ind = Dam,  Limit = Gen)]
    )[, .(Limit = min(Limit)), by = Ind]

    limit_vals <- limits$Limit[match(upd_dt$Ind, limits$Ind)]
    curr_gen   <- ped_dt$Gen[match(upd_dt$Ind, ped_dt$Ind)]
    to_change  <- (upd_dt$NewGen > curr_gen) &
                  (is.na(limit_vals) | upd_dt$NewGen < limit_vals)

    if (any(to_change)) {
      update_inds <- upd_dt$Ind[to_change]
      update_gens <- upd_dt$NewGen[to_change]
      ped_dt[Ind %in% update_inds, Gen := update_gens[match(Ind, update_inds)]]
    } else {
      break
    }
  }

  # 3. Final sibling re-alignment (mate push may have split siblings)
  ped_dt[!is.na(Family), Gen := min(Gen), by = Family]

  invisible(ped_dt)
}

library(testthat)
library(data.table)

# Load package functions if not available (for standalone testing)
if (!exists("tidyped")) {
  # Attempt to locate source files relative to this test file
  # If running from package root: "R/tidyped.R"
  # If running from tests/testthat: "../../R/tidyped.R"
  
  candidates <- c(
    file.path("R", "tidyped.R"),
    file.path("..", "..", "R", "tidyped.R")
  )
  
  for (f in candidates) {
    if (file.exists(f)) {
      # We also need utils for S3 class
      utils_f <- sub("tidyped.R", "utils-s3.R", f)
      inbreed_f <- sub("tidyped.R", "inbreed.R", f) # If inbreed=TRUE tested
      
      if (file.exists(utils_f)) source(utils_f)
      if (file.exists(inbreed_f)) source(inbreed_f)
      source(f)
      break
    }
  }
}

# Load simple_ped data for testing
if (!exists("simple_ped")) {
  data_candidates <- c(
    file.path("data", "simple_ped.rda"),
    file.path("..", "..", "data", "simple_ped.rda")
  )
  for (df in data_candidates) {
    if (file.exists(df)) {
      load(df)
      break
    }
  }
}

test_that("1. Basic Integrity and Types", {
  # Minimal pedigree
  ped_df <- data.frame(
    Ind = c("A", "B", "C"),
    Sire = c(NA, "A", "A"),
    Dam = c(NA, NA, "B")
  )
  
  # Should accept data.frame and return data.table/tidyped
  res <- tidyped(ped_df, addgen = TRUE)
  
  expect_s3_class(res, "tidyped")
  expect_s3_class(res, "data.table")
  expect_equal(nrow(res), 3)
  expect_true("Gen" %in% names(res))
  
  # Check NA handling
  expect_true(is.na(res[Ind == "A", Sire]))
})

test_that("2. Missing Value and Character Normalization", {
  ped_dirty <- data.table(
    Ind = c("X", "Y"),
    Sire = c("0", "X"),    # "0" -> NA
    Dam = c("*", "NA")     # "*" or "NA" -> NA
  )
  
  res <- tidyped(ped_dirty)
  
  # X should have NA parents
  expect_true(is.na(res[Ind == "X", Sire]))
  expect_true(is.na(res[Ind == "X", Dam]))
  
  # Y should have X as Sire
  expect_equal(res[Ind == "Y", Sire], "X")
})

test_that("3. Loop Detection", {
  # A -> B -> A
  ped_loop <- data.frame(
    Ind = c("A", "B"),
    Sire = c("B", NA),
    Dam = c(NA, "A")
  )
  
  expect_error(tidyped(ped_loop), "Pedigree loops detected")
  
  # Self loop A -> A
  ped_self <- data.frame( Ind="A", Sire="A", Dam=NA)
  expect_error(tidyped(ped_self), "Pedigree loops detected")
})

test_that("4. Generation Alignment Logic", {
  # Case: Alignment of introduced founder (E) to mate (C)
  # Gen 1: A, B
  # Gen 2: C (Child of A,B)
  # Founder E mates with C.
  # Gen 3: D (Child of C, E)
  
  # Standard behavior: Founders (no parents) get Gen 1
  # E is a founder, so E gets Gen 1
  # D is child of C (Gen 2) and E (Gen 1), so D gets max(parent_gen) + 1 = Gen 3
  
  ped_align <- data.table(
    Ind = c("A", "B", "C", "D", "E"),
    Sire = c(NA, NA, "A", "C", NA),
    Dam = c(NA, NA, "B", "E", NA)
  )
  
  res <- tidyped(ped_align)
  setkey(res, Ind)
  
  expect_equal(res["C", Gen], 2)
  # E is a founder (no parents), so E gets Gen 1 (standard behavior)
  expect_equal(res["E", Gen], 1)
  # D is child of C (Gen 2) and E (Gen 1) -> Gen 3
  expect_equal(res["D", Gen], 3)
})

test_that("5. Sibling Alignment (Full Sibs)", {
  # Full siblings should have same generation
  # Gen 1: P1, P2
  # Gen 2: C1 (P1xP2)
  # Gen 3: G3 (Child of C1)
  # C2 is full sib of C1 (P1xP2). C2 has no progeny. 
  # Without alignment, C2 might fall to bottom (Gen 3) in bottom-up mode
  # With Sibling/Height logic, C2 matches C1 (Gen 2).
  
  ped_sibs <- data.table(
    Ind = c("P1", "P2", "C1", "C2", "G3"),
    Sire = c(NA, NA, "P1", "P1", "C1"),
    Dam  = c(NA, NA, "P2", "P2", NA)
  )
  
  # Default (top)
  res_top <- tidyped(ped_sibs, genmethod = "top")
  setkey(res_top, Ind)
  expect_equal(res_top["C1", Gen], res_top["C2", Gen], info = "Top mode: Siblings aligned")
  
  # Bottom-up
  res_bottom <- tidyped(ped_sibs, genmethod = "bottom")
  setkey(res_bottom, Ind)
  expect_equal(res_bottom["C1", Gen], 2, info = "Bottom mode: C1 Gen")
  expect_equal(res_bottom["C2", Gen], 2, info = "Bottom mode: C2 aligned with C1")
  expect_equal(res_bottom["G3", Gen], 3, info = "Bottom mode: G3 Gen")
})

test_that("6. Tracing Up/Down", {
  # G1: F1, M1
  # G2: C1 (F1xM1)
  # G3: C2 (C1xUnknown)
  
  ped <- data.table(
    Ind = c("F1", "M1", "C1", "C2"),
    Sire = c(NA, NA, "M1", "C1"),
    Dam = c(NA, NA, "F1", NA)
  )
  
  # Trace UP from C2
  up <- tidyped(ped, cand = "C2", trace = "up")
  expect_true(all(c("F1", "M1", "C1", "C2") %in% up$Ind))
  
  # Trace DOWN from F1
  down <- tidyped(ped, cand = "F1", trace = "down")
  expect_true(all(c("F1", "C1", "C2") %in% down$Ind))
  
  # Trace Gen Limit (e.g., trace up 1 gen from C1)
  # Should include C1 and parents (F1, M1).
  up_lim <- tidyped(ped, cand = "C1", trace = "up", tracegen = 1)
  expect_true(all(c("F1", "M1") %in% up_lim$Ind))
  
  # Trace "all" (Up + Down) from C1
  # Ancestors: F1, M1. Descendants: C2. Self: C1.
  res_all <- tidyped(ped, cand = "C1", trace = "all")
  expect_true(all(c("F1", "M1", "C1", "C2") %in% res_all$Ind))
  expect_equal(sort(res_all$Ind), sort(c("F1", "M1", "C1", "C2")))
  
  # Trace Gen Depth (tracegen > 1) with C2
  # C2 -> C1 -> (F1, M1)
  # tracegen=1 from C2: Should get C2 + Parents (C1). Should missing Grandparents (F1, M1).
  up_gen1 <- tidyped(ped, cand = "C2", trace = "up", tracegen = 1)
  expect_true("C1" %in% up_gen1$Ind)
  expect_false(any(c("F1", "M1") %in% up_gen1$Ind))
  
  # tracegen=2 from C2: Should reach Grandparents
  up_gen2 <- tidyped(ped, cand = "C2", trace = "up", tracegen = 2)
  expect_true(all(c("F1", "M1") %in% up_gen2$Ind))
})

test_that("11. Parameter Validation", {
  ped <- data.table(
    Ind = c("A", "B"),
    Sire = c(NA, NA),
    Dam = c(NA, NA)
  )
  
  # addgen
  expect_error(tidyped(ped, addgen = 1), "The addgen parameter only is assigned using TRUE or FALSE")
  expect_error(tidyped(ped, addgen = "TRUE"), "The addgen parameter only is assigned using TRUE or FALSE")
  
  # addnum
  expect_error(tidyped(ped, addnum = 1), "The addnum parameter only is assigned using TRUE or FALSE")
  
  # inbreed
  expect_error(tidyped(ped, inbreed = "yes"), "The inbreed parameter only is assigned using TRUE or FALSE")
  
  # trace
  expect_error(tidyped(ped, trace = "backwards"), "The trace parameter must be one of: up, down, all")
  
  # tracegen
  expect_error(tidyped(ped, tracegen = "2"), "The tracegen parameter must be a single numeric value")
  
  # ped validation
  expect_error(tidyped(), "The ped parameter cannot be NULL or missing")
  expect_error(tidyped(NULL), "The ped parameter cannot be NULL or missing")
  expect_error(tidyped(data.frame()), "The ped parameter cannot be empty")
})

test_that("10. Deep Pedigree Tests (simple_ped)", {
  skip_if_not(exists("simple_ped"), "simple_ped data not available")
  
  # Ensure simple_ped is a data.table or data.frame
  ped_deep <- as.data.table(simple_ped)
  
  # Run full processing
  res <- tidyped(ped_deep, addgen = TRUE, addnum = TRUE)
  
  # 1. Structure check
  expect_s3_class(res, "tidyped")
  expect_true("Gen" %in% names(res))
  
  # 2. Generation Depth Check
  # Based on previous context, simple_ped has around 6 generations
  max_gen <- max(res$Gen)
  expect_gt(max_gen, 4) 
  
  # 3. Parent-Offspring Generation Constraint
  # Gen(Child) should generally be > Gen(Parent)
  # (Though with overlap generations and heurisitcs, strictly > is expected)
  
  # Merge Sire Gen
  res_sire <- merge(res[, .(IndNum, Gen, SireNum)], res[, .(IndNum, SireGen=Gen)], 
                    by.x="SireNum", by.y="IndNum", all.x=TRUE) # Use Num for speed/safety
  # Compare only known parents
  valid_sire <- res_sire[SireNum > 0]
  
  # Check if Child Gen > Sire Gen
  # Note: Due to alignment heuristics, Child Gen could be much larger, but never smaller or equal?
  # Actually, graph is DAG, so Child Gen MUST be > Sire Gen if purely topological.
  # The heuristic might pull parents UP (increase numeric value) -> making parent older.
  # Small Gen number = Old (Founders). Large Gen number = Young.
  # WAIT: Previous heuristic implementation:
  # Founders aligned to mate's generation (min mate gen).
  # Height propagation: Gen = MaxH - H + 1.
  # Topo sort implies Parents appear before Children.
  # If Gen 1 is old, Gen 6 is young.
  # Then Gen(Child) > Gen(Parent).
  
  expect_true(all(valid_sire$Gen > valid_sire$SireGen))
  
  # 4. Tracing a deep individual
  # Using J5X804 (Generation 6 in prior context)
  cand_id <- "J5X804"
  if (cand_id %in% res$Ind) {
    trace_res <- tidyped(ped_deep, cand = cand_id, trace = "up")
    
    # Should contain cand
    expect_true(cand_id %in% trace_res$Ind)
    # Should contain ancestors (at least one parent if known)
    # J5X804 has parents J4Y326/J4E185 in context
    parents <- c("J4Y326", "J4E185")
    if (any(parents %in% ped_deep$Ind)) {
       expect_true(any(parents %in% trace_res$Ind))
    }
  }
})

test_that("12. addgen=FALSE functionality", {
  # Regression test for bug where addgen=FALSE returned empty result
  ped <- data.table(
    Ind = c("C", "A", "B"),
    Sire = c("A", NA, NA),
    Dam = c("B", NA, NA)
  )
  
  res <- tidyped(ped, addgen = FALSE)
  
  expect_equal(nrow(res), 3)
  expect_false("Gen" %in% names(res))
  
  # Check topological sorting
  # Parents (A, B) should appear before Child (C)
  # But exact order of A vs B is not guaranteed, just before C.
  idx_A <- which(res$Ind == "A")
  idx_B <- which(res$Ind == "B")
  idx_C <- which(res$Ind == "C")
  
  expect_lt(idx_A, idx_C)
  expect_lt(idx_B, idx_C)
})

library(testthat)
library(data.table)

# Load package functions if not available
# if (!exists("tidyped")) {
#   devtools::load_all()
# }

test_that("Mate Alignment: Push Down Logic (pmax)", {
  # Case 1: Founder (F) mates with Descendant (D) at Gen 3.
  # A -> B -> D
  # F -> D (F is valid parent of D? No, F and D are mates? No wait)
  # Correction: Mate alignment is about optimizing shared children.
  # If A and B are parents of C.
  # A is Gen 1 (Founder).
  # B is Gen 3 (Child of ...).
  # C is Gen 4.
  # Old Logic (pmin): Pull B up to Gen 1. 
  #   -> If B has parents at Gen 2, B(1) < Parents(2) -> ERROR.
  # New Logic (pmax): Push A down to Gen 3.
  #   -> A(3), B(3), C(4). OK.
  
  ped <- data.table(
    Ind  = c("G1", "G2", "G3", "F_Mate", "Kind"),
    Sire = c(NA,   "G1", "G2", NA,       "G3"),     # G1 -> G2 -> G3. F_Mate x G3 -> Kind
    Dam  = c(NA,   NA,   NA,   NA,       "F_Mate")
  )
  
  # Expected generations (bottom-up or top-down, let's use bottom)
  # Bottom-up tends to push things deep.
  # Initial calculation:
  # G1 (Gen 1 or similar distance from bottom)
  # Kind is leaf.
  
  # Let's inspect generations directly.
  # G1 (Gen 1)
  # G2 (Gen 2)
  # G3 (Gen 3)
  # F_Mate (Gen 1 initially if treated as founder)
  # Kind (Gen 4)
  
  res <- tidyped(ped, genmethod = "bottom")
  setkey(res, Ind)
  
  # Check Baseline Generations
  expect_gt(res["G3", Gen], res["G2", Gen])
  expect_gt(res["G2", Gen], res["G1", Gen])
  
  # Check Mate Alignment
  # F_Mate should be aligned with G3
  expect_equal(res["F_Mate", Gen], res["G3", Gen], 
               info = "Founder Mate should be pushed down to Spouse's generation")
  
  # Check Topology
  # Kind must be > Parents
  expect_gt(res["Kind", Gen], res["F_Mate", Gen])
  expect_gt(res["Kind", Gen], res["G3", Gen])
})


test_that("Mate Alignment: Topological Constraint (Block Push)", {
  # Case 2: Founder (F) mates with G3 (Deep).
  # But F ALREADY has a child C at Gen 2.
  # If we push F to Gen 3 (to match G3), F(3) > C(2). VIOLATION.
  # Logic should block the push. F stays at 1.
  
  ped <- data.table(
    Ind  = c("G1", "G2", "G3", "F", "C", "Kind"),
    Sire = c(NA,   "G1", "G2", NA,  NA,  "G3"),
    Dam  = c(NA,   NA,   NA,   NA,  "F", "F") 
  )
  # Relationships:
  # G1 -> G2 -> G3
  # F -> C (F is parent of C)
  # G3 x F -> Kind (F is parent of Kind)
  
  # Initial Gen estimate:
  # G1: 1
  # G2: 2
  # G3: 3
  # F: 1 (Founder)
  # C: 2 (Child of F)
  # Kind: 4 (Child of G3(3) and F(1/3?))
  
  # Conflict:
  # Mate Alignment wants F and G3 to be same Gen.
  # G3 is Gen 3.
  # F ideal is Gen 3.
  # But F has child C at Gen 2.
  # F cannot be >= C. F must be < C.
  # F max valid Gen is 1 (if C is 2).
  # Result: F stays at 1. G3 stays at 3.
  
  res <- tidyped(ped, genmethod = "top")
  setkey(res, Ind)
  
  # F should NOT be pushed to 3
  expect_lt(res["F", Gen], res["C", Gen])
  expect_equal(res["F", Gen], 1, info = "F constrained to Gen 1 by child C")
  expect_equal(res["G3", Gen], 3, info = "G3 remains at Gen 3")
  
  # Mates are misaligned
  expect_true(res["F", Gen] != res["G3", Gen], info = "Mates remain misaligned due to constraints")
})

test_that("Reproduce Bug with Parents X, U, V", {
  # The small_ped bug.
  # X (Child) was Gen 1. U,V (Parents) were Gen 4.
  # Fixed logic should ensure X > U, V.
  
  # Load real simplified data if possible, or mock it closely.
  # The bug was triggered by X mating with a founder or similar.
  # Mock:
  # U(4) x V(4) -> X(5)
  # X(5) x N(1) -> Offspring
  # Old Bug: N(1) pulls X to 1.
  # New Fix: X(5) pushes N to 5.
  
  ped <- data.table(
    Ind = c("U", "V", "X", "N", "O"),
    Sire = c(NA, NA, "U", "X", "X"), # U->X, X->O
    Dam = c(NA, NA, "V", "N", "N"),  # V->X, N->O
    GenInit = c(4, 4, 5, 1, 6) # Simulated initial state or structural implication
  )
  # We provide structure, tidyped calculates Gen.
  # We need depth for U,V to be 4.
  # Chain: G1->G2->G3->U
  
  ped_chain <- data.table(
    Ind = c("A","B","C","U", "V", "X", "N", "O"),
    Sire = c(NA,"A","B","C", NA,  "U", NA,  "X"),
    Dam =  c(NA,NA,NA, NA,  NA,  "V", NA,  "N")
  )
  # A(1)->B(2)->C(3)->U(4).
  # V(1/4?) - lets assume V is deep or aligned.
  # X(5) from U(4).
  # N(1).
  # O(6) from X(5)xN(1).
  
  res <- tidyped(ped_chain, genmethod = "bottom")
  setkey(res, Ind)
  
  # Check X vs U
  expect_gt(res["X", Gen], res["U", Gen])
  
  # Check Mate Alignment (X and N)
  # X is 5 (approx). N should be pushed to 5.
  expect_equal(res["N", Gen], res["X", Gen], info = "Founder N aligned to Mate X")
})

---
title: "2. Efficient visPedigree Workflows"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{2. Efficient visPedigree Workflows}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6.5,
  fig.height = 6,
  dpi = 96,
  out.width = "100%"
)
```

This vignette summarizes efficient day-to-day workflows for `visPedigree`
after the `tidyped` architecture updates. The goal is simple:

1. tidy once,
2. reuse the resulting `tidyped` object many times,
3. subset safely,
4. trace candidates explicitly when pedigree completeness matters.

For basic tidying, see `tidy-pedigree`. For downstream statistics, see
`pedigree-analysis`.

## 1. Load packages and example data

```{r setup}
library(visPedigree)
library(data.table)

data(simple_ped, package = "visPedigree")
```

## 2. Tidy once, reuse many times

The most efficient workflow is to create a master `tidyped` object once and
reuse it for plotting, tracing, inbreeding, and matrix calculations.

```{r tidy-once}
tp_master <- tidyped(simple_ped)

class(tp_master)
is_tidyped(tp_master)
pedmeta(tp_master)
```

This avoids repeated validation, founder insertion, loop checking, generation
assignment, and integer re-indexing.

## 3. Fast repeated tracing from an existing `tidyped`

When the input is already a `tidyped` object and `cand` is supplied,
`tidyped()` now uses a fast path. It skips the expensive global preprocessing
steps and directly traces the requested candidates.

```{r fast-trace}
tp_up <- tidyped(tp_master, cand = "J5X804", trace = "up", tracegen = 2)
tp_down <- tidyped(tp_master, cand = "J0Z990", trace = "down")

has_candidates(tp_up)
tp_up[, .(Ind, Sire, Dam, Cand)]
```

Recommended pattern:

```{r fast-trace-pattern, eval = FALSE}
# expensive once
# tp_master <- tidyped(raw_ped)

# cheap many times
# tp_a <- tidyped(tp_master, cand = ids_a, trace = "up")
# tp_b <- tidyped(tp_master, cand = ids_b, trace = "all", tracegen = 3)
# tp_c <- tidyped(tp_master, cand = ids_c, trace = "down")
```

## 4. Safe `data.table` usage on `tidyped`

A `tidyped` object is also a `data.table`, so by-reference workflows remain
available.

### 4.1 Adding new columns is safe

```{r dt-modify}
tp_work <- copy(tp_master)
tp_work[, phenotype := seq_len(.N)]

class(tp_work)
head(tp_work[, .(Ind, phenotype)])
```

The `tidyped` class is preserved after `:=` operations.

### 4.2 Incomplete row subsetting now degrades safely

If row filtering removes required parents, the result is no longer a complete
pedigree. In that case the object is downgraded to a plain `data.table` with a
warning.

```{r incomplete-subset}
ped_year <- data.table(
  Ind = c("A", "B", "C", "D"),
  Sire = c(NA, NA, "A", "C"),
  Dam = c(NA, NA, "B", "B"),
  Year = c(2000, 2000, 2005, 2006)
)

tp_year <- tidyped(ped_year)
sub_dt <- tp_year[Year > 2005]

class(sub_dt)
sub_dt
```

This behavior prevents invalid integer pedigree indices from silently reaching
C++ code.

Completeness-sensitive analyses now fail fast on such truncated subsets:

```{r incomplete-subset-error, error = TRUE}
inbreed(sub_dt)
```

### 4.3 Use explicit tracing when you need a valid sub-pedigree

If the goal is to keep a structurally valid pedigree around focal individuals,
use candidate tracing instead of ad hoc row filtering.

```{r explicit-tracing}
valid_sub_tp <- tidyped(tp_year, cand = "D", trace = "up")

class(valid_sub_tp)
valid_sub_tp[, .(Ind, Sire, Dam, Cand)]
```

Then compute on the valid sub-pedigree and, if needed, filter the final result
back to the focal individuals:

```{r explicit-analysis}
inbreed(valid_sub_tp)[Ind == "D", .(Ind, f)]
```

## 5. `splitped()` versus `pedsubpop()`

These two functions serve different purposes.

- `splitped()` returns the actual split pedigree objects.
- `pedsubpop()` returns a summary table.

```{r split-vs-summary}
sub_tps <- splitped(tp_master)
length(sub_tps)
class(sub_tps[[1]])

pedsubpop(tp_master)
```

Use `splitped()` when you need downstream analysis on each component. Use
`pedsubpop()` when you only need the component summary.

## 6. Use accessors instead of manual attribute checks

The updated accessors are the preferred way to inspect object state.

```{r accessors}
tp_f <- inbreed(tp_master)

is_tidyped(tp_f)
has_inbreeding(tp_f)
has_candidates(tp_f)
pedmeta(tp_f)
```

This is preferable to hand-written checks such as `"f" %in% names(tp)` or
manual attribute access scattered throughout user code.

## 7. Recommended high-efficiency workflow

A practical pattern for large pedigrees is:

```{r best-practice, eval = FALSE}
# 1. build one validated master object
# tp_master <- tidyped(raw_ped)

# 2. add analysis-specific columns in place
# tp_master[, phenotype := pheno_vector]
# tp_master[, cohort := year_vector]

# 3. extract valid candidate sub-pedigrees explicitly
# tp_sel <- tidyped(tp_master, cand = selected_ids, trace = "up", tracegen = 3)

# 4. run downstream analysis on either the full master or traced sub-pedigree
# pedstats(tp_master)
# pedmat(tp_sel)
# inbreed(tp_sel)
# visped(tp_sel)

# 5. split only when disconnected components really matter
# comps <- splitped(tp_master)
```

## 8. Practical rules of thumb

1. Call `tidyped()` on raw pedigree data once.
2. Reuse the resulting `tidyped` object as the master pedigree.
3. Use `tidyped(tp_master, cand = ...)` for valid local extraction.
4. Use ordinary row filtering only when a plain `data.table` result is acceptable.
5. Use `splitped()` for actual component objects and `pedsubpop()` for summaries.
6. Use `pedmeta()`, `is_tidyped()`, `has_inbreeding()`, and `has_candidates()` to inspect object state.

These rules keep workflows fast, explicit, and structurally safe.

---
title: "5. Calculation and visualization of relationship matrix"
author: "Sheng Luan"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{5. Calculation and visualization of relationship matrix}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6.5,
  fig.height = 6.5,
  dpi = 300,
  out.width = "100%"
)
library(visPedigree)
library(Matrix)
```

1. [Calculating Relationship Matrices with pedmat()](#1)   
1.1 [Supported Methods](#1-1)   
1.2 [Basic Usage](#1-2)   
1.3 [Sparse Matrix Representation](#1-3)   
2. [Matrix-Free Products with pedprod()](#2)   
2.1 [When to Use pedprod vs pedmat](#2-1)   
2.2 [Basic Usage — $Ax$ (vector)](#2-2)   
2.3 [Products with $A^{-1}$](#2-3)    
2.4 [Multiple Contribution Schemes — $AX$](#2-4)   
2.5 [Practical Use Cases](#2-5)   
2.6 [Performance and Scalability](#2-6)   
3. [Inspecting the Matrix](#3)   
3.1 [Summary Statistics](#3-1)   
3.2 [Querying Specific Relationships](#3-2)   
4. [Compact Mode for Large Pedigrees](#4)   
4.1 [Using compact = TRUE](#4-1)   
4.2 [Expanding and Querying Compacted Matrices](#4-2)   
4.3 [When to Use Compact Mode](#4-3)   
5. [Visualizing Relationship Matrices with vismat()](#5)   
5.1 [Relationship Heatmaps](#5-1)   
5.2 [Inbreeding and Kinship Histograms](#5-2)   
6. [Performance Considerations](#6)   

Relationship matrices are fundamental tools in quantitative genetics and animal breeding. They quantify the genetic similarity between individuals due to shared ancestry, which is essential for estimating breeding values (BLUP) and managing genetic diversity. The `visPedigree` package provides efficient tools for calculating various relationship matrices and visualizing them through heatmaps and histograms.

## 1. Calculating Relationship Matrices with `pedmat()` {#1}

The `pedmat()` function is the primary tool for calculating relationship matrices. It supports both additive and dominance relationship matrices, as well as their inverses. The inverse additive relationship matrix (`Ainv`) follows Henderson's rules (Henderson, 1976), while inbreeding coefficients use the same optimized recursive engine as `inbreed()`, based on algorithms for large populations described by Meuwissen & Luo (1992) and Sargolzaei et al. (2005).

### 1.1 Supported Methods {#1-1}

The `method` parameter in `pedmat()` determines the type of matrix to calculate:

-   **"A"**: Additive relationship matrix (Numerator Relationship Matrix).
-   **"Ainv"**: Inverse of the additive relationship matrix.
-   **"D"**: Dominance relationship matrix.
-   **"Dinv"**: Inverse of the dominance relationship matrix.
-   **"AA"**: Additive-by-additive (epistatic) relationship matrix.
-   **"AAinv"**: Inverse of the epistatic relationship matrix.
-   **"f"**: Inbreeding coefficients vector (uses the same optimized engine as `tidyped(..., inbreed = TRUE)`).

### 1.2 Basic Usage {#1-2}

Most calculations require a pedigree tidied by `tidyped()`.

```{r basic_calc}
# Load example pedigree and tidy it
data(small_ped)
tped <- tidyped(small_ped)

# Calculate Additive Relationship Matrix (A)
mat_A <- pedmat(tped, method = "A")

# Calculate Dominance Relationship Matrix (D)
mat_D <- pedmat(tped, method = "D")

# Calculate inbreeding coefficients (f)
vec_f <- pedmat(tped, method = "f")
```

### 1.3 Sparse Matrix Representation {#1-3}

By default, `pedmat()` returns a sparse matrix (class `dsCMatrix` from the `Matrix` package) for relationship matrices. This is highly memory-efficient for large pedigrees where many individuals are unrelated.

```{r sparse_check}
class(mat_A)
```

## 2. Matrix-Free Products with `pedprod()` {#2}

For small and moderate pedigrees, matrices returned by `pedmat()` are ordinary
matrix-like R objects and can be used directly with R's `%*%`. For large
pedigrees, however, forming the dense additive relationship matrix can be the
dominant memory cost. `pedprod()` instead computes $Ax$, $AX$,
$A^{-1}x$, or $A^{-1}X$ directly from the pedigree — without ever constructing
the $n \times n$ relationship matrix.

The implementation uses the pedigree factorization $A = TDT'$ (Colleau, 2002),
where $T$ is the transmission matrix and $D$ is the diagonal matrix of Mendelian
sampling variances. The product is computed through backward and forward pedigree
traversals in $O(np)$ time for an $n \times p$ right-hand side, with $O(np)$
working memory — a fraction of the $O(n^2)$ storage that a dense $A$ would require.

### 2.1 When to Use `pedprod` vs `pedmat` {#2-1}

| Use `pedprod()` when ... | Use `pedmat()` when ... |
|:---|:---|
| You only need matrix-vector products ($Ax$, $A^{-1}x$) | You need individual relationship coefficients |
| The pedigree has >10 000 individuals | The pedigree is small enough that $A$ fits in memory |
| You are iterating (e.g., BLUP, OCS, MCMC) | You need the matrix for visualization or inspection |
| Memory is the bottleneck | You need $D$, $AA$, or dominance matrices |
| You are working with $A^{-1}$ products for mixed models | You need to query specific pairwise relationships |

The two functions are complementary: `pedmat()` gives you the full matrix when
you need to examine it; `pedprod()` applies the matrix implicitly when you only
need its action on vectors.

### 2.2 Basic Usage — $A x$ (vector) {#2-2}

The most common use is computing the additive relationship of every individual
to a weighted group — for example, a set of selection candidates.

```{r pedprod_Ax}
# Load example pedigree and tidy it
data(small_ped)
tped <- tidyped(small_ped)

# Equal contributions from two candidates; other individuals have zero weight
weights <- c(Z1 = 0.5, Z2 = 0.5)

# Relationship of every pedigree individual to the weighted candidate group
Ax <- pedprod(tped, weights)
head(Ax)

# Average additive relationship c' A c and average coancestry
mean_relationship <- sum(weights * Ax[names(weights)])
mean_coancestry <- mean_relationship / 2
c(mean_relationship = mean_relationship, mean_coancestry = mean_coancestry)
```

**Named vs unnamed vectors.** When `x` is named, individuals not listed
automatically receive value zero — you do not need to pad the vector yourself.
An unnamed vector must have exactly one entry per pedigree individual in
`IndNum` order:

```{r pedprod_named_unnamed}
# Named: only listed individuals get non-zero values
named_x <- c(Z1 = 0.3, Z2 = 0.4, A = 0.3)
pedprod(tped, named_x)[1:6]  # B, C, etc. are automatically zero

# Unnamed: must match the pedigree size
unnamed_x <- rep(1, nrow(tped))
length(pedprod(tped, unnamed_x))
```

Input validation catches common mistakes: duplicate or unknown IDs, missing
values, and non-finite entries are all rejected with informative messages.

### 2.3 Products with $A^{-1}$ {#2-3}

The inverse additive relationship matrix $A^{-1}$ is fundamental in mixed model
equations (Henderson's BLUP). `pedprod()` computes $A^{-1}x$ and $A^{-1}X$
efficiently using Henderson's rules without forming the $n \times n$ matrix:

```{r pedprod_Ainv}
# Ainv * vector
x <- rnorm(nrow(tped))
Ainv_x <- pedprod(tped, x, method = "Ainv")
head(Ainv_x)

# Verify against explicit computation (small pedigree only)
A <- pedmat(tped, method = "A", sparse = FALSE)
Ainv <- pedmat(tped, method = "Ainv", sparse = FALSE)
all.equal(
  unname(Ainv_x),
  unname(drop(Ainv %*% x)),
  tolerance = 1e-12
)
```

For large pedigrees where explicit inversion is impossible, `pedprod(tped, x,
method = "Ainv")` remains computationally feasible. The numerical equivalence
with the full matrix can be verified on a small pedigree:

```{r pedprod_Ainv_identity}
# Ainv * (A * x) should recover x (to machine precision)
A_x <- pedprod(tped, x, method = "A")
Ainv_Ax <- pedprod(tped, A_x, method = "Ainv")
all.equal(unname(Ainv_Ax), unname(x), tolerance = 1e-12)
```

### 2.4 Multiple Contribution Schemes — $A X$ {#2-4}

When `x` is a matrix, `pedprod()` computes all columns in a single backward/
forward traversal of the pedigree. This is significantly more efficient than
calling `pedprod()` repeatedly for each column.

```{r pedprod_AX}
# Three ways to weight the SAME two candidates (Z1 and Z2 are full sibs)
schemes <- cbind(
  Equal      = c(Z1 = 0.5, Z2 = 0.5),
  Z1_only    = c(Z1 = 1.0, Z2 = 0.0),
  Weighted   = c(Z1 = 0.7, Z2 = 0.3)
)
AX <- pedprod(tped, schemes)

# The schemes diverge at the candidates themselves (and their descendants);
# a shared ancestor is equally related to any weighting of two full sibs.
AX[c("Z1", "Z2"), ]

# Expected average relationship within each contributing group, 0.5 * c'A c.
# Under random mating this equals the mean inbreeding of the resulting progeny.
group_coancestry <- 0.5 * colSums(schemes * AX[rownames(schemes), ])
round(group_coancestry, 5)
```

The three schemes draw on the same two full sibs, yet the genetic consequence
differs sharply. Concentrating all contribution on a single animal
(`Z1_only`) exposes its full, partly inbred genome and inflates the group's
expected progeny inbreeding to
`r round(group_coancestry["Z1_only"], 3)`, whereas spreading contribution
evenly (`Equal`) drives it down to `r round(group_coancestry["Equal"], 3)`.
Trading a little short-term gain for a lower average relationship is exactly
the balance struck by optimal contribution selection. `pedprod()` supplies the
underlying $c'A c$ evaluation for every candidate column in a single $O(np)$
traversal — compared to $O(p \cdot n^2)$ for explicit products with a
precomputed $A$ — and never forms $A$ itself.

### 2.5 Practical Use Cases {#2-5}

#### Optimal Contribution Selection (OCS)

OCS maximizes genetic gain while constraining inbreeding. The core computation
repeatedly evaluates $c'A c$ for candidate contribution vectors $c$:

```{r pedprod_ocs}
# Weighted candidate contributions
candidates <- setNames(c(0.25, 0.25, 0.25, 0.25), c("Z1", "Z2", "A", "B"))
Ac <- pedprod(tped, candidates)

# Average coancestry of the selected group: 0.5 * c' A c
c_accepted <- sum(candidates * Ac[names(candidates)]) / 2
c_accepted
```

With `pedprod()`, this evaluation is fast enough to be embedded inside an
iterative optimization loop over thousands of candidate configurations.

#### Mixed Model Equations

The mixed model equations for BLUP require products $A^{-1} u$ and $A^{-1} Z$.
For an $n \times p$ design matrix $Z$, use:

```{r pedprod_blup}
# Simulated breeding values as a matrix right-hand side
set.seed(20260704)
Z_design <- cbind(
  trait1 = rnorm(nrow(tped)),
  trait2 = rnorm(nrow(tped))
)
rownames(Z_design) <- tped$Ind

# Ainv * Z in one traversal — no Ainv ever stored
Ainv_Z <- pedprod(tped, Z_design, method = "Ainv")
dim(Ainv_Z)
Ainv_Z[1:5, ]
```

#### Founder Contribution Analysis

Isolating one founder at a time turns `pedprod()` into a founder-origin
decomposition: the product $A e_f$ (a unit vector on founder $f$) returns that
founder's expected genetic contribution to every individual. Stacking the unit
vectors into a matrix decomposes the whole population in one traversal, exposing
which founders dominate the current gene pool and which have been lost:

```{r pedprod_geneflow}
# One unit contribution vector per founder: column f isolates founder f
founder_ids <- tped[Gen == 1, Ind]
founder_design <- diag(length(founder_ids))
dimnames(founder_design) <- list(founder_ids, founder_ids)

# A %*% e_f is founder f's expected genetic contribution to every individual
footprint <- pedprod(tped, founder_design, method = "A")

# Founder composition of one candidate: contributions sum to one
round(footprint["Z1", ], 4)

# Average founder contribution to the youngest generation, ranked
young <- tped[Gen == max(Gen), Ind]
founder_share <- sort(colMeans(footprint[young, , drop = FALSE]), decreasing = TRUE)
round(founder_share, 4)
```

Reading across a row (`footprint["Z1", ]`) decomposes an individual's genome
into founder origins that sum to one; reading down a column isolates a single
founder's footprint across the pedigree. Here founder **N** accounts for more
than half of the youngest generation, while founders **I** and **R** have
already fallen to zero — their founder genomes are lost. Such imbalance lowers
the effective number of founders and erodes genetic diversity faster than the
census size implies. Because `pedprod()` never forms $A$, the same
decomposition scales to pedigrees with millions of individuals, where it
underpins the monitoring of founder representation and effective population
size.

### 2.6 Performance and Scalability {#2-6}

The key advantage of `pedprod()` is scalability. The table below summarizes the
computational cost for a pedigree with $n$ individuals and a right-hand side
with $p$ columns:

| $n$ | $p$ | `pedmat("A")` memory | `pedprod()` memory | Feasible with `pedmat`? |
|:---|---:|:---|:---|:---|
| $10^3$ | 1 | ~8 MB | ~8 KB | Yes |
| $10^4$ | 1 | ~800 MB | ~80 KB | Marginal |
| $10^4$ | 10 | ~800 MB | ~800 KB | Marginal |
| $10^5$ | 1 | ~80 GB | ~800 KB | **No** |
| $10^5$ | 100 | ~80 GB | ~8 MB | **No** |
| $10^6$ | 1 | ~8 TB | ~8 MB | **No** |

For pedigrees exceeding ~25 000 individuals, `pedmat(method = "A")` will refuse
to construct a dense $A$ matrix. `pedprod()` has no such limit — it continues to
work as long as the pedigree and right-hand side fit in memory:

```{r pedprod_large_proof, eval=FALSE}
# Pedigrees beyond the dense-A guard still work with pedprod()
n <- 50000L
ids <- paste0("I", seq_len(n))
raw <- data.frame(
  Ind  = ids,
  Sire = c(NA_character_, ids[-n]),
  Dam  = NA_character_,
  stringsAsFactors = FALSE
)
tped_large <- tidyped(raw)

# pedprod works; pedmat would error
result <- pedprod(tped_large, setNames(1, tail(ids, 1)))
length(result)  # 50000
```

**Rule of thumb:** Use `pedprod()` whenever the analysis only requires a matrix
product. Construct $A$, compact matrices, or grouped summaries only when the
corresponding matrix entries are themselves required.

## 3. Inspecting the Matrix {#3}

### 3.1 Summary Statistics {#3-1}

Use the `summary()` method to get an overview of the calculated matrix, including size, density, and average relationship.

```{r matrix_summary}
tail(summary(mat_A),10)
```

### 3.2 Querying Specific Relationships {#3-2}

Instead of manually indexing the matrix, you can use `query_relationship()` to retrieve coefficients by individual IDs.

```{r query}
# Query relationship between Z1 and Z2
query_relationship(mat_A, "Z1", "Z2")

# Query multiple pairs
query_relationship(mat_A, c("Z1", "A"), c("Z2", "B"))
```

## 4. Compact Mode for Large Pedigrees {#4}

For large pedigrees with many full-sibling families (common in aquatic breeding populations), `pedmat()` can merge full siblings into representative nodes to save memory and time.

### 4.1 Using `compact = TRUE` {#4-1}

When `compact = TRUE`, the matrix is calculated for unique representative individuals from each full-sib family.

```{r compact_calc}
# Calculate compacted A matrix
mat_compact <- pedmat(tped, method = "A", compact = TRUE)

# The result is a 'pedmat' object containing the compacted matrix
print(mat_compact[11:20,11:20])
```

### 4.2 Expanding and Querying Compacted Matrices {#4-2}

If you need the full matrix after a compact calculation, use `expand_pedmat()`. For retrieving specific values, `query_relationship()` handles both standard and compact objects transparently.

```{r expand}
# Expand to full 28x28 matrix
mat_full <- expand_pedmat(mat_compact)
dim(mat_full)

# Query still works the same way
query_relationship(mat_compact, "Z1", "Z2")
```

### 4.3 When to Use Compact Mode {#4-3}

Compact mode is highly recommended for:

*   **Large Pedigrees**: More than 5,000 individuals with substantial full-sibling groups.
*   **High-fecundity species**: Such as aquatic animals or plants, where families often have hundreds or thousands of offspring.
*   **Memory-limited environments**: When the full matrix exceeds available RAM.

| Pedigree Size | Full-Sib Proportion | Recommended Mode |
| :--- | :--- | :--- |
| < 1,000 | Any | Standard |
| > 5,000 | < 20% | Standard / Compact |
| > 5,000 | > 20% | **Compact** |

## 5. Visualizing Relationship Matrices with `vismat()` {#5}

Visualization helps in understanding population structure, detecting family clusters, and checking the distribution of genetic relationships.

### 5.1 Relationship Heatmaps {#5-1}

The `"heatmap"` type (default) uses a Nature Genetics style color palette
(White–Orange–Red) to display relationships. Rows and columns are reordered
by hierarchical clustering (Ward.D2) by default, bringing closely related
individuals into contiguous blocks — full-sibs cluster tightly because they
share nearly identical relationship profiles with the rest of the population.

```{r heatmap, fig.width=6, fig.height=6}
# Heatmap of the A matrix (with default clustering reorder)
vismat(mat_A, labelcex = 0.5)
```

#### Compact Matrix — Direct Visualization

A compact `pedmat` object can be passed directly to `vismat()`. It is
automatically expanded to full dimensions before rendering.

```{r heatmap_compact, fig.width=6, fig.height=6}
# Compact matrix: expanded automatically (message printed)
vismat(mat_compact,labelcex=0.5)
```

#### Preserve Pedigree Order

Set `reorder = FALSE` to keep the original pedigree order instead of
re-sorting by clustering.

```{r heatmap_no_reorder, fig.width=6, fig.height=6}
vismat(mat_A, reorder = FALSE, labelcex = 0.5)
```

#### Display a Subset of Individuals

Use `ids` to focus on specific individuals.

```{r heatmap_ids, fig.width=5, fig.height=5}
target_ids <- rownames(as.matrix(mat_A))[1:8]
vismat(mat_A, ids = target_ids,
       main = "Relationship Heatmap — First 8 Individuals")
```

#### Grouping by Pedigree Column

For large populations, aggregate relationships to a group-level view using
the `by` parameter. The matrix is reduced to mean coefficients between groups.

```{r heatmap_group, fig.width=6, fig.height=6}
# Mean relationship between generations
vismat(mat_A, ped = tped, by = "Gen",
       main = "Mean Relationship Between Generations")
```

```{r heatmap_family, fig.width=6, fig.height=6}
# Mean relationship between full-sib families
# (founders without a family assignment are excluded automatically)
vismat(mat_A, ped = tped, by = "Family",
       main = "Mean Relationship Between Full-Sib Families")
```

### 5.2 Inbreeding and Kinship Histograms {#5-2}

The "histogram" type displays the distribution of relationship coefficients (lower triangle) or inbreeding coefficients.

```{r histogram, fig.width=6, fig.height=4}
# Distribution of relationship coefficients
vismat(mat_A, type = "histogram")
```

## 6. Performance Considerations {#6}

Calculation and visualization of large matrices can be resource-intensive.
`vismat()` applies the following automatic optimizations:

| Condition | Behavior |
| :--- | :--- |
| Compact + `by` | Group means are computed directly from the compact matrix (no full expansion) |
| Compact, no `by`, N > 5 000 | Uses compact representative view (labels show `ID (×n)`) |
| Compact, no `by`, N ≤ 5 000 | Matrix is automatically expanded via `expand_pedmat()` |
| N > 2 000 | Hierarchical clustering (reorder) is automatically skipped |
| N > 500 | Individual labels are automatically hidden |
| N > 100 | Grid lines are automatically hidden |

When a compact `pedmat` is used with `by`, `vismat()` computes the group-level
mean relationship matrix algebraically from the K×K compact matrix, including
a sibling off-diagonal correction. This avoids expanding to the full N×N matrix,
making family-level or generation-level visualization feasible even for pedigrees
with hundreds of thousands of individuals.

The example below uses `big_family_size_ped` (178 431 individuals, compact to 2 626)
and displays the mean additive relationship among **all** full-sib families in the
latest generation — a computation that would be infeasible with full expansion.

```{r large_ped_tip, fig.width=9, fig.height=8}
data(big_family_size_ped)

tp_big <- tidyped(big_family_size_ped)
last_gen <- max(tp_big$Gen, na.rm = TRUE)

# Compute the compact A matrix for the entire pedigree
mat_big_compact <- pedmat(tp_big, method = "A", compact = TRUE)

# Focus on all individuals in the last generation that belong to a family
ids_last_gen <- tp_big[Gen == last_gen & !is.na(Family), Ind]

# vismat() aggregates directly from the compact matrix — no expansion needed
vismat(
       mat_big_compact,
       ped = tp_big,
       ids = ids_last_gen,
       by = "Family",
       labelcex = 0.3,
       main = paste("Mean Relationship Between All Families in Generation", last_gen)
)
```

This family-level view reveals the genetic structure among all `r length(unique(tp_big[Gen == last_gen & !is.na(Family), Family]))` families
comprising `r length(ids_last_gen)` individuals, computed in seconds from the
compact matrix.

---
**See Also:**
- `vignette("tidy-pedigree", package = "visPedigree")`
- `vignette("draw-pedigree", package = "visPedigree")`

## References

- Colleau, J. J. (2002). An indirect approach to the extensive calculation of relationship coefficients. *Genetics Selection Evolution*, 34, 409-421.
- Henderson, C. R. (1976). A simple method for computing the inverse of a numerator relationship matrix used in prediction of breeding values. *Biometrics*, 32(1), 69-83.
- Meuwissen, T. H. E., & Luo, Z. (1992). Computing inbreeding coefficients in large populations. *Genetics Selection Evolution*, 24, 305-313.
- Sargolzaei, M., Iwaisaki, H., & Colleau, J.-J. (2005). A fast algorithm for computing inbreeding coefficients in large populations. *Journal of Animal Breeding and Genetics*, 122, 325-331.

---
title: "4. Pedigree Analysis and Population Genetics"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{4. Pedigree Analysis and Population Genetics}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette summarizes a practical workflow for pedigree analysis with
`visPedigree`, with emphasis on the interpretation of the main indicators used
in breeding and conservation genetics.

The discussion is organized around five questions:

1. How complete and deep is the pedigree?
2. How long is the generation cycle?
3. Has genetic diversity been eroded by unequal founder use, bottlenecks, or drift?
4. How large is the effective population size under different definitions?
5. Are relationship, inbreeding, subpopulation structure, and gene flow under control?

## 1. Setup and Data Preparation

Different package datasets are used for different analytical tasks.

- `deep_ped` is useful for pedigree depth and diversity summaries.
- `big_family_size_ped` contains a `Year` column and is suitable for generation intervals.
- `small_ped` is convenient for relationship examples.
- `inbred_ped` is convenient for inbreeding and classification examples.

```{r setup}
library(visPedigree)
library(data.table)

data(deep_ped, package = "visPedigree")
data(big_family_size_ped, package = "visPedigree")
data(small_ped, package = "visPedigree")
data(inbred_ped, package = "visPedigree")

tp_deep <- tidyped(deep_ped)
tp_small <- tidyped(small_ped)
tp_inbred <- tidyped(inbred_ped)
```

## 2. Pedigree Overview with `pedstats()`

`pedstats()` provides a compact structural summary with three components:

- `$summary`: pedigree size and parental structure.
- `$ecg`: pedigree completeness and ancestral depth.
- `$gen_intervals`: generation intervals, only when a usable time column is available.

Here `deep_ped` has no explicit birth-date column. Accordingly,
`pedstats()` returns `$summary` and `$ecg`, whereas `$gen_intervals` remains
`NULL`.

```{r pedstats-overview}
stats_deep <- pedstats(tp_deep)

stats_deep$summary
tail(stats_deep$ecg)
stats_deep$gen_intervals

# Visualize ancestral depth (Equivalent Complete Generations)
plot(stats_deep, type = "ecg", metric = "ECG")
```

Pedigree depth and pedigree time are related but distinct quantities. The former
is addressed by `pedecg()`, whereas the latter is addressed by `pedgenint()`.

## 3. Pedigree Completeness with `pedecg()`

Equivalent complete generations (ECG) summarize the amount of ancestral
information available for each individual:

$$
ECG_i = \sum_j \left(\frac{1}{2}\right)^{g_{ij}}
$$

where $g_{ij}$ is the number of generations between individual $i$ and ancestor
$j$. ECG increases with both pedigree depth and pedigree completeness.

In practice:

- `ECG` combines depth and completeness.
- `FullGen` counts how many fully known ancestral generations exist.
- `MaxGen` records the deepest known ancestral path.

`ECG` is especially useful because it provides the depth adjustment required by
realized effective population size estimators based on inbreeding or coancestry.

```{r pedecg-example}
ecg_deep <- pedecg(tp_deep)

ecg_deep[order(-ECG)][1:10]
```

## 4. Generation Intervals with `pedgenint()`

Generation interval is the age of a parent at the birth of its offspring:

$$
L = t_{offspring} - t_{parent}
$$

`pedgenint()` estimates this quantity for seven pathway summaries:

- `SS`, `SD`, `DS`, `DD`: sex-specific gametic pathways.
- `SO`, `DO`: sex-independent sire-offspring and dam-offspring pathways.
- `Average`: all parent-offspring pairs combined.

The function accepts `Date`/`POSIXct` columns, date strings, or numeric years.
When only an integer year is available, it is converted internally to
`YYYY-07-01` as a mid-year approximation.

```{r pedgenint-basic}
tp_time <- tidyped(big_family_size_ped)

genint_year <- pedgenint(tp_time, timevar = "Year", unit = "year")
genint_year

# Visualize generation intervals
# Note: we can also use stats <- pedstats(tp_time, timevar = "Year") followed by plot(stats)
plot(genint_year)
```

The optional `cycle` parameter adds `GenEquiv`, which compares the observed mean
interval with a target breeding cycle:

$$
GenEquiv_i = \frac{\bar{L}_i}{L_{cycle}}
$$

Values larger than 1 indicate that the realized generation interval exceeds the
target cycle.

```{r pedgenint-cycle}
genint_cycle <- pedgenint(tp_time, timevar = "Year", unit = "year", cycle = 1.2)

genint_cycle[Pathway %in% c("SS", "SD", "DS", "DD", "Average")]
```

## 5. Subpopulation Structure with `pedsubpop()`

Before interpreting diversity or relationship metrics, it is useful to check
whether the pedigree forms a single connected population or a mixture of
separate components.

`pedsubpop()` has two modes:

- `by = NULL`: summarize disconnected pedigree components.
- `by = "..."`: summarize the pedigree by an existing grouping variable.

```{r pedsubpop-example}
ped_demo <- data.table(
  Ind = c("A", "B", "C", "D", "E", "F", "G", "H"),
  Sire = c(NA, NA, "A", NA, NA, "E", NA, NA),
  Dam  = c(NA, NA, "B", NA, NA, NA, NA, NA),
  Sex = c("male", "female", "male", "female", "male", "female", "male", "female"),
  Batch = c("L1", "L1", "L1", "L1", "L2", "L2", "L3", "L3")
)

tp_demo <- tidyped(ped_demo)

pedsubpop(tp_demo)
pedsubpop(tp_demo, by = "Batch")
```

This is a compact summary tool. When the actual split pedigree objects are
needed for downstream analysis, use `splitped()`.

## 6. Diversity Indicators with `pediv()`

`pediv()` is the integrated diversity summary. It combines:

- founder and ancestor contributions (`fe`, `fa`),
- founder genome equivalents (`fg`),
- retained genetic diversity (`GeneDiv`),
- three effective population size estimates (`Ne`).

The gene-origin summaries follow the classical founder/ancestor framework of
Lacy (1989) and Boichard et al. (1997), while coancestry-based diversity and
effective population size calculations follow Caballero & Toro (2000) and
related pedigree-based estimators.

All of these quantities depend on the definition of the **reference
population**. In the present example, the reference population is defined as the
most recent two generations in `deep_ped`.

```{r pediv-setup}
ref_pop <- tp_deep[Gen >= max(Gen) - 1, Ind]
length(ref_pop)
```

```{r pediv-analysis}
div_res <- pediv(tp_deep, reference = ref_pop, top = 10, seed = 42L)

div_res$summary
div_res$ancestors
```

### 6.1 `fe`, `fa`, and `fg`: what do they measure?

These three indicators describe diversity loss from complementary angles.

#### Effective number of founders (`fe`)

$$
f_e = \frac{1}{\sum_{i=1}^{f} p_i^2}
$$

where $p_i$ is the expected contribution of founder $i$ to the reference
population. `fe` answers the question: if all founders had contributed equally,
how many founders would generate the same diversity?

Use `fe` when the main concern is **unequal founder representation**.

#### Effective number of ancestors (`fa`)

$$
f_a = \frac{1}{\sum_{j=1}^{a} q_j^2}
$$

where $q_j$ is the **marginal** contribution of ancestor $j$ after removing the
contributions already explained by more influential ancestors. `fa` is usually
smaller than `fe` when bottlenecks occurred.

Use `fa` when the main concern is **genetic bottlenecks caused by a limited set
of influential ancestors**.

#### Founder genome equivalents (`fg`)

In its simplest interpretation,

$$
f_g = \frac{1}{2\bar{C}}
$$

where $\bar{C}$ is the mean coancestry of the reference population. In
`visPedigree`, `fg` is estimated from a diagonal-corrected mean coancestry:

$$
\hat{\bar{C}} = \frac{N-1}{N}\cdot\frac{\bar{a}_{off}}{2} + \frac{1 + \bar{F}}{2N}
$$

$$
f_g = \frac{1}{2\hat{\bar{C}}}
$$

where $N$ is the size of the full reference population, $\bar{a}_{off}$ is the mean
off-diagonal additive relationship among sampled individuals, and $\bar{F}$
is their mean inbreeding coefficient.

Use `fg` when the main concern is **the amount of founder genome still surviving
after unequal use, bottlenecks, and drift**. In practice, `fg` is often the
most conservative diversity indicator.

#### Retained genetic diversity (`GeneDiv`)

`GeneDiv` is the pedigree-based retained genetic diversity of the reference
population:

$$
GeneDiv = 1 - \bar{C}
$$

where $\bar{C}$ is the same diagonal-corrected mean coancestry used to compute
$f_g$. Because coancestry increases as individuals become more related by
descent, $GeneDiv$ decreases towards 0 as the population becomes more uniform.
An unrelated, non-inbred population has $GeneDiv = 1$; a population of full
sibs from two unrelated founders has $GeneDiv \approx 0.75$.

`GeneDiv` and `fg` are both derived from $\bar{C}$, but they answer different
questions.  `fg` is measured in "equivalent number of founders" and is most
useful for between-programme comparisons.  `GeneDiv` is a dimensionless
proportion and is easiest to communicate to managers and stakeholders who need
a single intuitive index of diversity retention.

```{r pediv-genediv}
# GeneDiv is in the summary alongside MeanCoan
div_res$summary[, .(fg, MeanCoan, GeneDiv)]
```

### 6.2 Shannon-entropy effective numbers: `feH` and `faH`

The classical `fe` and `fa` are based on $\sum p_i^2$ (Hill number order
$q = 2$; Hill, 1973), which is disproportionately influenced by a few large
contributions. `visPedigree` also reports the information-theoretic counterpart
at order $q = 1$, derived from Shannon entropy (Jost, 2006):

$$
f_e^{(H)} = \exp\!\Bigl(-\sum_{i=1}^{f} p_i \ln p_i\Bigr), \qquad
f_a^{(H)} = \exp\!\Bigl(-\sum_{j=1}^{a} q_j \ln q_j\Bigr)
$$

All four metrics satisfy a strict monotone ordering:

$$
N_f \ge f_e^{(H)} \ge f_e, \qquad K \ge f_a^{(H)} \ge f_a
$$

where $N_f$ is the total number of founders and $K$ is the number of ancestors
considered.

#### Metric guide

| Metric | Order | Sensitivity | What it measures |
|---|---|---|---|
| `fe` | $q = 2$ | Dominated by common founders | Effective number of founders if contributions were equalized. Dominated by the largest contributions; many rare founders with small $p_i$ have negligible influence on the metric. |
| `feH` | $q = 1$ | Balanced; sensitive to rare founders | Shannon-entropy effective founders. Counts rare founders more equitably than $q = 2$; always $\ge$ `fe`. |
| `fa` | $q = 2$ | Dominated by common ancestors | Effective number of ancestors accounting for bottleneck structure. Typically lower than `fe`. |
| `faH` | $q = 1$ | Balanced; sensitive to rare ancestors | Shannon-entropy effective ancestors. Less dominated by the single most influential ancestor. |
| `fg` | — | Realized coancestry | Founder genome equivalents; the most conservative indicator because it captures drift as well. |
| `GeneDiv` | — | Retained diversity | Pedigree-based retained genetic diversity: $1 - \bar{C}$. Values on the $[0, 1]$ scale; higher values indicate more diversity retained relative to the base population. |

#### Interpreting the ratio $f_e^{(H)} / f_e$

The ratio $\rho = f_e^{(H)} / f_e$ is a **long-tail signal**:

- $\rho \approx 1$: founder contributions are already concentrated in a small
  number of individuals; there is no meaningful "hidden" diversity from minor
  founders.
- $\rho \gg 1$: many minor founders still carry genetic material into the
  reference population but are effectively invisible to the classical $f_e$
  metric because their contributions are small.  This situation is common early
  in a breeding programme when the pedigree is still shallow.

The analogous ratio $f_a^{(H)} / f_a$ diagnoses the same effect at the
**ancestor** level: a large ratio means that the bottleneck signal from the
dominant ancestor is masking genuine contributions from many secondary
ancestors.

#### Management implications

Use the two ratios together to set priorities:

1. If $f_e^{(H)} / f_e$ is large but $f_e$ itself is already high, the
   programme has good founder coverage; no immediate action needed.
2. If $f_e^{(H)} / f_e$ is large and $f_e$ is low, minor founders carry
   material that could be mobilized — consider deliberately increasing their
   representation in the next generation.
3. If $f_a^{(H)} / f_a$ is large, secondary ancestors are still present but
   marginalized; diversifying sire selection could rebalance the gene pool
   before their lineages are lost entirely.

```{r pediv-shannon}
# Shannon metrics are included in pediv() output
div_res$summary[, .(NFounder, feH, fe, NAncestor, faH, fa)]

# The ratio feH/fe reveals long-tail founder diversity
div_res$summary[, .(rho_founder = feH / fe, rho_ancestor = faH / fa)]
```

```{r pedcontrib-shannon}
# pedcontrib() provides the same metrics via its summary
contrib_res <- pedcontrib(tp_deep, reference = ref_pop, mode = "both")
contrib_res$summary[c("n_founder", "f_e_H", "f_e", "n_ancestor", "f_a_H", "f_a")]
```

## 7. Diversity Dynamics over Time with `pedhalflife()`

`pediv()` provides a static snapshot of diversity. When multiple time points
(generations, years, etc.) are available, `pedhalflife()` tracks how $f_e$,
$f_a$, and $f_g$ evolve and fits a log-linear decay model to quantify the
**rate of diversity loss**.

The total loss rate $\lambda_{total}$ is decomposed into three additive
components that each have a distinct biological meaning:

| Component | Symbol | Source of loss |
|---|---|---|
| Foundation | $\lambda_e$ | Unequal founder contributions |
| Bottleneck | $\lambda_b$ | Overuse of key ancestors |
| Drift | $\lambda_d$ | Random sampling loss |

The **diversity half-life** is the number of time units for $f_g$ to halve:
$$T_{1/2} = \frac{\ln 2}{\lambda_{total}}$$

```{r pedhalflife-analysis}
hl <- pedhalflife(tp_deep, timevar = "Gen")
print(hl)
```

The printed half-life shows the unit in parentheses — `(Gen)` here — taken
directly from the `timevar` argument.  The full object exposes three components:

- `hl$timeseries` — per-time-point table (see column guide below).
- `hl$decay` — single-row table with `LambdaE`, `LambdaB`, `LambdaD`,
  `LambdaTotal`, and `THalf`.
- `hl$timevar` — the `timevar` string passed to the call.

### 7.1 Column guide for `$timeseries`

The cascade rests on a simple algebraic identity:

$$
\ln f_g = \underbrace{\ln f_e}_{\texttt{lnfe}}
         + \underbrace{\ln\!\left(\frac{f_a}{f_e}\right)}_{\texttt{lnfafe}}
         + \underbrace{\ln\!\left(\frac{f_g}{f_a}\right)}_{\texttt{lnfgfa}}
$$

Because OLS is linear, the slope of $\ln f_g$ versus time equals the sum of the
slopes of the three right-hand terms.  This guarantees exact additivity:
$\lambda_{total} = \lambda_e + \lambda_b + \lambda_d$.

| Column | Formula | What it measures |
|---|---|---|
| `lnfe` | $\ln f_e$ | Log effective-founder diversity. Declines when founder contributions become more unequal over time. |
| `lnfa` | $\ln f_a$ | Log effective-ancestor diversity. Lies below `lnfe` whenever bottleneck ancestors have been over-used. |
| `lnfg` | $\ln f_g$ | Log founder-genome equivalents. The most comprehensive diversity signal; integrates founder use, bottlenecks, and drift. |
| `lnfafe` | $\ln(f_a/f_e)$ | **Bottleneck gap.** When this ratio decreases over time, a small number of ancestors dominate the gene pool even beyond what unequal founder use would predict. |
| `lnfgfa` | $\ln(f_g/f_a)$ | **Drift gap.** Captures random allele loss due to finite population size. Because $f_g \le f_a$ always, this term is $\le 0$ and becomes more negative as genetic drift accumulates. |
| `TimeStep` | numeric | OLS time axis (same as `Time` for numeric `timevar`; sequential indices otherwise). |

### 7.2 Interpreting the $\lambda$ components

The negative OLS slope of each log column gives the corresponding loss rate.
All three terms have a distinct policy implication:

**$\lambda_e$ (Foundation)** — rate of increase in founder-contribution
inequality.  A large $\lambda_e$ suggests that founder representation is still
diverging, possibly because some founders are reaching later generations
through many more breeding pathways than others.  Remedy: broaden founder
use in the breeding programme.

**$\lambda_b$ (Bottleneck)** — rate at which a few key ancestors capture
an ever-larger share of the gene pool beyond what founder inequality alone
explains.  A large $\lambda_b$ points to systematic over-selection of certain
families or sires.  Remedy: restrict the number of offspring per
high-ranking animal; rotate breeding males.

**$\lambda_d$ (Drift)** — rate of random allele loss in finite populations.
This component is always present and increases with small effective population
size.  It is the only component that cannot be reversed by changing selection
decisions; it can only be slowed by maintaining a larger breeding population
or by cryopreservation of genetic material.

When $\lambda_d \gg \lambda_e + \lambda_b$ (as in the example above), drift
is the dominant driver of diversity loss and should be the primary management
target.

### 7.3 Interpreting $T_{1/2}$

$T_{1/2}$ is the number of time units (generations, years, etc.) required for
$f_g$ to fall to half its current value, assuming the observed decay rate
continues.  It is a single headline figure that combines all three loss
channels.

As a rough benchmark: a half-life shorter than five to ten generations is
generally considered a cause for concern in conservation and aquaculture
genetics.  Breed registries and conservation programmes often use $T_{1/2}$
as a threshold indicator for triggering corrective management.

```{r pedhalflife-timeseries}
hl$timeseries
```

To focus on a specific time window, pass the desired values to the `at`
argument:

```{r pedhalflife-subset}
hl_recent <- pedhalflife(tp_deep, timevar = "Gen",
                         at = tail(sort(unique(tp_deep$Gen)), 4))
print(hl_recent)
```

Plotting the object shows the log-linear decay of each diversity metric:

```{r pedhalflife-plot, fig.width=6, fig.height=4}
plot(hl, type = "log")
```

A `type = "raw"` plot shows the equivalent numbers on the original scale:

```{r pedhalflife-plot-raw, fig.width=6, fig.height=4}
plot(hl, type = "raw")
```

## 8. Effective Population Size with `pedne()` and `pediv()`

`pediv()` reports three complementary effective population size definitions.
The pedigree-based estimators follow the individual-rate approach of Gutiérrez
et al. (2008, 2009) and the coancestry-rate approach of Cervantes et al. (2011).
Each addresses a distinct biological question.

### 8.1 Demographic `Ne`

$$
N_e = \frac{4 N_m N_f}{N_m + N_f}
$$

where $N_m$ and $N_f$ are the numbers of contributing males and females. This is
the easiest estimate to understand, but it ignores realized pedigree structure,
inbreeding, and drift. It is therefore often optimistic.

### 8.2 Inbreeding `Ne`

$$
\Delta F_i = 1 - (1 - F_i)^{1/(ECG_i - 1)}
$$

$$
N_e = \frac{1}{2\overline{\Delta F}}
$$

This estimator uses the realized rate of inbreeding (Gutiérrez et al., 2008,
2009). `ECG` standardizes individuals with different pedigree depths. Use this
estimate when the primary concern is **the rate of inbreeding accumulation**.

### 8.3 Coancestry `Ne`

$$
\Delta c_{ij} = 1 - (1 - c_{ij})^{1/(\frac{ECG_i + ECG_j}{2})}
$$

$$
N_e = \frac{1}{2\overline{\Delta c}}
$$

This estimator is based on the rate of coancestry among members of the reference
population (Cervantes et al., 2011). Because it captures the accumulation of
relatedness before it is fully expressed as realized inbreeding, it is often the
strictest and most sensitive warning signal in managed breeding populations.

## 9. Average Relationship Trends with `pedrel()`

`pedrel()` summarizes pairwise relatedness within groups. Two complementary
scales are available via the `scale` argument.

### 9.1 Mean additive relationship (`scale = "relationship"`)

The default scale returns the mean off-diagonal additive relationship:

$$
MeanRel = \frac{1}{n(n-1)} \sum_{i \ne j} a_{ij}
$$

where $a_{ij} = 2c_{ij}$ is the additive relationship and $c_{ij}$ is the
coancestry (kinship coefficient) between individuals $i$ and $j$.
A higher `MeanRel` means that individuals within the group are, on average,
more related by descent.

Internally, `pedrel()` creates an indicator vector $g$ for each group and
computes $Ag$ directly from the pedigree. The off-diagonal sum is obtained
from $g^\mathsf{T}Ag$ after subtracting
$\sum_{i:g_i=1}(1 + F_i)$. The dense relationship matrix is never
materialized, so `compact` is now retained only as a backward-compatible
no-op argument.

```{r pedrel-example}
tp_small$BirthYear <- 2010 + tp_small$Gen

rel_by_gen <- pedrel(tp_small, by = "Gen")
rel_by_gen
```

The `reference` argument lets you focus on a subset of interest inside each
group, such as candidate breeders.

```{r pedrel-reference}
ref_ids <- c("Z1", "Z2", "X", "Y")

pedrel(tp_small, by = "Gen", reference = ref_ids)
```

### 9.2 Corrected mean coancestry (`scale = "coancestry"`)

When `scale = "coancestry"`, `pedrel()` returns the diagonal-corrected
population mean coancestry following Caballero & Toro (2000):

$$
\bar{C} = \frac{g^\mathsf{T}Ag}{2N^2}
        = \frac{N - 1}{N} \cdot \frac{\bar{a}_{off}}{2}
          + \frac{1 + \bar{F}}{2N}
$$

where $\bar{a}_{off}$ is the mean off-diagonal relationship, $\bar{F}$ is the
mean inbreeding coefficient of the group, and $N$ is the group size. This
$\bar{C}$ includes both off-diagonal pairwise coancestry and diagonal
self-coancestry. It matches the
internal coancestry quantity used by `pediv()` to derive $f_g$.

```{r pedrel-coancestry}
coan_by_gen <- pedrel(tp_small, by = "Gen", scale = "coancestry")
coan_by_gen
```

Use `scale = "relationship"` to track mean pairwise relatedness trends.
Use `scale = "coancestry"` when you need a population-level coancestry
consistent with diversity metrics from `pediv()`.

`pedrel(scale = "coancestry")` applies the same diagonal-corrected formula
as `pediv()`. The resulting `MeanCoan` values are therefore directly
comparable to `pediv()$summary$MeanCoan`, and `GeneDiv = 1 - MeanCoan`
translates the trend table from `pedrel()` into a retained-diversity
timeline without any additional computation.

`MeanRel` and coancestry-based `Ne` are conceptually linked, but they are not
identical summaries. `MeanRel` is an average additive relationship within a
group, whereas coancestry-based `Ne` is derived from the *rate of increase* in
coancestry across pedigree depth.

## 10. Inbreeding Trends with `inbreed()` and `pedfclass()`

Wright's inbreeding coefficient $F$ (Wright, 1922, 1931) is the probability
that the two alleles of an individual are identical by descent. A practical
starting point is to inspect the mean trend by generation.

```{r inbreed-trend}
tp_inbred_f <- inbreed(tp_inbred)

f_trend <- tp_inbred_f[, .(MeanF = mean(f, na.rm = TRUE)), by = Gen]
f_trend
```

For classification by inbreeding severity, `pedfclass()` can be applied
directly to the pedigree object. If the inbreeding coefficient is not yet
present, the function computes it internally.

```{r pedfclass-example}
pedfclass(tp_inbred)
```

Custom reporting thresholds can be supplied through `breaks`.

```{r pedfclass-breaks}
pedfclass(tp_inbred, breaks = c(0.03125, 0.0625, 0.125, 0.25))
```

The default thresholds correspond approximately to familiar pedigree scenarios:

- $F = 0.0625$: half-sib mating.
- $F = 0.125$: avuncular or grandparent-grandchild mating.
- $F = 0.25$: full-sib or parent-offspring mating.

## 11. Gene Flow and Partial Inbreeding

### 11.1 `pedancestry()`: founder-line proportions

`pedancestry()` tracks how founder groups contribute to later descendants. This
is useful when founders are labeled by line, strain, or geographic origin.

```{r pedancestry-example}
ped_line <- data.table(
  Ind = c("A", "B", "C", "D", "E", "F", "G"),
  Sire = c(NA, NA, NA, NA, "A", "C", "E"),
  Dam  = c(NA, NA, NA, NA, "B", "D", "F"),
  Sex = c("male", "female", "male", "female", "male", "female", "male"),
  Line = c("Line1", "Line1", "Line2", "Line2", NA, NA, NA)
)

tp_line <- tidyped(ped_line)

anc <- pedancestry(tp_line, foundervar = "Line")
anc
```

### 11.2 `pedpartial()`: which ancestors explain inbreeding?

`pedpartial()` decomposes total inbreeding into contributions from selected
ancestors. It is useful for identifying which ancestors are most responsible
for the observed inbreeding.

```{r pedpartial-example}
partial_f <- pedpartial(tp_inbred, ancestors = c("A", "B"))
partial_f
```

## 12. Practical Interpretation

One useful interpretation order is:

1. Check pedigree depth and completeness with `pedecg()`.
2. Check generation timing with `pedgenint()`.
3. Quantify static diversity with `fe`, `fa`, `fg`, and `GeneDiv` via
   `pediv()`; compare with `feH` and `faH` to assess long-tail founder value.
   Use `GeneDiv` (= $1 - \bar{C}$) as the headline retained-diversity index
   for management reporting.
4. Track diversity dynamics over time with `pedhalflife()`. Inspect
   $\lambda_e$, $\lambda_b$, and $\lambda_d$ to identify the dominant driver
   of loss; use $T_{1/2}$ as a headline indicator for management reporting.
5. Compare demographic, inbreeding, and coancestry-based `Ne`.
6. Monitor `MeanRel` and `MeanF` over time.
7. Use `pedsubpop()`, `pedancestry()`, and `pedpartial()` to diagnose structure,
   introgression, and bottlenecks.

## References

- Boichard, D., Maignel, L., & Verrier, É. (1997). The value of using probabilities of gene origin to measure genetic variability in a population. *Genetics Selection Evolution*, 29(1), 5-23.
- Caballero, A., & Toro, M. A. (2000). Interrelations between effective population size and other pedigree tools for the management of conserved populations. *Genetical Research*, 75(3), 331-343.
- Cervantes, I., Goyache, F., Molina, A., Valera, M., & Gutiérrez, J. P. (2011). Estimation of effective population size from the rate of coancestry in pedigreed populations. *Journal of Animal Breeding and Genetics*, 128(1), 56-63.
- Gutiérrez, J. P., Cervantes, I., Molina, A., Valera, M., & Goyache, F. (2008). Individual increase in inbreeding allows estimating effective sizes from pedigrees. *Genetics Selection Evolution*, 40(4), 359-378.
- Gutiérrez, J. P., Cervantes, I., & Goyache, F. (2009). Improving the estimation of realized effective population sizes in farm animals. *Journal of Animal Breeding and Genetics*, 126(4), 327-332.
- Hill, M. O. (1973). Diversity and evenness: a unifying notation and its consequences. *Ecology*, 54(2), 427-432.
- Jost, L. (2006). Entropy and diversity. *Oikos*, 113(2), 363-375.
- Lacy, R. C. (1989). Analysis of founder representation in pedigrees: founder equivalents and founder genome equivalents. *Zoo Biology*, 8(2), 111-123.
- Wright, S. (1922). Coefficients of inbreeding and relationship. *The American Naturalist*, 56(645), 330-338.
- Wright, S. (1931). Evolution in Mendelian populations. *Genetics*, 16(2), 97-159.

---
title: "6. tidyped Class Structure and Extension Notes"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{6. tidyped Class Structure and Extension Notes}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

This document describes the structural contract of the `tidyped` class in
visPedigree 1.8.0. It is intended for maintenance and extension work.

## 1. Class identity

`tidyped` is an S3 class layered on top of `data.table`.

Expected class vector:

```r
c("tidyped", "data.table", "data.frame")
```

The class is created through `new_tidyped()` (internal constructor) and checked
with `is_tidyped()`.

## 2. Core design goals

`tidyped` is designed to be:

1. **safe for C++**: integer pedigree indices (`IndNum`, `SireNum`, `DamNum`)
  are always aligned with row order, so C++ routines can index directly
  without translation;
2. **fast for large pedigrees**: the fast path skips redundant validation when
  the input is already a `tidyped`;
3. **compatible with `data.table`**: in-place modification via `:=` and `set()`
  preserves class and metadata without copying;
4. **explicit about structural degradation**: row subsets that break pedigree
  completeness are downgraded to plain `data.table` with a warning.

## 3. The head invariant: IndNum == row index

The single most important structural rule in visPedigree:

> **`IndNum[i]` must equal `i` for every row.**

This means `SireNum` and `DamNum` are direct row pointers: the sire of
individual `i` lives at row `SireNum[i]`, and `0L` encodes a missing parent.

Every C++ function in visPedigree — inbreeding coefficients, relationship
matrices, BFS tracing, topological sorting — relies on this invariant. If it
breaks, C++ will read wrong parents.

This invariant is enforced at three levels:

- **`tidyped()`**: builds indices from scratch during construction.
- **`[.tidyped`**: rebuilds indices in-place after valid row subsets.
- **`ensure_tidyped()` / `ensure_complete_tidyped()`**: detect and repair
  stale indices when class was accidentally dropped.

## 4. Column contract

### 4.1 Minimal structural columns

These four columns define a valid `tidyped`:

| Column | Type      | Description                          |
|--------|-----------|--------------------------------------|
| `Ind`  | character | Unique individual ID                 |
| `Sire` | character | Sire ID, `NA` for unknown            |
| `Dam`  | character | Dam ID, `NA` for unknown             |
| `Sex`  | character | `"male"`, `"female"`, or `"unknown"` |

Checked by `validate_tidyped()`.

### 4.2 Integer pedigree columns

| Column    | Type    | Description                         |
|-----------|---------|-------------------------------------|
| `IndNum`  | integer | Row index (== row number, see §3)   |
| `SireNum` | integer | Row index of sire, `0L` for missing |
| `DamNum`  | integer | Row index of dam, `0L` for missing  |

These exist whenever `tidyped()` is called with `addnum = TRUE` (default).
They are the interface between R and C++.

### 4.3 Other common columns

| Column       | Description                                  |
|--------------|----------------------------------------------|
| `Gen`        | Generation number                            |
| `Family`     | Family group code                            |
| `FamilySize` | Number of offspring in the family            |
| `Cand`       | `TRUE` for candidate individuals             |
| `f`          | Inbreeding coefficient (added by `inbreed()`) |

### 4.4 Column naming convention

All data columns use **PascalCase** (`Ind`, `SireNum`, `MeanF`, `ECG`),
matching the core column style.

## 5. Metadata layer

Pedigree-level metadata is stored in a single attribute:

```r
attr(x, "ped_meta")
```

Built by `build_ped_meta()`, accessed by `pedmeta()`.

| Field              | Type      | Description                             |
|--------------------|-----------|-----------------------------------------|
| `selfing`          | logical   | Whether self-fertilization mode was used |
| `bisexual_parents` | character | IDs appearing as both sire and dam       |
| `genmethod`        | character | `"top"` or `"bottom"` generation numbering |

No other pedigree-level attributes should be added outside `ped_meta`.

## 6. Structural invariants

The following invariants must hold for a valid `tidyped`:

1. **IndNum == row index** (see §3).
2. **Ind is unique** — no duplicate individual IDs.
3. **Completeness** — every non-`NA` Sire and Dam appears in `Ind`.
4. **Acyclicity** — no individual is its own ancestor.
5. **SireNum / DamNum consistency** — `0L` for missing parents, valid row
  indices otherwise.
6. **ped_meta is the sole metadata container** — no scattered attributes.

Invariants 1–5 are established by `tidyped()` and guarded by `[.tidyped`.
Invariant 6 is a development convention.

## 7. Constructor pipeline

`tidyped()` currently has two distinct tracing paths:

- **Raw-input path** (`data.frame` / `data.table`) — uses igraph for loop
  detection, candidate tracing, and topological sorting before integer indices
  are finalized.
- **Fast path** (`tidyped` + `cand`) — skips graph rebuilding and uses C++ for
  candidate tracing and topological sorting on existing integer pedigree
  indices.

### 7.1 Full path: `tidyped(raw_input)`

When the input is a raw `data.frame` or `data.table`:

1. `validate_and_prepare_ped()` — normalize IDs, detect duplicates and
  bisexual parents, inject missing founders.
2. Loop detection — igraph builds a directed graph and checks `is_dag()`;
  `which_loop()` and `shortest_paths()` are used only on the error path to
  report informative loop diagnostics.
3. Candidate tracing — if `cand` is supplied, igraph neighborhood search is
  used on the raw-input path.
4. Topological sort — igraph `topo_sort()` on the raw-input path.
5. Generation assignment — C++ (`cpp_assign_generations_top` /
  `cpp_assign_generations_bottom`) using the pedigree implied by the sorted
  rows.
6. Sex inference — resolve unknowns from parental roles.
7. Build integer indices — `IndNum`, `SireNum`, `DamNum`.
8. `new_tidyped()` + attach `ped_meta`.

### 7.2 Fast path: `tidyped(tp, cand = ids)`

When the input is already a `tidyped` **and** `cand` is supplied:

- **Skipped**: ID validation, loop detection, sex inference, founder injection.
- **Executed**: C++ BFS tracing → C++ topo sort → C++ generation assignment →
  rebuild indices → `new_tidyped()` + `ped_meta`.

The fast path is the preferred workflow for repeated local tracing from a
previously validated master pedigree:

```r
tp_master <- tidyped(raw_ped)
tp_local  <- tidyped(tp_master, cand = ids, trace = "up", tracegen = 3)
```

### 7.3 `new_tidyped()` — internal constructor

`new_tidyped()` attaches the `"tidyped"` class via `setattr()` (no copy) and
clears data.table's invisible flag via `x[]`. It does **not** attach
`ped_meta` — that is the caller's responsibility. It should only be called when
the caller has already ensured structural validity.

## 8. Three-tier guard system

Analysis functions must guard their inputs. visPedigree provides three guard
levels, chosen based on what each function needs.

### 8.1 `validate_tidyped()` — visualization guard

- Attempts silent class recovery via `ensure_tidyped()`.
- Checks only that `Ind`, `Sire`, `Dam`, `Sex` exist.
- **Does not require** pedigree completeness.
- Used by: `visped()`, `plot.tidyped()`, `summary.tidyped()`.

### 8.2 `ensure_tidyped()` — structure-light guard

- If already `tidyped`: returns as-is.
- If class was dropped but 8 core columns (`Ind`, `Sire`, `Dam`, `Sex`, `Gen`,
  `IndNum`, `SireNum`, `DamNum`) are present: rebuilds `IndNum` if stale,
  restores class, emits a message.
- **Does not check** pedigree completeness.
- Used by: `pedsubpop()`, `splitped()`, `pedne(method = "demographic")`,
  `pedstats(ecg = FALSE, genint = FALSE)`, `pedfclass()` (when `f` column
  already exists).

### 8.3 `ensure_complete_tidyped()` — complete-pedigree guard

- Everything `ensure_tidyped()` does, **plus**:
- Calls `require_complete_pedigree()` — verifies that every non-`NA` Sire/Dam
  is present in `Ind`. Stops with an error if not.
- Required by any function that recurses through pedigree structure in C++.
- Used by: `inbreed()`, `pedecg()`, `pedgenint()`, `pedrel()`,
  `pedne(method = "inbreeding" | "coancestry")`, `pedcontrib()`,
  `pedancestry()`, `pedfclass()` (when `f` must be computed), `pedpartial()`,
  `pediv()`, `pedmat()`, `pedhalflife()`.

### 8.4 Choosing the right guard

| Guard                       | Recovers class? | Requires completeness? | When to use                   |
|-----------------------------|:---------------:|:----------------------:|-------------------------------|
| `validate_tidyped()`        | yes             | no                     | Visualization                 |
| `ensure_tidyped()`          | yes             | no                     | Summaries on existing columns |
| `ensure_complete_tidyped()` | yes             | **yes**                | Pedigree recursion in C++     |

Some functions are **conditionally guarded**: they use `ensure_tidyped()` by
default but escalate to `ensure_complete_tidyped()` when a parameter triggers
pedigree recursion (for example `pedstats(ecg = TRUE)`,
`pedne(method = "coancestry")`).

## 9. Safe subsetting contract

`[.tidyped` is the key protection layer.

### 9.1 `:=` operations

Modify-by-reference is passed through safely. Class and metadata are preserved
via `setattr()`. No copy occurs.

### 9.2 Column-only selections

If the selection removes core pedigree columns, the result is returned as a
plain `data.table` without warning.

### 9.3 Row subsets

After row subsetting, `[.tidyped` checks pedigree completeness:

- **Complete subset** (all referenced parents still present): `IndNum`,
  `SireNum`, `DamNum` are rebuilt in-place, class and `ped_meta` are preserved.
- **Incomplete subset** (parent records missing): result is downgraded to plain
  `data.table` with a warning guiding the user to
  `tidyped(tp, cand = ids, trace = "up")`.

This downgrade is deliberate. It prevents stale integer indices from reaching
C++ routines.

## 10. Computational boundaries: C++ vs igraph

visPedigree delegates heavy pedigree recursion to C++ and uses igraph where a
graph object is still the simplest representation.

### 10.1 C++ — core computation path

| Task                          | C++ function                                         |
|-------------------------------|------------------------------------------------------|
| Ancestry / descendant tracing | `cpp_trace_ancestors`, `cpp_trace_descendants`       |
| Topological sorting           | `cpp_topo_order`                                     |
| Generation assignment         | `cpp_assign_generations_top`, `cpp_assign_generations_bottom` |
| Inbreeding coefficients       | `cpp_calculate_inbreeding` (Meuwissen & Luo)         |
| Relationship matrices         | `cpp_addmat`, `cpp_dommat`, `cpp_aamat`, `cpp_ainv`  |

All C++ functions consume `SireNum` / `DamNum` integer vectors and assume the
head invariant (§3).

### 10.2 igraph — graph-specific tasks

| Task                   | Where                         | igraph functions                                     |
|------------------------|-------------------------------|------------------------------------------------------|
| Pedigree visualization | `visped()` pipeline           | `graph_from_data_frame`, `layout_with_sugiyama`, `plot.igraph` |
| Connected components   | `splitped()`                  | `graph_from_edgelist`, `components`                  |
| Loop detection         | `tidyped()` raw-input path    | `graph_from_edgelist`, `is_dag`                      |
| Loop diagnosis         | `tidyped()` error path        | `which_loop`, `shortest_paths`, `neighbors`, `components` |
| Candidate tracing      | `tidyped()` raw-input path    | `neighborhood`                                       |
| Topological sorting    | `tidyped()` raw-input path    | `topo_sort`                                          |

igraph is not used in the core numerical pedigree analysis routines such as
`inbreed()`, `pedmat()`, `pedecg()`, or `pedrel()`, but it is still part of
the preprocessing and visualization stack.

## 11. Extension rules

When extending the class, follow these rules.

### 11.1 Do not add new pedigree-level attributes

Prefer adding fields to `ped_meta` instead of scattering new standalone
attributes.

### 11.2 Keep computed state derivable

If a column can be rebuilt from pedigree structure, prefer derivation over
storing opaque cached state.

### 11.3 Preserve `data.table` semantics

Use `:=`, `set()`, and `setattr()` carefully. Avoid patterns that trigger full
copies unless unavoidable.

### 11.4 Respect downgrade semantics

Any future method that subsets rows must preserve the current rule:

valid complete subset -> `tidyped`; incomplete subset -> plain `data.table`.

### 11.5 Document C++ assumptions

Any feature using `IndNum`, `SireNum`, or `DamNum` should document whether it
requires:

- topologically ordered rows,
- dense consecutive indices,
- `0L` encoding for missing parents.

## 12. User-facing inspection helpers

| Function                  | Returns                           |
|---------------------------|-----------------------------------|
| `is_tidyped(x)`           | `TRUE` if class is present        |
| `is_complete_pedigree(x)` | `TRUE` if all Sire/Dam are in Ind |
| `pedmeta(x)`              | The `ped_meta` named list         |
| `has_inbreeding(x)`       | `TRUE` if `f` column exists       |
| `has_candidates(x)`       | `TRUE` if `Cand` column exists    |

Future extensions should prefer helper functions over direct attribute access.

## 13. Maintenance checklist

Before merging a structural change to `tidyped`, check:

1. Does class identity remain `c("tidyped", "data.table", "data.frame")`?
2. Is the head invariant `IndNum == row index` preserved after every code path?
3. Are `ped_meta` fields preserved correctly?
4. Does `[.tidyped` still handle `:=` without copy issues?
5. Do incomplete row subsets still downgrade with warning?
6. Are integer pedigree columns rebuilt whenever a subset remains valid?
7. Does `tidyped(tp_master, cand = ...)` match the full path result?
8. After `setorder()` or `merge()`, are indices rebuilt before reaching C++?
9. Do package tests and vignettes build cleanly?

## 14. Recommended workflow

For large pedigrees, the intended usage pattern is:

```r
# build one validated master pedigree
tp_master <- tidyped(raw_ped)

# reuse it for repeated local tracing (fast path)
tp_local <- tidyped(tp_master, cand = ids, trace = "up", tracegen = 3)

# modify analysis columns in place
tp_master[, phenotype := pheno]

# split only when disconnected components matter
parts <- splitped(tp_master)
```

This keeps workflows explicit, fast, and safe.

---
title: "3. How to draw a pedigree"
author: "Sheng Luan"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{3. How to draw a pedigree}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6.5,
  fig.height = 6.5,
  dpi = 300,
  out.width = "100%"
)
library(visPedigree)
```

1. [Drawing the pedigree graph](#1)   
1.1 [A simple pedigree graph](#1-1)   
1.1.1 [Highlighting specific individuals](#1-1-1)   
1.1.2 [Showing inbreeding coefficients](#1-1-2)   
1.1.3 [Drawing a plant pedigree with monoecious individuals](#1-1-3)   
1.2 [A reduced pedigree graph](#1-2)   
1.3 [An outlined pedigree graph](#1-3)   
1.4 [How to use this package in a selective breeding program](#1-4)   
1.4.1 [Analysis of founders for an individual](#1-4-1)   
1.4.2 [The contribution of different families in a selective breeding program](#1-4-2) 

## 1 Drawing the pedigree {#1}
The **visped** function takes a pedigree tidied by the `tidyped` function and outputs a hierarchical graph for all individuals in the pedigree. The graph can be displayed on the default graphics device and saved as a PDF file. The graph in the PDF file is a vector drawing, which is legible and avoids overlapping. It is especially useful when the number of individuals is large or when individual labels are long. This function can visualize very large pedigrees (> 10,000 individuals per generation) by compacting full-sib individuals. It is particularly effective for aquatic animal pedigrees, which typically include many full-sib families per generation in the nucleus breeding population. A pedigree outline without individual labels is shown if the graph width exceeds the maximum PDF width (200 inches). This helps breeders quickly review the population construction process and identify any introduction of new genetic material.

**Important Note:** It is strongly recommended to set the `cand` parameter when tidying a pedigree. Pruning the pedigree by specifying candidates allows for more accurate generation inference and a more logical layout in the resulting pedigree tree.

Additionally, isolated individuals (those with no parents and no progeny) are automatically filtered out by `visped()` to prevent cluttering the graph. These individuals are assigned Generation 0 during the tidying process.

A small pedigree is drawn in the following figure. The following code also demonstrates how to save the graph as a high-quality vector graphic in a PDF file.

```{r smallped, fig.show="hold"}
tidy_small_ped <-
  tidyped(ped = small_ped,
          cand = c("Y", "Z1", "Z2"))
visped(tidy_small_ped, 
        compact = TRUE, 
        cex=0.5, 
        symbolsize=10, 
        file = tempfile(fileext = ".pdf"))
```

By default, `shapeby = "sex"` encodes individual sex by shape. Females are
circles, males are squares, individuals of unknown sex are diamonds, and
monoecious individuals are hexagons. Node color also identifies sex: dark sky
blue indicates males, dark goldenrod indicates females, teal indicates
monoecious individuals, and neutral grey indicates unknown sex. Purple is
reserved for highlighting. A compact family label such as `FS×10` means that
the rectangle represents 10 terminal, non-parent full siblings; it is not an
individual identifier. The ancestors are drawn at the top and descendants at
the bottom. Parents and offspring are connected via virtual nodes. Lines from
offspring to virtual nodes are dark grey, while parental edges use the
corresponding male, female, or monoecious color.

Set `shapeby = "role"` to use the legacy role-based symbol scheme, where
individual records are circles and compact full-sib family summaries are
green-grey rectangles. Compact family summaries remain green-grey
rectangles in both modes. In `shapeby = "role"` mode, real individuals remain
circles, but their fill colors still encode sex.

```{r symbol-schemes, fig.show="hold"}
visped(tidy_small_ped, 
      compact = TRUE,
      cex=0.5, 
      symbolsize=10)
visped(tidy_small_ped, 
        compact = TRUE, 
        shapeby = "role",
        cex=0.5, 
        symbolsize=10)
```

### 1.1 A simple pedigree graph {#1-1}
The trimmed **simple_ped** pedigree is drawn and displayed on the default graphics device. The **addgen** and **addnum** parameters need to be set to TRUE when tidying the pedigree using the **tidyped** function.
```{r vissimpleped, fig.show="hold"}
tidy_simple_ped <- tidyped(simple_ped)
visped(tidy_simple_ped, cex=0.3, symbolsize=10)
```

Generation labels can be added on the left margin of the pedigree graph by setting `genlab = TRUE`. This is useful for deep pedigrees when you want to quickly identify the generation of each row.

```{r vissimpleped_genlab, fig.show="hold"}
visped(tidy_simple_ped, cex = 0.3, symbolsize = 10, genlab = TRUE)
```

Instead of the default `G1`, `G2`, ... labels, an unnamed character vector
can provide one custom label for each displayed generation, from top to
bottom. This is useful when the plotted pedigree layers have an
application-specific interpretation. The labels are not inferred from a
birth-year column, which may not correspond one-to-one with pedigree
generations in overlapping cohorts.

```{r vissimpleped_custom_genlab, fig.show="hold"}
generation_labels <- paste0("Gen_", sort(unique(tidy_simple_ped$Gen)))
visped(tidy_simple_ped, cex = 0.3, symbolsize = 10, genlab = generation_labels, genlabcex=0.6)
```

In deep pedigrees the generation labels may appear small because their size is tied to the node scaling. The `genlabcex` parameter lets you set the generation-label size independently of the individual-label size (`cex`). For example, when `cex` is deliberately small to fit many individuals, you can still keep the generation labels readable:

```{r vissimpleped_genlabcex, fig.show="hold"}
# cex controls individual label size; genlabcex controls generation label size
visped(tidy_simple_ped, cex = 0.3, symbolsize = 15, genlab = TRUE, genlabcex = 0.8)
```

Figures displayed in the RStudio Plots panel often have limited resolution. Individual IDs may overlap if the pedigree is large and the plot area is restricted. This can be resolved by saving the graph as a vector graphic in a PDF file. The **visped** function will suppress output to the default device if `showgraph = FALSE`.

```{r}
suppressMessages(visped(tidy_simple_ped, showgraph = FALSE, file = tempfile(fileext = ".pdf")))
```

By setting the **file** parameter, you can generate a high-definition PDF version of the pedigree.

SVG output is also supported by using a file name ending in `.svg`. This is useful when the pedigree graph needs to be edited in vector-graphics software or embedded in web documents.

```{r}
suppressMessages(visped(tidy_simple_ped, showgraph = FALSE, file = tempfile(fileext = ".svg")))
```

Custom node labels can be displayed either from a selected pedigree column or from a character vector with one value per input row using `labelvar`. Row-aligned vectors remain attached to the correct individuals if the pedigree is internally tidied or reordered. Both forms keep the underlying individual IDs available for tracing and highlighting, while changing the text shown on individual nodes. Compact full-sib family nodes still use the explicit `FS×N` family-size label when `compact = TRUE`.

```{r custom-labels, fig.show="hold"}
tidy_simple_ped[, ShortLabel := paste0("N", seq_len(.N))]
visped(tidy_simple_ped, labelvar = "ShortLabel", cex = 0.3, symbolsize = 10)

# A row-aligned vector is also accepted
visped(tidy_simple_ped, labelvar = rep("1x", nrow(tidy_simple_ped)),
       cex = 0.3, symbolsize = 10)
```

#### 1.1.1 Highlighting specific individuals {#1-1-1}
Specific individuals can be highlighted in the pedigree graph using the **highlight** parameter. This is useful for marking candidates, founders, or any individuals of interest.

You can provide a character vector of individual IDs to use the default highlight colors (purple border and light purple fill):
```{r highlight1}
visped(tidyped(small_ped), highlight = c("Y", "Z1"))
```

You can also highlight an individual and its relatives (ancestors and descendants) by setting `trace = TRUE`:
```{r highlight_trace}
# Highlight individual "Y" and all its ancestors and descendants
visped(tidyped(small_ped), highlight = "Y", trace = TRUE)
```

Alternatively, you can customize the colors by providing a list with **ids**, **frame.color**, and **color**:
```{r highlight2}
visped(tidyped(small_ped), 
       highlight = list(ids = c("Y", "Z1"), 
                        frame.color = "#4caf50", 
                        color = "#81c784"))
```

### 1.1.2 Showing inbreeding coefficients {#1-1-2}
Inbreeding coefficients can be displayed on the pedigree graph using the **showf** parameter in the **visped** function. This requires that the pedigree has been processed with inbreeding coefficients calculated using the **inbreed** parameter in the **tidyped** function.
```{r showinbreed}
library(data.table)
test_ped <- data.table(
  Ind = c("A", "B", "C", "D", "E"),
  Sire = c(NA, NA, "A", "C", "C"),
  Dam = c(NA, NA, "B", "B", "D"),
  Sex = c("male", "female", "male", "female", "male")
)
tidy_test_ped_inbreed <- tidyped(test_ped, inbreed = TRUE)
visped(tidy_test_ped_inbreed, showf = TRUE)
```

### 1.1.3 Drawing a plant pedigree with monoecious individuals {#1-1-3}

In plant breeding, many crop species are **monoecious** — a single plant produces both male and female gametes, enabling **self-fertilization** (selfing). When a plant serves as both sire and dam, `tidyped()` with `selfing = TRUE` assigns it `Sex = "monoecious"`. In the graph, monoecious individuals appear as **hexagons** with teal edges and fill, and selfing edges (sire = dam) are drawn in teal.

The following example shows a multi-generation self-pollinating pedigree with inbreeding coefficients displayed:

```{r plant-selfing-draw}
library(data.table)

# P1 × P2 → F1;  F1 selfs → F2_1, F2_2;  F2_1 selfs → F3
plant_ped <- data.table(
  Ind  = c("P1", "P2", "F1", "F2_1", "F2_2", "F3"),
  Sire = c(NA,   NA,   "P1", "F1",   "F1",   "F2_1"),
  Dam  = c(NA,   NA,   "P2", "F1",   "F1",   "F2_1")
)

tp_plant <- tidyped(plant_ped, selfing = TRUE, inbreed = TRUE)

visped(tp_plant, compact = TRUE, showf = TRUE, cex=0.5, symbolsize=10)
```

Self-fertilization produces rapid inbreeding: *f* = 0.5 after one generation of selfing (`F2_1`, `F2_2`), and *f* = 0.75 after two (`F3`). `F1` and `F2_1` are drawn as hexagons because they act as both sire and dam; selfing edges are shown in teal. For the general shape and color legend, see the introduction of Section 1.

See `vignette("tidy-pedigree", package = "visPedigree")`, section 3.9, for preparing plant pedigrees and calculating inbreeding coefficients.

### 1.2 A reduced pedigree graph {#1-2}

Warning messages will be shown when you try to draw the pedigree graph of the deep_ped dataset.
```{r deepped, eval=FALSE}
cand_J11_labels <- deep_ped[(substr(Ind, 1, 3) == "K11"), Ind]
visped(
  tidyped(deep_ped,
    cand = cand_J11_labels,
    tracegen = 3),
  cex=0.08,
  symbolsize=5.5
  )
```

```
  Too many individuals (>=3362) in one generation!!! Two choices:
1. Removing full-sib individuals using the parameter compact = TRUE; or, 
2. Visualizing all nodes without labels using the parameter outline = TRUE.
Rerun visped() function!
```

The function indicates that there are too many individuals in a single generation to draw a standard pedigree graph. It is recommended to use the **compact** or **outline** parameters to simplify the pedigree.

First, let's try the **compact** parameter and output it to a PDF file. The plot on the default device may suffer from significant overlapping due to the high density of individuals.
```{r reduceped1}
cand_J11_labels <- deep_ped[(substr(Ind,1,3) == "K11"),Ind]
visped(
  tidyped(
    deep_ped,
    cand = cand_J11_labels,
    trace = "up",
    tracegen = 3
  ),
  cex=0.08, 
  symbolsize=5.5,
  compact = TRUE,
  showgraph = TRUE,
  file = tempfile(fileext = ".pdf")
)
```

You can open the generated PDF file to view the high-definition pedigree vector graph. Most rectangles at the bottom are compact family summaries, and labels such as `FS×10` report the number of represented full-sib individuals. The summary is sex-neutral because a compacted family may contain individuals of different or unknown sex. Individual labels may be shorter than their shapes and might not align perfectly. Labels can be resized using `cex`; increasing it makes labels larger, while decreasing it makes them smaller. The value typically ranges from 0 to 1, but can be larger, and adjustments of 0.1 are usually sufficient. The **visped** function reports the `cex` value used to draw the graph.

```{r reduceped2}
visped(
  tidyped(
    deep_ped,
    cand = cand_J11_labels,
    trace = "up",
    tracegen = 3
  ),
  compact = TRUE,
  cex = 0.83,
  showgraph = FALSE,
  file = tempfile(fileext = ".pdf")
)
```
You can open the generated PDF file to view the high-definition pedigree vectorgraph. The labels align better with the shapes compared to the previous version. You can continue to adjust `cex` until the labels are sized appropriately.

### 1.3 An outlined pedigree graph {#1-3}
Setting `outline = TRUE` produces an outlined pedigree graph. Individual labels will not be shown in the graph. This is highly effective for large pedigrees with many individuals.

The following code generates an outlined pedigree graph in a PDF file.

```{r reduceped3}
suppressMessages(visped(
  tidyped(
    deep_ped,
    cand = cand_J11_labels, 
    tracegen = 3),
  compact = TRUE,
  outline = TRUE,
  showgraph = TRUE,
  file = tempfile(fileext = ".pdf")
))
```

### 1.4 How to use this package in a selective breeding program {#1-4}    
#### 1.4.1 An analysis of founders for an individual {#1-4-1}    
Selective breeding is a process of enriching desirable minor genes from multiple founders through successive generations of mating. This is supported by the well-known infinitesimal model (or minor polygene hypothesis).

We select the individual "K110550H" in the deep_ped dataset to visualize its pedigree. The following code generates the pedigree graph for a specific individual in a PDF file.
```{r pedofoneind}
suppressWarnings(
  K110550H_ped <- tidyped(deep_ped, cand = "K110550H")
)
if(interactive()) {
  visped(K110550H_ped, compact = TRUE)
}
```

As you can see from the figure above, the number of founder individuals (without parents) of the K110550H individual is `r nrow(K110550H_ped[is.na(Sire) & is.na(Dam)])`. This indicates that the individual has accumulated favorable genes from many founders, contributing to genetic gain in the target traits.

#### 1.4.2 The contribution of different families in a selective breeding program {#1-4-2}

Under optimum contribution theory, families contribute different numbers of individuals to the next generation, with higher-indexing families contributing more. By visualizing pedigree, we can directly see the contribution ratio of different families.

The code below shows the parental composition of 106 families born in the nucleus breeding population in 2007. Setting `tracegen = 2` limits the graph to two generations (parents and grandparents).
```{r optiMate}
cand_2007_G8_labels <-
  big_family_size_ped[(Year == 2007) & (substr(Ind, 1, 2) == "G8"), Ind]
suppressWarnings(
  cand_2007_G8_tidy_ped_ancestor_2 <-
    tidyped(
      big_family_size_ped,
      cand = cand_2007_G8_labels,
      trace = "up",
      tracegen = 2
    )
)
sire_label <-
  unique(cand_2007_G8_tidy_ped_ancestor_2[Ind %in% cand_2007_G8_labels,
                                          Sire])
dam_label <-
  unique(cand_2007_G8_tidy_ped_ancestor_2[Ind %in% cand_2007_G8_labels,
                                          Dam])
sire_dam_label <- unique(c(sire_label, dam_label))
sire_dam_label <- sire_dam_label[!is.na(sire_dam_label)]
sire_dam_ped <-
  cand_2007_G8_tidy_ped_ancestor_2[Ind %in% sire_dam_label]
sire_dam_ped <-
  sire_dam_ped[, FamilyID := paste(Sire, Dam, sep = "")]
family_size <- sire_dam_ped[, .N, by = c("FamilyID")]
fullsib_family_label <- unique(sire_dam_ped$FamilyID)
suppressMessages(
  visped(
    cand_2007_G8_tidy_ped_ancestor_2,
    compact = TRUE,
    outline = TRUE,
    showgraph = TRUE
  )
)
```

In the above figure, 106 families are shown at bottom, the parents are shown in middle, and the grandparents are shown at top. It can be seen that the parents are composed of `r length(sire_label[!is.na(sire_label)])` sires and `r length(dam_label[!is.na(dam_label)])` dams. The parents are from `r length(fullsib_family_label)` full-sib families in the generation of grandparent. Approximately `r family_size$N[1]+family_size$N[2]` parents originate from just two full-sib families due to the application of optimum contribution theory, accounting for `r round((family_size$N[1]+family_size$N[2])/sum(family_size$N),4)*100`% of all parents.

---
title: "1. How to tidy a pedigree"
author: "Sheng Luan"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{1. How to tidy a pedigree}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6.5,
  fig.height = 6.5,
  dpi = 96,
  out.width = "100%"
)
library(visPedigree)
```

Pedigrees are fundamental to both animal and plant breeding. They are used to improve the accuracy of breeding value estimation, to monitor and manage inbreeding, and to support a wide range of downstream analyses based on ancestry and relatedness. In applied settings, however, pedigree data are often incomplete or poorly ordered: founders may be omitted, parents may appear after offspring, and sex annotations may be missing or inconsistent. In addition, pedigrees are commonly stored in a simple three-column form (individual, sire, dam), which is convenient for storage but less convenient for checking structure, tracing relatives, or producing readable graphical displays.

Several software tools address parts of this workflow. On Windows, pedigraph and [pedigree viewer](https://bkinghor.une.edu.au/pedigree.htm) provide facilities for pedigree trimming and display. Within R, packages such as `pedigree`, `nadiv`, and `optiSel` support pedigree preparation and analysis, while `kinship2` can be used to draw pedigree trees. Nevertheless, large pedigrees continue to pose practical difficulties for data cleaning, loop detection, multi-generation tracing, and graphical display with limited node overlap.

The [visPedigree](https://github.com/luansheng/visPedigree) package was developed to provide an integrated workflow for pedigree tidying, analysis, and visualization. Built on `data.table`, C++, and `igraph`, it supports pedigree standardization, loop detection, candidate tracing, integer pedigree construction, generation assignment, and optional inbreeding calculation, together with scalable pedigree visualization. The package is designed for both animal and plant pedigrees, including selfing or monoecious mating systems. This guide focuses on the tidying workflow and introduces the main arguments and outputs of `tidyped()`.

The main contents of this guide are as follows:

1. [Installation of the visPedigree package](#1)   
2. [The specification of pedigree format](#2)   
3. [Checking and tidying pedigree](#3)   
3.1 [Introduction](#3-1)   
3.2 [Pedigree loop detection](#3-2)   
3.3 [Tracing the pedigree of a specific individual](#3-3)   
3.4 [Creating an integer pedigree](#3-4)   
3.5 [Calculating inbreeding coefficients](#3-5)   
3.6 [Customizing generation assignment](#3-6)   
3.7 [Summarizing the pedigree](#3-7)   
3.8 [Splitting large pedigrees](#3-8)   
3.9 [Working with plant pedigrees and selfing](#3-9)

## 1. Installation of the visPedigree package {#1}

The visPedigree package can be installed from CRAN:
```r
install.packages("visPedigree")
```

Or from GitHub:
```r
# install.packages("devtools")
devtools::install_github("luansheng/visPedigree")
```

## 2. Pedigree format specification {#2}
The first three columns of pedigree data must be in the order of individual, sire, and dam IDs. The column names can be customized, but their order must remain unchanged. Individual IDs should not be coded as `""`, `" "`, `"0"`, `*`, or `NA`; otherwise, they will be removed from the pedigree. Missing parents should be denoted by `NA`, `0`, or `*`. Spaces and empty strings (`""`) will also be treated as missing parents, though this is not recommended. Additional columns, such as sex and generation, can also be included.

## 3. Checking and tidying pedigree {#3}

### 3.1 Introduction {#3-1}
The pedigree can be checked and tidied through the `tidyped()` function.

This function takes a pedigree, checks for duplicates and bisexual individuals, detects loops, adds missing founders, sorts the pedigree, and traces candidate pedigrees. 

If the `cand` parameter is provided, only those individuals and their ancestors or descendants are retained. 

Tracing direction and the number of generations can be specified using the `trace` and `tracegen` parameters. 

Virtual generations are inferred and assigned when `addgen = TRUE`. 

A numeric pedigree is generated when `addnum = TRUE`. 

Sex will be inferred for all individuals if sex information is missing. If a `Sex` column is present, values should be coded as `'male'`, `'female'`, or `NA` (unknown). Missing sex information will be inferred from the pedigree structure where possible.

The visPedigree package comes with multiple datasets. You can check through the following command.
```{r gettingdataset,eval=FALSE}
data(package="visPedigree")
```

The following code displays the `simple_ped` dataset, which contains four columns: individual, sire, dam, and sex. Missing parents are denoted by `'NA'`, `'0'`, or `*`. Founders are not explicitly listed, and some parents appear after their offspring in the original data.
```{r simpleped}
head(simple_ped)
tail(simple_ped)
# The number of individuals in the pedigree dataset
nrow(simple_ped)
# Individual records with missing parents
simple_ped[Sire %in% c("0", "*", "NA", NA) |
             Dam %in% c("0", "*", "NA", NA)]
```
Example: If we incorrectly set the female `J0Z167` as the sire of `J2F588`, `tidyped()` will detect this bisexual conflict.
```{r, error=TRUE}
x <- data.table::copy(simple_ped)
x[ID == "J2F588", Sire := "J0Z167"]
y <- tidyped(x)
```

The `tidyped()` function sorts the pedigree, replaces missing parents with `NA`, ensures parents precede their offspring, and adds missing founders.
```{r tidyped}
tidy_simple_ped <- tidyped(simple_ped)
head(tidy_simple_ped)
tail(tidy_simple_ped)
nrow(tidy_simple_ped)
```

In the resulting `tidy_simple_ped`, founders are added with their inferred sex, and parents are sorted before their offspring. The number of individuals increases from 31 to 59. The columns are renamed to `Ind`, `Sire`, and `Dam`. Missing parents are uniformly replaced with `NA`, and `tidyped()` provides informative messages during processing. By default, `tidy_simple_ped` includes new columns: `Gen`, `IndNum`, `SireNum`, and `DamNum`. These can be disabled by setting `addgen = FALSE` and `addnum = FALSE`.

If the input dataset lacks a `Sex` column, it will be automatically added to the tidied output.

```{r}
tidy_simple_ped_no_gen_num <-
  tidyped(simple_ped, addgen = FALSE, addnum = FALSE)
    head(tidy_simple_ped_no_gen_num)
```
Once tidied, you can use `data.table::fwrite()` to export the pedigree for genetic evaluation software like ASReml.

### 3.2 Pedigree loop detection {#3-2}
A pedigree loop occurs when an individual is its own ancestor (e.g., A is the parent of B, B is the parent of C, and C is the parent of A). This is a biological impossibility and a serious error in pedigree records. The `tidyped()` function automatically detects these cycles using graph theory algorithms. If a loop is detected, the function will stop and provide information about the individuals involved in the loop.

The following code demonstrates what happens when a pedigree with loops is processed:
```{r loop_detection, error=TRUE}
# loop_ped contains cycles (e.g., V -> T -> R -> P -> M -> V)
# Attempting to tidy it will result in an error
try(tidyped(loop_ped))
```
Detecting loops early is crucial for ensuring the integrity of genetic evaluations.

When saving the pedigree, missing parents should typically be replaced with `0`. 
```{r writeped,eval=FALSE}
saved_ped <- data.table::copy(tidy_simple_ped)
saved_ped[is.na(Sire), Sire := "0"]
saved_ped[is.na(Dam), Dam := "0"]
data.table::fwrite(
  x = saved_ped,
  file = tempfile(fileext = ".csv"),
  sep = ",",
  quote = FALSE
)
```

### 3.3 Tracing the pedigree of a specific individual {#3-3}

To trace the pedigree of specific individuals, use the `cand` parameter. This adds a `Cand` column where `TRUE` identifies the specified candidates. If `cand` is provided, only the candidates and their ancestors/descendants are retained.
```{r}
tidy_simple_ped_J5X804_ancestors <-
  tidyped(ped = tidy_simple_ped_no_gen_num, cand = "J5X804")
  tail(tidy_simple_ped_J5X804_ancestors)
```

By default, the function traces ancestors. You can limit the number of generations using `tracegen`. If `tracegen` is `NULL`, all available generations are traced.

```{r}
tidy_simple_ped_J5X804_ancestors_2 <-
  tidyped(ped = tidy_simple_ped_no_gen_num,
  cand = "J5X804",
  tracegen = 2)
  print(tidy_simple_ped_J5X804_ancestors_2)
```
The code above traces the ancestors of `J5X804` back two generations.

To trace descendants, set `trace = 'down'`.

There are three options for the **trace** parameter:

* "up"-trace candidates' pedigree to ancestors;
* "down"-trace candidates' pedigree to descendants;
* "all"-trace candidaes' pedigree to ancestors and descendants simultaneously.

```{r}
tidy_simple_ped_J0Z990_offspring <-
  tidyped(ped = tidy_simple_ped_no_gen_num, cand = "J0Z990", trace = "down")
  print(tidy_simple_ped_J0Z990_offspring)
```
Tracing the descendants of `J0Z990` reveals a total of 5 individuals.

### 3.4 Creating an integer pedigree {#3-4}
Certain genetic evaluation programs require integer-coded pedigrees, where individuals are numbered consecutively to facilitate the calculation of the additive genetic relationship matrix.

By default, `tidyped()` adds `IndNum`, `SireNum`, and `DamNum` columns. This can be disabled with `addnum = FALSE`.
```{r intped}
tidy_simple_ped_with_int <-
  tidyped(ped = tidy_simple_ped_no_gen_num, addnum = TRUE)
head(tidy_simple_ped_with_int)
```


### 3.5 Calculating inbreeding coefficients {#3-5}
The inbreeding coefficient (F) of each individual can be calculated using tidyped() or inbreed() functions. There are two options to add the inbreeding coefficients to a tidied pedigree:

1. Set `inbreed = TRUE` in the `tidyped()` function. This will calculate the inbreeding coefficients using an optimized C++ implementation of algorithms for large populations (Meuwissen & Luo, 1992; Sargolzaei et al., 2005), including the LAP bucket strategy described by Sargolzaei et al. (2005), and add an `f` column to the tidied pedigree.
2. Or call `inbreed()` directly on a tidied pedigree to add the `f` column. 

Both options use the same high-performance engine as `pedmat(method = "f")`, ensuring consistent results across the package.

```{r inbreed}
# Create a simple inbred pedigree
library(data.table)
test_ped <- data.table(
  Ind = c("A", "B", "C", "D", "E"),
  Sire = c(NA, NA, "A", "C", "C"),
  Dam = c(NA, NA, "B", "B", "D"),
  Sex = c("male", "female", "male", "female", "male")
)
# Option 1: Calculate during tidying
tidy_test <- tidyped(test_ped, inbreed = TRUE)
head(tidy_test)

# Option 2: Calculate after tidying
tidy_test <- inbreed(tidyped(test_ped))
```

### 3.6 Customizing generation assignment {#3-6}
Generation inference is essential for pedigree visualization. `tidyped()` provides two methods for assigning generation numbers via the `genmethod` parameter:

*   **"top" (default)**: Top-aligned (depth-based). Founders are assigned to Generation 1. This is the optimal scheme for most biological pedigrees as it ensures all founders start at the top, preventing them from "drifting" to lower generations if they have fewer descendants.
*   **"bottom"**: Bottom-aligned (height-based). Generations are counted from the bottom up, aligning terminal nodes (offspring with no further descendants) at the highest generation number. This is useful when you want to show that all current populations are at the same level, or when introducing unrelated exogenous parents in later years.

```{r genmethod}
# Default behavior (Top-Down): J2Y434 is at Gen 3
tidy_top <- tidyped(simple_ped, genmethod = "top")
tidy_top[Ind == "J2Y434"]

# Bottom-Up behavior: J2Y434 is at Gen 6
tidy_bottom <- tidyped(simple_ped, genmethod = "bottom")
tidy_bottom[Ind == "J2Y434"]
```

### 3.7 Summarizing the pedigree {#3-7}
The `summary()` method provides a quick overview of the pedigree statistics, including the number of individuals, sex distribution, founders, and isolated individuals. If inbreeding coefficients have been calculated (column `f`), the summary will also include descriptive statistics of inbreeding.

```{r summary}
# Summarize the tidied pedigree
summary(tidy_simple_ped)
```

### 3.8 Splitting large pedigrees {#3-8}
For extremely large pedigrees, it is sometimes useful to split them into disconnected subsets or "sub-pedigrees". The `splitped()` function automatically detects disconnected components (families that share no ancestors) and splits the pedigree into a list of smaller `tidyped` objects.

```{r splitped}
# Split the pedigree into components
sub_pedigrees <- splitped(tidy_simple_ped)

# View summary of the split result
summary(sub_pedigrees)

# Access a specific sub-pedigree
# first_sub <- sub_pedigrees[[1]]
```

### 3.9 Working with plant pedigrees and selfing {#3-9}

In plant breeding, many species are **monoecious** — a single plant produces both male and female gametes and can serve as both the male (sire) and female (dam) parent. Self-fertilization (selfing) is a common practice in crops such as rice, wheat, soybean, and maize. In a pedigree, selfing appears as the same individual listed as both Sire and Dam for one or more offspring.

The `selfing` parameter in `tidyped()` controls how such pedigrees are handled:

- **`selfing = FALSE`** (default): Treats any individual appearing as both Sire and Dam as a sex conflict and returns an error. This mode is suitable for animal pedigrees where bisexuality is a data error.
- **`selfing = TRUE`**: Allows the same individual to be both Sire and Dam. These individuals are automatically assigned `Sex = "monoecious"`, and the pedigree metadata records this status.

The following example demonstrates a typical self-pollinating plant breeding scenario:

```{r plant-selfing}
library(data.table)

# A plant pedigree with self-fertilization:
# P1 (male) × P2 (female) → F1
# F1 selfs → F2_1, F2_2
# F2_1 selfs → F3
plant_ped <- data.table(
  Ind  = c("P1", "P2", "F1", "F2_1", "F2_2", "F3"),
  Sire = c(NA,   NA,   "P1", "F1",   "F1",   "F2_1"),
  Dam  = c(NA,   NA,   "P2", "F1",   "F1",   "F2_1")
)

# Without selfing = TRUE: error — F1 and F2_1 appear as both Sire and Dam
try(tidyped(plant_ped))
```

Setting `selfing = TRUE` resolves the issue. The monoecious individuals (`F1`, `F2_1`) are automatically identified, and their sex is assigned accordingly:

```{r plant-selfing-tidy}
tp_plant <- tidyped(plant_ped, selfing = TRUE, inbreed = TRUE)
tp_plant[]

# Check metadata
attr(tp_plant, "ped_meta")
```

The `summary()` method reports the count of monoecious individuals:

```{r plant-selfing-summary}
summary(tp_plant)
```

In the tidied output:
- **`P1`** → `male` (only appears as Sire)
- **`P2`** → `female` (only appears as Dam)
- **`F1`** → `monoecious` (appears as both Sire and Dam for `F2_1`, `F2_2`)
- **`F2_1`** → `monoecious` (appears as both Sire and Dam for `F3`)

Self-fertilization produces inbred offspring: the inbreeding coefficient *f* = 0.5 for individuals derived from a single generation of selfing, and increases with successive selfing generations (*f* = 0.75 for `F3`).

```{r plant-selfing-inbreed}
# Inbreeding coefficients
tp_plant[, .(Ind, Sex, Gen, f)]
```

See `vignette("draw-pedigree", package = "visPedigree")` for visualizing plant pedigrees — monoecious individuals are drawn as **hexagons** with teal edges and fill.

**Tip:** When working with large plant breeding datasets where selfing is expected, always set `selfing = TRUE` to avoid false sex-conflict errors. The `selfing` attribute is preserved through `splitped()` and other downstream operations, ensuring consistent handling across the entire workflow.

---
**See Also:**
- `vignette("draw-pedigree", package = "visPedigree")`
- `vignette("relationship-matrix", package = "visPedigree")`

## References

- Meuwissen, T. H. E., & Luo, Z. (1992). Computing inbreeding coefficients in large populations. *Genetics Selection Evolution*, 24, 305-313.
- Sargolzaei, M., Iwaisaki, H., & Colleau, J.-J. (2005). A fast algorithm for computing inbreeding coefficients in large populations. *Journal of Animal Breeding and Genetics*, 122, 325-331.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedmatrix.R
\name{query_relationship}
\alias{query_relationship}
\title{Query Relationship Coefficients from a Pedigree Matrix}
\usage{
query_relationship(x, id1, id2 = NULL)
}
\arguments{
\item{x}{A pedmat object created by \code{\link{pedmat}}.}

\item{id1}{Character, first individual ID.}

\item{id2}{Character, second individual ID. If \code{NULL}, returns 
the entire row of relationships for \code{id1}.}
}
\value{
\itemize{
  \item If \code{id2} is provided: numeric value (relationship coefficient)
  \item If \code{id2} is \code{NULL}: named numeric vector (id1's row)
  \item Returns \code{NA} if individual not found
}
}
\description{
Retrieves relationship coefficients between individuals from a pedmat object.
For compact matrices, automatically handles lookup of merged full-siblings.
}
\details{
For compact matrices (\code{compact = TRUE}), this function automatically
maps individuals to their family representatives. For methods A, D, and AA,
it can compute the correct relationship even between merged full-siblings
using the formula:
\itemize{
  \item \strong{A}: \eqn{a_{ij} = 0.5 * (a_{is} + a_{id})} where s, d are parents
  \item \strong{D}: \eqn{d_{ij} = a_{ij}^2} (for full-sibs in same family)
  \item \strong{AA}: \eqn{aa_{ij} = a_{ij}^2}
}
}
\note{
Inverse matrices (Ainv, Dinv, AAinv) are \strong{not supported} because
inverse matrix elements do not represent meaningful relationship coefficients.
}
\examples{
tped <- tidyped(small_ped)
A <- pedmat(tped, method = "A", compact = TRUE)

# Query specific pair
query_relationship(A, "A", "B")

# Query merged full-siblings (works with compact)
query_relationship(A, "Z1", "Z2")

# Get all relationships for one individual
query_relationship(A, "A")

}
\seealso{
\code{\link{pedmat}}, \code{\link{expand_pedmat}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-s3.R
\name{require_complete_pedigree}
\alias{require_complete_pedigree}
\title{Error on row-truncated pedigree subsets}
\usage{
require_complete_pedigree(ped, fun)
}
\arguments{
\item{ped}{A pedigree-like object.}

\item{fun}{Character scalar. Calling function name for the error message.}
}
\value{
The input object, invisibly.
}
\description{
Error on row-truncated pedigree subsets
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped_layout.R
\name{prepare_ped_graph}
\alias{prepare_ped_graph}
\title{Internal layout engine for pedigree visualization}
\usage{
prepare_ped_graph(
  ped,
  compact = FALSE,
  outline = FALSE,
  cex = NULL,
  highlight = NULL,
  trace = FALSE,
  showf = FALSE,
  labelvar = NULL,
  shapeby = "sex",
  pagewidth = 200,
  symbolsize = 1,
  maxiter = 1000,
  ...
)
}
\description{
Internal layout engine for pedigree visualization
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedanalysis.R
\name{pedstats}
\alias{pedstats}
\title{Pedigree Statistics}
\usage{
pedstats(
  ped,
  timevar = NULL,
  unit = "year",
  cycle = NULL,
  ecg = TRUE,
  genint = TRUE,
  ...
)
}
\arguments{
\item{ped}{A \code{tidyped} object.}

\item{timevar}{Optional character. Name of the column containing the
\strong{birth date} (or hatch date) of each individual.
Accepted column formats:
\itemize{
  \item \code{Date} or \code{POSIXct} (recommended).
  \item A date string parseable by \code{as.POSIXct}
    (e.g., \code{"2020-06-15"}).  Use \code{format} via \code{...}
    for non-ISO strings.
  \item A numeric year (e.g., \code{2020}).  Automatically converted
    to \code{Date} (\code{"YYYY-07-01"}) with a message.
}
If \code{NULL}, attempts auto-detection from common column names
(\code{"BirthYear"}, \code{"Year"}, \code{"BirthDate"}, etc.).}

\item{unit}{Character. Time unit for reporting generation intervals:
\code{"year"} (default), \code{"month"}, \code{"day"}, or \code{"hour"}.}

\item{cycle}{Numeric. Optional target generation cycle length in
\code{unit}s. When provided, \code{gen_intervals} will include a
\code{GenEquiv} column (observed Mean / cycle). See
\code{\link{pedgenint}} for details.}

\item{ecg}{Logical. Whether to compute equivalent complete generations
for each individual via \code{\link{pedecg}}. Default \code{TRUE}.}

\item{genint}{Logical. Whether to compute generation intervals via
\code{\link{pedgenint}}. Requires a detectable \code{timevar} column.
Default \code{TRUE}.}

\item{...}{Additional arguments passed to \code{\link{pedgenint}},
e.g., \code{format} for custom date parsing or \code{by} for grouping.}
}
\value{
An object of class \code{pedstats}, which is a list containing:
\itemize{
  \item \code{summary}: A \code{data.table} with one row summarising the
    whole pedigree.  Columns:
    \itemize{
      \item \code{N} — total number of individuals.
      \item \code{NSire} — number of unique sires.
      \item \code{NDam} — number of unique dams.
      \item \code{NFounder} — number of founder individuals
        (both parents unknown).
      \item \code{MaxGen} — maximum generation number.
    }
  \item \code{ecg}: A \code{data.table} with one row per individual
    (\code{NULL} if \code{ecg = FALSE}).  Columns:
    \itemize{
      \item \code{Ind} — individual identifier.
      \item \code{ECG} — equivalent complete generations.
      \item \code{FullGen} — number of fully known generations.
      \item \code{MaxGen} — maximum traceable generation depth.
    }
  \item \code{gen_intervals}: A \code{data.table} of generation intervals
    (\code{NULL} if no \code{timevar} is detected or
    \code{genint = FALSE}).  Columns:
    \itemize{
      \item \code{Pathway} — gametic pathway label.  Seven values:
        \code{"SS"} (sire to son), \code{"SD"} (sire to daughter),
        \code{"DS"} (dam to son), \code{"DD"} (dam to daughter) —
        require offspring sex; \code{"SO"} (sire to offspring) and
        \code{"DO"} (dam to offspring) — sex-independent; and
        \code{"Average"} — all parent-offspring pairs combined.
      \item \code{N} — number of parent-offspring pairs.
      \item \code{Mean} — mean generation interval.
      \item \code{SD} — standard deviation of the interval.
      \item \code{GenEquiv} — \code{Mean / cycle} (only present when
        \code{cycle} is supplied).
    }
}
}
\description{
Calculates comprehensive statistics for a pedigree, including population structure,
generation intervals, and ancestral depth.
}
\examples{
\donttest{
# ---- Without time variable ----
tp <- tidyped(simple_ped)
ps <- pedstats(tp)
ps$summary
ps$ecg

# ---- With annual Year column (big_family_size_ped) ----
tp2 <- tidyped(big_family_size_ped)
ps2 <- pedstats(tp2, timevar = "Year")
ps2$summary
ps2$gen_intervals
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped_graph.R
\name{compact_pedigree}
\alias{compact_pedigree}
\title{Compact pedigree by merging full siblings}
\usage{
compact_pedigree(ped_node, compact, h_ids)
}
\arguments{
\item{ped_node}{A data.table of nodes.}

\item{compact}{Logical, whether to compact.}

\item{h_ids}{Highlighted IDs to exempt from compaction.}
}
\description{
Compact pedigree by merging full siblings
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedanalysis.R
\name{pedfclass}
\alias{pedfclass}
\title{Summarize Inbreeding Levels}
\usage{
pedfclass(ped, breaks = c(0.0625, 0.125, 0.25), labels = NULL)
}
\arguments{
\item{ped}{A \code{tidyped} object.}

\item{breaks}{Numeric vector of strictly increasing positive upper bounds for
inbreeding classes. Default is \code{c(0.0625, 0.125, 0.25)}, corresponding
approximately to half-sib, avuncular/grandparent, and full-sib/parent-offspring
mating thresholds. The class \code{"F = 0"} is always kept as a fixed first
level. A final open-ended class \code{"F > max(breaks)"} is always appended
automatically.}

\item{labels}{Optional character vector of interval labels. If \code{NULL},
labels are generated automatically from \code{breaks}. When supplied, its
length must equal \code{length(breaks)}, with each element naming the
bounded interval \code{(breaks[i-1], breaks[i]]}. The open-ended tail
class is always auto-generated and cannot be overridden.}
}
\value{
A \code{data.table} with 3 columns:
\describe{
  \item{\code{FClass}}{An ordered factor. By default it contains 5 levels:
  \code{"F = 0"}, \code{"0 < F <= 0.0625"}, \code{"0.0625 < F <= 0.125"},
  \code{"0.125 < F <= 0.25"}, and \code{"F > 0.25"}. The number of levels
  equals \code{length(breaks) + 2} (the fixed zero class plus one class per
  bounded interval plus the open-ended tail).}
  \item{\code{Count}}{Integer. Number of individuals in each class.}
  \item{\code{Percentage}}{Numeric. Percentage of individuals in each class.}
}
}
\description{
Classifies individuals into inbreeding levels based on their inbreeding 
coefficients (F) according to standard or user-defined thresholds.
}
\details{
The default thresholds follow common pedigree interpretation rules:
\itemize{
  \item \code{F = 0.0625}: approximately the offspring of half-sib mating.
  \item \code{F = 0.125}: approximately the offspring of avuncular or grandparent-grandchild mating.
  \item \code{F = 0.25}: approximately the offspring of full-sib or parent-offspring mating.
}
Therefore, assigning \code{F = 0.25} to the class \code{"0.125 < F <= 0.25"}
is appropriate. If finer reporting is needed, supply custom \code{breaks},
for example to separate \code{0.25}, \code{0.375}, or \code{0.5}.
}
\examples{
tp <- tidyped(simple_ped, addnum = TRUE)
pedfclass(tp)

# Finer custom classes (4 breaks, labels auto-generated)
pedfclass(tp, breaks = c(0.03125, 0.0625, 0.125, 0.25))

# Custom labels aligned to breaks (3 labels for 3 breaks; tail is auto)
pedfclass(tp, labels = c("Low", "Moderate", "High"))

\donttest{
tp_inbred <- tidyped(inbred_ped, addnum = TRUE)
pedfclass(tp_inbred)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedanalysis.R
\name{pedrel}
\alias{pedrel}
\title{Calculate Mean Relationship or Coancestry Within Groups}
\usage{
pedrel(
  ped,
  by = "Gen",
  reference = NULL,
  compact = FALSE,
  scale = c("relationship", "coancestry")
)
}
\arguments{
\item{ped}{A \code{tidyped} object.}

\item{by}{Character. The column name to group by (e.g., "Year", "Breed", "Generation").}

\item{reference}{Character vector. An optional vector of reference individual IDs to calculate
relationships for. If provided, only individuals matching these IDs in each group
will be used. Default is NULL (use all individuals in the group).}

\item{compact}{Logical. Retained for backward compatibility. It is ignored
because \code{pedrel()} now uses matrix-free pedigree products for all
calculations.}

\item{scale}{Character. One of \code{"relationship"} or \code{"coancestry"}.
\code{"relationship"} returns the pairwise off-diagonal mean additive
relationship (current \code{pedrel()} behavior). \code{"coancestry"}
returns the corrected population mean coancestry used for pedigree-based
diversity calculations.}
}
\value{
A \code{data.table} with columns:
\itemize{
  \item A grouping identifier column, named after the \code{by} parameter (e.g., \code{Gen}, \code{Year}).
  \item \code{NTotal}: Total number of individuals in the group.
  \item \code{NUsed}: Number of individuals used in calculation (could be subset by reference).
  \item \code{MeanRel}: Present when \code{scale = "relationship"}; average of off-diagonal
    elements in the Additive Relationship (A) matrix for this group
    (\eqn{a_{ij} = 2f_{ij}}).
  \item \code{MeanCoan}: Present when \code{scale = "coancestry"}; diagonal-corrected
    population mean coancestry for this group.
  \item \code{Status}: \code{"ok"}, \code{"skipped"}, or \code{"failed"}.
  \item \code{Message}: Empty for successful groups; otherwise a diagnostic
    explaining why \code{NA} was returned.
}
}
\description{
Computes either the average pairwise additive genetic relationship
coefficients (\eqn{a_{ij}}) within cohorts, or the corrected population mean
coancestry used for pedigree-based diversity summaries.
}
\details{
Let \eqn{g} be a zero-one indicator vector for the \eqn{N} selected
individuals and let \eqn{S = g^\prime A g}. When
\code{scale = "relationship"}, the returned value is
\deqn{\bar{a}_{off} =
  \frac{S - \sum_{i:g_i=1}(1 + F_i)}{N(N - 1)},}
the mean of the off-diagonal additive relationship coefficients. When
\code{scale = "coancestry"}, the returned value is
\deqn{\bar{C} = \frac{S}{2N^2},}
which includes both pairwise coancestry and diagonal self-coancestry. This
is equivalent to
\deqn{\bar{C} = \frac{N - 1}{N} \cdot
  \frac{\bar{a}_{off}}{2} + \frac{1 + \bar{F}}{2N}.}
The implementation computes \eqn{A g} directly from the pedigree and does
not construct the dense relationship matrix \eqn{A}.
}
\examples{
\donttest{
library(data.table)
# Use the sample dataset and simulate a birth year
tp <- tidyped(small_ped)
tp$Year <- 2010 + tp$Gen

# Example 1: Calculate average relationship grouped by Generation (default)
rel_by_gen <- pedrel(tp, by = "Gen")
print(rel_by_gen)

# Example 2: Calculate average relationship grouped by Year
rel_by_year <- pedrel(tp, by = "Year")
print(rel_by_year)

# Example 3: Calculate corrected mean coancestry
coan_by_gen <- pedrel(tp, by = "Gen", scale = "coancestry")
print(coan_by_gen)

# Example 4: Filter calculations with a reference list in a chosen group
candidates <- c("N", "O", "P", "Q", "T", "U", "V", "X", "Y")
rel_subset <- pedrel(tp, by = "Gen", reference = candidates)
print(rel_subset)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-s3.R
\name{print.summary.tidyped}
\alias{print.summary.tidyped}
\title{Print method for summary.tidyped}
\usage{
\method{print}{summary.tidyped}(x, ...)
}
\arguments{
\item{x}{A summary.tidyped object.}

\item{...}{Additional arguments (ignored).}
}
\value{
The input object, invisibly.
}
\description{
Print method for summary.tidyped
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{big_family_size_ped}
\alias{big_family_size_ped}
\title{A large pedigree with big family sizes}
\format{
A data.table with 8 columns:
\describe{
  \item{Ind}{Individual ID}
  \item{Sire}{Sire ID}
  \item{Dam}{Dam ID}
  \item{Sex}{Sex of the individual}
  \item{Year}{Year of birth}
  \item{IndNum}{Numeric ID for individual}
  \item{SireNum}{Numeric ID for sire}
  \item{DamNum}{Numeric ID for dam}
}
}
\usage{
big_family_size_ped
}
\description{
A dataset containing a pedigree with many full-sib individuals per family.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-s3.R
\name{[.tidyped}
\alias{[.tidyped}
\title{Subset a tidyped object}
\usage{
\method{[}{tidyped}(x, ...)
}
\arguments{
\item{x}{A \code{tidyped} object.}

\item{...}{Arguments passed to the \code{data.table} \code{[} method.}
}
\value{
A \code{tidyped} object if the result is still a complete pedigree,
  otherwise a plain \code{data.table}.
}
\description{
Intercepts \code{data.table}'s \code{[} method for \code{tidyped} objects.
After subsetting, the method checks whether the result is still a valid
pedigree (all referenced parents still present). If so, \code{IndNum},
\code{SireNum}, and \code{DamNum} are rebuilt and the \code{tidyped} class
is preserved. If the pedigree becomes structurally incomplete (missing parent
records), the result is degraded to a plain \code{data.table} with a warning.
Column-only selections (missing core columns) also return a plain
\code{data.table}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-s3.R
\name{ensure_complete_tidyped}
\alias{ensure_complete_tidyped}
\title{Internal helper to ensure ped is a complete tidyped object}
\usage{
ensure_complete_tidyped(ped, fun)
}
\arguments{
\item{ped}{An object expected to be a complete tidyped pedigree.}

\item{fun}{Character scalar. Calling function name for the error message.}
}
\value{
A valid, structurally complete tidyped object.
}
\description{
Like \code{ensure_tidyped()}, but also rejects row-truncated pedigree
subsets whose referenced parents are no longer present.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-s3.R
\name{is_complete_pedigree}
\alias{is_complete_pedigree}
\title{Check whether all referenced parents are present}
\usage{
is_complete_pedigree(x)
}
\arguments{
\item{x}{A pedigree-like object.}
}
\value{
Logical scalar.
}
\description{
Check whether all referenced parents are present
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped_style.R
\name{apply_node_styles}
\alias{apply_node_styles}
\title{Apply node styles (color, shape, highlighting)}
\usage{
apply_node_styles(ped_node, highlight_info, shapeby = "sex")
}
\description{
Apply node styles (color, shape, highlighting)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedanalysis.R
\name{print.pedstats}
\alias{print.pedstats}
\title{Print Pedigree Statistics}
\usage{
\method{print}{pedstats}(x, ...)
}
\arguments{
\item{x}{A \code{pedstats} object.}

\item{...}{Additional arguments.}
}
\description{
Print Pedigree Statistics
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-s3.R
\name{as_tidyped}
\alias{as_tidyped}
\title{Restore the tidyped class to a manipulated pedigree}
\usage{
as_tidyped(x)
}
\arguments{
\item{x}{A \code{data.table} or \code{data.frame} that was previously a
tidyped object. It must still contain the core columns: \code{Ind},
\code{Sire}, \code{Dam}, \code{Sex}, \code{Gen}, \code{IndNum},
\code{SireNum}, \code{DamNum}.}
}
\value{
A \code{tidyped} object.
}
\description{
Rapidly restores the \code{tidyped} class to a \code{data.table} or
\code{data.frame} that was previously processed by \code{\link{tidyped}()}
but lost its class attributes due to data manipulation.
}
\details{
This is a lightweight operation that only checks for the required columns
and re-attaches the class---it does \strong{not} re-run the full pedigree
sorting, generation inference, or loop detection.


This helper is intended for objects that still contain the core pedigree
columns and numeric indices, but no longer inherit from \code{tidyped}.
A common reproducible case is \code{rbind()} on two \code{tidyped}
fragments, which typically returns a plain \code{data.table}. Converting a
\code{tidyped} object to a plain \code{data.frame} and then subsetting it
also drops the class.

Some operations, such as \code{merge()} or certain dplyr workflows, may or
may not preserve the \code{tidyped} class depending on the versions of
\pkg{data.table}, \pkg{dplyr}, and the exact method dispatch path used in
the current R session. Therefore, \code{as_tidyped()} should be viewed as a
safe recovery helper rather than something only needed after one specific
verb.

Typical class-loss scenarios include:
\itemize{
  \item \code{rbind(tped1, tped2)} --- often returns plain \code{data.table}
  \item \code{as.data.frame(tped)[rows, ]} --- returns plain \code{data.frame}
  \item manual class removal or serialization / import workflows
}
After such operations, downstream analysis functions (e.g.,
\code{\link{pedstats}}, \code{\link{pedne}}) will either error or
automatically restore the class. You can also call \code{as_tidyped()}
explicitly to restore the class yourself.
}
\examples{
library(visPedigree)
tp <- tidyped(simple_ped)
class(tp)
# [1] "tidyped"    "data.table" "data.frame"

# Simulate class loss via rbind()
tp2 <- rbind(tp[1:5], tp[6:10])
class(tp2)
# [1] "data.table" "data.frame"

# Restore the class
tp3 <- as_tidyped(tp2)
class(tp3)
# [1] "tidyped"    "data.table" "data.frame"

# It can also restore from a plain data.frame if core columns are intact
tp_df <- as.data.frame(tp)
tp4 <- tp_df[tp_df$Gen > 1, ]
class(tp4)
# [1] "data.frame"

tp5 <- as_tidyped(tp4)
class(tp5)
# [1] "tidyped"    "data.table" "data.frame"

}
\seealso{
\code{\link{tidyped}}, \code{\link{new_tidyped}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{complex_ped}
\alias{complex_ped}
\title{A complex pedigree}
\format{
A data.table with a standard pedigree structure.
}
\usage{
complex_ped
}
\description{
A dataset containing a large, complex pedigree covering about 100 generations,
useful for testing the performance and accuracy of partial inbreeding and similar calculations.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-s3.R
\name{summary.tidyped}
\alias{summary.tidyped}
\title{Summary method for tidyped objects}
\usage{
\method{summary}{tidyped}(object, ...)
}
\arguments{
\item{object}{A tidyped object.}

\item{...}{Additional arguments (ignored).}
}
\value{
A summary.tidyped object (list) containing:
\itemize{
  \item \code{n_ind}: Total number of individuals.
  \item \code{n_male}, \code{n_female}, \code{n_unknown_sex}: Sex composition counts.
  \item \code{n_founders}: Number of individuals with no known parents.
  \item \code{n_both_parents}: Count of individuals with complete parentage.
  \item \code{max_gen}, \code{gen_dist}: (Optional) Maximum generation and its distribution.
  \item \code{n_families}, \code{family_sizes}, \code{top_families}: (Optional) Family statistics.
  \item \code{f_stats}, \code{n_inbred}: (Optional) Inbreeding coefficient statistics.
  \item \code{n_cand}, \code{cand_f_stats}: (Optional) Candidate-specific statistics.
}
}
\description{
Summary method for tidyped objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedmatrix.R
\name{expand_pedmat}
\alias{expand_pedmat}
\title{Expand a Compact Pedigree Matrix to Full Dimensions}
\usage{
expand_pedmat(x)
}
\arguments{
\item{x}{A pedmat object from \code{\link{pedmat}}.}
}
\value{
Matrix or vector with original pedigree dimensions:
\itemize{
  \item Matrices: Row and column names set to all individual IDs
  \item Vectors (e.g., method="f"): Names set to all individual IDs
}
The result is \strong{not} a pedmat object (S3 class stripped).
}
\description{
Restores a compact pedmat to its original dimensions by mapping each
individual to their family representative's values. For non-compact matrices,
returns the matrix unchanged.
}
\details{
For compact matrices, full-siblings within the same family will have
identical relationship values in the expanded matrix because they shared
the same representative during calculation.
}
\examples{
tped <- tidyped(small_ped)

# Compact matrix
A_compact <- pedmat(tped, method = "A", compact = TRUE)
dim(A_compact)  # Reduced dimensions

# Expand to full size
A_full <- expand_pedmat(A_compact)
dim(A_full)  # Original dimensions restored

# Non-compact matrices are returned unchanged
A <- pedmat(tped, method = "A", compact = FALSE)
A2 <- expand_pedmat(A)
identical(dim(A), dim(A2))  # TRUE

}
\seealso{
\code{\link{pedmat}}, \code{\link{query_relationship}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inbreed.R
\name{inbreed}
\alias{inbreed}
\title{Calculate inbreeding coefficients}
\usage{
inbreed(ped, ...)
}
\arguments{
\item{ped}{A \code{tidyped} object.}

\item{...}{Additional arguments (currently ignored).}
}
\value{
A \code{tidyped} object with an additional column \strong{f}.
}
\description{
\code{inbreed} function calculates the inbreeding coefficients for all individuals in a tidied pedigree.
}
\details{
This function takes a pedigree tidied by the \code{\link{tidyped}} function and
calculates the inbreeding coefficients using an optimized C++ implementation of
the Sargolzaei & Iwaisaki (2005) LAP (Longest Ancestral Path) bucket algorithm.
This method is the fastest known direct algorithm for computing all inbreeding
coefficients: it replaces the O(\eqn{N^2}) linear scan of Meuwissen & Luo (1992)
with O(1) bucket pops and selective ancestor clearing, giving
\eqn{O(\sum m_i)} total work where \eqn{m_i} is the number of distinct
ancestors of individual \eqn{i}. At \eqn{N = 1{,}000{,}000}, the kernel
completes in approximately 0.12 s — over 10\eqn{\times} faster than the previous
Meuwissen & Luo (1992) implementation and on par with the \pkg{pedigreemm}
reference C implementation of the same algorithm.
It is the core engine used by both \code{tidyped(..., inbreed = TRUE)} and
\code{pedmat(..., method = "f")}, ensuring consistent results across the package.
}
\examples{
library(visPedigree)
data(simple_ped)
ped <- tidyped(simple_ped)
ped_f <- inbreed(ped)
ped_f[f > 0, .(Ind, Sire, Dam, f)]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped.R
\name{prepare_visped_genlab}
\alias{prepare_visped_genlab}
\title{Normalize generation-label display settings}
\usage{
prepare_visped_genlab(genlab)
}
\description{
Normalize generation-label display settings
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-s3.R
\name{new_tidyped}
\alias{new_tidyped}
\title{Internal constructor for tidyped class}
\usage{
new_tidyped(x)
}
\arguments{
\item{x}{A data.table object}
}
\value{
A tidyped object
}
\description{
Internal constructor for tidyped class
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedanalysis.R
\name{print.pediv}
\alias{print.pediv}
\title{Print Genetic Diversity Summary}
\usage{
\method{print}{pediv}(x, ...)
}
\arguments{
\item{x}{A \code{pediv} object.}

\item{...}{Additional arguments.}
}
\description{
Print Genetic Diversity Summary
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedmatrix.R
\name{compact_ped_for_matrix}
\alias{compact_ped_for_matrix}
\title{Compact pedigree by merging full siblings for matrix calculation}
\usage{
compact_ped_for_matrix(ped)
}
\arguments{
\item{ped}{A tidyped object or pedigree data.}
}
\value{
A list containing:
\itemize{
  \item \code{ped_compact}: The compacted pedigree (tidyped object)
  \item \code{compact_map}: Individual-level mapping table
  \item \code{family_summary}: Family-level summary table
  \item \code{compact_stats}: Compression statistics
}
}
\description{
This internal function identifies full siblings (individuals sharing the same
sire and dam) and selects one representative per family. This can dramatically
reduce memory requirements when calculating relationship matrices for pedigrees
with large full-sibling families.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped.R
\name{prepare_visped_geninfo}
\alias{prepare_visped_geninfo}
\title{Attach display labels to the final plotted generation layers}
\usage{
prepare_visped_geninfo(gen_info, labels = NULL)
}
\description{
Attach display labels to the final plotted generation layers
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedhalflife.R
\name{pedhalflife}
\alias{pedhalflife}
\alias{print.pedhalflife}
\alias{plot.pedhalflife}
\title{Calculate Information-Theoretic Diversity Half-Life}
\usage{
pedhalflife(
  ped,
  timevar = "Gen",
  at = NULL,
  nsamples = 1000,
  ncores = 1,
  seed = NULL
)

\method{print}{pedhalflife}(x, ...)

\method{plot}{pedhalflife}(x, type = c("log", "raw"), ...)
}
\arguments{
\item{ped}{A \code{tidyped} object.}

\item{timevar}{Character.
Column name in \code{ped} that defines time points
(e.g. \code{"Gen"}, \code{"Year"}).  Default: \code{"Gen"}.}

\item{at}{Optional vector of values selecting which time points to include
(e.g., \code{2:4}, \code{2010:2020}, or non-consecutive \code{c(2015, 2018, 2022)}).
Values must match entries in the \code{timevar} column.  Non-numeric values
are accepted but the OLS time axis will fall back to sequential indices.
If \code{NULL} (default), all non-NA unique values in \code{timevar} are used.}

\item{nsamples}{Integer.
Sample size per time point for coancestry estimation
(passed to the internal coancestry engine).  Default: 1000.}

\item{ncores}{Integer.
Number of OpenMP threads for C++ backends.  Default: 1.}

\item{seed}{Integer or \code{NULL}.
Random seed for reproducible coancestry sampling.
Default: \code{NULL}.}

\item{x}{A \code{pedhalflife} object.}

\item{...}{Additional arguments (ignored).}

\item{type}{Character.
\code{"log"} for log-transformed values;
\code{"raw"} for \eqn{f_e}, \eqn{f_a}, \eqn{f_g}.}
}
\value{
A list of class \code{pedhalflife} with two \code{data.table}
  components:
  \describe{
    \item{\code{timeseries}}{Per-time-point tracking with columns
      \code{Time} (time-point label from \code{timevar}), \code{NRef},
      \code{fe}, \code{fa}, \code{fg} and their log transformations
      (\code{lnfe}, \code{lnfa}, \code{lnfg}, \code{lnfafe},
      \code{lnfgfa}), plus \code{TimeStep} (numeric OLS time axis).}
    \item{\code{decay}}{Single-row table with \code{lambda_e},
      \code{lambda_b}, \code{lambda_d}, \code{lambda_total}, and
      \code{thalf}.}
  }
}
\description{
Calculates the diversity half-life (\eqn{T_{1/2}}) of a pedigree across
time points using a Renyi-2 entropy cascade framework.  The total loss rate of
genetic diversity is partitioned into three additive components:
\itemize{
  \item \eqn{\lambda_e}: foundation bottleneck (unequal founder
    contributions).
  \item \eqn{\lambda_b}: breeding bottleneck (overuse of key ancestors).
  \item \eqn{\lambda_d}: genetic drift (random sampling loss).
}

The function rolls over time points defined by \code{timevar}, computing
\eqn{f_e} and \eqn{f_a} (via \code{\link{pedcontrib}}) and \eqn{f_g}
(via the internal matrix-free coancestry engine) for each time point. Only
the \eqn{f_g} statistic is requested; no redundant Ne calculations are
performed.
}
\details{
The mathematical identity underlying the cascade is:
\deqn{\ln f_g = \ln f_e + \ln(f_a / f_e) + \ln(f_g / f_a)}
Taking the negative time-slope of each term gives the \eqn{\lambda}
components, which sum exactly by linearity of OLS:
\deqn{\lambda_{total} = \lambda_e + \lambda_b + \lambda_d}

\eqn{T_{1/2} = \ln 2 / \lambda_{total}} is the number of time-units
(time points, years, generations) for diversity to halve.
}
\examples{
\donttest{
library(visPedigree)
data(simple_ped)
tp <- tidyped(simple_ped)

# 1. Calculate half-life using all available generations
hl <- pedhalflife(tp, timevar = "Gen")
print(hl)

# 2. View the underlying log-linear decay plot
plot(hl, type = "log")

# 3. Calculate half-life for a specific time window (e.g., Generations 2 to 4)
hl_subset <- pedhalflife(tp, timevar = "Gen", at = c(2, 3, 4))
print(hl_subset)
}

}
\seealso{
\code{\link{pediv}}, \code{\link{pedcontrib}}, \code{\link{tidyped}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{deep_ped}
\alias{deep_ped}
\title{A deep pedigree}
\format{
A data.table with 4 columns:
\describe{
  \item{Ind}{Individual ID}
  \item{Sire}{Sire ID}
  \item{Dam}{Dam ID}
  \item{Sex}{Sex of the individual}
}
}
\usage{
deep_ped
}
\description{
A dataset containing a pedigree with many generations.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-s3.R
\name{print.tidyped}
\alias{print.tidyped}
\title{Print method for tidyped pedigree}
\usage{
\method{print}{tidyped}(x, ...)
}
\arguments{
\item{x}{A tidyped object}

\item{...}{Additional arguments passed to the data.table print method}
}
\value{
The input object, invisibly.
}
\description{
Print method for tidyped pedigree
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped_graph.R
\name{prepare_initial_nodes}
\alias{prepare_initial_nodes}
\title{Prepare initial node table for igraph conversion}
\usage{
prepare_initial_nodes(ped, labelvar = NULL)
}
\arguments{
\item{ped}{A data.table containing pedigree info.}
}
\description{
Prepare initial node table for igraph conversion
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped_style.R
\name{fade_cols}
\alias{fade_cols}
\title{Fade colors by appending a reduced alpha value}
\usage{
fade_cols(x)
}
\arguments{
\item{x}{Character vector of colors.}
}
\value{
Character vector of faded hex colors.
}
\description{
Converts any R color specification to `#RRGGBB4D` form.
Handles hex colors (`#RRGGBB`, `#RRGGBBAA`) and named colors (e.g. `"red"`).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedanalysis.R
\name{pedgenint}
\alias{pedgenint}
\title{Calculate Generation Intervals}
\usage{
pedgenint(
  ped,
  timevar = NULL,
  unit = c("year", "month", "day", "hour"),
  format = NULL,
  cycle = NULL,
  by = NULL
)
}
\arguments{
\item{ped}{A \code{tidyped} object.}

\item{timevar}{Character. The name of the column containing the \strong{birth date}
(or hatch date) of each individual. The column must be one of:
\itemize{
  \item \code{Date} or \code{POSIXct} (recommended).
  \item A date string parseable by \code{as.POSIXct} (e.g., \code{"2020-03-15"}).
    Use \code{format} for non-ISO strings.
  \item A numeric year (e.g., \code{2020}). Automatically converted to
    \code{Date} (\code{"YYYY-07-01"}) with a message. For finer precision,
    convert to \code{Date} beforehand.
}
If \code{NULL}, auto-detects columns named \code{"BirthYear"}, \code{"Year"},
\code{"BirthDate"}, or \code{"Date"}.}

\item{unit}{Character. Output time unit for the interval:
\code{"year"} (default), \code{"month"}, \code{"day"}, or \code{"hour"}.}

\item{format}{Character. Optional format string for parsing \code{timevar}
when it contains non-standard date strings (e.g., \code{"\%d/\%m/\%Y"}
for \code{"15/03/2020"}).}

\item{cycle}{Numeric. Optional target (designed) length of one generation
cycle expressed in \code{unit}s. When provided, an additional column
\code{GenEquiv} is appended to the result, defined as:
\deqn{GenEquiv_i = \frac{\bar{L}_i}{L_{cycle}}}
where \eqn{\bar{L}_i} is the observed mean interval for pathway \eqn{i} and
\eqn{L_{cycle}} is \code{cycle}. A value > 1 means the observed
interval exceeds the target cycle (lower breeding efficiency).
Example: for Pacific white shrimp with a 180-day target cycle, set
\code{unit = "day", cycle = 180}.}

\item{by}{Character. Optional grouping column (e.g., \code{"Breed"}, \code{"Farm"}).
If provided, intervals are calculated within each group.}
}
\value{
A \code{data.table} with columns:
\itemize{
  \item \code{Group}: Grouping level (if \code{by} is used).
  \item \code{Pathway}: One of \code{"SS"}, \code{"SD"}, \code{"DS"}, \code{"DD"},
    \code{"SO"}, \code{"DO"}, or \code{"Average"}.
    SS/SD/DS/DD require offspring sex; SO (Sire-Offspring) and DO (Dam-Offspring)
    are computed from all parent-offspring pairs regardless of offspring sex.
  \item \code{N}: Number of parent-offspring pairs used.
  \item \code{Mean}: Average generation interval in \code{unit}.
  \item \code{SD}: Standard deviation of the interval.
  \item \code{GenEquiv}: (Optional) Generation equivalents based on \code{cycle}.
}
}
\description{
Computes the generation intervals for the four gametic pathways:
Sire to Son (SS), Sire to Daughter (SD), Dam to Son (DS), and Dam to Daughter (DD).
The generation interval is defined as the age of the parents at the birth of their offspring.
}
\details{
Parent-offspring pairs with zero or negative intervals are excluded from
the calculation because they typically indicate data entry errors or
insufficient time resolution. If many zero intervals are expected (e.g.,
when using \code{unit = "year"} with annual spawners), consider using a
finer time unit such as \code{"month"} or \code{"day"}.

Numeric year columns (e.g., \code{2020}) are automatically converted to
\code{Date} by appending \code{"-07-01"} (mid-year) as a reasonable default.
For more precise results, convert to \code{Date} before calling this function.
}
\examples{
\donttest{
# ---- Basic usage with package dataset (numeric Year auto-converted) ----
tped <- tidyped(big_family_size_ped)
gi <- pedgenint(tped, timevar = "Year")
gi

# ---- Generation equivalents with cycle ----
gi2 <- pedgenint(tped, timevar = "Year", cycle = 2)
gi2
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-s3.R
\name{pedmeta}
\alias{pedmeta}
\title{Access pedigree metadata from a tidyped object}
\usage{
pedmeta(x)
}
\arguments{
\item{x}{A tidyped object.}
}
\value{
The \code{ped_meta} list, or \code{NULL} if not set.
}
\description{
Access pedigree metadata from a tidyped object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{inbred_ped}
\alias{inbred_ped}
\title{A highly inbred pedigree}
\format{
A data.table with 5 columns:
\describe{
  \item{Ind}{Individual ID}
  \item{Sire}{Sire ID}
  \item{Dam}{Dam ID}
  \item{Sex}{Sex of the individual}
  \item{Gen}{Generation number}
}
}
\usage{
inbred_ped
}
\description{
A simulated pedigree designed to demonstrate high levels of inbreeding and partial inbreeding decomposition.
Contains full-sib mating and backcrossing.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped_render.R
\name{plot_ped_igraph}
\alias{plot_ped_igraph}
\title{Render pedigree graph using Two-Pass strategy}
\usage{
plot_ped_igraph(
  g,
  l,
  node_size,
  gen_info = NULL,
  genlab = FALSE,
  genlabcex = NULL,
  ...
)
}
\description{
Render pedigree graph using Two-Pass strategy
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedanalysis.R
\name{pediv}
\alias{pediv}
\title{Calculate Genetic Diversity Indicators}
\usage{
pediv(
  ped,
  reference = NULL,
  top = 20,
  nsamples = 1000,
  ncores = 1,
  seed = NULL
)
}
\arguments{
\item{ped}{A \code{tidyped} object.}

\item{reference}{Character vector. Optional subset of individual IDs defining the reference population.
If NULL, uses all individuals in the most recent generation.}

\item{top}{Integer. Number of top contributors to return in founder/ancestor tables. Default is 20.}

\item{nsamples}{Integer. Number of individuals sampled per cohort for the coancestry
Ne method and for \eqn{f_g} estimation. Very large cohorts are sampled down to this
size to control memory usage (default: 1000).}

\item{ncores}{Integer. Number of cores for parallel processing in the coancestry method. Default is 1.}

\item{seed}{Integer or NULL. Random seed passed to \code{set.seed()} before sampling
in the coancestry method, ensuring reproducible \eqn{f_g} and \eqn{N_e} estimates.
Default is \code{NULL} (no fixed seed).}
}
\value{
A list with class \code{pediv} containing:
\itemize{
  \item \code{summary}: A single-row \code{data.table} with columns
    \code{NRef}, \code{NFounder}, \code{feH}, \code{fe}, \code{NAncestor},
    \code{faH}, \code{fa}, \code{fafe}, \code{fg}, \code{MeanCoan},
    \code{GeneDiv}, \code{NSampledCoan}, \code{NeCoancestry}, \code{NeInbreeding},
    \code{NeDemographic}.
    Here \code{feH} and \code{faH} are the Shannon-entropy (\eqn{q=1})
    effective numbers of founders and ancestors, respectively.
    \code{GeneDiv} is the pedigree-based retained genetic diversity,
    computed as \eqn{1 - \bar{C}}, where \eqn{\bar{C}} is the
    diagonal-corrected population mean coancestry (\code{MeanCoan}).
  \item \code{founders}: A \code{data.table} of top founder contributions.
  \item \code{ancestors}: A \code{data.table} of top ancestor contributions.
}
}
\description{
Combines founder/ancestor contributions ($f_e$, $f_a$) and effective population
size estimates (Ne) from three methods into a single summary object.
}
\details{
Internally calls \code{\link{pedcontrib}} for \eqn{f_e} and \eqn{f_a}.
The coancestry method is called via the internal \code{calc_ne_coancestry()}
function directly so that \eqn{f_g} and the Ne estimate can be obtained from
the same traced pedigree and the same batched matrix-free relationship
products without duplication.
The inbreeding and demographic Ne methods are obtained via \code{\link{pedne}}.
All calculations use the same \code{reference} population.
If any method fails (e.g., insufficient pedigree depth), its value is \code{NA}
rather than stopping execution.

\eqn{f_g} (founder genome equivalents, Caballero & Toro 2000) is estimated from
the diagonal-corrected mean coancestry of the reference population:
\deqn{\hat{\bar{C}} = \frac{N-1}{N} \cdot \frac{\bar{a}_{off}}{2} + \frac{1 + \bar{F}_s}{2N}}
\deqn{f_g = \frac{1}{2 \hat{\bar{C}}}}
where \eqn{N} is the full reference cohort size, \eqn{\bar{a}_{off}} is the
off-diagonal mean relationship among sampled individuals, and \eqn{\bar{F}_s}
is their mean inbreeding coefficient.
}
\examples{
\donttest{
tp <- tidyped(small_ped)
div <- pediv(tp, reference = c("Z1", "Z2", "X", "Y"), seed = 42L)
print(div)

# Access Shannon effective numbers from summary
div$summary$feH   # Shannon effective founders (q=1)
div$summary$faH   # Shannon effective ancestors (q=1)

# Founder diversity profile: NFounder >= feH >= fe
with(div$summary, c(NFounder = NFounder, feH = feH, fe = fe))
}

}
\seealso{
\code{\link{pedcontrib}}, \code{\link{pedne}}, \code{\link{pedstats}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedmatrix.R
\name{pedmat}
\alias{pedmat}
\title{Genetic Relationship Matrices and Inbreeding Coefficients}
\usage{
pedmat(
  ped,
  method = "A",
  sparse = TRUE,
  invert_method = "auto",
  threads = 0,
  compact = FALSE
)
}
\arguments{
\item{ped}{A tidied pedigree from \code{\link{tidyped}}. Must be a single
pedigree, not a splitped object. For splitped results, use 
\code{pedmat(ped_split$GP1, ...)} to process individual groups.}

\item{method}{Character, one of:
\itemize{
  \item \code{"A"}: Additive (numerator) relationship matrix (default)
  \item \code{"f"}: Inbreeding coefficients (returns named vector)
  \item \code{"Ainv"}: Inverse of A using Henderson's rules (O(n) complexity)
  \item \code{"D"}: Dominance relationship matrix
  \item \code{"Dinv"}: Inverse of D (requires matrix inversion)
  \item \code{"AA"}: Additive-by-additive epistatic matrix (A # A)
  \item \code{"AAinv"}: Inverse of AA
}}

\item{sparse}{Logical, if \code{TRUE} returns sparse Matrix (recommended for
large pedigrees). Default is \code{TRUE}.}

\item{invert_method}{Character, method for matrix inversion (Dinv/AAinv only):
\itemize{
  \item \code{"auto"}: Auto-detect and use optimal method (default)
  \item \code{"sympd"}: Force Cholesky decomposition (faster for SPD matrices)
  \item \code{"general"}: Force general LU decomposition
}}

\item{threads}{Integer. Number of OpenMP threads to use. Use 0 to keep
  the system/default setting. Currently, multi-threading is explicitly implemented for:
\itemize{
  \item \code{"D"}: Dominance relationship matrix (significant speedup).
  \item \code{"Ainv"}: Inverse of A (only for large pedigrees, n >= 5000).
}
For \code{"Dinv"}, \code{"AA"}, and \code{"AAinv"}, parallelism depends on the 
linked BLAS/LAPACK library (e.g., OpenBLAS, MKL, Accelerate) and is not 
controlled by this parameter. Methods \code{"A"} and \code{"f"} are single-threaded.}

\item{compact}{Logical, if \code{TRUE} compacts full-sibling families by
selecting one representative per family. This dramatically reduces matrix
dimensions for pedigrees with large full-sib groups. See Details.}
}
\value{
Returns a matrix or vector with S3 class \code{"pedmat"}.

\strong{Object type by method:}
\itemize{
  \item \code{method="f"}: Named numeric vector of inbreeding coefficients
  \item All other methods: Sparse or dense matrix (depending on \code{sparse})
}

\strong{S3 Methods:}
\itemize{
  \item \code{print(x)}: Display matrix with metadata header
  \item \code{\link{summary_pedmat}(x)}: Detailed statistics (size, compression, mean, density)
  \item \code{dim(x)}, \code{length(x)}, \code{Matrix::diag(x)}, \code{t(x)}: Standard operations
  \item \code{x[i, j]}: Subsetting (behaves like underlying matrix)
  \item \code{as.matrix(x)}: Convert to base matrix
}

\strong{Accessing Metadata (use \code{attr()}, not \code{$}):}
\itemize{
  \item \code{attr(x, "ped")}: The pedigree used (or compact pedigree if \code{compact=TRUE})
  \item \code{attr(x, "ped_compact")}: Compact pedigree (when \code{compact=TRUE})
  \item \code{attr(x, "method")}: Calculation method used
  \item \code{attr(x, "call_info")}: Full calculation metadata (timing, sizes, etc.)
  \item \code{names(attributes(x))}: List all available attributes
}

\strong{Additional attributes when \code{compact = TRUE}:}
\itemize{
  \item \code{attr(x, "compact_map")}: data.table mapping individuals to representatives
  \item \code{attr(x, "family_summary")}: data.table summarizing merged families
  \item \code{attr(x, "compact_stats")}: Compression statistics (ratio, n_families, etc.)
}
}
\description{
Optimized calculation of additive (A), dominance (D), epistatic (AA) 
relationship matrices, their inverses, and inbreeding coefficients (f).
Uses Rcpp with Meuwissen & Luo (1992) algorithm for efficient computation.
}
\details{
\strong{API Design:}

Only a single method may be requested per call. This design prevents
accidental heavy computations. If multiple matrices are needed, call
\code{pedmat()} separately for each method.

\strong{Compact Mode (\code{compact = TRUE}):}

Full-siblings share identical relationships with all other individuals.
Compact mode exploits this by selecting one representative per full-sib
family, dramatically reducing matrix size. For example, a pedigree with
170,000 individuals might compact to 1,800 unique relationship patterns.

Key features:
\itemize{
  \item \code{\link{query_relationship}(x, id1, id2)}: Query any individual
        pair, including merged siblings (automatic lookup)
  \item \code{\link{expand_pedmat}(x)}: Restore full matrix dimensions
  \item \code{\link{vismat}(x)}: Visualize directly (auto-expands compact)
}

\strong{Performance Notes:}
\itemize{
  \item \strong{Ainv}: O(n) complexity using Henderson's rules. Fast for any size.
  \item \strong{Dinv/AAinv}: O(n³) matrix inversion. Practical limits:
    \itemize{
      \item n < 500: ~10-20 ms
      \item n = 1,000: ~40-60 ms  
      \item n = 2,000: ~130-150 ms
      \item n > 2,000: Consider using \code{compact = TRUE}
    }
  \item \strong{Memory}: Sparse matrices use ~O(nnz) memory; dense use O(n²)
}
}
\examples{
# Basic usage with small pedigree
library(visPedigree)
tped <- tidyped(small_ped)

# --- Additive Relationship Matrix (default) ---
A <- pedmat(tped)
A["A", "B"]      # Relationship between A and B
Matrix::diag(A)  # Diagonal = 1 + F (inbreeding)

# --- Inbreeding Coefficients ---
f <- pedmat(tped, method = "f")
f["Z1"]  # Inbreeding of individual Z1

# --- Using summary_pedmat() ---
summary_pedmat(A)   # Detailed matrix statistics

# --- Accessing Metadata ---
attr(A, "ped")              # Original pedigree
attr(A, "method")           # "A"
names(attributes(A))        # All available attributes

# --- Compact Mode (for large full-sib families) ---
A_compact <- pedmat(tped, method = "A", compact = TRUE)

# Query relationships (works for any individual, including merged sibs)
query_relationship(A_compact, "Z1", "Z2")  # Full-sibs Z1 and Z2

# View compression statistics
attr(A_compact, "compact_stats")
attr(A_compact, "family_summary")

# Expand back to full size
A_full <- expand_pedmat(A_compact)
dim(A_full)  # Original dimensions restored

# --- Inverse Matrices ---
Ainv <- pedmat(tped, method = "Ainv")  # Henderson's rules (fast)

# --- Dominance and Epistatic ---
D <- pedmat(tped, method = "D")
AA <- pedmat(tped, method = "AA")

# --- Visualization (requires display device) ---
\donttest{
vismat(A)                       # Heatmap of relationship matrix
vismat(A_compact)               # Works with compact matrices
vismat(A, by = "Gen")     # Group by generation
}

}
\references{
Meuwissen, T. H. E., & Luo, Z. (1992). Computing inbreeding coefficients 
in large populations. Genetics Selection Evolution, 24(4), 305-313.

Henderson, C. R. (1976). A simple method for computing the inverse of a
numerator relationship matrix used in prediction of breeding values.
Biometrics, 32(1), 69-83.
}
\seealso{
\code{\link{tidyped}} for preparing pedigree data,
\code{\link{pedprod}} for matrix-free products with A or Ainv,
\code{\link{query_relationship}} for querying individual pairs,
\code{\link{expand_pedmat}} for restoring full dimensions,
\code{\link{vismat}} for visualization,
\code{\link{inbreed}} for simple inbreeding calculation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{half_founder_ped}
\alias{half_founder_ped}
\title{A pedigree with half founders}
\format{
A data.frame with 4 columns:
\describe{
  \item{Ind}{Individual ID}
  \item{Sire}{Sire ID}
  \item{Dam}{Dam ID}
  \item{Sex}{Sex of the individual}
}
}
\usage{
half_founder_ped
}
\description{
A dataset from ENDOG containing individuals with a single missing parent (half founders).
Useful for testing genetic algorithms correctly conserving probability mass for missing lineages.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vismat.R
\name{vismat}
\alias{vismat}
\title{Visualize Relationship Matrices}
\usage{
vismat(
  mat,
  ped = NULL,
  type = "heatmap",
  ids = NULL,
  reorder = TRUE,
  by = NULL,
  grouping = NULL,
  labelcex = NULL,
  ...
)
}
\arguments{
\item{mat}{A relationship matrix. Can be one of the following types:
\itemize{
  \item A \code{pedmat} object returned by \code{\link{pedmat}} — including
        compact matrices. When \code{by} is specified, group-level means are
        computed directly from the compact matrix (no full expansion needed).
        Without \code{by}, compact matrices are automatically expanded to
        full dimensions before plotting (see Details).
  \item A \code{\link{tidyped}} object (automatically calculates additive
        relationship matrix A).
  \item A named list containing matrices (preferring A, D, AA).
  \item A standard \code{matrix} or \code{Matrix} object.
}
\strong{Note}: Inverse matrices (Ainv, Dinv, AAinv) are not supported for 
visualization because their elements do not represent meaningful relationship
coefficients.}

\item{ped}{Optional. A tidied pedigree object (\code{tidyped}), used for 
extracting labels or grouping information. Required when using the 
\code{by} parameter with a plain matrix input. If \code{mat} is a 
\code{pedmat} object, the pedigree is extracted automatically.}

\item{type}{Character, type of visualization. Supported options:
\itemize{
  \item \code{"heatmap"}: Relationship matrix heatmap (default). Uses a 
        Nature Genetics style color palette (white-orange-red-dark red),
        with optional hierarchical clustering and group aggregation.
  \item \code{"histogram"}: Distribution histogram of relationship coefficients.
        Shows the frequency distribution of lower triangular elements (pairwise kinship).
}}

\item{ids}{Character vector specifying individual IDs to display. Used to 
filter and display a submatrix of specific individuals. If \code{NULL} 
(default), all individuals are shown.}

\item{reorder}{Logical. If \code{TRUE} (default), rows and columns are 
reordered using hierarchical clustering (Ward.D2 method) to bring closely 
related individuals together. Only affects heatmap visualization. 
Automatically skipped for large matrices (N > VISMAT_REORDER_MAX, default 2 000)
to improve performance.

\strong{Clustering principle}: Based on relationship profile distance (Euclidean
distance between rows). Full-sibs have nearly identical relationship profiles
with the whole population, so they cluster tightly together and appear as
contiguous blocks in the heatmap.}

\item{by}{Optional. Column name in \code{ped} to group by (e.g., 
\code{"Family"}, \code{"Gen"}, \code{"Year"}). When grouping is enabled:
\itemize{
  \item Individual-level matrix is aggregated to a group-level matrix 
        (computing mean relationship coefficients between groups).
  \item For \code{"Family"} grouping, founders without family assignment are excluded.
  \item For other grouping columns, NA values are assigned to an \code{"Unknown"} group.
}
Useful for visualizing population structure in large pedigrees.}

\item{grouping}{\code{[Deprecated]} Use \code{by} instead.}

\item{labelcex}{Numeric. Manual control for font size of individual labels. 
If \code{NULL} (default), uses a dynamic font size that adjusts automatically 
based on matrix dimensions (range 0.2–0.7). Labels are hidden automatically
when N > VISMAT_LABEL_MAX (default 500).}

\item{...}{Additional arguments passed to the underlying plotting function:
\itemize{
  \item Heatmap uses \code{\link[lattice]{levelplot}}: can set \code{main},
        \code{xlab}, \code{ylab}, \code{col.regions}, \code{colorkey},
        \code{scales}, etc.
  \item Histogram uses \code{\link[lattice]{histogram}}: can set \code{main},
        \code{xlab}, \code{ylab}, \code{nint} (number of bins), etc.
}}
}
\value{
Invisibly returns the \code{lattice} plot object. The plot is 
rendered on the current graphics device.
}
\description{
\code{vismat} provides visualization tools for relationship matrices (A, D, AA),
supporting individual-level heatmaps and relationship coefficient histograms.
This function is useful for exploring population genetic structure, identifying
inbred individuals, and analyzing kinship between families.
}
\details{
\subsection{Compact Matrix Handling}{
When \code{mat} is a compact \code{pedmat} object (created with
\code{pedmat(..., compact = TRUE)}):
\itemize{
  \item \strong{With \code{by}}: Group-level mean relationships are computed
        algebraically from the K×K compact matrix, including a correction for
        sibling off-diagonal values. This avoids expanding to the full N×N
        matrix, making family-level or generation-level visualization feasible
        even for pedigrees with hundreds of thousands of individuals.
  \item \strong{Without \code{by}, N > VISMAT_EXPAND_MAX (5 000)}: The compact K\eqn{\times}K matrix
        is plotted directly using representative individuals. Labels show the
        number of individuals each representative stands for, e.g.,
        \code{"ID (\u00d7350)"}. This avoids memory-intensive full expansion.
  \item \strong{Without \code{by}, N \eqn{\le} 5 000}: The compact matrix is
        expanded via \code{\link{expand_pedmat}} to restore full dimensions.
}
}

\subsection{Heatmap}{
\itemize{
  \item Uses a Nature Genetics style color palette (white to orange to red to dark red).
  \item Hierarchical clustering reordering (Ward.D2) is enabled by default.
  \item Grid lines shown when N \eqn{\le} VISMAT_GRID_MAX (100);
        labels shown when N \eqn{\le} VISMAT_LABEL_MAX (500).
  \item \code{mat[1,1]} is displayed at the top-left corner.
}
}

\subsection{Histogram}{
\itemize{
  \item Shows the distribution of lower-triangular elements (pairwise kinship).
  \item X-axis: relationship coefficient values; Y-axis: frequency percentage.
}
}

\subsection{Performance}{
The following automatic thresholds are defined as package-internal
constants (\code{VISMAT_*}) at the top of \code{R/vismat.R}:
\itemize{
  \item Passing a \code{tidyped} object together with \code{by} uses
        matrix-free additive relationship products to compute the grouped
        heatmap directly, without constructing the full individual-level
        relationship matrix.
  \item \code{VISMAT_EXPAND_MAX} (5 000): compact matrices with original
        N above this are shown in representative view instead of expanding.
  \item \code{VISMAT_REORDER_MAX} (2 000): hierarchical clustering is
        automatically skipped.
  \item \code{VISMAT_LABEL_MAX} (500): individual labels are hidden.
  \item \code{VISMAT_GRID_MAX} (100): cell grid lines are hidden.
  \item \code{by} grouping uses vectorized \code{rowsum()} algebra — suitable
        for large matrices.
}
}

\subsection{Interpreting Relationship Coefficients}{
For the additive relationship matrix A:
\itemize{
  \item Diagonal elements = 1 + F (F = inbreeding coefficient).
  \item Off-diagonal elements = 2 × kinship coefficient.
  \item 0: unrelated; 0.25: half-sibs / grandparent–grandchild;
        0.5: full-sibs / parent–offspring; 1.0: same individual.
}
}
}
\examples{
library(visPedigree)
data(small_ped)
ped <- tidyped(small_ped)

# ============================================================
# Basic Usage
# ============================================================

# Method 1: from tidyped object (auto-computes A)
vismat(ped)

# Method 2: from pedmat object
A <- pedmat(ped)
vismat(A)

# Method 3: from plain matrix
vismat(as.matrix(A))

# ============================================================
# Compact Pedigree (auto-expanded before plotting)
# ============================================================

# For pedigrees with large full-sib families, compute a compact matrix
# first for efficiency, then pass directly to vismat() — it automatically
# expands back to full dimensions.
A_compact <- pedmat(ped, compact = TRUE)
vismat(A_compact)   # prints: "Expanding compact matrix (N -> M individuals)"

# For very large pedigrees, aggregate to a group-level view instead
vismat(A, ped = ped, by = "Gen",
       main = "Mean Relationship Between Generations")

# ============================================================
# Heatmap Customization
# ============================================================

# Custom title and axis labels
vismat(A, main = "Additive Relationship Matrix",
       xlab = "Individual", ylab = "Individual")

# Preserve original pedigree order (no clustering)
vismat(A, reorder = FALSE)

# Custom label font size
vismat(A, labelcex = 0.5)

# Custom color palette (blue-white-red)
vismat(A, col.regions = colorRampPalette(c("blue", "white", "red"))(100))

# ============================================================
# Display a Subset of Individuals
# ============================================================

target_ids <- rownames(A)[1:8]
vismat(A, ids = target_ids)

# ============================================================
# Histogram of Relationship Coefficients
# ============================================================

vismat(A, type = "histogram")
vismat(A, type = "histogram", nint = 30)

# ============================================================
# Group-level Aggregation
# ============================================================

# Group by generation
vismat(A, ped = ped, by = "Gen",
       main = "Mean Relationship Between Generations")

# Group by full-sib family (founders without a family are excluded)
vismat(A, ped = ped, by = "Family")

# ============================================================
# Other Relationship Matrices
# ============================================================

# Dominance relationship matrix
D <- pedmat(ped, method = "D")
vismat(D, main = "Dominance Relationship Matrix")

}
\seealso{
\code{\link{pedmat}} for computing relationship matrices,
\code{\link{expand_pedmat}} for manually restoring compact matrix dimensions,
\code{\link{query_relationship}} for querying individual pairs,
\code{\link{tidyped}} for tidying pedigree data,
\code{\link{visped}} for visualizing pedigree structure graphs,
\code{\link[lattice]{levelplot}}, \code{\link[lattice]{histogram}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedprod.R
\name{pedprod}
\alias{pedprod}
\title{Matrix-Free Pedigree Relationship Products}
\usage{
pedprod(ped, x, method = c("A", "Ainv"))
}
\arguments{
\item{ped}{A complete pedigree accepted by \code{\link{tidyped}}, preferably
an existing \code{tidyped} object.}

\item{x}{A numeric or logical vector, or a numeric or logical matrix.
An unnamed vector must have one value per pedigree individual in
\code{IndNum} order. A named vector may contain a subset of individuals;
omitted individuals receive value zero. For a matrix, the same rules apply
to its rows and \code{rownames}.}

\item{method}{Character scalar. \code{"A"} (default) computes products with
the additive relationship matrix. \code{"Ainv"} computes products with its
inverse.}
}
\value{
For vector input, a named numeric vector in pedigree order. For
  matrix input, a numeric matrix whose rows are in pedigree order; input
  column names are preserved.
}
\description{
Computes products of the form \eqn{A x}, \eqn{A X}, \eqn{A^{-1} x}, or
\eqn{A^{-1} X} directly from an ordered pedigree, without constructing the
full additive relationship matrix \eqn{A}. This is useful for large
pedigrees where explicitly storing the dense \eqn{n \times n} matrix is
infeasible.
}
\details{
For an ordered pedigree, the additive relationship matrix can be written as
\deqn{A = T D T^\prime,}
where \eqn{T} is the transmission matrix and \eqn{D} contains Mendelian
sampling variances. \code{pedprod()} applies these factors through backward
and forward pedigree traversals. For an \eqn{n \times p} right-hand side,
computation requires \eqn{O(np)} time and working memory rather than
materializing the \eqn{O(n^2)} relationship matrix.

Named inputs are aligned by individual ID. Every supplied name must occur in
\code{ped$Ind}, and duplicate, missing, or empty names are rejected. Missing
values and non-finite input values are also rejected to avoid propagating an
invalid value through all descendants or relatives.
}
\examples{
tped <- tidyped(small_ped)

# Equal contributions from two candidates; unspecified individuals are zero.
contributions <- c(Z1 = 0.5, Z2 = 0.5)
relationship_to_group <- pedprod(tped, contributions)

# Average additive relationship among the weighted candidates: c' A c.
sum(contributions * relationship_to_group[names(contributions)])

# Multiple contribution schemes can be evaluated in one call.
schemes <- cbind(
  Equal = c(A = 0.5, B = 0.5),
  SireA = c(A = 1.0, B = 0.0)
)
pedprod(tped, schemes)

}
\references{
Colleau, J. J. (2002). An indirect approach to the extensive calculation of
relationship coefficients. \emph{Genetics Selection Evolution}, 34,
409--421. \doi{10.1051/gse:2002015}

Colleau, J. J., Palhiere, I., Rodriguez-Ramilo, S. T., & Legarra, A.
(2017). A fast indirect method to compute functions of genomic relationships
concerning genotyped and ungenotyped individuals, for diversity management.
\emph{Genetics Selection Evolution}, 49, 87.
\doi{10.1186/s12711-017-0363-9}
}
\seealso{
\code{\link{pedmat}} for explicitly constructing relationship
  matrices and \code{\link{pedrel}} for group relationship summaries.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-s3.R
\name{is_tidyped}
\alias{is_tidyped}
\title{Test if an object is a tidyped}
\usage{
is_tidyped(x)
}
\arguments{
\item{x}{An object to test.}
}
\value{
Logical scalar.
}
\description{
Test if an object is a tidyped
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyped.R
\name{tidyped}
\alias{tidyped}
\title{Tidy and prepare a pedigree}
\usage{
tidyped(
  ped,
  cand = NULL,
  trace = "up",
  tracegen = NULL,
  addgen = TRUE,
  addnum = TRUE,
  inbreed = FALSE,
  selfing = FALSE,
  genmethod = "top",
  ...
)
}
\arguments{
\item{ped}{A data.table or data frame containing the pedigree. The first three columns must be \strong{individual}, \strong{sire}, and \strong{dam} IDs. Additional columns, such as sex or generation, can be included. Column names can be customized, but their order must remain unchanged. Individual IDs should not be coded as "", " ", "0", "*", or "NA"; otherwise, they will be removed. Missing parents should be denoted by "NA", "0", or "*". Spaces and empty strings ("") are also treated as missing parents but are not recommended.}

\item{cand}{A character vector of individual IDs, or NULL. If provided, only the candidates and their ancestors/descendants are retained.}

\item{trace}{A character value specifying the tracing direction: "\strong{up}", "\strong{down}", or "\strong{all}". "up" traces ancestors; "down" traces descendants; "all" traces the union of ancestors and descendants. This parameter is only used if \code{cand} is not NULL. Default is "up".}

\item{tracegen}{An integer specifying the number of generations to trace. This parameter is only used if \code{trace} is not NULL. If NULL or 0, all available generations are traced.}

\item{addgen}{A logical value indicating whether to generate generation numbers. Default is TRUE, which adds a \strong{Gen} column to the output.}

\item{addnum}{A logical value indicating whether to generate a numeric pedigree. Default is TRUE, which adds \strong{IndNum}, \strong{SireNum}, and \strong{DamNum} columns to the output.}

\item{inbreed}{A logical value indicating whether to calculate inbreeding coefficients. Default is FALSE. If TRUE, an \strong{f} column is added to the output. This uses the same optimized engine as \code{pedmat(..., method = "f")}.}

\item{selfing}{A logical value indicating whether to allow the same individual to appear as both sire and dam. This is common in plant breeding (monoecious species) where the same plant can serve as both male and female parent. If TRUE, individuals appearing in both the Sire and Dam columns will be assigned Sex = "monoecious" instead of triggering an error. Default is FALSE.}

\item{genmethod}{A character value specifying the generation assignment method: "\strong{top}" or "\strong{bottom}". "top" (top-aligned) assigns generations from parents to offspring, starting founders at Gen 1. "bottom" (bottom-aligned) assigns generations from offspring to parents, aligning terminal nodes at the bottom. Default is "top".}

\item{...}{Additional arguments passed to \code{\link{inbreed}}.}
}
\value{
A \code{tidyped} object (which inherits from \code{data.table}). Individual, sire, and dam ID columns are renamed to \strong{Ind}, \strong{Sire}, and \strong{Dam}. Missing parents are replaced with \strong{NA}. The \strong{Sex} column contains \code{"male"}, \code{"female"}, \code{"monoecious"}, or \code{NA}. The \strong{Cand} column is included if \code{cand} is not NULL. The \strong{Gen} column is included if \code{addgen} is TRUE. The \strong{IndNum}, \strong{SireNum}, and \strong{DamNum} columns are included if \code{addnum} is TRUE. The \strong{Family} and \strong{FamilySize} columns identify full-sibling families (for example, \code{"AxB"} for offspring of sire \code{A} and dam \code{B}). The \strong{f} column is included if \code{inbreed} is TRUE.
}
\description{
This function standardizes pedigree records, checks for duplicate IDs and
incompatible parental roles, detects pedigree loops, injects missing
founders, assigns generation numbers, sorts the pedigree, and optionally
traces the pedigree of specified candidates. If the \code{cand} parameter
contains individual IDs, only those individuals and their ancestors or
descendants are retained. Tracing direction and the number of generations
can be specified using the \code{trace} and \code{tracegen} parameters.
}
\details{
Compared to the legacy version, this function reports cyclic pedigrees more
clearly and uses a mixed implementation. There are two candidate-tracing
paths: when the input is a raw pedigree, \code{igraph} is used for loop
checking, candidate tracing, and topological sorting; when the input is an
already validated \code{tidyped} object and \code{cand} is supplied,
tracing and topological sorting use integer-indexed C++ routines. Generation
assignment can be performed using either a top-down approach (default,
aligning founders at the top) or a bottom-up approach (aligning terminal
nodes at the bottom).
}
\examples{
library(visPedigree)
library(data.table)

# Tidy a simple pedigree
tidy_ped <- tidyped(simple_ped)
head(tidy_ped)

# Trace ancestors of a specific individual within 2 generations
tidy_ped_tracegen <- tidyped(simple_ped, cand = "J5X804", trace = "up", tracegen = 2)
head(tidy_ped_tracegen)

# Trace both ancestors and descendants for multiple candidates
# This is highly optimized and works quickly even on 100k+ individuals
cand_list <- c("J5X804", "J3Y620")
tidy_ped_all <- tidyped(simple_ped, cand = cand_list, trace = "all")

# Check for loops (will error if loops exist)
try(tidyped(loop_ped))

# Example with a large pedigree: extract 2 generations of ancestors for 2007 candidates
cand_2007 <- big_family_size_ped[Year == 2007, Ind]
\donttest{
tidy_big <- tidyped(big_family_size_ped, cand = cand_2007, trace = "up", tracegen = 2)
summary(tidy_big)
}

}
\seealso{
\code{\link{summary.tidyped}} for summarizing tidyped objects
\code{\link{visped}} for visualizing pedigree structure
\code{\link{pedmat}} for computing relationship matrices
\code{\link{vismat}} for visualizing relationship matrices
\code{\link{splitped}} for splitting pedigree into connected components
\code{\link{inbreed}} for calculating inbreeding coefficients
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped.R
\name{prepare_visped_labels}
\alias{prepare_visped_labels}
\title{Normalize custom labels before pedigree tidying}
\usage{
prepare_visped_labels(ped, labelvar)
}
\description{
Normalize custom labels before pedigree tidying
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-s3.R
\name{ensure_tidyped}
\alias{ensure_tidyped}
\title{Internal helper to ensure ped is a tidyped object}
\usage{
ensure_tidyped(ped)
}
\arguments{
\item{ped}{An object expected to be a tidyped.}
}
\value{
A valid tidyped object.
}
\description{
If the object has lost its \code{tidyped} class (e.g., after \code{merge()},
\code{rbind()}, or dplyr operations) but still contains the required columns,
the class is automatically restored with an informational message. Otherwise,
an error is raised guiding the user to call \code{tidyped()} or
\code{as_tidyped()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedanalysis.R
\name{pedecg}
\alias{pedecg}
\title{Calculate Equi-Generate Coefficient}
\usage{
pedecg(ped)
}
\arguments{
\item{ped}{A \code{tidyped} object.}
}
\value{
A \code{data.table} with columns:
\itemize{
  \item \code{Ind}: Individual ID.
  \item \code{ECG}: Equi-Generate Coefficient.
  \item \code{FullGen}: Fully traced generations (depth of complete ancestry).
  \item \code{MaxGen}: Maximum ancestral generations (depth of deepest ancestor).
}
}
\description{
Estimates the number of distinct ancestral generations using the Equi-Generate Coefficient (ECG).
The ECG is calculated as 1/2 of the sum of the parents' ECG values plus 1.
}
\examples{
tp <- tidyped(simple_ped)
ecg <- pedecg(tp)

# ECG combines pedigree depth and completeness
head(ecg)

# Individuals with deeper and more complete ancestry have larger ECG values
ecg[order(-ECG)][1:5]

}
\references{
Boichard, D., Maignel, L., & Verrier, E. (1997). The value of using probabilities of gene origin to measure genetic variability in a population. Genetics Selection Evolution, 29(1), 5.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedanalysis.R
\name{pedne}
\alias{pedne}
\title{Calculate Effective Population Size}
\usage{
pedne(
  ped,
  method = c("coancestry", "inbreeding", "demographic"),
  by = NULL,
  reference = NULL,
  nsamples = 1000,
  ncores = 1,
  seed = NULL
)
}
\arguments{
\item{ped}{A \code{tidyped} object.}

\item{method}{Character. The method to compute Ne. One of \code{"coancestry"} (default), \code{"inbreeding"}, or \code{"demographic"}.}

\item{by}{Character. The name of the column used to group cohorts (e.g., "Year", "BirthYear").
If NULL, calculates overall Ne for all individuals.}

\item{reference}{Character vector. Optional subset of individual IDs defining the reference cohort.
If NULL, uses all individuals in the pedigree.}

\item{nsamples}{Integer. Number of individuals to randomly sample per cohort when using the \code{"coancestry"} method. Very large cohorts will be sampled down to this size to save memory and time (default: 1000).}

\item{ncores}{Integer. Number of cores for parallel processing. Currently only effective for \code{method = "coancestry"} (default: 1).}

\item{seed}{Integer or NULL. Random seed passed to \code{set.seed()} before sampling
in the coancestry method, ensuring reproducible \eqn{N_e} estimates.
Default is \code{NULL} (no fixed seed).}
}
\value{
A \code{data.table} with columns:
\itemize{
  \item \code{Cohort}: Cohort or grouping variable value.
  \item \code{N}: Number of individuals in the cohort.
  \item \code{Ne}: Calculated effective population size.
  \item \code{...}: Additional columns depending on the selected method (e.g., \code{NSampled}, \code{DeltaC}, \code{MeanF}, \code{DeltaF}, \code{Nm}, \code{Nf}).
}
}
\description{
Calculates the effective population size (Ne) based on the rate of
coancestry, the rate of inbreeding, or demographic parent numbers.
}
\details{
The effective population size can be calculated using one of three methods:

\itemize{
  \item \strong{"coancestry"} (Default): Based on the rate of coancestry (\eqn{f_{ij}}) between 
  pairs of individuals. In this context, \eqn{f_{ij}} is the probability that two alleles 
  randomly sampled from individuals $i$ and $j$ are identical by descent, which is equivalent 
  to half the additive genetic relationship (\eqn{f_{ij} = a_{ij} / 2}). 
  This method is generally more robust as it accounts for full genetic drift and 
  bottlenecks (Cervantes et al., 2011).
  \deqn{\Delta c_{ij} = 1 - (1 - c_{ij})^{1/(\frac{ECG_i + ECG_j}{2})}}
  \deqn{N_e = \frac{1}{2 \overline{\Delta c}}}
  To handle large populations, this method samples \code{nsamples} individuals per cohort 
  and computes the mean rate of coancestry among them. Sampled relationship
  coefficients are obtained from batched matrix-free \eqn{AX} products, so
  the full additive relationship matrix is not constructed.
  
  \item \strong{"inbreeding"}: Based on the individual rate of inbreeding ($F_i$) (Gutiérrez et al., 2008, 2009).
  \deqn{\Delta F_i = 1 - (1 - F_i)^{1/(ECG_i - 1)}}
  \deqn{N_e = \frac{1}{2 \overline{\Delta F}}}
  
  \item \strong{"demographic"}: Based on the demographic census of breeding males and females (Wright, 1931).
  \deqn{N_e = \frac{4 N_m N_f}{N_m + N_f}}
  Where \eqn{N_m} and \eqn{N_f} are the number of unique male and female parents contributing to the cohort.
}
}
\examples{
\donttest{
# Coancestry-based Ne (default) using a simple pedigree grouped by year
tp_simple <- tidyped(simple_ped)
tp_simple$BirthYear <- 2000 + tp_simple$Gen
ne_coan <- suppressMessages(pedne(tp_simple, by = "BirthYear", seed = 42L))
ne_coan

# Inbreeding-based Ne using an inbred pedigree
tp_inbred <- tidyped(inbred_ped)
ne_inb <- suppressMessages(pedne(tp_inbred, method = "inbreeding", by = "Gen"))
ne_inb

# Demographic Ne from the number of contributing sires and dams
ne_demo <- suppressMessages(pedne(tp_simple, method = "demographic", by = "BirthYear"))
ne_demo
}

}
\references{
Cervantes, I., Goyache, F., Molina, A., Valera, M., & Gutiérrez, J. P. (2011). Estimation of effective population size from the rate of coancestry in pedigreed populations. \emph{Journal of Animal Breeding and Genetics}, 128(1), 56-63.

Gutiérrez, J. P., Cervantes, I., Molina, A., Valera, M., & Goyache, F. (2008). Individual increase in inbreeding allows estimating effective sizes from pedigrees. \emph{Genetics Selection Evolution}, 40(4), 359-370.

Gutiérrez, J. P., Cervantes, I., & Goyache, F. (2009). Improving the estimation of realized effective population sizes in farm animals. \emph{Journal of Animal Breeding and Genetics}, 126(4), 327-332.

Wright, S. (1931). Evolution in Mendelian populations. \emph{Genetics}, 16(2), 97-159.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedanalysis.R
\name{pedsubpop}
\alias{pedsubpop}
\title{Pedigree Subpopulations}
\usage{
pedsubpop(ped, by = NULL)
}
\arguments{
\item{ped}{A \code{tidyped} object.}

\item{by}{Character. The name of the column to group by. 
If NULL, summarizes disconnected components via \code{\link{splitped}}.}
}
\value{
A \code{data.table} with columns:
\itemize{
  \item \code{Group}: Subpopulation label.
  \item \code{N}: Total individuals.
  \item \code{N_Sire}: Number of distinct sires.
  \item \code{N_Dam}: Number of distinct dams.
  \item \code{N_Founder}: Number of founders (parents unknown).
}
}
\description{
Summarizes pedigree subpopulations and group structure.
}
\details{
When \code{by = NULL}, this function is a lightweight summary wrapper around
\code{\link{splitped}}, returning one row per disconnected pedigree component
plus an optional \code{"Isolated"} row for individuals with no known parents
and no offspring. When \code{by} is provided, it instead summarizes the pedigree
directly by the specified column (e.g. \code{"Gen"}, \code{"Year"}, \code{"Breed"}).


Use \code{pedsubpop()} when you want a compact analytical summary table.
Use \code{\link{splitped}} when you need the actual re-tidied sub-pedigree
objects for downstream plotting or analysis.
}
\examples{
tp <- tidyped(simple_ped)

# Summarize disconnected pedigree components
pedsubpop(tp)

# Summarize by an existing grouping variable
pedsubpop(tp, by = "Gen")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/splitped.R
\name{splitped}
\alias{splitped}
\title{Split Pedigree into Disconnected Groups}
\usage{
splitped(ped)
}
\arguments{
\item{ped}{A tidyped object created by \code{\link{tidyped}}.}
}
\value{
A list of class "splitped" containing:
  \item{GP1, GP2, ...}{tidyped objects for each disconnected group (with at least
    2 individuals), with renumbered IndNum, SireNum, DamNum}

  The returned object has the following attributes:
  \item{n_groups}{Number of disconnected groups found (excluding isolated individuals)}
  \item{sizes}{Named vector of group sizes}
  \item{total}{Total number of individuals in groups (excluding isolated)}
  \item{isolated}{Character vector of isolated individual IDs (Gen = 0)}
  \item{n_isolated}{Number of isolated individuals}
}
\description{
Detects and splits a tidyped object into disconnected groups (connected components).
Uses igraph to efficiently find groups of individuals that have no genetic
relationships with each other. Isolated individuals (Gen = 0, those with no
parents and no offspring) are excluded from group splitting and stored separately.
}
\details{
This function identifies connected components in the pedigree graph where
edges represent parent-offspring relationships. Two individuals are in the
same group if they share any ancestry (direct or indirect).

Isolated individuals (Gen = 0 in tidyped output) are those who:
\itemize{
  \item Have no known parents (Sire and Dam are both NA)
  \item Are not parents of any other individual in the pedigree
}
These isolated individuals are excluded from splitting and stored in the
\code{isolated} attribute. Each resulting group contains at least 2 individuals
(at least one parent-offspring relationship).

The function always returns a list, even if there is only one group (i.e.,
the pedigree is fully connected). Groups are sorted by size in descending
order.

Each group in the result is a valid tidyped object with:
\itemize{
  \item Renumbered IndNum (1 to n for each group)
  \item Updated SireNum and DamNum referencing the new IndNum
  \item Recalculated Gen (generation) based on the group's structure
}
}
\examples{
# Load example data
library(visPedigree)
data(small_ped)

# First tidy the pedigree
tped <- tidyped(small_ped)

# Split into groups
result <- splitped(tped)
print(result)

# Access individual groups (each is a tidyped object)
result$GP1

# Check isolated individuals
attr(result, "isolated")

}
\seealso{
\code{\link{tidyped}} for pedigree tidying
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped_style.R
\name{get_highlight_ids}
\alias{get_highlight_ids}
\title{Styling and finalizing pedigree graph}
\usage{
get_highlight_ids(ped, highlight, trace)
}
\description{
Styling and finalizing pedigree graph
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-s3.R
\name{build_ped_meta}
\alias{build_ped_meta}
\title{Build a unified metadata list for a tidyped object}
\usage{
build_ped_meta(
  selfing = FALSE,
  bisexual_parents = character(0),
  genmethod = "top"
)
}
\arguments{
\item{selfing}{Logical: whether selfing/monoecious mode was used.}

\item{bisexual_parents}{Character vector of IDs that appear as both sire and dam.}

\item{genmethod}{Character: generation assignment method ("top" or "bottom").}
}
\value{
A named list of pedigree metadata.
}
\description{
Build a unified metadata list for a tidyped object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped_style.R
\name{finalize_graph}
\alias{finalize_graph}
\title{Finalize graph and reindex IDs}
\usage{
finalize_graph(ped_node, ped_edge, highlight_info, trace, showf)
}
\description{
Finalize graph and reindex IDs
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedanalysis.R
\name{pedancestry}
\alias{pedancestry}
\title{Calculate Ancestry Proportions}
\usage{
pedancestry(ped, foundervar, target_labels = NULL)
}
\arguments{
\item{ped}{A \code{tidyped} object.}

\item{foundervar}{Character. The name of the column containing founder-group
labels (e.g., "Breed", "Origin").}

\item{target_labels}{Character vector. Specific founder-group labels to track.
If NULL, all unique labels in \code{foundervar} among founders are used.}
}
\value{
A \code{data.table} with columns:
\itemize{
  \item \code{Ind}: Individual ID.
  \item One column per tracked label (named after each unique value in
    \code{foundervar} among founders, or as specified by
    \code{target_labels}).
    Each value gives the proportion of genes (0--1) originating from that
    founder group. Row sums across all label columns equal 1.
}
}
\description{
Estimates the proportion of genes for each individual that originates from 
specific founder groups (e.g., breeds, source populations).
}
\examples{
\donttest{
library(data.table)
# Create dummy labels for founders
tp <- tidyped(small_ped)
tp_dated <- copy(tp)
founders <- tp_dated[is.na(Sire) & is.na(Dam), Ind]
# Assign 'LineA' and 'LineB'
tp_dated[Ind \%in\% founders[1:(length(founders)/2)], Origin := "LineA"]
tp_dated[is.na(Origin), Origin := "LineB"]

# Calculate ancestry proportions for all individuals
anc <- pedancestry(tp_dated, foundervar = "Origin")
print(tail(anc))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped_graph.R
\name{ped2igraph}
\alias{ped2igraph}
\title{Convert pedigree to igraph structure}
\usage{
ped2igraph(
  ped,
  compact = FALSE,
  highlight = NULL,
  trace = FALSE,
  showf = FALSE,
  labelvar = NULL,
  shapeby = "sex"
)
}
\description{
Convert pedigree to igraph structure
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped_layout.R
\name{repeloverlap}
\alias{repeloverlap}
\title{Repel overlapping nodes on the x-axis}
\usage{
repeloverlap(x)
}
\arguments{
\item{x}{A numeric vector of x positions.}
}
\value{
A numeric vector of unique x positions.
}
\description{
Repel overlapping nodes on the x-axis
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped_graph.R
\name{inject_missing_parents}
\alias{inject_missing_parents}
\title{Inject missing parents for subsetted pedigrees}
\usage{
inject_missing_parents(ped)
}
\arguments{
\item{ped}{A data.table containing pedigree info.}
}
\description{
Inject missing parents for subsetted pedigrees
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vispstat.R
\name{vispstat}
\alias{vispstat}
\alias{plot.pedstats}
\title{Visualize Pedigree Statistics (internal)}
\usage{
vispstat(x, type = c("genint", "ecg"), metric = "ECG", ...)

\method{plot}{pedstats}(x, ...)
}
\arguments{
\item{x}{A \code{pedstats} object returned by \code{\link{pedstats}}.}

\item{type}{Character. The type of plot to generate:
\itemize{
  \item \code{"genint"}: Bar chart of mean generation intervals.
  \item \code{"ecg"}: Histogram of ancestral depth metrics (ECG, FullGen, or MaxGen).
}}

\item{metric}{Character. Specific metric to plot when \code{type = "ecg"}.
Options: \code{"ECG"} (default), \code{"FullGen"}, \code{"MaxGen"}.}

\item{...}{Additional arguments passed to \code{\link[lattice]{barchart}} or 
\code{\link[lattice]{histogram}}.}
}
\value{
A lattice plot object.
}
\description{
Internal plotting backend for \code{\link{plot.pedstats}}.
Users should call \code{plot(stats, ...)} instead of this function directly.
}
\seealso{
\code{\link{pedstats}}, \code{\link{plot.pedstats}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedmatrix.R
\name{summary_pedmat}
\alias{summary_pedmat}
\title{Summary Statistics for Pedigree Matrices}
\usage{
summary_pedmat(x)
}
\arguments{
\item{x}{A pedmat object from \code{\link{pedmat}}.}
}
\value{
An object of class \code{"summary.pedmat"} with statistics including
  method, dimensions, compression ratio (if compact), mean relationship, 
  and matrix density.
}
\description{
Computes and displays summary statistics for a pedmat object.
}
\details{
Since pedmat objects are often S4 sparse matrices with custom attributes,
use this function instead of the generic \code{summary()} to ensure proper
display of pedigree matrix statistics.
}
\examples{
tped <- tidyped(small_ped)
A <- pedmat(tped, method = "A")
summary_pedmat(A)

}
\seealso{
\code{\link{pedmat}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedanalysis.R
\name{pedcontrib}
\alias{pedcontrib}
\title{Calculate Founder and Ancestor Contributions}
\usage{
pedcontrib(
  ped,
  reference = NULL,
  mode = c("both", "founder", "ancestor"),
  top = 20
)
}
\arguments{
\item{ped}{A \code{tidyped} object.}

\item{reference}{Character vector. Optional subset of individual IDs defining the reference population.
If NULL, uses all individuals in the most recent generation.}

\item{mode}{Character. Type of contribution to calculate:
\itemize{
  \item \code{"founder"}: Founder contributions ($f_e$).
  \item \code{"ancestor"}: Ancestor contributions ($f_a$).
  \item \code{"both"}: Both founder and ancestor contributions.
}}

\item{top}{Integer. Number of top contributors to return. Default is 20.}
}
\value{
A list with class \code{pedcontrib} containing:
\itemize{
  \item \code{founders}: A \code{data.table} of founder contributions (if mode includes "founder", or "both").
  \item \code{ancestors}: A \code{data.table} of ancestor contributions (if mode includes "ancestor", or "both").
  \item \code{summary}: A \code{list} of summary statistics including:
    \itemize{
      \item \code{f_e}: Classical effective number of founders (\eqn{q=2}, Lacy 1989).
      \item \code{f_e_H}: Information-theoretic effective number of founders
        (\eqn{q=1}, Shannon entropy): \eqn{f_e^{(H)} = \exp(-\sum p_i \ln p_i)}.
      \item \code{f_a}: Classical effective number of ancestors (\eqn{q=2}, Boichard 1997).
      \item \code{f_a_H}: Information-theoretic effective number of ancestors
        (\eqn{q=1}): \eqn{f_a^{(H)} = \exp(-\sum q_k \ln q_k)}.
    }
}

Each contribution table contains:
\itemize{
  \item \code{Ind}: Individual ID.
  \item \code{Contrib}: Contribution to the reference population (0-1).
  \item \code{CumContrib}: Cumulative contribution.
  \item \code{Rank}: Rank by contribution.
}
}
\description{
Calculates genetic contributions from founders and influential ancestors.
Implements the gene dropping algorithm for founder contributions and Boichard's 
algorithm for ancestor contributions to estimate the effective number of founders ($f_e$) 
and ancestors ($f_a$).
}
\details{
**Founder Contributions ($f_e$):**
Calculated by probabilistic gene flow from founders to the reference cohort. 
When individual ancestors with one unknown parent exist, "phantom" parents are temporarily injected correctly conserving the probability mass.

**Ancestor Contributions ($f_a$):**
Calculated using Boichard's iterative algorithm (1997), accounting for:
\itemize{
  \item Marginal genetic contribution of each ancestor
  \item Long-term contributions through multiple pathways
}
The parameter $f_a$ acts as a stringent metric since it identifies the bottlenecks of genetic variation in pedigrees.
}
\examples{
\donttest{
library(data.table)
# Load a sample pedigree
tp <- tidyped(small_ped)

# Calculate both founder and ancestor contributions for reference population
ref_ids <- c("Z1", "Z2", "X", "Y")
contrib <- pedcontrib(tp, reference = ref_ids, mode = "both")

# Print results including f_e, f_e(H), f_a, and f_a(H)
print(contrib)

# Access Shannon-entropy effective numbers directly
contrib$summary$f_e_H   # Information-theoretic effective founders (q=1)
contrib$summary$f_e     # Classical effective founders (q=2)
contrib$summary$f_a_H   # Information-theoretic effective ancestors (q=1)
contrib$summary$f_a     # Classical effective ancestors (q=2)

# Diversity ratio rho > 1 indicates long-tail founder value
contrib$summary$f_e_H / contrib$summary$f_e
}

}
\references{
Boichard, D., Maignel, L., & Verrier, É. (1997). The value of using probabilities 
of gene origin to measure genetic variability in a population. Genetics Selection 
Evolution, 29(1), 5-23.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped_graph.R
\name{generate_graph_structure}
\alias{generate_graph_structure}
\title{Generate edges and virtual family nodes}
\usage{
generate_graph_structure(ped_node, h_ids)
}
\arguments{
\item{ped_node}{A data.table of nodes.}

\item{h_ids}{Highlighted IDs.}
}
\description{
Generate edges and virtual family nodes
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped.R
\name{visped}
\alias{visped}
\title{Visualize a tidy pedigree}
\usage{
visped(
  ped,
  compact = FALSE,
  outline = FALSE,
  cex = NULL,
  showgraph = TRUE,
  file = NULL,
  highlight = NULL,
  trace = FALSE,
  showf = FALSE,
  labelvar = NULL,
  shapeby = c("sex", "role"),
  pagewidth = 200,
  symbolsize = 1,
  maxiter = 1000,
  genlab = FALSE,
  genlabcex = NULL,
  ...
)
}
\arguments{
\item{ped}{A \code{tidyped} object (which inherits from \code{data.table}). It is recommended that the pedigree is tidied and pruned by candidates using the \code{\link{tidyped}} function with the non-null parameter \code{cand}.}

\item{compact}{A logical value indicating whether terminal, non-parent
full-sib individuals will be replaced by one family display node. For
example, 100 eligible full siblings are shown as \code{"FS×100"} when
\code{compact = TRUE}. Highlighted individuals are not compacted, and the
underlying pedigree is unchanged. The default value is FALSE.}

\item{outline}{A logical value indicating whether shapes without labels will be shown. A graph of the pedigree without individual labels is shown when setting \code{outline = TRUE}. This is useful for viewing the pedigree outline and identifying immigrant individuals in each generation when the graph width exceeds the maximum PDF width (500 inches). The default value is FALSE.}

\item{cex}{NULL or a numeric value changing the size of individual labels shown in the graph. \emph{cex} is an abbreviation for 'character expansion factor'. The \code{visped} function will attempt to estimate (\code{cex=NULL}) the appropriate cex value and report it in the messages. Based on the reported cex from a previous run, this parameter should be increased if labels are wider than their shapes in the PDF; conversely, it should be decreased if labels are narrower than their shapes. The default value is NULL.}

\item{showgraph}{A logical value indicating whether a plot will be shown in the default graphic device (e.g., the Plots panel in RStudio). This is useful for quick viewing without opening a PDF file. However, the graph on the default device may not be legible (e.g., overlapping labels or aliasing lines) due to size restrictions. It is recommended to set \code{showgraph = FALSE} for large pedigrees. The default value is TRUE.}

\item{file}{NULL or a character value specifying whether the pedigree graph will be saved as a vector file. Files ending in '.svg' are written with the SVG device; all other file names use PDF output. The vector output is legible and avoids overlapping labels even with many individuals or long labels. It is recommended to save the pedigree graph as a vector file. The default value is NULL.}

\item{highlight}{NULL, a character vector of individual IDs, or a list specifying individuals to highlight. If a character vector is provided, individuals will be highlighted with a purple border while preserving their sex-based fill color. If a list is provided, it should contain:
\itemize{
  \item \code{ids}: (required) character vector of individual IDs to highlight.
  \item \code{frame.color}: (optional) hex color for the border of focal individuals.
  \item \code{color}: (optional) hex color for the fill of focal individuals.
  \item \code{rel.frame.color}: (optional) hex color for the border of relatives (used when \code{trace} is not NULL).
  \item \code{rel.color}: (optional) hex color for the fill of relatives (used when \code{trace} is not NULL).
}
For example: \code{c("A", "B")} or \code{list(ids = c("A", "B"), frame.color = "#9c27b0")}. The function will check if the specified individuals exist in the pedigree and issue a warning for any missing IDs. The default value is NULL.}

\item{trace}{A logical value or a character string. If TRUE, all ancestors and descendants of the individuals specified in \code{highlight} will be highlighted. If a character string, it specifies the tracing direction: "\strong{up}" (ancestors), "\strong{down}" (descendants), or "\strong{all}" (union of ancestors and descendants). This is useful for focusing on specific families within a large pedigree. The default value is FALSE.}

\item{showf}{A logical value indicating whether inbreeding coefficients will be shown in the graph. If \code{showf = TRUE} and the column \strong{f} is missing, \code{visped()} will try to compute it automatically with \code{\link{inbreed}} on a structurally complete pedigree. If automatic computation is not possible, a warning is issued and labels are drawn without \strong{f}. The default value is FALSE.}

\item{labelvar}{NULL, a single non-empty character value naming a column in
\code{ped}, or a character vector with one displayed label per input
pedigree row. Row-aligned vectors are attached before any internal
tidying or reordering. Compact full-sib family nodes still show family
size when \code{compact = TRUE}. Missing or empty values fall back to
individual IDs. The default value is NULL.}

\item{shapeby}{Character string controlling what node shapes encode.
\code{"sex"} (default) uses circles for females, squares for males,
diamonds for unknown sex, and hexagons for monoecious individuals.
\code{"role"} distinguishes individual records from compact family
summaries by shape while keeping sex-based fill colors for real
individuals. Compact family summaries are rectangles in both modes.}

\item{pagewidth}{A numeric value specifying the width of the PDF file in inches. This controls the horizontal scaling of the layout. The default value is 200.}

\item{symbolsize}{A numeric value specifying the scaling factor for node size relative to the label size. Values greater than 1 increase the node size (adding padding around the label), while values less than 1 decrease it. This is useful for fine-tuning the whitespace and legibility of dense graphs. The default value is 1.}

\item{maxiter}{An integer specifying the maximum number of iterations for the Sugiyama layout algorithm to minimize edge crossings. Higher values (e.g., 2000 or 5000) may result in fewer crossed lines for complex pedigrees but will increase computation time. The default value is 1000.}

\item{genlab}{FALSE to omit generation labels; TRUE to draw the default
labels G1, G2, ...; or an unnamed, non-empty character vector of custom
labels. Custom labels are assigned from top to bottom to the final
displayed generation layers and must contain exactly one label per layer.
The default value is FALSE.}

\item{genlabcex}{NULL or a numeric value controlling the size of displayed
generation labels. If \code{NULL}, \code{visped()} uses an automatic size
based on node scaling. Set a larger value to keep generation labels
readable in deep pedigrees. The default value is NULL.}

\item{...}{Additional arguments passed to \code{\link[igraph:plot.igraph]{plot.igraph}}.}
}
\value{
The function mainly produces a plot on the current graphics device and/or a vector file. It invisibly returns a list containing the graph object, layout coordinates, and node sizes.
}
\description{
\code{visped} function draws a graph of a full or compact pedigree.
}
\details{
This function takes a pedigree tidied by the \code{\link{tidyped}} function and outputs a hierarchical graph for all individuals in the pedigree. The graph can be shown on the default graphic device or saved as a PDF or SVG file. The vector output is legible and avoids overlapping labels. It is especially useful when the number of individuals is large and individual labels are long.

Rendering is performed using a Two-Pass strategy: edges are drawn first to ensure center-to-center connectivity, followed by nodes and labels. This ensures perfect visual alignment in high-resolution vector outputs. The function also supports real-time ancestry and descendant highlighting.

This function can draw the graph of a very large pedigree (> 10,000 individuals per generation) by compacting full-sib individuals. It is highly effective for aquatic animal pedigrees, which usually include many full-sib families per generation in nucleus breeding populations. The outline of a pedigree without individual labels is still shown if the width of a pedigree graph exceeds the maximum width (500 inches) of the PDF file.

By default, \code{shapeby = "sex"} encodes individual sex with node shape:
females are circles, males are squares, individuals of unknown sex are
diamonds, and monoecious individuals are hexagons. Set
\code{shapeby = "role"} to use the legacy role-based scheme, where circles
represent individual records and green-grey rectangles represent compact
full-sib family summaries. Compact family summaries remain green-grey
rectangles in both modes. In \code{shapeby = "role"} mode, real individuals
remain circles, but their fill colors still encode sex. Dark sky blue
indicates males, dark goldenrod indicates females, teal indicates monoecious
individuals, and neutral grey indicates unknown sex. Purple borders indicate
highlighted individuals rather than sex.
}
\note{
Isolated individuals (those with no parents and no progeny, assigned Gen 0) are automatically filtered out and not shown in the plot. A message will be issued if any such individuals are removed.
}
\examples{
library(visPedigree)
library(data.table)
# Drawing a simple pedigree
simple_ped_tidy <- tidyped(simple_ped)
visped(simple_ped_tidy, 
       cex=0.25, 
       symbolsize=5.5)

# Use application-specific labels for the displayed generations.
# Custom labels are assigned from top to bottom.
generation_labels <- paste("Generation", sort(unique(simple_ped_tidy$Gen)))
visped(simple_ped_tidy,
       genlab = generation_labels,
       cex = 0.25,
       symbolsize = 5.5)

# Highlighting an individual and its ancestors and descendants
visped(simple_ped_tidy, 
       highlight = "J5X804", 
       trace = "all", 
       cex=0.25, 
       symbolsize=5.5)

# Showing inbreeding coefficients in the graph
simple_ped_tidy_inbreed <- tidyped(simple_ped, inbreed = TRUE)
visped(simple_ped_tidy_inbreed,
       showf = TRUE, 
       cex=0.25, 
       symbolsize=5.5)

# visped() will automatically compute inbreeding coefficients if 'f' is missing
visped(simple_ped_tidy,
       showf = TRUE,
       cex=0.25,
       symbolsize=5.5)

# Use the role-based legacy symbol scheme
visped(simple_ped_tidy,
       shapeby = "role",
       cex=0.25,
       symbolsize=5.5)

# Adjusting page width and symbol size for better layout
# Increase pagewidth to spread nodes horizontally in the pdf file
# Increase symbolsize for more padding around individual labels
visped(simple_ped_tidy, 
       cex=0.25, 
       symbolsize=5.5, 
       pagewidth = 100, 
       file = tempfile(fileext = ".pdf"))

# Saving pedigree as SVG with larger labels and tighter symbols
visped(simple_ped_tidy,
       cex=0.8,
       symbolsize=1,
       pagewidth = 100,
       file = tempfile(fileext = ".svg"))

# Highlighting multiple individuals with custom colors
visped(simple_ped_tidy,
       highlight = list(ids = c("J3Y620", "J1X971"),
                        frame.color = "#4caf50",
                        color = "#81c784"),
       cex=0.25,
       symbolsize=5.5)

# Handling large pedigrees: Saving to PDF is recommended for legibility
# The 'trace' and 'tracegen' parameters in tidyped() help prune the graph
cand_labels <- big_family_size_ped[(Year == 2007) & (substr(Ind,1,2) == "G8"), Ind]
\donttest{
big_ped_tidy <- tidyped(big_family_size_ped, 
                        cand = cand_labels, 
                        trace = "up", 
                        tracegen = 2)
# Use compact = TRUE for large families
visped(big_ped_tidy, 
       compact = TRUE, 
       cex=0.08, 
       symbolsize=5.5, 
       file = tempfile(fileext = ".pdf"))

# Use outline = TRUE if individual labels are not required
visped(big_ped_tidy, 
       compact = TRUE, 
       outline = TRUE, 
       file = tempfile(fileext = ".pdf"))
}

}
\seealso{
\code{\link{tidyped}} for tidying pedigree data (required input)
\code{\link{vismat}} for visualizing relationship matrices as heatmaps
\code{\link{pedmat}} for computing relationship matrices
\code{\link{splitped}} for splitting pedigree into connected components
\code{\link[igraph:plot.igraph]{plot.igraph}} underlying plotting function
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pedanalysis.R
\name{pedpartial}
\alias{pedpartial}
\title{Calculate Partial Inbreeding}
\usage{
pedpartial(ped, ancestors = NULL, top = 20)
}
\arguments{
\item{ped}{A \code{tidyped} object.}

\item{ancestors}{Character vector. IDs of ancestors to calculate partial 
inbreeding for. If NULL, the top ancestors by marginal contribution are used.}

\item{top}{Integer. Number of top ancestors to include if \code{ancestors} is NULL.}
}
\value{
A \code{data.table} with the first column as \code{Ind} and subsequent 
  columns representing the partial inbreeding ($pF$) from each ancestor.
}
\description{
Decomposes individuals' inbreeding coefficients into marginal contributions from 
specific ancestors. This allows identifying which ancestors or lineages are 
responsible for the observed inbreeding.
}
\details{
The sum of all partial inbreeding coefficients for an individual (including 
contributions from founders) equals $1 + f_i$, where $f_i$ is the total 
inbreeding coefficient. This function specifically isolates the terms in 
the Meuwissen & Luo (1992) decomposition that correspond to the selected ancestors.
}
\examples{
\donttest{
library(data.table)
tp <- tidyped(inbred_ped)
# Calculate partial inbreeding originating from specific ancestors
target_ancestors <- inbred_ped[is.na(Sire) & is.na(Dam), Ind]
pF <- pedpartial(tp, ancestors = target_ancestors)
print(tail(pF))
}

}
\references{
Lacey, R. C. (1996). A formula for determining the partial inbreeding coefficient, 
\eqn{F_{ij}}. Journal of Heredity, 87(4), 337-339.

Meuwissen, T. H., & Luo, Z. (1992). Computing inbreeding coefficients in 
large populations. Genetics Selection Evolution, 24(4), 305-313.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-s3.R
\name{has_candidates}
\alias{has_candidates}
\title{Check whether a tidyped object contains candidate flags}
\usage{
has_candidates(x)
}
\arguments{
\item{x}{A tidyped object.}
}
\value{
Logical scalar.
}
\description{
Check whether a tidyped object contains candidate flags
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-s3.R
\name{plot.tidyped}
\alias{plot.tidyped}
\title{Plot a tidy pedigree}
\usage{
\method{plot}{tidyped}(x, ...)
}
\arguments{
\item{x}{A \code{tidyped} object.}

\item{...}{Additional arguments passed to \code{\link{visped}}.}
}
\value{
Invisibly returns a list of graph data from \code{\link{visped}} (node/edge
  data and layout components) used to render the pedigree; the primary result is
  the plot drawn on the current device.
}
\description{
Plot a tidy pedigree
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visped_render.R
\name{resolve_visped_genlabcex}
\alias{resolve_visped_genlabcex}
\title{Resolve generation-label text size}
\usage{
resolve_visped_genlabcex(node_size, genlabcex = NULL)
}
\description{
Resolve generation-label text size
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-s3.R
\name{validate_tidyped}
\alias{validate_tidyped}
\title{Internal validator for tidyped class}
\usage{
validate_tidyped(x)
}
\arguments{
\item{x}{A tidyped object}
}
\value{
The (possibly restored) object if valid, otherwise an error
}
\description{
Validates a tidyped object. If the object has lost its class but retains the
required columns, it is automatically restored via \code{ensure_tidyped()}.
Fatal structural problems (e.g., missing core columns) raise an error.
}
\keyword{internal}
